#include "gtk.h"
/* XPM */
static const char *const xpm_icon_0[] = {
/* columns rows colors chars-per-pixel */
"128 128 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX0X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX0XiXxXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXzXzXlXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXzXzXzXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XxXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXjXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXCXNXUXtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXkXiXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXsXqX'.PXSXuXsXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXdX XtXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXiX6X9X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X7XwX}.~.UXZX4X0X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8XqX1XL.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X^.UXZX7XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrX4XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuXpX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuXOX}.pX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XcXa.o ].pX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XxXs.5 F R.aX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqX0X9X9XqXqXqX0X9X9X9X0XqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqX0X9X9X0XqXqX0X9X9X9X0XqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqX0X8X9X9X9X9X9X9X5XxX7.^ B.p ~.aX9XqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqX8XyXgXgXrXqX0XqXaXfXdXrX9XqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqX8XeXfXhXuX0X0XwXsXfXdXeX9XqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqX0XyXkXlXkXkXkXkXkXfXLXd.P 2Xw.d ~.aX8XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXdX&XP.I.1XwXeXwX].Q.(.1XpX9XqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqX0XsX5XT.J.@XtXrX6X_.Q.(.2XyXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqX0XtX%XB.M.M.M.M.M.M.b.U.Y I &X'.y.d !.sX7XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXeXK.d     R.hX1X#.#     P |.iX9XqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqX0XrX}.T     <.vX%X)       H XXrXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqX6XnX&.$ F b n n n n m m m d.{.!.}.r.g !.sX8XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqX0XrX|.$ = t   /.gX6.  p t.P   A 8X0X0XqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqX0XqX6Xv   f   8.CX(   N l.F   P tXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqX6XMXo.;.rX$X*X*X*X*X*X*X*X[.^./.!.|.t.f !.aX0XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqX0XyXP./ W.n   `.rX, * Z.MX<Xw + C.dX8XqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqX9XtX@X~ H.X.  8.nXb r '.BX].5 + wXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqX6XmXO.[ oXR.W.W.W.W.W.Q.W.~./././.!.|.y.f ~.aX0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXwX4XeX<Xc   '.@X  z :XqXeXX.  <.xX6XqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXwX8X5XaX}   0.lX-X,X7XuXT.= t tXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqX6XMXo.| $X!././././././././././././.~.}.y.d ~.aX9XqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXiX%Xc   '.^.  V 9X0XwX7.  #.xX6XqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXwXrXwX{   0.zX0XwXrX<XM   h.yXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqX6XMXo.| $X!./././././././././././././.^.{.u.l W.iX9XqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqX0XyX*Xc   '._.  N 5X0XqX1.  &.cX6XqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXeX{   0.zX4XtX:XA   o.rXwXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqX6XMXo.| $X!./././././././././././././.^.`.Q.f 0.kX7XqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqX0XyX*Xc   '.<X  u oXuXyXU   w.lX6XqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXwXeX{   0.lXwX*Xm   { 0XqX9XqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqX6XMXo.| $X!././././././././././././.^._.Q.M ;.yX9XqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqX0XyX*Xv   _.pXv   -.sXK.  2 /.iX9XqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXwXeX}   0.cXoXc   $.sXsXuXeXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqX6XMXo.} #XQ.^.^.^.^.^.^.^././././.^._.Q.M -.uXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqX0XyX&Xh   ~.kXK.O   3 +   =.sX8XqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXwXeX`   3.ZXl     > 7 > 0 |.rX0XqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqX6XMXO.@.1X[.|.|.|.|.|.|.}._././.~.`.W.M -.uXwX0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqX0XtX-XT 7 ].aXrXW.K 0 c h.rXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXeX:., a.MXB 5 r 6 3 2 q |.rX0XqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqX6XMX .Z C.s.f.f.f.f.f.d.k.Y.).!.`.E.N -.uXwX0XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXwXsXgXeX0XqXtXdXgXfXuXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXiXgXuX9XdXgXfXgXhXhXgXeXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqX8XgXV.z N M m m m m n D > ` @X/.W.M -.pXqX0XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqX9X8X0XqXqX0X9X8X9X0XqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqX9X8X0XqX9X8X8X8X8X8X8XqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqX9XaXnXnXnXnXnXnXnXxXUXj.L ;XY.N =.iX0X0XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqX9X5X6X6X6X6X6X6X2XkX6.R #Xv -.iX0X0XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XcXw.F W %.uXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XxXf.  w.tXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XhXS.>.iXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XiXfX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX X~.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXaX7X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XrX X~.UXZX6XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XwX3XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXsXwXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXpX.X~.UXZX0XyXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXiX6XP.aXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XkXUXqX}. X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X.X).^.UXVXoX X.X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X.X|.U.pXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XxXjX'.~.^.~.~.~.~.~.~.^./.^.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.^.[.CX%X^.^.~.~.~.~.~.~.~./.^.^.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.(.W.sXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XxXUXIXUXUXUXUXUXUXUXUXGXFXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXUXUXUXUXUXUXUXUXCXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX8XfXDXZXZXZXZXZXZXZXZXGX3XgXUXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXSXzX;XUXGXCXAXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXAXCXJXUXtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX9X6X7X7X7X7X7X7X6X9X1XY.gXUX8X6X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X9X`..XUXkX3X8X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X2XnXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXwXwXwXwXwXwXwXqXtX2XP.hXUXeXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXrX^.oXUXxX8XeXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXeX8XMXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXpXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XlXv.E eXrX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.  $.7XrX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XvX0.T  .W 0XeX0XqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqX9X8X9XqXqX9X8X8X8X8X9XqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqX0X8X8X0XqXqX0X9X8X9X0XwX7XmXUXrX0XqXqXqXqXqXqX0X8X8X8X8X8X8X8X5XxX8.U :XJ ^ qXeX0XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXrXdXgXsX0XeXdXgXhXhXhXdXeXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXwXpXgXgXrX0XqXrXsXgXfXiXeX6XMXUXrX0XqXqXqXqXqX0XuXfXsXdXdXdXdXdXiXDXy.K =X/.U ^ qXeX9XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqX9XuXoXF 2 P fX}.C 4 6 6 , M oXuX9XqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXeXwX5.8 w W.sXrX[.E 0 j u.rX9XmXUXrX0XqXqXqXqX7XkXf.2 a y y y y y y d * O.oX~._.I / qXrX9XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqX0XuXJ.=     i lXT.=   2 1 - m OXuX9XqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXwX4XG       J.zX_.:   ; o   ( eXbXUXrX0XqXqXqXqX6XMX .~ [.J.P.P.P.P.P.L.Y././.^.).`.I / 0XrX9XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXwXeXO.p ;.% s cXJ.   .BXHXNXdXwXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqX0XyXW.7 +.V   Y.jXU   &.jX@Xb m {.VXUXrX0XqXqXqXqX6XMXO...*X/._._._._._._._.(././.^.(.`.I / 0XeX0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXwX0XR.>XL.X s bXB.  *.v.1.M.7XrX0XqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqX0XrX+XQ.9XV   R.rX  h  Xl.w.n.,X,XNXUXrX0XqXqXqXqX6XMXo.| $X!./././././././././././.^./.'.I ^ 0XeX0XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXyXgXn.+ a NXu.  2 o     ;.7XwX0XqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXrXfX-XV   W.OX  r A     @ 0.iXbXUXrX0XqXqXqXqX6XMXo.| $X!././././././././././././.^./.`.U ) 8XqX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqX6XpXB.O f jX].X.=.[.Y.7   L.dX9XqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqX9XwX>XV   W._..   ' $Xm.  , ).CXUXrX0XqXqXqXqX6XMXo.| $X!./././././././././././././.~._._.f 1.zX7XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqX8XsXB.O h iX0X9XyXuXgXq.  t.iX0XqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqX0XtX>XV   E.$X  0  XdXpXD + h.JXUXrX0XqXqXqXqX6XMXo.| $X!./././././././././././././.^.{.;.B XXtX9XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqX8XsXB.O g gXz.A K.lXdX%.  g.sX9XqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqX0XtX>XV   R.iX& , I.MX9Xh $ B.FXUXrX0XqXqXqXqX6XMXo.| $X!././././././././././././.~.|.-.v @XiX9XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqX8XsXV.% h cXf.  8 i.,.  > {.pX9XqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqX0XtX>XA   T.kX:.  z l.)   x <XMXUXrX0XqXqXqXqX6XMXO.[ OXR.Q.Q.Q.Q.Q.Q.Q.~./././.!.|.-.b @XuX0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqX8XdXv.  9 sX6X0.;     p S.rX0XqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqX0XtX:Xh   K.kX,XX.o     Z ).tXnXUXrX0XqXqXqXqX6XMXX.%.8XXX@X@X@X@X@X+X@X].^./.^.|.=.b +XiX9XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqX0XeX-XI.^.wXrXeX|.W.!.&XuXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXwX8X).K.,XeXeXwX'.W./.>XsX5XMXUXrX0XqXqXqXqX6XvX1.X g y y y y y u i 0 7. X!.}.*.n @XiX8XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqX0XtXgXsXqX0X0XpXfXfXyX9XqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXwXaXhXrX0X0XqXaXgXdXrX0X7XmXUXrX0XqXqXqXqXqXwX3X X{.}.}.}.}.}.'.,XO.L *X'.;.v @XiX8XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqX0X9X9XqXqXqX9X9X9X0XqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqX9X8X0XqXqX0X9X9X9X0XwX7XmXUXrX0XqXqXqXqXqXqXeXpXpXpXpXpXpXpXtXZXy.P 2X%.v #XiX9XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqX0X0X0X0X0X0X0X7XvX8.^ u.z @XuX0XqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXf.@ T XXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XxXd.8 =XiX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrX,X<XuX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX1XP.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX/.oXUXzX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX6XmXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwXrXeXeXeXeXeXeXeXyX3XP.hXUXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXyX/.oXUXzX9XrXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXrX9XMXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXwX8X3X4X3X3X3X3X3X3X6X;XY.jXUXqX2X4X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X5X(.OXUXbX1X4X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X5X>XvXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XyX$XG.I.P.P.P.P.P.P.P.Y.G.sXbX(.L.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.I.L.|.KX&XI.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.Y.A.rXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX0XtXsXaXaXaXaXaXaXaXaXaXtXZXzXyXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXuXkXAXyXpXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXsXuXVXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwX7XvXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwXrXeXeXeXeXeXeXeXeXeXeXtXrXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXrXtXrXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXrXeXrXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X6X6X6X6X6X6X6X6X6X6X8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X6X6X6X6X6X6X6X6X6X6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXnXMXmXMXMXMXMXMXMXMXgX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XkXMXMXMXMXMXMXMXMXMXvXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX%X&.o.O.o.o.o.o.o.O. .V.aX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuXf. .O.o.o.o.o.o.O.X.1.3XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XkXB.$ ;.[ | | | | } @.Z z nX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XfX2 ~ ..| | | | | [ %.X  XpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XlXM.F rXoX$X$X$X$X#X1XC.N nX6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XsXa [.*X$X$X$X$X$XOX8Xg {.pX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XkXM.b $XR.!.!.!.!.Q.[.s.M nX6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XdXy J./.!.!.!.!.!.R.XXy }.pX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XkXM.n *XW././././.^.|.f.m nX6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XdXy P._./././././.Q.@Xy }.pX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XkXM.n *XW././././.^.|.f.m nX6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XdXy P._./././././.Q.@Xy }.pX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XkXM.n *XW././././.^.|.f.m nX6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XdXy P._./././././.Q.@Xy }.pX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XkXM.n *XW././././.^.|.f.m nX6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XdXy P._./././././.Q.@Xy }.pX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X7X7X5XfXb.m *XW././././.^.|.f.n xX2X7X7X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X7X7X5XiXy P._./././././.Q.@Xu '.tX7X7X7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuXcXxXxXLXU.m *XQ././././.^.|.d.D UXkXcXxXhX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XkXzXvXxXDXd L._./././././.Q.+Xi ,XZXcXzXxXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuXOXs.s.7.d.Y m *XW././././.^.}.k., j.6.w.f.D.iX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXb.g.0.8.u.* Y._./././././.Q.@X0 O.y.8.f.f.,XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XiX|.+ 3 ^ P I d.[.~./././././._.Y.' K R F   >.fX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX!   U U K O./.(./././././.~.].8.L P ^ # 8 ,XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XpX{.G n.2X&X{.^./././././././.).+X;X#XW q.uX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXrX&.{ ;X-XoX/./././././././.^. X*X2Xu.Y *XyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XpX!.a 9.'.!././././././././.!./.Y.v %.rXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX9X^ G ~.^.~././././././././.!.'.%.z XXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XaX).g e.|.!././././././.~.`.W.N -.uXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XeXwX_ L ).).^././././././.^.}.;.v @XiX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XaX(.g w.|.!././././.^.`.W.N =.iXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XwXeX_ L ).(.^././././.!.|.-.b #XuX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XaX(.h w.|.~././.^._.W.N =.iX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXeX_ L _./.^././.~.|.=.b @XiX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XsX/.h w.}.^.^._.Q.M -.iX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXwX_ K _.(.~.^.|.=.b +XiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XaX(.g e.{.`.Q.M -.uXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XeXwX` K )._.{.-.b +XpX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XaX).f r.Q.M -.uXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XrXwX_ L _.;.v @XiX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX^.l d :.iXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XeXqX_ d B $XuX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XaXW.0.uXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX9X,.oXuX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XiXkX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XzXtX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X6X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_1[] = {
/* columns rows colors chars-per-pixel */
"96 96 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXtXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXtXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXLXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXMXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXKXVXJX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUXvXsXgXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXjX0X=XUXhXfXgXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXkX,X0XUX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUXsX5X9X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X7XeX}.+XUXqX7X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8XwX~.9XUX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUXgX9XwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXyXXX@XUXrXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXtX(.9XUX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqXqXqXqXqXqXqXqXqXqXrXqX0XqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqXqXqXqXqXqXqXqXqXwX1XtXwX0XqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqXqXqXqXqXqXqXqX7XxX} | gXqX0XqXqXqXqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqXqXqX0X0XqXqXqX0X0X0XqXqXqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqXqX0X0XqXqXqX0X0X0XqXqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqXqX0X0X0X0X0X0X5XlX| v 3.pXqX0XqXqXqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqX9XqXiXyXqX0X0XyXuXwX9XqXqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqX9XyXiXeX0X0XyXiXeX9XqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqX0XrXsXsXsXsXsXuXCXX.N...=.aXqX0XqXqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqX0XiXrXoX%XqXeXtX*X$X0XuX9XqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXwXdX-XXX4XtXyX-X#X5XuXqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqXeX<X/././././.Q.+XF x.@X_ =.aXqX0XqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqX0XrX$XL   k iX5X[     D #XuX9XqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqX0XeXp.%   K.fX;.O   u E.yX0XqXqXqXqXqXqXtX(.9XUX8XqXqXqX7XxX%.k ^ T T T T Y ~ Y.).{._ -.aXqX0XqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqX9XpX] q e s zX2.  6.N.  U tX0X0XqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqX0XiXQ.1 B   ~.L.  $.E.1 y rXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqX7XcXo.n.>XOX+X+X+X+XOX(.^./.{.` =.aXqX0XqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqX0XT.;Xb i jXq k eXbX` . @XuX0XqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqX0XtX|.|.b.  R.XX7.5XnX_ # wXwXqXqXqXqXqXqXtX(.9XUX8XqXqXqX7XcXO.u.[.W.Q.Q.Q.Q.!././.^./.{.` =.aXwX0XqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqX0XsX0Xh f 9X- T uXyXe.  '.aX9XqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqX0XeXzXs.  J.xXsXfXQ.  X.pX0XqXqXqXqXqXqXtX(.9XUX8XqXqXqX7XcXO.a. X^././././././././.^./.[.{ :.rXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqX0XqX7Xj d wX; I yXuX8.  [.pX9XqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqX8XaXf.  L.pXtX~.o x 3XeX0XqXqXqXqXqXqXtX(.9XUX8XqXqXqX7XcXO.a. X^./././././././././.Q.XXi.A 8XeXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqX0XeX7Xl a xXh 7 4XbXS - =XyX0XqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqX9XdXf.  J.vXY.- P gXbXwXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqX7XcXO.p.}.!.^.^.^.^.^./././.!. Xs.J &XyX0XqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqX0XeX7Xa q lXv.  C ~   %.aX9XqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqX9XfXp.  R.{.    ` %.$.:XeX0XqXqXqXqXqXtX(.9XUX8XqXqXqX7XcX+.c.*X|..X.X X.X}././.!. Xs.H $XpX9XqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqX0XwX8XT K uXtXz.a 9 2.8XeX0XqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqX9XaXM.6 }.A.; d 2 = 9 OXtX0XqXqXqXqXqXtX(.9XUX8XqXqXqX6XvX{ ' h.t.y.y.y.r.l.^.~. Xs.H %XpX8XqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqXqXqXiXiX0XqXpXfXfXaXeX0XqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqXqXrXaXwXtXdXaXfXgXfXeXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqX9XaXI.X.&.$.$.$.@.,.d m.#Xa.G %XpX8XqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqXqXqX0X0XqXqX9X9X9X9X0XqXqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqXqX0X9XqX0X9X9X9X9X9XqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqX9XaXvXcXcXcXcXkXIXo.B.x.G %XpX8XqXqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqXqX9X6X7X7X7X7X2XgX..O.R $XpX8XqXqXqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqXqXqXqXqXqXqXqX6XvXQ f :XpX8XqXqXqXqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqXqXqXqXqXqXqXqX9XpXK.oXaX8XqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqXqXqXqXqXqXqXqXqX9XsXyX9XqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqXqXqXqXqXqXqXqXqXqX9X0XqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUXfX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX.X+XUXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX(.9XUX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUXfX7XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XtX.X+XUXeX9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XrX/.9XUX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUXjXeXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXtXaXOX@XUXuXtXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXtXaX).9XUX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUX0X}.XX.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.XOX!.OXUX;X|.XX.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.XOXY.8XUX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XyXMX=X+X@X+X+X+X+X#X@X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X@XoX0XcX#X+X@X+X+X+X+X#X@X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X@XoXfXIX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXUXUXUXUXUXUXUXUXCXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXnXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXGX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXtXrXrXrXrXrXrXtX`.bXZX9XrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXtXrXrXrXrXeXtX8X`.KXlXqXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrX0XnXLX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX0X0X0X0X0X0XqX7XT.BXVX5XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XeX:X~.UXjX7XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX6XvXPX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX9XE.NXCX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrX,X^.UXkX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XvXPX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XvXPX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XvXPX9XqXqXqXqXqXqXqXqXqXqXqXqXqX9X0XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XvXPX9XqXqXqXqXqXqXqXqXqXqXqXqX0XiXeX0XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XvXPX9XqXqXqXqXqXqXqXqXqXqXqX0XyX[.1XyX9XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XvXPX9XqXqXqXqXqXqXqXqXqXqXqX7XvX~ Z wXtX9XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqXqX9X7X9XqX9X7X8X8X8X0XqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqXqX9X7X9XqXqX9X8X8XqX7XvXPX9XqXqXqXqXqX9X7X8X8X8X8X3XhX..E ' 6XtX9XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqX9XtXkXdX0XpXkXjXkXjXrX0XqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqX8XpXlXpX0X9XyXjXgXeX6XvXPX9XqXqXqXqX0XiXxXzXzXzXzXhXKXO.C.9.Q 7XtX9XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqX9XfX*Xz.Y.uX_.c.b.c.m.-XrX0XqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqX0XgX].j.].tXuX$XM.D.7XrXcXPX9XqXqXqX0XuX[.u.s.a.a.a.i.b.x v.%X,.Q 7XtX9XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqX9XuXJ.5   l cXQ         Y.aX8XqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqX0XtXy.@   X.NXL.2     g oXMXKX9XqXqXqX6XvX| G ,.#.$.$.$.@.3.!.^..X<.! 7XyX9XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqX0XqXrXQ T p M vXA D hXjXrX9XqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqX0XrX;Xn ' = *.jXu q 2X!.p j.CXKX9XqXqXqX7XcX+.n.>XOX+X+X+X@XXX(./.~..X<.Q 7XtX0XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXwX8X$X7Xi v vXw k [ F I.pX9XqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXwX1X:XOX. :.>X  9.;.G v.6XvXLX9XqXqXqX7XcXO.i.{.Q.!.!.!.!.!./././.~..X<.Q 7XtX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqX0XpX1Xr b zXK e ) q X W.aX9XqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqX0XgX_.. >._.  w ( g   C.CXKX9XqXqXqX7XcXO.a. X^./././././././././.~..X3.Q 0XwXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqX0XeX1Xt n gX2X<XvX:X  @.hX8XqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqX9XuX'.. >.|.  X.MX2X= R ZXKX9XqXqXqX7XcXO.a. X^./././././././././.~. X:._ qXwXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqX0XrX2Xi B yXU q.bX X  >.hX8XqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqX9XpX'.+ :.qX  Q bXOXX [ AXKX9XqXqXqX7XcXO.u.[.Q.!.!.!.!.!./././.~..X=.) qXrX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqX0XrX1X1 a xXI   x - 2 |.iX9XqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqX9XpX).  } NX&.  l 1 * ^.BXKX9XqXqXqX7XcXO.m.,X+X@X@X@X#XOX(./.~..X=._ qXrX0XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqX0XeX5X%.2.sX0Xp.I o..XuX9XqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqX0XyX#X( b.fXqXx.U } [.rXcXPX9XqXqXqX6XcX..C $.|  . . .} =.W.^. X*._ qXrX9XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqXqXwXkXhX0XqXgXvXzXtX0XqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqX0XrXzXsX9XqXdXvXzXuX6XvXPX9XqXqXqX0XyXXXz.c.c.c.c.l.F.b c.$X*.) qXrX9XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqXqXqX7X8XqXqX8X7X7X0XqXqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqXqX0X7X9XqXqX9X7X7XqX7XvXPX9XqXqXqXqX0XuXlXlXlXlXlXfXJXO.C.3.) qXrX9XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XvXPX9XqXqXqXqXqX0X8X8X8X8X8X4XjX..P ..0XrX9XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XvXPX9XqXqXqXqXqXqXqXqXqXqXqX7XvX/ J yXrX9XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XvXPX9XqXqXqXqXqXqXqXqXqXqXqX0XtXoX6XtX9XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XvXPX9XqXqXqXqXqXqXqXqXqXqXqXqX0XuXeX0XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XvXPX9XqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX8XE.NXCX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX>X~.UXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XvXPX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqX0XwX8XE.NXVX6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX>X~.UXjX7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XvXPX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXtXtXtXtXtXtXyXwXW.BXSX0XyXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXiX2X/.UXxXqXyXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXyXqXnXLX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXwX9X)./.(.(.(.(.(.(.K.bXaX~.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(./.(.!.E.HX3X!.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.).R.qXUX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX9X9X9X9X9X9X9X9X9XVXuX8X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9XrXCXqX9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X0X5XvXIX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XrXUXUXUXUXUXUXUXUXUXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXCX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX8X8X8X8X8X8X8X8X8X9X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X9X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X7X7X7X7X7X7X6X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X6X7X7X7X7X7X7X6X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXxXcXcXcXcXcXcXvXaX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuXvXcXcXcXcXcXcXcXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqX0XrX<X%.o.O.O.O.O.+.{ I.aX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XiX[.| +.O.O.O.O.O...XXuX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqX0XsX/.k n.u.a.a.p.c.' X.vX6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XxXu.G n.i.a.a.u.m.C z.lX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqX0XsX/.^ >X[. X X}.*Xh.&.cX7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XzXs.,.>X{. X X[.,X$.c.lX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqX0XsX/.T OXW.^.^.!.|.t.$.cX7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XzXa.#.OXQ.^.^.Q.+X| c.lX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqX0XsX/.T +XQ././.^..Xy.$.cX7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XzXa.$.+X!././.!.@X .c.lX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqX0XsX/.T +XQ././.^..Xy.$.cX7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XzXa.$.+X!././.!.@X .c.lX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqX7X5XuXQ.T +XQ././.^. Xy.@.kX2X6X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X7X3XhXi.$.+X!././.!.@X..l.fX4X7X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXwXxXlXCX+XY +XQ././.^..Xr.,.IXgXvXpX9XqXqXqXqXqXqXqXqXqXqXqXqX0XyXvXhXKXb.@.@X!././.!.#X} F.JXjXcXtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqX0XrX2X} } X.F ~ OX!././.^.}.l.d o...! L.sX9XqXqXqXqXqXqXqXqXqXqX9XiX{.^ ..O.x 3.XX!././.!.oX=.b O...( oXuX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqX0XtX..x M.x.Y.(./././././.^.m.B.O.f XXyX0XqXqXqXqXqXqXqXqXqXqX0XeX2XS T C.v.!.(././././.(.W.c.C.P H 5XeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqX0XwXgX6.| +X).^./././././.~.#Xx.R :XaX9XqXqXqXqXqXqXqXqXqXqXqXqX0XyXeX{ 7.$X^././././././.^.$X3...tXtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqX0X0XaX:.( [.(.^./././.!. Xa.G $XpX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XtX8X~ :..X~././././.~..X*.) 0XrX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqX0XqXsX>.) {./.^./.!. Xs.H %XpX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XtX9X^ >..X~././.~..X=._ qXrX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqX0XqXsX>.) {./.Q. Xs.H %XpX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XtX9X^ >..X~.~..X=._ qXrX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXsX>.) ].XXa.H %XpX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XtX9X^ >..X X=._ qXrX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXsX;.[ i.J $XpX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XtX8X~ 1.>._ qXrX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXaX:.Z *XpX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XtX8XQ _ wXrX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXrX8XyX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX0XqXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_2[] = {
/* columns rows colors chars-per-pixel */
"88 88 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX0X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X8X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XuXlXlXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXkXkXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXlXdX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXwX7XlXUXJXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXPXVXlXUXHXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXIXmXbXHX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXwX7XlXJX6X9X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8XeX{.5XKX4X9X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8XeX'.wXHX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXwX7XzXKX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXyX].8XPX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXyX_.tXJX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX_.rXJX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX_.rXJX6XwXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX_.rXJX6XwXqXqXqXqXqXqXqXqXqXqXwXrX0XqXqXqXqXqXqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX_.rXJX6XwXqXqXqXqXqXqXqXqXqXwX6X>XiX9XqXqXqXqXqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX_.rXJX6XwXqXqXqXqXqXqXqXqX0XuX#Xu `.dX8XqXqXqXqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXqXqX8X7X0XqX0X7X7X9XqXqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqX0X7X8XqX0X8X7X8XqXqXqXqXqXqXqX0XtX_.rXJX6XwXqXqXqX0X7X7X7X7X6XrX{.A ( W.fX8XqXqXqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXqXqXhXxXtX9XrXzXcXiX9XqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqX0XyXcXgX0XwXjXbXfXqXqXqXqXqXqXqX0XtX_.rXJX6XwXqXqX0XyXxXlXzXzXlXNX*X[ ~.H ^.fX8XqXqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqX0XeXrXi. .OXfX$X>.X./.pX0XqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqX9XaX X .p.pX7Xy.` x.wXwXqXqXqXqXqX0XtX_.rXJX6XwXqX9XiX|.X.%.$.$.$.-.~ / &XD.K ^.fX8XqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXwXE     T.6X9 1 i O U.pX9XqXqXqXqX0XyX].7XPX7XwXqXqXqXqX9XiXT.<   j hXS O m   -.aX9XqXqXqXqX0XtX_.rXJX6XwXqX8XlXk.E S.k.z.z.z.x.T.^.[.H.L ^.fX8XqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqX0XrX=X3.a.X }.e.  XXjX0 V pX0XqXqXqXqX0XyX].7XPX7XwXqXqXqXqX0XpXm.s.G m eXc m.FX~ 1 sX0XqXqXqXqX0XtX_.rXJX6XwXqX8XkXc.2.-X].{.{.{.{.)./.~.[.H.L !.gX9XqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqX8XxXH.  './ r wXsX} ; eXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqX0XrXjXB x fXeXaX5Xq E dX9XqXqXqXqX0XtX_.rXJX6XwXqX8XkXx.=.oX!.^.^.^.^./././.~.[.F.R !.uX0XqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqX9XaXC.  _.~ y wXiX..= eXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqX0XeXN c sXeX7Xx 2 &XtX0XqXqXqXqX0XtX_.rXJX6XwXqX8XkXx.-.OX!././././././././.W.%X,.) fX9XqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqX9XsXS.   X1.X :XbXj k iX0XqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXtXC v jX3XB d 0XMX0XqXqXqXqXqX0XtX_.rXJX6XwXqX8XkXx.*..XW.!.!.!.!.^././.!..Xs.~ :XrX0XqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqX8XdXB.  Q.&X+ h I   v.aX9XqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXyXk a fXd   H X.#.4XwXqXqXqXqX0XtX_.rXJX6XwXqX8XkXx.6.2X.XOXOXOXOX].^.~..Xd.E :XiX9XqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqX9XpX!.A '.jX~.P M n.yXqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXrX%.#.9XD F b a H 1XeX0XqXqXqX0XtX_.rXJX6XwXqX8XkXc.M 8.-.;.:.:.=.N.).|.d.E >XiX8XqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXqX0XyXkXtX9XuXlXxXaX0XqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXqXfXfXqXlXkXzXcXlXwXqXqXqXqXqX0XtX_.rXJX6XwXqX0XrX;Xz.k.l.l.k.n.1.T :Xp.R >XiX8XqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXqXqX0X7X0XqX9X8X8X9XqXqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXqX8X8XqX8X7X7X7X8XqXqXqXqXqXqX0XtX_.rXJX6XwXqXqX0XrXkXlXlXlXkXMX*XX.M.Y >XiX8XqXqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX_.rXJX6XwXqXqXqX0X8X8X8X8X7XtX Xa { -XiX8XqXqXqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX_.rXJX6XwXqXqXqXqXqXqXqXqX0XyX%XL ,XiX8XqXqXqXqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX_.rXJX6XwXqXqXqXqXqXqXqXqXqXqX0XiXrX9XqXqXqXqXqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX_.rXJX6XwXqXqXqXqXqXqXqXqXqXqXqX0X0XqXqXqXqXqXqXqXqXqXqX",
"qXwX7XzXKX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX].7XPX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX_.rXJX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXwX7XzXKX8XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XtX].7XLX6XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XtX_.rXJX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXwX7XzXPXeXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXtXsX}.8XUXqXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXtXaX'.tXHX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXwX7XzXBX{.].].].].].].[.].].].].].].].].].].].].].].].].].}.I.5XVX'.].].].].].].[.].].].].].].].].].].].].].].].].].}.P.eXJX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXwX7XlXxX5X8X7X7X7X8X6X4X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X6XvXkX5X8X7X7X7X8X6X5X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X6XNXSX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXwX9XfXLXLXPXPXPXLXUXdXUXUXLXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXHXHXPXPXPXPXLXIXdXUXUXLXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXKXUXCX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX0X7X7X7X7X7X8X2X~.HXaX4X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X7X7X7X7X6X9X,X/.LXyX4X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X2XnXDX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXwXwXwXwXqXeX5X^.KXgX9XwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXrX1X).IXaX9XwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwX7XMXDX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X/.KXgX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXDX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXDX7XwXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXDX7XwXqXqXqXqXqXqXqXqXqXqXeXyX9XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXDX7XwXqXqXqXqXqXqXqXqXqXeX4X$XpX9XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXDX7XwXqXqXqXqXqXqXqXqX0XuX@X0 Q.fX8XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqXqXqX8X7X0X0X7X7X7X7X0XqXqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqXqXqX8X7X0XqX0X7X7X0X7XMXDX7XwXqXqXqX0X7X7X7X7X6XtX{.J ) K.gX8XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqXqXwXhXcXtXrXzXvXbXcXtX0XqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqX0XwXjXxXtX9XtXzXcXaX6XMXDX7XwXqXqX0XyXlXjXkXkXjXmX%X' `.J I.gX8XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqX0XeXwX7.Y oX@X) K F E +XyX0XqXqXqXqXeX1X).IXaX9XqXqXqXqXqX0XrX8X<.Q oXfXXX} R I.tXmXDX7XwXqX9XpX].( ..| | | O.Y ` $XL.L U.gX8XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqX0XeXT X   R.J.  F ! W oXyX0XqXqXqXqXeX1X).IXaX9XqXqXqXqX0XqX9XD .   ~.2X1 w C   b.FXZX7XwXqX8XlXk.( U.M.B.B.B.V.Q.^.].U.P I.gX8XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqX0XrX>Xt.c.X /.x.9 %X-XkXpX9XqXqXqXqXqXeX1X).IXaX9XqXqXqXqX0XtX&Xy.g.  +X3.: oX[.q.R.VXSX7XwXqX8XkXc.<.&X_.].].].].(./.~.].U.P P.hX9XqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqX9XxXY.  ).u.  a   A 2XeX0XqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqX9XvXC.  }.! , K + s  XZXAX7XwXqX8XkXx.=.oX!./././.^./././.~.].I.E P.uX0XqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqX9XpXJ.  P.8XV.*XE.  d.hX8XqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqX9XaXm.  '.~   E.=X& ! GXAX7XwXqX8XkXx.-.OX!././././././././.W.$X:.] fX9XqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqX9XsXL.  T.I.i.mXeXX 6.jX8XqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqX9XdXB.  +X@.& rXmXi J FXAX7XwXqX8XkXx.*..XE.!.!.!.!.~././.!..Xe.) 3XeX0XqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqX9XdXF.  R.M.  J f = |.iX9XqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqX8XfXv.  ).OX+ z A   C.FXAX7XwXqX8XlXx.6.3XXX+X+XOX+X[.^.~..Xr.^ 4XuX0XqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqX0XuX].E `.hXV.K / `.uX9XqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqX9XiX~.E .XhX/.! I K.rXmXSX7XwXqX8XkXb.v >.O.@.@.#.X.c._. Xr.^ 4XyX9XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqXqX0XtXzXyX9XsXcXxXuX0XqXqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqXqX0XyXzXtX9XuXxXcXaX6XMXDX7XwXqX0XeX<XA.V.C.C.V.H.w.Y >Xq.~ 5XyX9XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqXqXqX0X7X0XqX9X7X7X0XqXqXqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqXqXqX0X7X0XqX9X7X7X0X7XMXDX7XwXqXqX0XrXhXjXjXjXhXnX%XX.l.Q 4XyX9XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXDX7XwXqXqXqX0X8X8X8X8X7XtXXXr X.1XyX9XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXDX7XwXqXqXqXqXqXqXqXqX0XyX&X^ 4XyX9XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXDX7XwXqXqXqXqXqXqXqXqXqXqXqXsXeX0XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXDX7XwXqXqXqXqXqXqXqXqXqXqXqX9X0XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXeX4X^.KXgX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX1X).IXaX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXDX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX0X0X0X0X0XwX4X^.KXfX8XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XeX1X).IXpX8XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX6XmXDX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXeXyXtXtXtXtXuX8X(.PXkXwXyXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXiX5X_.UXgXeXyXtXtXtXtXtXtXtXtXtXtXtXtXtXtXyXqXBXSX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qX0XrX:X/.`._._._.`.(.R.SX>X/.`._._._._._._._._._._._._._._._._._._._._._._.`.^.!.FX=X/.`._._._._._._._._._._._._._._.'.Q.iXGX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXwXrXrXrXrXrXrXrXpXZXyXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXsXCXtXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXtX0XVXFX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXwX8XhXPXHXJXJXJXJXJXHXSXHXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXGXSXHXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXHXJXGXxX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX9X6X6X6X6X6X6X6X6X7X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X7X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X7X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqX9X8X8X8X8X8X8X8X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X8X8X8X8X8X8X8X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqX0XiXlXkXkXkXkXkXkXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXlXkXkXkXkXlXkXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqX0XyX|.k.c.x.x.x.x.c.;XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX].k.c.x.x.x.x.b.<XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqX7XxXX.E 2.=.-.*.6.M z.kX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XlX( ( <.=.-.*.6.v A.hX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqX7XlX%.S.-XoXOX.X2X8.k.lX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XjX..U.&XoXOX.X3X>.V.jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqX7XzX$.k.].!.!.W..X-.l.lX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XkX| M._.!.!.E.XXO.C.jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqX7XzX$.z.{.^./.!.OX;.l.lX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XkX| B.]././.!.+X@.C.jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqX0X6XlX$.z.{.^./.!.OX:.k.kX7X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X6XjX| B.]././.!.+X@.V.hX7X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwXuXrXNX-.z.{.^./.!.OX:.n.MXtXyXqXqXqXqXqXqXqXqXqXqXqXqXqXeXuXtXmXO.B.]././.!.+X#.H.nXtXyXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXwX6X#X}.*X~ x.{.^./.!.OX=.1.*X X%X0XqXqXqXqXqXqXqXqXqXqXqXeX4X@X{.%XY V.].^./.!.+Xo.w.%XXX&XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX0XrX,Xi Z { / T.)././.^.].N.T X.s L iX0XqXqXqXqXqXqXqXqXqX0XyX%Xq H ] ` Q.(././.~.[.c.Y X.t ^ sX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqX0XiX].( Q.&X^././././.^.(.:XM.{ >XrX0XqXqXqXqXqXqXqXqXqXqX9XpX!.) ).$X~././././.^._.>Xl.X.3XeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqX9XdX~.H A.[.~./././.~.|.p.Y -XiX9XqXqXqXqXqXqXqXqXqXqXqXqX9XfXI.H K.].~./././.~. Xq.Q 1XyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqX8XfX(.L F.{.~./.!. Xd.E >XiX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XgXT.L I.].~./.!..Xr.^ 4XyX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqX8XfX(.L F.{.W..Xd.E >XiX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XgXT.L I.].W..Xr.^ 4XyX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqX8XfX(.L D.%Xs.E >XiX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XgXT.P L.$Xe.^ 4XyX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqX8XfX/.T >.^ :XiX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XgXU.R ;._ 4XyX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XfX~.) :XiX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XgXI.' 4XuX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XyXfXrX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XuXfXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_3[] = {
/* columns rows colors chars-per-pixel */
"64 64 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX9X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X9X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXeXsXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXsXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXgXaXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX9XsXPXBXCXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXZXdXGXZXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXAXhXVXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX8XfXBX4X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X7X|.BX8X7X8X8X8X8X8X8X8X8X8X8X8X8X8X8X7XqX{.xXaX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX8XfXCX8XwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXoXCXwXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXtX.XcXpX9XqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqX",
"qXqX8XfXVX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XXXVXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX.XcXaX9XqXqXqXqXqXqXqXtX0XqXqXqXqXqXqXqXqX",
"qXqX8XfXVX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XXXVXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX.XcXaX9XqXqXqXqXqXqXwX;XyX0XqXqXqXqXqXqXqX",
"qXqX8XfXVX8XwXqXqXqX0X7X9XqX8X7X9XqXqXqXqXqX0XXXVXwXqXqXqXqXqX9X7X0X0X7X8X0XqXqXqX0XrX.XcXaX9XqXqX9X9X9X7XpX' l.hX9XqXqXqXqXqXqX",
"qXqX8XfXVX8XwXqXqX0XyXzXaX0XhXxXpX0XqXqXqXqX0XXXVXwXqXqXqXqX0XsXlXeXiXzXjXeXqXqXqX0XrX.XcXaX9XqXqXaXpXiXuXfX4.;.n.gX9XqXqXqXqXqX",
"qXqX8XfXVX8XwXqX0XtX&X/ d.gX8.K Y.pX9XqXqXqX0XXXVXwXqXqXqX0XuXF.P +X(.L ;.6XwX0XqX0XrX.XcXaX9X0XyXi.<.7.6.2.$..X<.b.gX9XqXqXqXqX",
"qXqX8XfXVX8XwXqX9XiXi.< X.m.u i.8 %XyX0XqXqX0XXXVXwXqXqX0XrX>X_ $ J.H r.M 1.zX7XqX0XrX.XcXaX9XqXwX&.Y.T.T.Y././.}.1.v.gX0XqXqXqX",
"qXqX8XfXVX8XwXqXqX7XgX2.O.` 2.HXg n.jX8XqXqX0XXXVXwXqXqXqXqX8XyXn A.tXlXv d.jX8XqX0XrX.XcXaX9XqXqX3.(.).).)./.^.^.{.6.v.iX0XqXqX",
"qXqX8XfXVX8XwXqXqX8XlX&.@.{ &.FXy C.hX9XqXqX0XXXVXwXqXqXqX0XwXeXj K.nXR >.vX8XqXqX0XrX.XcXaX9XqXwX1.~.^.^.^.~./.!..Xp.9.pX9XqXqX",
"qXqX8XfXVX8XwXqXqX8XlXX.~ Q.9 ^ f 2XeX0XqXqX0XXXVXwXqXqXqXqXqXeXq J.I   ;.D.uX0XqX0XrX.XcXaX9XqXqX>.`.'.'.'.[.~. Xi.0.iXqXqXqXqX",
"qXqX8XfXVX8XwXqXqX0XiXW.F.jXH.<.#XuX0XqXqXqX0XXXVXwXqXqXqXqXqXwXv.oXf.q.$.D.iX0XqX0XrX.XcXaX9X0XyX>.=.<.<.;.1.oXi.0.iXwX0XqXqXqX",
"qXqX8XfXVX8XwXqXqXqX0XpXsX8XdXxXtX0XqXqXqXqX0XXXVXwXqXqXqXqXqXqXgXyXjXlXvXdX0XqXqX0XrX.XcXaX9XqXwXqX2X3X2X5X3.t.0.iXwX0XqXqXqXqX",
"qXqX8XfXVX8XwXqXqXqXqX9X9XqX9X7X0XqXqXqXqXqX0XXXVXwXqXqXqXqXqXqX8X0X8X7X7X9XqXqXqX0XrX.XcXaX9XqXqXqXeXeXqXfX) 6.pXwX0XqXqXqXqXqX",
"qXqX8XfXVX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XXXVXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX.XcXaX9XqXqXqXqXqX0XrX/.0XeX0XqXqXqXqXqXqX",
"qXqX8XfXVX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XXXVXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX.XcXaX9XqXqXqXqXqXqX0XpXqX0XqXqXqXqXqXqXqX",
"qXqX8XfXVX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XoXVXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX.XcXaX9XqXqXqXqXqXqXqX9XqXqXqXqXqXqXqXqXqX",
"qXqX8XfXZX7XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9X.XZXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XrX|.cXaX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX8XgXfX|.oXXXoXoX.XoXXXXXXXXXXXXXXXXXXXoX.X+XxXoXXXXXXXoX.XXXXXXXXXXXXXXXXXXXXXXXoX[.mXpX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX9XpXZXVXVXCXBXkXDXVXCXVXVXVXVXVXVXVXVXVXVXAXZXVXCXVXCXjXAXCXVXVXVXVXVXVXVXVXVXVXVXCXLXtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXwXwXqXrX<X<XCX7XeXwXwXwXwXwXwXwXwXwXwXqXqXwXwXeX8X#XAX0XwXwXwXwXwXwXwXwXwXwXwX0XSXuX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX0XrX>X,XBX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX6X@XZX9XqXqXqXqXqXqXqXqXqXqXqX9XSXuX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXrX>X,XVX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7X@XZX9XqXqXqXqXqXqXqXqXqXqXqX0XSXuX0XqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXrX>X,XVX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7X@XZX9XqXqXqXqXqXqXqXqXqXqXqX0XSXuX0XqXqXqXqXqXqX0XwXuX9XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXrX>X,XVX6XwXqXqXqXqX9X8XqX8X7X8XqXqXqXqXwX7X@XZX9XqXqXqXqXqX9X7XqX0X8X8X9XSXuX0XqXqX9X8X8X6XuX#.U.hX8XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXrX>X,XVX6XwXqXqXqX0XdXjXwXjXvXzXeXqXqXqXwX7X@XZX9XqXqXqXqX0XiXlXwXwXzXjX0XSXuX0XqX0XdXhXgXfXcX>.o.!.gX8XqXqXqXqXqX",
"qXqXqXqXqXqXqXrX>X,XVX6XwXqXqX0XuXG.d.4Xi.%.6.3XeX0XqXwX7X@XZX9XqXqXqX0XuX`.9.3X6X7.h.8XDXuX0X0XtXA.p.h.g.d.@._.#.Q.gX8XqXqXqXqX",
"qXqXqXqXqXqXqXrX>X,XVX6XwXqX9XpX!.w e (.0 3.a.3XeXqXqXwX7X@XZX9XqXqX0XtX*XV    X/ R H k.UXtX0XqXeXO.n.N.N.m.L.`.(.$.E.gX9XqXqXqX",
"qXqXqXqXqXqXqXrX>X,XVX6XwXqXqXwX4XI.M I.t ] a.uX0XqXqXwX7X@XZX9XqXqXqXwX3X+Xc C.u &.U ~.KXtX0XqXqX4.].[.].[._.~._./.=.W.uX0XqXqX",
"qXqXqXqXqXqXqXrX>X,XVX6XwXqXqX9XfX^.f %XF. X7 P.dX9XqXwX7X@XZX9XqXqXqX0XyX4Xj m.< ^...R UXrX0XqXwX<.Q.!.!.!.~.(.Q.XXz.=.aX9XqXqX",
"qXqXqXqXqXqXqXrX>X,XVX6XwXqXqX9XaXT.r E.T S.w /.aX9XqXwX7X@XZX9XqXqXqX0XrX<X6 !.l p.R | UXrX0XqXqX3.|.|.|. X{.!.}.N.-.5XeX0XqXqX",
"qXqXqXqXqXqXqXrX>X,XVX6XwXqXqX0XuXXX%.7Xq.F m.iX0XqXqXwX7X@XZX9XqXqXqX0XeX5X@.-X|.U #.3XFXyX0X0XtX+.1.7.6.3.i..XN.=.5XyX0XqXqXqX",
"qXqXqXqXqXqXqXrX>X,XVX6XwXqXqXqX0XrXgXwXhXzXsX0XqXqXqXwX7X@XZX9XqXqXqXqXqXwXgXeXuXzXkXwXAXuX0XqXeX*X].{.[.{.:.M.=.5XyX9XqXqXqXqX",
"qXqXqXqXqXqXqXrX>X,XVX6XwXqXqXqXqX0X8XqX8X7X9XqXqXqXqXwX7X@XZX9XqXqXqXqXqXqX8XqX0X8X8X9XSXuX0XqXqXtXpXpXuXlX} $.6XyX9XqXqXqXqXqX",
"qXqXqXqXqXqXqXrX>X,XVX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7X@XZX9XqXqXqXqXqXqXqXqXqXqXqX0XSXuX0XqXqX0X0X0X9XtXn.:XiX9XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXrX>X,XVX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7X@XZX9XqXqXqXqXqXqXqXqXqXqXqX0XSXuX0XqXqXqXqXqXqX0XsXrX0XqXqXqXqXqXqXqX",
"qXqXqXqX0X0X0XeX>X,XBX6XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX6X@XCX8XqX0X0X0X0X0X0X0X0X0XqX9XSXuX0XqXqXqXqXqXqXqX9X0XqXqXqXqXqXqXqXqX",
"qXqXqXwXtXrXrXuX1X<XAX0XtXrXrXrXrXrXrXrXrXrXrXrXrXrXrXtX0X@XSXeXtXrXrXrXrXrXrXrXrXrXtXeXSXuX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXwX5X X.X.XXX_.>XdX[.XX.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X{.XXzX|..X.X.X.X.X.X.X.X.X.X.X}.NXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XtXvXcXcXcXcXVXNXxXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXNXVXxXcXcXcXcXcXcXcXcXcXcXcXcXGXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXeXaXpXaXaXaXiXiXaXpXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXiXiXaXpXaXaXaXaXaXaXaXaXaXpXaXyXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXyXwXqXwXqXyXwXqXqXqXqXqXqXqXqXqXqXqXqX0XtXeXqXwXqXtXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqX9XaXi.&.3.1.>.>.qXqXqXqXqXqXqXqXqXqXqXqX9XdXA.O.4.<.3.+.*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqX9XpX<.Y.(.~.`.=.2XeXqXqXqXqXqXqXqXqXqXqX8XhXp.n.].Q.|.1.].pX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqX9XiX7.T.).^.'.<.3XeXqXqXqXqXqXqXqXqXqXqX8XgXh.N.[.!.|.7.{.pX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqX7XuX6.T.).^.'.<.2XqX0XqXqXqXqXqXqXqXqXqX6XfXg.N.].!.|.6.[.uX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXwXpXfX2.Y.).^.'.;.5XfXrX0XqXqXqXqXqXqXqX0XuXcXd.m.[.!. X3.{.lXtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX0XtX;X' 4.%././.~.[.1.3.) /.pX9XqXqXqXqXqXqXwX$.>.@.L._.~.{.i.:.} m.sX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX0XyXx.;. X/.^./.~.oXt.6.0XwXqXqXqXqXqXqX0XuXY.o.).`.~.(.!..XM.#.:XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX0XhXm.<.}./.!. Xi.0.pXeX0XqXqXqXqXqXqXqX9XhX~.#./._.Q.}.N.=.6XiX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqX9XgXn.<.{..Xi.0.iXwX0XqXqXqXqXqXqXqXqXqX8XgX!.#.^.XXN.*.5XyX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqX9XgXb.5.p.0.iXwX0XqXqXqXqXqXqXqXqXqXqXqX8XfXQ.=.z.;.5XyX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqX9XgXb.9.pXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XgXQ.*.6XyX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqX0XiXpXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XuXaXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqX0X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_4[] = {
/* columns rows colors chars-per-pixel */
"48 48 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX0X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXuXpXpXpXpXpXpXpXpXpXpXpXpXaXiXiXpXpXpXpXpXpXpXpXpXpXpXpXpXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XuXAXcXbXvXvXvXvXvXvXvXvXvXnXhXxXNXcXvXvXvXvXvXvXvXvXvXvXvXgXzX9XwXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX9XpXcX5X9X8X8X8X8X8X8X8X8X8XqX=XrXiX6X8X8X8X8X8X8X8X8X8X8X7X*XzX9XwXqXqXqXqX9XqXqXqXqXqXqXqX",
"qXqX9XpXbX9XwXwXwXwXwXwXwXwXwXqXtX>XyXsX0XwXwXwXwXwXwXwXwXwXwXqX:XzX9XwXqXqXqXwXpX0XqXqXqXqXqXqX",
"qXqX9XpXvX8XwXqX0X8X0X9X8X0XqXqXrX:XyXaX9XqXqX0X8X0X8X9XqXqXqX0X:XzX9XwX0X0XqX7XQ.yXqXqXqXqXqXqX",
"qXqX9XpXvX8XwX0XtXjXrXsXhXtX0XqXrX:XyXaX9XqX0XpXgXuXhXdXwXqXqX0X:XzX9XqXyXtXyX8X .S.aX0XqXqXqXqX",
"qXqX9XpXvX8XqXtX&X` |.j.{ -XtX0XrX:XyXaX9X0XiX!.| {...r.qXqXqX0X:XzX7XpXP.y.z.u.n.c.N.aXqXqXqXqX",
"qXqX9XpXvX8X0XrX#XH =.O.t.@.kX7XrX:XyXaX9X0XtXW.N b.I.Z {.iX0X0X:XzX7XsXx.G.^.!.(.{.k.A.rXqXqXqX",
"qXqX9XpXvX8XwX7XxX9._ 6.(./ kX7XrX:XyXaX9XqX9XhX! ;Xn.9.aX9XqX0X:XzX7XaXB.E.|.[./.(.`.,.<XrX0XqX",
"qXqX9XpXvX8XwX8XgX>.N.$.A U.aX9XrX:XyXaX9XqX0XeX+.h.6 +.*XtX0X0X:XzX6XdXk.h.A.N.W.).0.XXuX0XqXqX",
"qXqX9XpXvX8XwXqXwX4XwX9X:XiX0XqXrX:XyXaX9XqXqXqX5X2X:XOX6XwXqX0X:XzX8XyXoXH.Y.A.w.u.}.aX9XqXqXqX",
"qXqX9XpXvX8XwXqXqXeXqXwXtX9XqXqXrX:XyXaX9XqXqXqXeXeXtXuXwXqXqX0X:XzX9XqXyXdXfXyX%..XaX8XqXqXqXqX",
"qXqX9XpXvX8XqXqXqX0X0X0X0XqXqX0XrX:XyXaX8XqXqXqX0X0X0X0X0XqXqX0X:XzX9XwX0X9X9X7X3XuX9XqXqXqXqXqX",
"qXqX9XpXnXqXtXrXrXrXrXrXrXrXrXrXuX<XuXfXwXrXrXrXrXrXrXrXrXrXrXeX>XzX9XwXqXqXqXqXeX0XqXqXqXqXqXqX",
"qXqX9XaXjX=X>X>X>X:X>X:X:X:X:X:X<XoXeX9X-X>X>X:X>X:X:X:X:X:X>X;X$XzX9XwXqXqXqXqX0XqXqXqXqXqXqXqX",
"qXqX0XuXzXrXuXwXrXuXyXyXyXyXyXyXyXtXcXpXtXyXqXuXyXyXyXyXyXyXyXrXhXvX8XwXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXiXaXfX4XvXdXaXaXaXaXaXaXaXsXuXpXdXrXrXnXuXsXaXaXaXaXsXpXkXvX8XwXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX9X8XwX&XgXqX9X9X9X9X9X9X9X9X9X9XqX1X4XgX6X9X9X9X9X9X9X7XiXvX8XwXqXqXqXqXqX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXrX=XjXeXqXqXqX0X0X0X9X0XqXqXeX4X6XjX8XwXqXqX0XqXqX8XaXvX8XwX0X9X0X8XwXtX9XqXqXqXqXqX",
"qXqXqXqXqX0XrX=XhXeXqXqX0XtXuXrXfXpX0XqXeX3X6XjX8XwXqXqXpXwXrXpXpXvX8XqXtXaXsXtX<.*XpX9XqXqXqXqX",
"qXqXqXqXqX0XrX=XhXeXqX0XrX;X.X,XP.`.eX0XeX3X6XjX8XqXqXeX[.6X>X!.pXvX7XrX=X~.`.T.4.u.&XiX9XqXqXqX",
"qXqXqXqXqX0XrX=XhXeXqX0XuXX.U ,.B G.yX9XeX3X6XjX8X0XuX[.r b.} D '.CX4XdXz.u.n.k.T.!.9.-XuX0XqXqX",
"qXqXqXqXqX0XrX=XhXeXqXqXtX}.} j.*.w.tX0XeX3X6XjX8XqX9XiX6. .T $.A.SX4XaXN.E.|.[.(.).^.4.2XeXqXqX",
"qXqXqXqXqX0XrX=XhXeXqX0XiXoXY B.u.F 9XqXwX3X6XjX8XwX7XlX&.*.[ +.8.JX3XsXv.P.`.).).[.M.x.qXqXqXqX",
"qXqXqXqXqX0XrX=XhXeXqXqXrX>XS.[.;.`.uX0XeX3X6XjX8XqX0XiXK.OXP.5.eXnX5XaXA.q.h.r.C.V.j.uXeX0XqXqX",
"qXqXqXqXqX0XrX=XhXeXqXqXqXeXaXuXzXiX0XqXeX3X6XjX8XqXqX0XaXtXsXhXsXcX8XwXeX6X9X,Xo.b.yXwX0XqXqXqX",
"qXqXqXqXqX0XrX=XhXeXqXqXqXqX9X0X7X9XqXqXeX3X6XjX8XqXqXqX9X0X9X6XpXvX8XwXqXwXeX8XA.eXwX0XqXqXqXqX",
"qXqXqXqXqX0XrX=XhXeXqXqXqXqXqXqXqXqXqXqXeX4X6XjX8XwXqXqXqXqXqX9XaXvX8XwXqXqXqXqXaXqX0XqXqXqXqXqX",
"qXqXqXqX0X9XeX&XhXwX0X0X0X0X0X0X0X0X0X0XwX2X4XjX7XqX0X0X0X0X0X8XpXvX8XwXqXqXqXqX9XqXqXqXqXqXqXqX",
"qXqXqX9X:X:X:X*XgX>X:X:X:X:X:X:X:X:X:X:X>X%XqXrX*X:X:X:X:X:X:X=X8XbX8XwXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XeXzXzXzXxXxXzXzXzXzXzXzXzXzXzXzXzXzXzXcXzXzXzXzXzXzXzXzXzXcXdX9XqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX9X9X9X8X9X7X7X7X6X8X9X9X9X9X9X9X9X9X8X8X6X7X7X7X9X9X9X9X8X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXwXwXwXwXqXpXsXaXdXyXqXwXwXwXwXwXwXwXqXrXdXaXsXaXqXwXwXwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX0XyXP.x.B.k.oXyX0XqXqXqXqXqX0XtX=Xz.N.v.A.eXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX0XtXy.G.E.h.H.dX9XqXqXqXqXqX9XaX~.u.E.P.q.6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXyXz.^.|.A.Y.fX9XqXqXqXqXqX0XsX`.n.|.`.h.9XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXwX7X8Xu.!.[.N.A.yX7XqXqXqXqXqX8XtXT.k.[.).r.,X8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX9XpXQ.| n.(./.W.w.%.3XeX0XqXqXqXwX1.3.T.(.).C.o.A.aX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX0XyXD.x.{.(.).i..XuX0XqXqXqX0XtX=Xu.!._.[.V.b.eXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXaXB.j.`.0.}.aX9XqXqXqXqXqX9XiX*X9.~.M.j.yXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX0XsXS.,.XXaX8XqXqXqXqXqXqXqX9XiX;X4.x.uXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXrX<XuX9XqXqXqXqXqXqXqXqXqX9XuX2XqXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXrX0XqXqXqXqXqXqXqXqXqXqXqX0XeXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_5[] = {
/* columns rows colors chars-per-pixel */
"44 44 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX9X8X8X8X8X8X8X8X8X8X8X8X9X9X8X8X8X8X8X8X8X8X8X8X8X8X0XqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXeXdXfXfXfXfXfXfXfXfXfXdXfXsXsXfXfXfXfXfXfXfXfXfXfXfXdXrX0XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX9XdXBXhXkXkXkXkXkXkXkXkXkXjXuXMXhXkXkXkXkXkXkXkXkXkXjXiXjX9XwXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX8XfXhX5X9X8X8X8X8X8X8X8X8X6X>XgX5X9X8X8X8X8X8X8X8X8X7X;XjX9XwXqXqXqXqX0XqXqXqXqXqXqX",
"qXqX8XfXkX9XwXwX0X0X0X9XqXwXwX0X1XjX9XwXwXqX9XqX9XqXwXwXqX<XjX9XqX9X9X9XqXyX9XqXqXqXqXqX",
"qXqX8XfXkX8XwX0XsXuXyXhXqXqXqX9X1XjX8XwX0XuXsXyXjXeXqXqX0X<XjX9XwXaXpXzXd./.dX8XqXqXqXqX",
"qXqX8XfXkX8XqXtXT. XOXN.eXqXqX9X1XjX8XqXtXXXY.#Xc.3XeXqX0X<XjX9XqXE.L.U.r.i.].sX0XqXqXqX",
"qXqX8XfXkX8XwXrX_ o.-.' k.gX9X9X1XjX8X0XaXf.F B.{ <.kX8X0X<XjXqX1Xw.G.C.T.`.p.{.tX0XqXqX",
"qXqX8XfXkX8X9XfXE.Q 3.W.+.lX8X9X1XjX8XqXeX3X] wX3.!.pX0X0X<XjXwX<Xn.+X|.`._.~.7.0XqXqXqX",
"qXqX8XfXkX8X0XdXJ.w.q.K ~.pX9X9X1XjX8XqXrX;X<.,.s Y.yX0X0X<XjXqX2X0.Z.m.I.).u.+XtX0XqXqX",
"qXqX8XfXkX8XwXqX8X0XwX<XuX0XqX9X1XjX8XwXqX0X7X2X-X;XeXqX0X<XjX9XwX_.W.^.q.y.oXaX9XqXqXqX",
"qXqX8XfXkX8XwXqXwXqXqXrX0XqXqX9X1XjX8XwXqXqXwXeXtXrXqXqX0X<XjX9XwXpXpXlXl.{.sX8XqXqXqXqX",
"qXqX8XdXkX8XwXqXqXqXqXqXqXqXqX0X1XjX9XwXqXqXqXqX0XqXqXqX0X<XjX9XwX9X9X8XrXyX9XqXqXqXqXqX",
"qXqX8XfXjX6X0X0X9X9X9X9X9X9X0X7X>XhX7X0X0X9X9X9X9X9X9X9X8X:XjX9XwXqXqXqX0X0XqXqXqXqXqXqX",
"qXqX9XdXpX:X2X,X<X1X1X1X1X1X1X>X7XaX:X2X,X<X1X1X1X1X1X1X>X6XlX8XwXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XrXjXkXdXsXbXgXjXjXjXjXjXjXkXhXkXfXaXnXgXjXjXjXjXjXgXnXzX8XwXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX0X8X0X1X6XsX6X9X8X8X8X8X8X8X8X0X2X5XdX5X9X8X8X8X9X6XiXzX8XwXqXqXqXqX0XqXqXqXqXqXqX",
"qXqXqXqXqXeX5X8XfX9XwXwX0X0X0X9XqXwXeX6X7XgX9XwXwX0X0XqX7XsXzX8XwX9X9X8XwXyX9XqXqXqXqXqX",
"qXqXqXqXqXeX4X8XfX8XwX0XpXiXpXhXrX0XeX5X7XgX8XwX0XpXiXrXsXdXlX8XwXpXpXlXg._.dX8XqXqXqXqX",
"qXqXqXqXqXeX4X8XfX8XqXtX_.].`.N.:XrXwX5X7XgX8XqXyX`.].1XS.wXxX8XqX~.Y.E.e.u.}.aX0XqXqXqX",
"qXqXqXqXqXeX4X8XfX8XqXuX .| R  .1XeXwX5X7XgX8XqXiX+.[ ;.m _.NX9X1Xq.D.B.Y.`.i. XtX0XqXqX",
"qXqXqXqXqXeX4X8XfX8X0XsX/.o.F.@.C.fX0X5X7XgX8X0XpX[.` Q >.d.AX8X<Xn.+X|.`._.~.7.0XqXqXqX",
"qXqXqXqXqXeX4X8XfX8X0XsXQ.>.j.^ W.aXqX5X7XgX8X0XaX/.<.p.W J.VX8X2Xq.S.N.U._.i..XtX0XqXqX",
"qXqXqXqXqXeX4X8XfX8XqXwX8X7X1X=XiX0XeX5X7XgX8XqXwX8X6XwX@XkXlX8XqX^.T.W.w.u.|.aX9XqXqXqX",
"qXqXqXqXqXeX4X8XfX8XqXqXwXwXeXtX0XqXwX5X7XgX8XqXqXwXwXqXeXpXzX8XwXpXpXlXh.`.dX8XqXqXqXqX",
"qXqXqXqXqXeX4X8XfX8XqXqXqXqXqX0XqXqXwX5X7XgX8XqXqXqXqXqX8XaXzX8XwX9X9X8XeXyX9XqXqXqXqXqX",
"qXqXqXqX0XwX3X6XdX7X0X0X0X0X0X0X0X0XqX4X5XfX7X0X0X0X0X0X8XpXlX8XwXqXqXqXqX0XqXqXqXqXqXqX",
"qXqXwX9X,X1X-XeXwX:X<X<X<X<X<X<X<X<X1X-XwXrX:X<X<X<X<X<X:XqXxX8XwXqXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XrXkXjXjXlXjXjXjXjXjXjXjXjXjXjXjXjXlXjXjXjXjXjXjXjXjXlXiX0XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX9X9X9X8X9XqXwXqX9X9X9X9X9X9X9X9X8X9XqXwXqX9X9X9X9X8X0XqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXwXwXqXwXqX1X<X2XwXwXwXwXwXwXwXwXwXqX1X<X1XqXwXwXwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX9XaXE.w.n.0._.pX9XqXqXqXqX9XpX~.q.n.q.^.pX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX9XpXL.G.+XZ.W.pX9XqXqXqXqX9XpXY.D.+XS.T.pX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX9XzXU.C.|.m.^.lX8XqXqXqXqX8XlXE.N.|.N.W.lX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXd.r.T.`.I.q.l.rX0XqXqXqXwXg.e.Y.`.U.w.h.eXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XyX(.i.`._.).y.[.yX0XqXqX0XyX`.u._._._.u.`.yX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX9XdX].p.~.u.oXsX9XqXqXqXqX9XdX|.i.~.i.|.dX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX8XsX{.7.+XaX8XqXqXqXqXqXqX8XaX.X7.XXaX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX0XtX0XtX9XqXqXqXqXqXqXqXqX9XtX0XtX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqX0XqX0XqXqXqXqXqXqXqXqXqXqX0XqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_6[] = {
/* columns rows colors chars-per-pixel */
"32 32 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqX9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X0XqXqXqXqXqXqXqXqXqX",
"qXwXiXpXpXpXpXpXpXpXpXpXiXpXpXpXpXpXpXpXpXaXrXqXqXqXqXqXqXqXqXqX",
"9XiXbXdXgXgXgXgXgXgXhXuXzXgXgXgXgXgXgXgXhXiXdX0XqXqXqX9XqXqXqXqX",
"9XpXdX6X9X7X7X6X8X8X9X<XpX7X9X8X7X7X7X8X0X,XpXqX0X9X0XiXqXqXqXqX",
"9XpXgX9XqXsXyXdXtX0XwX3XsXqXqXuXaXsXpX0XeX3XaX0XtXsXrXT.rXqXqXqX",
"9XpXgX8XtXQ.+XJ.*XtXqX3XaX9XyXoX~.Y.'.tXwX2XpXtX%XR.Y.q.L.yXwXqX",
"9XpXgX7XaXw._ 7.q.iX0X3XaX8XaXL.X.S.-.uXqX2XiXpXI.V.Y._.F.F.eXqX",
"9XpXgX6XdXQ...1.y.pX0X3XaX0XtX;X0.:.p.aXqX2XiXpXI.D.Q.].K.B.eXqX",
"9XpXgX8XrX;X>X~.<XrXqX3XaX0XwX7XOXP.).wXwX2XiXyXoXJ.J.y.D.eXwXqX",
"9XpXgX8XqXrXrXaXeX0XwX3XaX0XqXqXyXdXpX0XwX2XaX0XtXaXwXD.wXeX0XqX",
"9XpXhX9XwXqXqX0XqXwXeX2XsXqXwXwXqX0X0XqXeX2XpXqX0X9X0XiXqX0XqXqX",
"9XpXiX<X3X2X3X3X3X3X2X1XuX1X3X2X3X3X3X2X3X,XdXqXqXqXqX0XqXqXqXqX",
"0XeXdXaXtXhXaXsXaXaXaXdXdXsXrXhXaXsXaXaXpXhXzX9XqXqXqX0XqXqXqXqX",
"qXqX0X9X3XaX8XqX9X9X8X8X9X0X2XaX9XqX9X9X8XwXkX9XqX9X0XiX0XqXqXqX",
"qXqXqX0X5XsX0XqXuXuXaXyXqXwX3XsX0XqXyXuXyXsXkX9XrXaXrX).yXqXqXqX",
"qXqXqX0X4XsX8XtXOX.X~.#XrXqX3XsX9XtX#X+X+X#XxXqX;X_.^.7.W.iXqXqX",
"qXqXqX0X4XsX7XdXk. .) L.iX0X3XsX8XsXZ.[ ) 8.cXyXI.n.J./.Z.U.rX0X",
"qXqXqX0X4XsX8XpX}.:.0.5.sX9X3XsX9XyX-X} { ;.fXiXU.K.).[.U.v.wXqX",
"qXqXqX0X4XsX9XrX>X|.U.#XtXqX3XsX9XeX1XoX/.{.cXrX}.C.A.d.V.9XeX0X",
"qXqXqX0X4XaX9XqXeXuXsXtX0XqX3XaX9XqXwXyXiXsXkX9XtXpX0Xb.9XrX0XqX",
"qXqXwXwX5XsXqXeXwXqX0XqXwXeX3XsXqXeXwXqX0XrXkX9XqX0X0XuXwX0XqXqX",
"qX0X2X<X3XtX<X3X2X2X2X2X2X1X<XyX<X3X2X2X1X6XkX9XqXqXqX0XqXqXqXqX",
"qXwXaXpXfXdXuXiXiXaXpXpXaXpXdXdXuXiXiXaXpXsXsX0XqXqXqXqXqXqXqXqX",
"qXqXqXqX0XrXpXpXyX0XqXqXqXqX0XeXpXpXyX0XqXqX0XqXqXqXqXqXqXqXqXqX",
"qXqXqX0XtX%XI.I.oXtX0XqXqX0XrX;XI.I.}.tX0XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX9XsXR.V.D.J.aX9XqXqX9XaX_.n.K.C.pX0XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX0XrXY.Y.Q.J.wX0XqXqX0XrX^.J.).A.0X0XqXqXqXqXqXqXqXqXqXqXqX",
"qXqX9XiXT.0._.].y.D.iX0X0XiX_.7./.[.f.b.uX0XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXtXP.F.K.D.wXqXqXqX0XyXW.Z.U.V.8XwXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXyXG.B.eXeX0XqXqXqX0XiXY.v.9XrX0XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXwXeXeXwX0XqXqXqXqXqXqXrXwXeX0XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqX0XqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_7[] = {
/* columns rows colors chars-per-pixel */
"24 24 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXyXuXuXuXuXuXuXuXyXuXuXuXuXuXuXuXwXqXqXqXqXqXqX",
"yXlXsXdXsXsXdXdXiXhXsXdXsXsXdXdXiXuX0X0XqXqXqXqX",
"uXsX7XqXqXeX9X9X4XyX8XqXwXeX9X9X4XuX0XuXqXwXwX0X",
"uXdXqX4X4X;XwXwX7XuXwX6X>X=X0XwX7XiX6X>XW.B.yXwX",
"uXaXuX~.O.1.~.pX4XyXuX[.5.t.).pX4XfXoXM.U.L.L.0X",
"uXsXeX,X6.1.(.pX4XuXwX3Xf.-.oXyX4XfXXXM.E.T.D.9X",
"uXdX9X9X0X1XeXqX6XuX0XqX4X&X6XwX6XiX4X$XU.b.eXeX",
"uXdX9XwXwXrXqXwX5XuX0XwXeXyXwXqX6XuXqXiX0X8XeX0X",
"uXiX4X6X6X6X6X5X6XtX5X6X6X5X6X5X7XiX9X0XqXwX0XqX",
"wXiXeXpXpXuXuXyXiXpXeXpXiXuXuXtXgXsX9X0XqXqX0XqX",
"qXqX5XrXwX0XrXtXeXqX5XtXqX0XrXqXaXaX0XiX0X0XeX0X",
"qXwX6XtXeX0X-X#X>XtX6XtXeX9X;X>X8XfX4X=XT.m.rXeX",
"qXwX6XrXwXwX6.{ M.sX4XtXwX9X1.^ z.mX|.M.R.U.G.9X",
"qXwX6XrXqXpXG.w.z.pX4XtX0XiXZ.$.n.vX XM.T.I.H.0X",
"qXwX6XrXeX0X0X5X0XeX6XtXwX0XqX9XuXdX4X-XR.m.tXeX",
"qXwX5XrXeX0XqXwXqXwX5XrXwXqXqX0XpXaX0XuXqX0XeX0X",
"qX7X4XtX6X5X6X6X6X7X4XtX6X5X7X4XtXsX9X0XqXqX0XqX",
"qXuXiXpXdXfXiXuXuXuXiXpXfXfXiXuXiXeXqXqXqXqXqXqX",
"qX0X0X6XoXXX4XqX0X0X0X5XXXXX5XqX0XqXqXqXqXqXqXqX",
"qX0XuX>XM.M.$XiX0X0XiX=XM.M.-XuX0XqXqXqXqXqXqXqX",
"qXqXqXW.U.E.U.0XqXqX0XT.R.T.R.qXqXqXqXqXqXqXqXqX",
"qXqXwXB.K.T.b.8XwXqX0Xm.I.I.m.0XqXqXqXqXqXqXqXqX",
"qXqXwXyXL.D.eXeX0X0XeXrXH.H.tXeX0XqXqXqXqXqXqXqX",
"qXqX0XwX0X9XeX0XqXqX0XeX9X0XeX0XqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_8[] = {
/* columns rows colors chars-per-pixel */
"16 16 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXyXyXtXyXyXyXyXtXyXyXwX0X0X0XqX",
"yXsXiXdXpXtXiXiXdXpXtXrXrXtXeXqX",
"yXiX>XOX4X9XrX2XOX1X9XeX,X!.*XrX",
"rXgX].X.).tXuX*X,.!.uXeXH.D.L.9X",
"yXuX9X-X6X0XeXwX%X2X0XeX:XE.%XrX",
"yXtX8XeX9X7XeX8XrXqX7XrXrXeXeXqX",
"wXeXrXrXuXpXeXrXrXyXsXiXwXtXeXqX",
"qX9XwXwX&X-X0XqXwX-X,XiX1X/.;XrX",
"wX9XqXiXs.0.eX0XaXf.=.iXH.D.L.9X",
"wX0XwXrX1X*X0XqXrX1X=XiX*XT.#XrX",
"qX7XwX9XqXrX7XqX0X0XuXiXrXeXeXqX",
"qXeXtXtXeXwXeXtXtXeXeXeX0XqXqXqX",
"0XrX,XH.:XrXeX2XH.=XtX0XqXqXqXqX",
"0XtX!.D.E.eXtX/.D.T.eXqXqXqXqXqX",
"0XeX=XL.%XeXeX;XL.#XeXqXqXqXqXqX",
"qXqXrX9XrXqXqXrX9XrXqXqXqXqXqXqX"
};

const char *const *const xpm_icons[] = {
    xpm_icon_0,
    xpm_icon_1,
    xpm_icon_2,
    xpm_icon_3,
    xpm_icon_4,
    xpm_icon_5,
    xpm_icon_6,
    xpm_icon_7,
    xpm_icon_8,
};
const int n_xpm_icons = 9;
