#include "gtk.h"
/* XPM */
static const char *const xpm_icon_0[] = {
/* columns rows colors chars-per-pixel */
"128 128 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXzXxXcXcXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXxXcXcXxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXcXcXxXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXcXcXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXfX6X,X1XtXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjX0X<X<X0XhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXtX1X,X6XfXlXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXlXaX3X<XgXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXfX.XK 3 t t.0XzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkX,Xo.7 7 ..>XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXqXi.t 3 H {.qXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXyX8XH.l q <XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXxX6XP =   . w s.lXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXsX+.2     2 o.aXgXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXzXg.e .   # 6 0 q q q q q q q q q q q q q q q q q q q q q q q q q q q q q 0 2 o = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXcX,X5         X.tX9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X6Xm         b 5X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9XtX+.                                                                            = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXcX1Xu X     @ ;.dXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXrXJ O     O F wXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXdX,.#                                                                           = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXtXu.r   @ R /.zXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXhXC.n X . v B.gXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXzX).K     = #.w.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.r.0.' : % :XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXkX0Xf.X.;./.gXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXpXS.#.@.Z.pXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXlX0.    7 3XmXNXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXNXvX-X7.@.9XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXzXzXtXfXxXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXxXpXiXxXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXNXi.    9 rXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXkXeXjXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXzXvXmXMXMXnXcXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXxXbXmXMXmXbXxXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXdX`.c.y.a.G.:XcXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXxX7XT.f.r.d.L.3XzXzXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXzXyXB.P ,   . p ;..XlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXjX2Xw.c #   + h 2.;XhXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXzXsXE.w . 0 c d = % ( 9XkXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXzXaXe.< @ a N s @ & 6.wXxXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXcX:X^ @ < k.8X}./   8 N.sXzXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXzX0Xz @ D E.gX^.N O y :XcXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXvX].2 # { >XxXcXD.w * ) tXzXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXkX3XQ N m.wXvX4X5.X ; &XcXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXzXfXS.X & k.rXxXbX{.I   d qXzXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXhX9X6XiXkXcX-X..+ p >XcXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXxXyXl.  * K.aXzXvX%X}   6 7XzXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXxXxXzXxXaXm.3 * :.qXxXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXxXeXf.  * Y.sXzXcX*XX.  3 5XzXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXxXsXC.j + D @XgXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXzXuXc.  * D.iXzXvX@X`   0 8XzXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXxXiXM.z O B T.hXzXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXzXkXI.+ % w.qXcXmX).B o n eXzXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXzXtXb.h o V G.lXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXbXXXu @ F }.kXpXs.> ; *.uXzXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXkXwXb.a O Z H.bXcXxXzXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXcX4X;.& O $.T.l.m   d Q.fXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXsXb.w   5 :.}.{.{.|.wXjXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXjX#XB ; @ + + @ w 1.dXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlX8Xe       o + + O = H.pXzXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXfX$X-.h , 6 L m.tXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlX7Xr @ $ $ $ $ $ $ > J.pXzXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXhX9X:X&X=X2XyXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXgX-X$X%X%X%X%X%X%X&XyXkXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXxXcXvXvXcXzXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXvXvXvXvXvXvXvXvXvXzXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXkXdX6XrXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX0X9XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXu.    8 wXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXpX5XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXjX,XO.n G V.pXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXwXw.B B 0.eXcXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXbXhX5.    6 *XjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfX].Q C 4XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXxX0X+.2   X v Z.zXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXfXe.y     8 J R R R R R R R R R R R R R R R R R R R R R R R R R R R R R R I u     & ..6XkXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXxX XY # & :XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXcX<X9 .     . @.iXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwX9XC                                                                               . 6 XXtXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXpXB.$   = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXcX<X9 .     . @.iXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwX9XC .                                                                               6 XXtXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXiXB.%   = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXxX0Xo.3   o v C.zXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXfXw.8     < F I I I I I I I I I I I I I I I I I I I I I I I I I I I I I U J a O   <  .6XkXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXxX XL   = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXhX>XO.v F B.pXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXeXJ     F eXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXvXsXZ.G v X.:XhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXhXL.. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXkXsX5XeXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXeX5XsXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbX'.. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXlXzXcXbXnXnXbXxXlXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXzXsX<X}.).`.OX0XjXzXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXdX/...w $ ; v y.>XjXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXzX!.Z X * < : # # <.1XcXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXzXX.# 0 5.;X).A # 0 T.nXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXpX..0 X.$XxXyXg.<   V.nXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXgX+X~.*XfXxXiXd.< $ F.mXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXvXnXvXxXkX$XS + S }.bXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXkX*X#.# 7 v.rXxXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXlXzXjX$X*.$ 4 <.6XvXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXzXhX@X@.@ 4 $.8XzXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXhX+X{ O 4 @.wXkXlXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXzXOX~ O ; ` -XqXwXwXaXkXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXkX[ @   X : t t t w f.rXzXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXhXS                 2.qXxXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXkXQ.v.m.m.m.m.m.m.b.OXdXzXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXnXMXMXMXMXMXMXMXMXvXzXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXgX9XuXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXR     I cXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXuX9XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXzXxXtXfXxXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXR     U vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXxXgXtXzXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXnX].. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXkX0Xf.O.:.(.lXNXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXhXI     J sXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXgX).>.O.d.qXnXNXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXNXmX!.. = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXtXi.w   + J 0.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.u.5.u     p C.gXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXzX_.W +   8 *.t.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.i.r._   = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXcX1Xu .                                                                             O F eXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXfX,.@                                                                           = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXcX,X5 .                                                                               b 4X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8XtX+.                                                                            = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXxX6XK #     = 7 9 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 6 * .   1 X.aXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXlXg.0     o 3 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 9 8 1 o - >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXfX{.6     #.3XrXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXwX*Xo.7 6 ..:XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXqX=.    3 _.0XeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXrX7XG.l 0 ,XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXqX0     w.mXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXjX7XXXXX7XhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXnXt.    8 0XxXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXlXpX-X}.sXzXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.NXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXyXtXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXNXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX0XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXlXzXzXxXzXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXgXyXwX0XwXyXfXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXaX_...k w h ..!.jXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXzXuXJ.f - .   X & N Z.jXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXvX&XD   r w.K.>.e X P uXkXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXvX-X} ^ f.7XmX&X`   p 3XjXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXzXuX XXXwXvXvX|.E   n 9XkXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXzXcXxXpXW.n.[ t x i.hXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXjX>Xk     4 -.,XlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXkX7X@.n e % 0 5.tXkXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXjXfXfXhXpXiX.X_   t D.dXzXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXzXgX-Xv.M.>XhXxXxXs.  1 7.iXzXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXzXkX~.e y o.OXsXXX_   4 f.aXzXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXxX@XD   > Z E B <   N $XhXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXzXiXE.~ 9       8 ~ F.hXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXxXjXoXg.=.o.=.g..XzXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXxXbXnXnXnXbXxXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXtXq     r.MXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXeXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXMXi.    8 eXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXdX7XhXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXzXyXq     r.NXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXxXpXpXxXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXNXi.    9 rXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXxXlXrXjXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlX8X0     0.mXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXbXhX XN.N.|.hXbXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXmXr.    8 7XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkX8XP.m.yXzXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXaXJ.2     / Q.].].].].].].].].].].].].].].].].].].].].].].].].].].].].].'.P.P # # K L.'.].].].].].].].].].].].].].].].].].].].].].].].].].].].].].!._     1 G.pXlXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXzX8Xd.7 - :XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXcX3Xx O       . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .         . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .       o k =XfXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXkXI.8   = >XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXcX<Xe - = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = - 0 }.0X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7XrXM.: = 6 ,XcXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXgX<X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X,XsXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXjXyX>X>X,XgXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXcXcXcXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX"
};

/* XPM */
static const char *const xpm_icon_1[] = {
/* columns rows colors chars-per-pixel */
"96 96 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXjXgXgXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXfXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXgXgXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXfXjXlXlXlX",
"lXlXzXjXmXJXGXzXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXZXKXAXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXlXFXKXNXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXvXLXMXjXlXlX",
"lXlXjXMX-Xx.V.aXcXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXmX/.j.R.bXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXxXhXS.l.@XPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXUX0Xf.&XNXjXlX",
"lXjXmX;Xv     $.hXxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXMXK.:   # M.MXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXxX3.    p 1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.4.J   a.PXfXlX",
"lXgXJXc.  * O   ;XrX9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X4XxXD   &   j kX5X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9XqX8X$   @                                                         c.JXgXlX",
"lXgXGXZ.  O   ; 5XsXyXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXqXmX~   o   D bXwXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXiXyX8   .                                                         z.JXgXlX",
"lXkXzXaX&.  ; G.cXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXkXcX<XN   h &XvXlXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXbXU.    D ;X%X&X*X*X*X*X*X*X*X*X*X*X*X*X*X*X*X*X*X*X*X&X&X;Xt.  l.LXfXlX",
"lXlXkXcXjX;X5XcXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXvXiX*XtXvXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnXeX    | UXvXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXxX:XwXvXkXlX",
"lXlXlXkXxXrXdXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXqXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { PXaXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXjXxXuXpXxXlXlX",
"lXlXlXlXlX9XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX6XfXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXkXgXdXdXgXkXlXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXhXdXdXfXhXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXjXZXLXLXSXlXjXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXjXMXJXPXHXnXjXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXzXzXn.X.{ s.yXmXhXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXkXxXvX^.#.] ;.|.bXzXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlX/ O       F 3XMXjXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXzXzXz.%       , K.cXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXjXmXh.  = W.'.x   3.kXzXkXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXkXmX6XO   i.eX+.  t fXcXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXhXCXj   0.BXHXD.  r 2XMXkXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXbX;Xi #.aXAX%Xq   iXvXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXkXlXeX1   [.zXmX$Xg   +XCXjXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXzXbXvXgXCXG.# w fXcXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXkXxX:X> X -XlXvX<XM   }.CXjXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXkXgXmX>Xs   M.cXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXkXxX2X< . #XkXvX-Xb   .XCXjXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXhXmX-Xj   =.zXcXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXjXzX9   H.cXBX].5 o *XBXjXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXkXmX#Xi   o.dXhXhXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXhXZX~   D hXMX=.  P eXbXkXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXbXXXr   q.GXKXDXbXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXkXxX%X=   M Z O   P.NXjXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXkXnX7X9   = A E K P *XcXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXnX`.A     f H.xXzXkXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXkXNX%X              ^.vXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXkXnXiX9X8XyXvXxXkXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXzXhX5X4X4X6X7X7X7XfXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXkXvXmXmXbXlXkXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXzXMXmXmXmXmXmXmXxXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXkXkXkXlXlXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXkXkXkXkXkXkXkXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXlXlXhX4XwXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXsX1XpXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkX6XtXcXkXlX",
"lXlXlXjXNXxXMXcXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXCXlXCXjXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXzX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXmXnXhXzXlXlX",
"lXlXjXmXJ.A Q <XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXzX7.M ;.cXHXFXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXIX3X    _ FXfXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXmX[.x I.DXhXlX",
"lXhXCX/.>     M uXlXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXaXBX8.      i d d d d d d d d d d d d d d d d d d d d d d d f w     * G.BXsXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXfXbX$Xy   g.LXfXlX",
"lXfXKXl.  % o   &XqX6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X1XkXB   +                                                       . @   =.kX1X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X4XaXF.    z.JXgXlX",
"lXhXAXE.%     h rXjXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXiXBX>.    X 2 6 5 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 ,     o B.BXpXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXsXbXoX8   x.KXgXlX",
"lXlXjXbXN.j D &XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXcXi   2 kXVXBXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXAX4XP f g.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXAXz.  n.KXgXlX",
"lXlXlXkXMXkXbXvXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXHXd   6 VXfXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXhXxXmXjXNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXJXS.  m.KXgXlX",
"lXlXlXlXjX5XeXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXFXd   5 BXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXtX4XhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXHXZ.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXlXlXlXlXkXkXkXkXlXlXlXlXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXlXlXlXcXbXnXmXnXvXlXlXlXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXlXjXbX,X| 2 ; F ~.bXkXlXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXjXnX<Xx   3 6   % R.mXjXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXgXGXj.  A 0XVX]   J GXfXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXkXnX)...].zXzXH.  M HXfXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXlXkXMXKXnXjXxXP   2.AXgXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXlXlXjXsXgXvX<.  l wXlXkXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXlXlXkXlXzX;.  t %XMXkXlXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXlXkXzXjX#.  t %XVXjXlXlXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXkXzXaXo.  t +XMXfXjXlXlXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXgXHXs.      X     p eXlXkXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXgXDXv.- s t w e 3 F tXlXkXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXlXkXmXGXSXDXDXDXFXZXzXlXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXlXlXjXgXgXgXgXgXgXhXlXlXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXlX0XuXzXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXpX8XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXJXA.  m.KXgXlX",
"lXlXlXlXzXqXpXzXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXfXGXd   6 VXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXaX0XlXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXfXHXA.  m.KXgXlX",
"lXlXkXxXcX8XyXbXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXzXIXf   6 AXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXpX7XxXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXxXUXF.  m.KXgXlX",
"lXkXlXhX3.@ 5 U.eX9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X3Xw   , 3XxXlXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXvX].w O +.9X8X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X8XiX9.  b.KXgXlX",
"lXgXFXD.                                                            .   L nXrXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXaXiXe                                                         O   l.KXgXlX",
"lXgXJXz.  @ .                                                       @   f hX4X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X9X7X@   X                                                         x.JXgXlX",
"lXjXMX#Xa     D | { { { { { { { { { { { { { { { { { { { { { { { _ =   X g.MXgXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXlXzX#.    9 { { { { { { { { { { { { { { { { { { { { { { { { | m   s.PXfXlX",
"lXlXdXPX1.    ;XUXPXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXFXH.=.V.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXnX9X    { UXHXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXUX3X*.[.CXjXlX",
"lXlXdXPX1.    %XvXaXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXfXVXlXVXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX8X    { HXtXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXsXbXnXhXzXlXlX",
"lXlXdXPX1.    &XBXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXsX1XaXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkX6XtXcXkXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXlXlXlXkXjXjXkXlXlXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXlXkXlXMXCXCXmXlXkXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXlXcXxX<X$X#X6XcXcXkXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXlXhXy.3     p H.lXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXjXMX6.  , %.^   # {.VXjXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXjXBX..D Z.HXzX]   v.VXjXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXzXwX8XbXKX2XN   P.ZXjXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXlXvXNX|.j r 3 j.gXxXkXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXlXkXbX`.# .   E 7XnXjXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXkXzXzXzXdXqX~   $.fXxXkXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXaX8.1.,XCXUXa.  S 0XnXkXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXjXbXf   Z !.N.p   -.jXxXkXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXkXbX[.x       X G >XMXjXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXkXNX1Xk.5.7.B.iXmXjXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXlXhXvXFXPXLXSXlXjXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXlXlXjXfXdXdXgXkXlXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    *XBXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX7XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXwXuXcXlXlX",
"lXlXdXPX1.    &XBXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfX4XdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXnX9X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX9XyXcXlXlX",
"lXlXdXPX1.    %XnXfXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXhXmXsXnXjXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXfXxX8X    { IXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXvXkXdXzXlXlX",
"lXlXdXUX4.    <XUXGXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXHXJXDX$XD.oXAXJXHXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXHXUXiX    | UXsXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXbXiXG.:XmXkXlX",
"lXjXvX0XJ   . ,.G.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.Z.A.n.y   6 z.S.Z.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.F.9.O   m 2XbXkXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXlXvXiX$.  h.LXfXlX",
"lXfXLXg.                                                                                                                          &.nX6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XjXG.    s.LXfXlX",
"lXjXMX*Xa.c.l.b.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.n.x.l.x.n.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.b.l.x.d.{.hXtXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXyXdX:Xj.s.%XMXjXlX",
"lXlXjXMXPXJXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXJXPXCXzXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXzXmXLXLXMXjXlXlX",
"lXlXlXjXfXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXfXjXlXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXfXfXjXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX"
};

/* XPM */
static const char *const xpm_icon_2[] = {
/* columns rows colors chars-per-pixel */
"88 88 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXgXfXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXfXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXfXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXgXlXlXlXlX",
"lXlXlXzXSXJXNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXAXJXVXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXZXJXCXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXZXGXlXlXlXlX",
"lXlXzXyXX.n B.bXkXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXlXxXdX*.b k.vXlXlXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXlXzXjX2.v y.vXcXcXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXcXvXxX7.L vXlXlXlX",
"lXgXSXO.    + /.vXiXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXyXMX>.      L.nXuXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXyXNXe.    . 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5   + vXlXlXlX",
"lXfXJXM   ;   b.gX5X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X4XzXH   :   y.kX5X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X3XxX(   $                                                   7 vXlXlXlX",
"lXjXNXV.#   b 3XvXhXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXgXNXL.-   f =XnXhXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXgXMX^.,   q 1.<.<.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.<.1.<.4   bXlXlXlX",
"lXlXlXbX(.b.3XmXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXmX[.v.>XmXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXJXn   Q UXGXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLX+XV.xXlXlXlX",
"lXlXlXkXvXhXbXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXbXgXbXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXDXv   E GXuXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXfXnXfXlXlXlXlX",
"lXlXlXzXiX6XjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXaX5XhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfX6XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXkXlXzXzXlXkXlXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXkXlXzXzXlXkXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXkXcXbXcXcXcXbXlXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXvXvXcXcXvXvXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXkXcXpX0.6 O D {.bXkXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXjXbX;X^ $ & ( ;XvXjXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXkXkXxXO.  d D X $ +XnXkXlXlXlXlXlXzXfX8XkXlXlXlXlXlXjXbX8Xd   F Z   x qXvXkXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXhXCX{.X q 1XPX7.   .xXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXjXCX~.  | VXcXY   `.ZXhXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXjXCXn.  2.lXNXoX1 3 jXxXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXxXrX`.0XlXgX`   |.CXjXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXkXnXe.  h.vXcX<Xa   sXcXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXkXvXCXxXBXJ.  J uXcXkXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXkXnXr.  g.vXcX>Xi   sXcXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXkXfXMXP.  8 $XNXhXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXhXZXB.  =.lXZX|.- 0 jXxXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXkXkXNXA.  2  XNXgXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXhXVXoX& ; XXNX@.  =.cXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXkXvXM.  y :XIXNXbXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXjXnX4.  ; 7   9 :XnXkXlXlXlXlXlXzXfX8XkXlXlXlXlXlXjXVX].    # y 0 ; V.nXkXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXkXzXlXn.m p { :XnXkXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXjXVX`.3 k i 0 t > C.nXkXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXzXMXSXDXCXvXkXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXkXvXFXAXSXFXDXGXmXkXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXjXhXhXjXkXlXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXkXgXhXhXhXhXhXkXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXlXuX4XhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXpX3XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXb   E LXfXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdX4XlXlXlXlX",
"lXlXlXxXNXxXNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXBXxXBXhXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXsXFXb   W PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXVXkXkXlXlXlX",
"lXlXzXdX:.G K.mXkXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXlXzXjX6.D V.AXVXVXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXVXBXAXl   I SXhXlXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXlXkXvXa.Q cXlXlXlX",
"lXhXAX=.    * ].vXpXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXiXNX7.    # p i i i i i i i i i i i i i i i i i i i i p y #   O A.MXiXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXiXMXM.  + bXlXlXlX",
"lXfXJXm   :   v.gX5X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X3XzXG   $                                                   *   :.lX3X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X3XzX#.  8 vXlXlXlX",
"lXjXBXz..   f :XbXgXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXfXNXC.#   w X. . ................................... ...{ :   4 [.mXfXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXfXMX).- ; bXlXlXlX",
"lXlXlXvXP.y.-XmXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXZXp   X.UXGXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXJX/.r. XMXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXPXA o mXlXlXlX",
"lXlXlXlXnXkXnXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXVXi    .GXyXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXfXmXkXmXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXKXC o mXlXlXlX",
"lXlXlXlXiX5XjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXVXi    .PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXsX4XfXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXKXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXlXlXlXlXkXkXkXkXkXlXlXlXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXlXlXlXcXMXBXBXMXcXlXlXlXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXlXkXvX0X&.8 5 { >XbXjXlXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXkXcXpXG   g k   l 7XbXjXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXkXNXoX  L kXmXQ   P.DXhXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXlXxXrXI.,XzXzX-.  T.SXhXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXlXkXcXSXcXNX{.  j 7XbXjXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXlXlXkXsXmX X, $ ^.VXhXlXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXlXlXjXNX_.& X L.NXfXlXlXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXlXkXbX!.* 1 {.UXAXMXlXlXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXkXMX-X    = m x t n.bXkXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXkXNXOX# e 3 & >   i.nXkXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXlXlXxXBXMXNXVXVXCXbXlXlXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXlXlXlXkXkXkXkXkXkXkXlXlXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXzXsX9XlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXCXi   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgX8XjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXLXZ o mXlXlXlX",
"lXlXlXlXuX4XhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXVXi    .PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXaX3XdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXKXC o mXlXlXlX",
"lXlXlXzXBXcXNXfXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXsXBXp   ..PXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXBXxXBXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXiXJXZ o mXlXlXlX",
"lXlXlXjXw./ ~.JXDXFXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXFXFXAXy   { JXgXlXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXlXkXvXz.! A.FXFXFXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXFXSXUXB O mXlXlXlX",
"lXhXZX3.    > n v b b b b b b b b b b b b b b b b b b b b l #   ; ^.nXaXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXaXNXx.    $ b b b b b b b b b b b b b b b b b b b b b b v @ > bXlXlXlX",
"lXfXJXn   $                                                   *   r.jX4X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X2XzX^   &                                                   5 vXlXlXlX",
"lXhXCXi..   q Q E E E E E E E E E E E E E E E E E E E E W I +   4 |.nXfXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXdXNXS.%   1 E E E E E E E E E E E E E E E E E E E E E E E $   bXlXlXlX",
"lXlXlXcX5   1.UXGXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXPXSXS.:.[.MXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXb   E UXGXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXKXT.8.xXlXlXlX",
"lXlXlXcX5   <.GXuXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXfXdXhXMXzXMXjXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXb   E GXuXfXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXfXdXfXNXkXkXlXlXlX",
"lXlXlXcX5   <.LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXpX3XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXb   E LXfXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdX4XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXlXlXlXlXjXhXhXjXlXlXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXlXlXlXxXBXDXDXVXcXlXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXlXjXbXyX7.g s $.6XbXkXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXkXcXeXB   9 8   n wXvXkXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXhXCX].7 ` sXzXE   _.NXkXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXlXxXaX-XsXKXaXH   XXMXkXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXlXlXzXnXtXJ 7 > v.cXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXlXlXzXcXtXF 5 . L 9XcXkXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXlXlXsX0XkXMXCX;.  3.vXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXjXNXL.t w.kXFX1.  X.xXzXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXjXNX.X@   B B o   ^.mXkXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXlXjXnXOXT % @ A Q.vXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXlXlXlXvXvXcXcXcXnXlXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXlXlXlXkXlXzXzXlXkXlXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXvX5   1.LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfX8XkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXGXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhX9XlXlXlXlX",
"lXlXlXcX5   <.LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXpX3XgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXb   E LXfXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdX4XlXlXlXlX",
"lXlXlXvX5   1.GXuXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXhXNXxXNXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXiXSXb   E LXdXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXjXBXlXkXlXlXlX",
"lXlXlXxX6   <.UXGXKXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXKXPXAXm.@.(.PXKXKXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXKXJXUXv   E KXfXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXMXJ.,.cXlXlXlX",
"lXhXZX8.    1 A C Z Z Z Z Z Z Z Z Z Z Z Z Z Z Z Z Z Z C Z M     = A C Z Z Z Z Z Z Z Z Z Z Z Z Z Z Z Z Z Z C Z B @   $ Y.MXsXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXsXNXK.    bXlXlXlX",
"lXgXGXP + 3 = X o o o o o o o o o o o o o o o o o o o o o O < 5 ; o o o o o o o o o o o o o o o o o o o o o o O > 5   8.kX4X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X4XkX1.  h vXlXlXlX",
"lXlXlXvXvXvXbXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXbXvXbXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXbXvXbXxXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXcXbXvXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX"
};

/* XPM */
static const char *const xpm_icon_3[] = {
/* columns rows colors chars-per-pixel */
"64 64 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXlXlXlXlX",
"lXlXlXkXlXvXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXlXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXbXlXfXgXgXgXgXgXgXgXgXgXgXgXgXgXfXhXvXlXlXlXlX",
"lXlXkXcXzXwXbXzXzXzXzXzXzXzXzXzXzXzXzXzXlXxXcXwXcXxXlXzXzXzXzXzXzXzXzXzXzXzXzXzXbXwXxXKXHXHXHXHXHXHXHXHXHXHXHXHXHXJXZXtXjXlXlXlX",
"lXlXlXzX'   i.nXdXhXhXhXhXhXhXhXhXhXhXgXfXcX:.  ;.cXfXgXhXhXhXhXhXhXhXhXhXhXdXnXi.  Y s.r.t.t.t.t.t.t.t.t.t.t.t.t.p.-.7 qXbXkXlX",
"lXkXvXeXX   d fX0XwXwXwXwXwXwXwXwXwXwXwXeXeX1   < wXeXwXwXwXwXwXwXwXwXwXwXwX0XfXh                                       qXbXkXlX",
"lXlXkXbXp.f R.mXjXlXlXlXlXlXlXlXlXlXlXlXjXmXB.a N.mXjXlXlXlXlXlXlXlXlXlXlXlXhXVXZ.  { &X].}.}.}.}.}.}.}.}.}.}.}.{.XXL.D eXvXkXlX",
"lXlXlXzXnXgXMXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXfXmXlXlXlXlXlXlXlXlXlXlXlXlXlXhXCX+X  p.UXMXCXCXCXCXCXCXCXCXCXCXCXCXCXZXgXhXzXlXlX",
"lXlXlXzXfXqXjXlXlXlXlXlXlXjXhXkXlXlXlXlXlXlXhX0XhXlXlXlXlXlXlXjXhXkXlXlXlXlXhXCX|.  w.DXsXhXhXhXhXhXhXhXhXhXhXhXhXhXjXwXhXzXlXlX",
"lXlXlXzXhXeXlXlXlXlXlXlXjXmXAXvXjXlXlXlXlXlXkXeXkXlXlXlXlXlXkXNXAXvXjXlXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXlXzXhXeXlXlXlXlXlXlXMX>X!.qXMXkXlXlXlXlXkXeXkXlXlXlXlXzXmX#X!.eXmXkXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXlXzXhXeXlXlXlXlXjXBXb.& * 7 |.MXjXlXlXlXkXeXkXlXlXlXkXnX1.# - r %XnXjXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXlXzXhXeXlXlXlXkXvXrX$ %.CXk m nXlXlXlXlXkXeXkXlXlXkXcXpXr v.nX2 X.KXgXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXlXzXhXeXlXlXlXjXBX'.  XXUXf.@ pXcXkXlXlXkXeXkXlXlXlXlXkXiXCXrX: s.DXhXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXlXzXhXeXlXlXlXjXBX~.  @XFXx.o uXvXkXlXlXkXeXkXlXlXlXlXjXBXiXy N kXlXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXlXzXhXeXlXlXlXjXmX<X  m.UX` 0 zXzXlXlXlXkXeXkXlXlXlXkXbXtXx T KXGXjXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXlXzXhXeXlXlXlXlXkXmX' # I   k.VXjXlXlXlXkXeXkXlXlXkXzXjXe   I U i.mXkXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXlXzXhXeXlXlXlXlXkXcXhXg.` K.nXlXlXlXlXlXkXeXkXlXlXlXxXpX&.%.[ Y j.nXkXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXlXzXhXeXlXlXlXlXlXkXzXDXPXCXkXlXlXlXlXlXkXeXkXlXlXlXlXzXJXJXLXUXSXkXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXlXzXhXeXlXlXlXlXlXlXlXhXgXhXlXlXlXlXlXlXkXwXkXlXlXlXlXlXgXfXgXfXhXlXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXrXhXzXlXlX",
"lXlXlXlXgXeXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXeXgXgXgXgXgXgXgXgXgXgXgXgXgXgXdXMX{.  w.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXrXhXzXlXlX",
"lXlXlXxXcXrXmXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXnXeXNXFXDXDXDXDXDXDXDXDXDXDXDXDXZXUX*X  p.KXgXlXlXlXlXlXlXlXlXlXlXlXlXlXMXuXgXzXlXlX",
"lXlXlXcX:.< N.mXgXjXjXjXjXjXjXjXjXjXjXjXhXnXu., 1.U.F.G.G.G.G.G.G.G.G.G.G.G.D.Y.5.  _ bXhXjXjXjXjXjXjXjXjXjXjXjXgXMXT.i wXbXkXlX",
"lXkXbXwX    a fX0XwXwXwXwXwXwXwXwXwXwXwXwXwX<                                         -XuXqXwXwXwXwXwXwXwXwXwXwX8XzXH   tXbXkXlX",
"lXlXlXcX:.< N.mXgXjXjXjXjXjXjXjXjXjXjXjXgXNX1.  X.Y.A.D.D.D.D.D.D.D.D.D.D.D.S.K.r.2 *.zXjXjXjXjXjXjXjXjXjXjXjXjXgXNXH.X wXbXkXlX",
"lXlXlXxXcXeXmXlXlXlXlXlXlXlXlXlXlXlXlXlXhXFXI.  Y.UXCXDXDXDXDXDXDXDXDXDXDXDXDXFXCXwXcXcXlXlXlXlXlXlXlXlXlXlXlXlXkXmX5X@ qXbXkXlX",
"lXlXlXlXgXeXkXlXlXlXlXlXlXlXlXlXlXlXlXlXgXDXF.  A.CXsXgXgXgXgXgXgXgXgXgXgXgXgXgXhXrXgXlXlXlXlXlXlXkXjXkXlXlXlXlXjXMX>X@ qXbXkXlX",
"lXlXlXzXhXeXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXDXG.  D.DXgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXeXhXzXlXlXlXlXjXvXMXxXjXlXlXlXkXMX,X@ qXbXkXlX",
"lXlXlXzXhXeXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXDXG.  D.DXgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXeXhXzXlXlXlXkXNXtX-XdXmXkXlXlXkXMX,X@ qXbXkXlX",
"lXlXlXzXhXeXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXDXG.  D.DXgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXeXhXzXlXlXjXMXH.3 @ a @XnXjXlXkXMX,X@ qXbXkXlX",
"lXlXlXzXhXeXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXDXG.  D.DXgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXeXhXzXlXlXkXbXa | dXu P HXgXlXkXMX,X@ qXbXkXlX",
"lXlXlXzXhXeXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXDXG.  D.DXgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXeXhXzXlXlXlXkX3XkXSXh [ GXgXlXkXMX,X@ qXbXkXlX",
"lXlXlXzXhXeXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXDXG.  D.DXgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXeXhXzXlXlXlXkXMXVX( > 5XvXkXlXkXMX,X@ qXbXkXlX",
"lXlXlXzXhXeXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXDXG.  D.DXgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXeXhXzXlXlXkXzXnX| y iXUXjXlXlXkXMX,X@ qXbXkXlX",
"lXlXlXzXhXeXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXDXG.  D.DXgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXeXhXzXlXlXjXMX!   W 4.g.xXlXlXkXMX,X@ qXbXkXlX",
"lXlXlXzXhXeXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXDXG.  D.DXgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXeXhXzXlXlXlXzX' L F f O.xXlXlXkXMX,X@ qXbXkXlX",
"lXlXlXzXhXeXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXDXG.  D.DXgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXeXhXzXlXlXlXlXFXGXHXPXDXlXlXlXkXMX,X@ qXbXkXlX",
"lXlXlXzXhXeXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXDXG.  D.DXgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXeXhXzXlXlXlXlXhXgXgXgXhXlXlXlXkXMX,X@ qXbXkXlX",
"lXlXlXzXfXqXhXhXhXhXhXhXhXhXhXhXhXhXhXhXdXZXD.  S.DXgXlXlXlXlXlXlXlXlXlXlXlXlXlXjXqXdXjXhXhXhXhXhXhXhXhXhXhXhXhXgXbX:X@ qXbXkXlX",
"lXlXlXzXmXgXCXCXCXCXCXCXCXCXCXCXCXCXCXCXMXUXY.  K.FXgXlXlXlXlXlXlXlXlXlXlXlXlXkXMXgXMXZXCXCXCXCXCXCXCXCXCXCXCXCXBXJX9X@ qXbXkXlX",
"lXlXkXbXi.g Z.+X|. X X X X X X X X X X X{.*X5.  r.CXhXlXlXlXlXlXlXlXlXlXlXlXjXMXE.f q.#X}. X X X X X X X X X X X}.#XC.+ qXbXkXlX",
"lXkXvXwX.                                       2 wXeXwXwXwXwXwXwXwXwXwXwXwX0XgXg                                     o qXbXkXlX",
"lXlXlXcXY   { p.w.e.e.e.e.e.e.e.e.e.e.e.w.p._   *.cXfXgXgXgXgXgXgXgXgXgXgXgXdXMXw.  A a.q.e.e.e.e.e.e.e.e.e.e.e.w.u.*.4 0XbXkXlX",
"lXlXfXKXs.  &XUXDXHXHXHXHXHXHXHXHXHXHXHXHXKXbX;XxXcXlXzXzXzXzXzXzXzXzXzXzXzXjXZX#X  a.UXSXHXHXHXHXHXHXHXHXHXHXHXHXJXAX1XfXzXlXlX",
"lXlXgXHXr.  ].MXsXgXgXgXgXfXfXfXgXgXgXgXgXgXjXuXjXlXlXlXlXlXlXlXlXlXlXlXlXlXhXCX|.  q.SXpXgXgXgXgXgXgXgXgXgXgXgXgXfXjXiXhXzXlXlX",
"lXlXgXHXt.  }.CXhXlXlXlXjXzXcXkXkXlXlXlXlXlXjXwXjXlXlXlXlXlXlXlXlXlXlXlXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXrXhXzXlXlX",
"lXlXgXHXt.  }.CXhXlXlXlXBXkXpXbXvXkXlXlXlXlXkXeXkXlXlXlXlXlXlXlXlXlXlXlXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXgXHXt.  }.CXhXlXkXbXM.7 @ ^ tXcXkXlXlXlXkXeXkXlXlXlXlXlXlXlXlXlXlXlXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXgXHXt.  }.CXhXkXvXtXi 9.-X  7.GXgXlXlXlXkXeXkXlXlXlXlXlXlXlXlXlXlXlXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXgXHXt.  }.CXhXlXzXgXqXuXR.y F.ZXhXlXlXlXkXeXkXlXlXlXlXlXlXlXlXlXlXlXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXgXHXt.  }.CXhXlXkXlXUXl.  u <XnXjXlXlXlXkXeXkXlXlXlXlXlXlXlXlXlXlXlXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXgXHXt.  }.CXhXlXxXtX[.xXJXd D MXkXlXlXlXkXeXkXlXlXlXlXlXlXlXlXlXlXlXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXgXHXt.  }.CXhXkXnX3X& W c.O } BXjXlXlXlXkXeXkXlXlXlXlXlXlXlXlXlXlXlXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXgXHXt.  }.CXhXlXkXbXE.m w &.uXcXkXlXlXlXkXeXkXlXlXlXlXlXlXlXlXlXlXlXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXgXHXt.  }.CXhXlXlXlXNXAXAXCXcXkXlXlXlXlXkXeXkXlXlXlXlXlXlXlXlXlXlXlXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXgXHXt.  }.CXhXlXlXlXjXjXjXhXkXlXlXlXlXlXkXeXkXlXlXlXlXlXlXlXlXlXlXlXlXlXhXCX X  e.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXzXtXhXzXlXlX",
"lXlXgXHXt.  {.VXgXkXkXkXkXkXkXkXkXkXkXkXjXkXgX9XgXkXjXkXkXkXkXkXkXkXkXkXkXkXgXBX}.  w.HXgXlXlXlXlXlXlXlXlXlXlXlXlXlXlXqXhXzXlXlX",
"lXlXfXJXp.  +XJXbXMXMXMXMXMXMXMXMXMXMXMXMXmXBXxXBXmXMXMXMXMXMXMXMXMXMXMXMXMXbXJX#X  u.JXfXlXlXlXlXlXlXlXlXlXlXlXlXkXmXcXjXlXlXlX",
"lXlXhXZX-.  V.0X:X,X,X,X,X,X,X,X,X,X,X,X>X5XH.H G.5X>X,X,X,X,X,X,X,X,X,X,X,X:X9XC.  &.AXjXzXzXzXzXzXzXzXzXzXzXzXkXmX&X( tXvXkXlX",
"lXkXvXtX5 o @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ .   . @ @ @ @ @ @ @ @ @ @ @ @ @ @ @ + o 4 1XiXeXrXrXrXrXrXrXrXrXrXrX0XxX)   rXbXkXlX",
"lXlXlXjXqXqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXtXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXfXhXhXhXhXhXhXhXhXhXhXhXhXhXjXtXrXkXlXlXlX",
"lXlXlXlXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXzXzXzXzXzXzXzXzXzXzXzXzXzXzXlXvXbXlXlXlXlX",
"lXlXlXlXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXkXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX"
};

/* XPM */
static const char *const xpm_icon_4[] = {
/* columns rows colors chars-per-pixel */
"48 48 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXhXkXlXlXlXlXlXlXlXlXlXlXlXhXkXlXlXlXlXlXlXlXlXlXlXlXhXjXlXkXkXkXkXkXkXkXkXkXkXhXlXlXlXlX",
"lXlXlXzXAXbXkXlXlXlXlXlXlXlXlXlXzXAXbXkXlXlXlXlXlXlXlXlXlXzXZXmXzXxXxXxXxXxXxXxXxXzXxXVXzXlXlXlX",
"lXlXzXhXT.5XbXlXzXzXzXzXzXzXzXxXjXR.3XbXlXzXzXzXzXzXzXzXzXkXY.<XcXhXjXjXjXjXjXjXjXkXfX`.gXzXlXlX",
"lXhXAXR.  R vXeXyXyXyXyXyXyXeXzXP.  I cXeXyXyXyXyXyXyXrXlXI.  o = & & & & & & & & = X + sXxXkXlX",
"lXkXbX5XE H.nXfXhXhXhXhXhXhXgXxX5XW D.nXfXhXhXhXhXhXhXgXxX5X$ #.Q.F.J.J.J.J.J.J.H.K.A.*.sXxXlXlX",
"lXlXkXbXvXmXlXzXzXlXjXhXlXzXzXlXbXvXmXzXzXzXlXhXhXlXzXlXzXmX> K.UXCXSXSXSXSXSXSXSXSXFXxXjXlXlXlX",
"lXlXlXlXrXfXzXlXlXkXSXJXxXkXlXlXlXrXfXzXlXkXxXGXGXzXlXlXzXxX; N.ZXdXhXhXhXhXhXhXhXhXjXtXjXzXlXlX",
"lXlXlXzXuXhXzXlXlXvXN.q.uXvXkXlXzXuXhXzXkXxXsXp.a.gXxXkXzXcX; V.FXhXlXlXlXlXlXlXlXlXzXiXjXzXlXlX",
"lXlXlXzXuXhXzXhXSXi.n _ Y MXkXlXzXuXhXzXhXDX-.E T $.FXgXzXcX; V.FXhXlXlXlXlXlXlXlXlXzXiXjXzXlXlX",
"lXlXlXzXuXhXlXjXCXM U.bXq 8XbXkXzXuXhXzXlXxX>XUX-...HXgXzXcX; V.FXhXlXlXlXlXlXlXlXlXzXiXjXzXlXlX",
"lXlXlXzXuXhXlXjXZXM P.vX0 8XbXkXzXuXhXzXlXhXIXk.W BXzXkXzXcX; V.FXhXlXlXlXlXlXlXlXlXzXiXjXzXlXlX",
"lXlXlXzXuXhXzXhXSXs.x Q E MXkXlXzXuXhXzXhXAXu.  *.P.cXkXzXcX; V.FXhXlXlXlXlXlXlXlXlXzXiXjXzXlXlX",
"lXlXlXzXuXhXzXlXlXbXS.i.pXvXkXlXzXuXhXzXjXmXL.u.4.V.bXkXzXcX; V.FXhXlXlXlXlXlXlXlXlXzXiXjXzXlXlX",
"lXlXlXzXrXgXzXlXlXkXSXHXxXkXlXlXzXrXgXlXlXjXCXHXKXAXkXkXlXxX; V.FXhXlXlXlXlXlXlXlXlXzXyXjXzXlXlX",
"lXlXlXxXzXxXlXlXlXlXhXgXkXlXlXkXzXzXcXzXzXzXkXhXhXjXzXzXxXvX: C.DXgXlXlXlXlXlXlXlXkXxXlXkXlXlXlX",
"lXlXzXjXP.5XbXlXzXzXzXzXzXzXzXzXlXK.1XbXlXzXzXzXzXzXzXlXxXlX: B.FXjXzXzXzXzXzXzXzXxXjX~.dXzXlXlX",
"lXhXAXR.  E vXeXyXyXyXyXyXyXrXzXK.  + ; = = = = = = = = - -   L xXeXyXyXyXyXyXyXrXzXK.& gXxXkXlX",
"lXkXbX4XU D.nXfXhXhXhXhXhXhXfXxX2X+ #.Y.Z.S.S.S.S.S.S.S.D.S.T N.nXfXhXhXhXhXhXhXgXcX>Xh pXcXkXlX",
"lXlXkXbXvXmXzXzXzXzXzXzXzXzXlXzXbX; Y.UXCXDXDXDXDXDXDXDXDXFXcXmXzXzXzXlXhXhXlXzXlXxXkXZ tXcXkXlX",
"lXlXlXlXrXfXzXlXlXlXlXlXlXlXkXzXlX= Z.CXdXhXhXhXhXhXhXhXhXjXrXfXzXlXlXzXGXGXzXlXlXxXdXB yXcXkXlX",
"lXlXlXzXuXhXzXlXlXlXlXlXlXlXlXzXzX= S.DXhXlXlXlXlXlXlXlXlXzXuXhXzXkXxXgXg.d.gXxXkXxXfXB tXcXkXlX",
"lXlXlXzXuXhXzXlXlXlXlXlXlXlXlXzXzX= S.DXhXlXlXlXlXlXlXlXlXzXuXhXzXhXDX,.I T @.DXgXxXfXB tXcXkXlX",
"lXlXlXzXuXhXzXlXlXlXlXlXlXlXlXzXzX= S.DXhXlXlXlXlXlXlXlXlXzXuXhXzXkXcX-XUX1.] HXgXxXfXB tXcXkXlX",
"lXlXlXzXuXhXzXlXlXlXlXlXlXlXlXzXzX= S.DXhXlXlXlXlXlXlXlXlXzXuXhXzXlXhXUXn.U mXxXkXxXfXB tXcXkXlX",
"lXlXlXzXuXhXzXlXlXlXlXlXlXlXlXzXzX= S.DXhXlXlXlXlXlXlXlXlXzXuXhXzXhXZXf.  *.U.xXkXzXfXB tXcXkXlX",
"lXlXlXzXuXhXzXlXlXlXlXlXlXlXlXzXzX= S.DXhXlXlXlXlXlXlXlXlXzXuXhXzXjXMXK.w.1.n.bXkXzXfXB tXcXkXlX",
"lXlXlXzXrXgXlXkXlXlXlXlXlXlXkXzXlX= S.DXhXlXlXlXlXlXlXlXlXzXrXgXlXlXjXCXHXKXSXkXkXzXdXB tXcXkXlX",
"lXlXlXzXzXxXlXzXzXzXzXzXzXzXlXxXxX- D.DXhXlXlXlXlXlXlXlXkXzXzXxXzXzXzXjXgXgXjXzXlXxXgXV tXcXkXlX",
"lXlXzXkXI.5XmXxXcXcXcXcXcXcXxXvXlX- S.FXjXzXzXzXzXzXzXzXzXzXU.3XMXxXcXcXcXcXcXcXxXbXfXV yXcXkXlX",
"lXhXZXY.  # > ; ; ; ; ; ; ; ; : :   T cXeXuXuXuXuXuXuXrXlXU.  @ > ; ; ; ; ; ; ; ; > @ : fXxXkXlX",
"lXjXnX<Xo #.K.N.V.V.V.V.V.V.V.C.B.L N.nXfXhXhXhXhXhXhXfXxX3X@ ..L.N.V.V.V.V.V.V.V.C.b.o.sXxXlXlX",
"lXlXzXcX= Q.UXZXFXDXZXZXDXFXDXDXFXcXmXzXzXzXzXzXzXzXzXzXzXMX> L.UXZXFXFXFXFXFXFXFXDXGXxXjXzXlXlX",
"lXkXxXhX& F.CXdXgXjXAXSXjXgXhXgXjXrXfXzXlXlXlXlXlXlXlXkXzXxX; N.ZXsXhXhXhXhXhXhXhXgXhXtXjXzXlXlX",
"lXkXxXjX& J.SXgXzXfXk.d.fXxXkXlXzXuXhXzXlXlXlXlXlXlXlXlXzXcX; V.FXhXlXlXlXlXlXlXlXlXzXiXjXzXlXlX",
"lXkXxXjX& J.SXdXAXr.' ~ [ DXhXlXzXuXhXzXlXlXlXlXlXlXlXlXzXcX; V.FXhXlXlXlXlXlXlXlXlXzXiXjXzXlXlX",
"lXkXxXjX& J.SXhXlXsX*XH i.SXhXlXzXuXhXzXlXlXlXlXlXlXlXlXzXcX; V.FXhXlXlXlXlXlXlXlXlXzXiXjXzXlXlX",
"lXkXxXjX& J.SXhXkXfX>X'  .mXkXlXzXuXhXzXlXlXlXlXlXlXlXlXzXcX; V.FXhXlXlXlXlXlXlXlXlXzXiXjXzXlXlX",
"lXkXxXjX& J.SXdXZX:.[ X.N mXkXkXzXuXhXzXlXlXlXlXlXlXlXlXzXcX; V.FXhXlXlXlXlXlXlXlXlXzXiXjXzXlXlX",
"lXkXxXjX& J.SXgXcXtXy.9.tXvXkXlXzXuXhXzXlXlXlXlXlXlXlXlXzXcX; V.FXhXlXlXlXlXlXlXlXlXzXiXjXzXlXlX",
"lXkXxXjX& H.SXhXkXcXGXHXcXkXlXkXlXrXgXlXlXlXlXlXlXlXlXkXlXxX; V.FXhXlXlXlXlXlXlXlXlXzXyXjXzXlXlX",
"lXkXzXkX= L.FXjXxXzXjXjXzXxXxXzXcXxXcXzXxXxXxXxXxXxXxXzXxXbX> C.DXhXlXlXlXlXlXlXlXkXxXlXkXlXlXlX",
"lXkXxXfXo V.BXaXfXfXfXfXfXfXfXgXsXK.>XkXdXfXfXfXfXfXfXdXgXfX@ b.GXhXzXzXzXzXzXzXzXxXhX~.fXzXlXlX",
"lXjXVX).> l Z B B B B B B B B V N , f Z B B B B B B B B V V : o.zXtXiXiXiXiXiXiXtXlX~.i fXxXkXlX",
"lXlXzXgXdXpXtXyXtXtXtXtXtXtXyXtXyXhXpXtXyXtXtXtXtXtXtXtXtXyXfXsXjXjXjXjXjXjXjXjXjXkXfXfXzXlXlXlX",
"lXlXlXzXxXcXcXcXcXcXcXcXcXcXcXcXcXxXcXcXcXcXcXcXcXcXcXcXcXcXxXxXlXzXzXzXzXzXzXzXzXlXzXxXlXlXlXlX",
"lXlXlXlXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXlXlXlXlXlXlXlXlXlXlXlXlXkXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX"
};

/* XPM */
static const char *const xpm_icon_5[] = {
/* columns rows colors chars-per-pixel */
"44 44 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXkXlXlXlXlXlXlXlXlXlXlXlXkXlXlXlXlXlXlXlXlXlXlXlXkXlXlXlXlXlXlXlXlXlXlXkXlXlXlXlX",
"lXlXlXkXzXzXkXlXlXlXlXlXlXlXlXkXxXkXlXlXlXlXlXlXlXlXkXlXxXgXgXgXgXgXgXgXgXgXgXzXzXlXlXlX",
"lXlXkXcXjXlXcXzXzXzXzXzXzXxXzXbXfXvXzXzXzXzXzXzXzXzXcXlXgXLXJXJXJXJXJXJXJXJXKXkXjXzXlXlX",
"lXkXzXjXy P xXiXaXaXaXaXaXuXcXW.  C.bXuXaXaXaXaXaXuXvXO.2 P H H H H H H H J G p qXvXkXlX",
"lXlXzXzXI $.vXaXdXdXdXdXdXaXcX|.p W.bXaXdXdXdXdXdXpXBXO.% 8.,.<.<.<.<.<.<.<.,.T eXvXkXlX",
"lXlXkXcXcXvXcXzXzXlXjXjXzXzXlXnXzXnXlXzXzXjXjXlXzXjXKXz.m UXGXKXKXKXKXKXKXKXPXcXhXzXlXlX",
"lXlXlXzXiXsXzXlXkXmXGXZXkXlXlXkXtXjXlXkXxXDXDXnXkXgXHXd.l CXsXgXgXgXgXgXgXgXhXiXhXzXlXlX",
"lXlXlXzXaXdXzXjXMX+X~ x.nXkXlXlXuXkXlXxXsX3.[ ,XbXfXHXf.z DXhXlXlXlXlXlXlXlXzXaXhXzXlXlX",
"lXlXlXzXaXfXzXlXzXm M.V Q.CXjXlXuXkXjXVX).2.e.R HXsXHXf.z DXhXlXlXlXlXlXlXlXzXaXhXzXlXlX",
"lXlXlXzXaXfXlXnX2Xc UX>.r.JXgXlXuXkXlXjXbXIX| C.ZXdXHXf.z DXhXlXlXlXlXlXlXlXzXaXhXzXlXlX",
"lXlXlXzXaXfXzXzXhXz T.A P.AXhXlXuXkXkXzXgXF ' fXkXgXHXf.z DXhXlXlXlXlXlXlXlXzXaXhXzXlXlX",
"lXlXlXzXaXfXzXjXNX(.G w.vXkXlXlXuXkXjXBX|.v ! w.mXfXHXf.z DXhXlXlXlXlXlXlXlXzXaXhXzXlXlX",
"lXlXlXzXiXsXzXlXkXMXZXCXzXlXlXkXrXjXkXjXxXZXmXbXkXgXGXf.z DXhXlXlXlXlXlXlXlXzXuXhXzXlXlX",
"lXlXlXzXcXcXlXlXlXjXjXhXlXlXlXxXcXbXvXvXvXxXxXcXvXzXPXk.x FXhXlXlXlXlXlXlXlXxXxXjXlXlXlX",
"lXlXkXbXW.|.nXkXlXlXlXlXlXkXxXuXC.2XtXeXeXeXeXeXeX0XxXe.k CXjXlXlXlXlXlXlXlXlXQ.uXcXlXlX",
"lXkXxXgX  i lXtXuXuXuXuXuXrXxXC.  - < > > > > > > > 1     >XdXyXuXuXuXuXyXfX*X< 9XbXkXlX",
"lXlXkXvXZ.W.nXjXkXkXkXkXkXjXbX2X- W.3X-X;X;X;X;X;X-X2XE.b.lXkXkXkXjXjXkXkXkXzXP 3XnXkXlX",
"lXlXlXzXbXbXlXlXlXlXlXlXlXkXvXtX< 3XDXnXMXMXMXMXMXMXMXmXbXxXlXlXkXbXxXkXzXjXBX~ <XmXkXlX",
"lXlXlXzXuXaXzXlXlXlXlXlXlXkXvXeX> -XnXhXjXjXjXjXjXjXlXsXyXzXlXkXBX0XdXnXkXjXMXQ 1XmXkXlX",
"lXlXlXzXaXfXzXlXlXlXlXlXlXkXvXeX> ;XMXjXlXlXlXlXlXlXzXgXpXzXhXCXc.b R 1XnXhXNXQ 1XmXkXlX",
"lXlXlXzXaXfXzXlXlXlXlXlXlXkXvXeX> ;XMXjXlXlXlXlXlXlXzXgXpXzXjXnXL.sXD G.DXfXNXQ 1XmXkXlX",
"lXlXlXzXaXfXzXlXlXlXlXlXlXkXvXeX> ;XMXjXlXlXlXlXlXlXzXgXpXzXlXhXUXv.-.VXjXjXNXQ 1XmXkXlX",
"lXlXlXzXaXfXzXlXlXlXlXlXlXkXvXeX> ;XMXjXlXlXlXlXlXlXzXgXpXzXhXAXt.  q.:XbXhXNXQ 1XmXkXlX",
"lXlXlXzXaXfXzXlXlXlXlXlXlXkXvXeX> ;XMXjXlXlXlXlXlXlXzXgXpXzXjXnXW.c.i.&XnXhXNXQ 1XmXkXlX",
"lXlXlXzXiXpXjXgXgXgXgXgXgXfXzX0X> -XMXjXlXlXlXlXlXlXzXdXyXjXgXfXNXAXDXcXfXfXbXW 1XmXkXlX",
"lXlXkXcXvXBXJXHXHXHXHXHXHXGXPXxX1 2XMXlXzXzXzXzXzXzXzXnXnXJXGXHXFXSXSXGXHXFXUX) <XmXkXlX",
"lXlXlXzXo.o.z.d.f.f.f.f.f.f.k.e.  R.mXsXgXgXgXgXgXsXnXg.E l.f.f.f.f.f.f.f.f.h.l 7XbXkXlX",
"lXkXxXgX2 % m l z z z z z z x k   v.vXyXpXpXpXpXpXyXbXE   n z z z z z z z x j e wXvXkXlX",
"lXlXgXLXP 8.UXCXDXSXZXAXDXDXFXCX>XlXxXzXzXzXzXzXzXjXJXl.n UXAXDXDXDXDXDXDXDXGX8XsXxXlXlX",
"lXlXgXJXH ,.GXdXgXcXFXbXgXjXhXjXdXkXlXlXlXlXlXlXlXgXHXf.z AXdXhXhXhXhXhXhXhXjXfXjXzXlXlX",
"lXlXgXJXH <.KXfXvX1Xp.$XnXkXlXlXyXkXlXlXlXlXlXlXlXgXHXf.z DXhXlXlXlXlXlXlXlXzXpXhXzXlXlX",
"lXlXgXJXH <.KXgXcXo.2.n 0XvXkXlXuXkXlXlXlXlXlXlXlXgXHXf.z DXhXlXlXlXlXlXlXlXzXaXhXzXlXlX",
"lXlXgXJXH <.KXgXkXBXi.K hXzXkXlXuXkXlXlXlXlXlXlXlXgXHXf.z DXhXlXlXlXlXlXlXlXzXaXhXzXlXlX",
"lXlXgXJXH <.KXhXfX/.].k  XBXjXlXuXkXlXlXlXlXlXlXlXgXHXf.z DXhXlXlXlXlXlXlXlXzXaXhXzXlXlX",
"lXlXgXJXH <.KXgXcXO.K Q 9XvXkXlXuXkXlXlXlXlXlXlXlXgXHXf.z DXhXlXlXlXlXlXlXlXzXaXhXzXlXlX",
"lXlXgXJXH <.KXgXxXmX3XlXnXkXlXlXyXkXlXlXlXlXlXlXlXgXHXf.z DXhXlXlXlXlXlXlXlXzXpXhXzXlXlX",
"lXlXgXJXJ <.HXdXjXhXvXjXhXjXjXkXfXkXjXjXjXjXjXjXjXfXFXf.x DXhXlXlXlXlXlXlXlXzXgXjXzXlXlX",
"lXlXgXKXF <.UXbXNXNXMXMXNXNXBXbX*XxXBXMXNXNXNXNXNXbXUXh.j GXjXzXzXzXzXzXzXzXvX3XsXxXlXlX",
"lXkXzXhXj a ) W Q Q Q Q Q Q ! Y 7 L ~ Q Q Q Q Q Q W ) l e 8XdXiXpXpXpXpXiXfX3XM 0XbXkXlX",
"lXlXzXhXwX8X<X1X1X1X1X1X1X1X<X2XqX3X<X1X1X1X1X1X1X1X<X7XwXsXjXhXhXhXhXhXhXjXsX0XlXlXlXlX",
"lXlXlXzXvXbXmXmXmXmXmXmXmXmXmXmXbXnXmXmXmXmXmXmXmXmXmXbXvXxXzXzXzXzXzXzXzXzXxXbXlXlXlXlX",
"lXlXlXlXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXlXlXlXlXlXlXlXlXlXlXkXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX"
};

/* XPM */
static const char *const xpm_icon_6[] = {
/* columns rows colors chars-per-pixel */
"32 32 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"lXlXcXlXlXlXlXlXlXlXkXxXzXlXlXlXlXlXlXkXxXlXhXjXjXjXjXjXhXzXzXlX",
"lXzXdXlXzXzXzXzXzXzXxXfXkXzXzXzXzXzXlXcXaXcXCXBXVXVXVXBXCXkXhXzX",
"xXdX2 N.nXuXaXaXpXaXsX0 f.mXyXaXaXpXpXfXV d J S D D D D F a 7XnX",
"lXlXB.>XvXhXgXfXhXjXzXZ.*XvXhXfXfXhXfXZXJ q.;X}. X X X XoXL.yXcX",
"lXzXnXvXlXlXCXSXvXlXzXnXvXlXzXAXAXvXgXJX( R.UXMXVXVXVXBXCXmXkXlX",
"lXzXuXhXkXmXM.;.yXcXlXiXgXzXzXa.6.tXkXFX~ G.AXgXjXjXjXjXkXpXhXzX",
"lXzXaXhXNX#XO.k.q.FXjXaXhXcXeX`.| K.VXDX~ H.DXhXlXlXlXlXzXdXhXzX",
"lXzXaXhXVX|.*.P.<.GXhXaXjXkXBXE.&.lXhXGX~ H.DXhXlXlXlXlXzXdXhXzX",
"lXzXaXjXlXcXe.` 2XnXlXaXhXxXdXL ] %XcXFX~ H.DXhXlXlXlXlXzXdXhXzX",
"lXlXsXjXlXzXMXnXbXkXzXsXgXjXgXvXgXfXdXSX~ H.SXhXlXlXlXlXzXdXhXzX",
"kXxXdXzXzXzXkXlXlXzXvXpXmXFXSXAXSXDXCXUX] P.FXjXzXzXzXzXcXdXgXzX",
"xXfX0 C.nXuXaXaXaXsXpXB m R I I I I P ~ . 9.NXyXaXaXaXpXgX] 2XmX",
"zXkXf.*XvXgXhXhXhXgXmXm w.}.Q.~.~.~.!.).j..XbXgXfXdXhXfXMXO.-XNX",
"lXzXMXvXlXzXzXzXzXhXFXR }.UXVXZXZXZXZXAXNXvXlXzXAXDXvXhXSX9.=XNX",
"lXzXuXhXzXlXlXlXlXhXSXI Q.VXgXjXjXjXjXkXiXfXlXvXm.0.tXlXZX5.=XNX",
"lXzXaXjXlXlXlXlXlXhXSXI ~.ZXjXlXlXlXlXzXsXgXcXuXE.X.M.AXVX5.=XNX",
"lXzXaXjXlXlXlXlXlXhXSXI ~.ZXjXlXlXlXlXzXsXhXkXNX+X%.sXkXZX5.=XNX",
"lXzXaXjXlXlXlXlXlXhXSXI ~.ZXjXlXlXlXlXzXsXgXzXzXP E +XvXCX5.=XNX",
"lXlXaXfXhXgXgXgXgXsXCXP !.ZXjXlXlXlXlXzXsXdXhXfXjXiXpXdXBX4.=XNX",
"kXcXfXZXJXGXGXGXGXSXUX~ ).AXkXzXzXzXzXcXgXBXJXGXHXJXHXSXUXt.*XNX",
"xXsXV J ( ~ ~ ~ ~ ~ ] . j.MXuXaXaXaXaXgXT Z _ ~ ~ ~ ~ ~ ) p 3XmX",
"lXcXd q.R.G.F.D.H.H.P.0..XvXfXhXhXgXdXBXZ $.Q.F.H.H.H.H.K.j.tXvX",
"hXCXJ ;XUXAXUXUXGXSXFXNXbXlXzXzXzXzXhXJX_ Q.UXZXDXDXDXSXFXNXjXlX",
"jXBXS }.MXgXN.k.uXkXjXuXgXzXlXlXlXlXgXGX~ F.ZXfXhXhXhXhXjXpXhXzX",
"jXVXD  XCXiXV.K }.VXkXaXjXzXlXlXlXlXgXGX~ H.DXhXlXlXlXlXzXdXhXzX",
"jXVXD  XVXdX3X^ T.CXkXaXjXzXlXlXlXlXgXGX~ H.DXhXlXlXlXlXzXdXhXzX",
"jXVXD  XZXrX-.' {.NXkXaXhXzXlXlXlXlXgXGX~ H.DXhXlXlXlXlXzXdXhXzX",
"jXBXD }.mXkXdXuXbXfXjXaXfXhXhXhXhXhXsXSX~ H.SXhXlXlXlXlXzXsXhXzX",
"hXCXD *XUXVXSXDXCXAXDXjXMXSXAXAXAXAXBXUX) K.FXjXzXzXzXzXcXjXhXzX",
"zXhXk | q.5.5.5.5.5.9.R #.9.5.5.5.5.4.t.p j.NXpXdXdXsXsXjX4.6XnX",
"zXhX9X:X=X=X=X=X=X=X=X,X;X=X=X=X=X=X=X*X3XtXjXhXhXhXhXhXhX6XhXzX",
"lXzXbXMXNXNXNXNXNXNXNXMXMXNXNXNXNXNXNXNXmXvXlXzXzXzXzXzXzXnXzXlX"
};

/* XPM */
static const char *const xpm_icon_7[] = {
/* columns rows colors chars-per-pixel */
"24 24 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"hX3XjXlXkXlXlXlX3XdXzXkXlXkXzX9XqXuXyXyXyXuX8XsX",
"3XF 0XgXiXiXdXsXH .XxXuXiXiXbX) Z ..' ] ' } L *X",
"kXqXlXkXVXVXkXzXqXhXlXVXCXhXGX-.^.CXhXkXkXzXiXsX",
"lXgXkXmX~.E.mXkXgXkXbX~.I.hXFX>.(.ZXjXlXlXzXjXhX",
"lXaXnX1X5.4.-XMXaXkXlX{.@.7XLX;./.ZXjXlXlXzXgXgX",
"lXaXbX7X1.<.2XmXaXhXVXg.-.lXFX:./.ZXjXlXlXzXgXgX",
"lXdXkXmXOX.XmXkXdXhXzXE.Y.yXFX;./.ZXjXlXlXzXgXhX",
"lXsXzXkXNXBXkXxXsXnXMXJXKXnXUX1.(.ZXkXzXlXxXgXgX",
"3XH qXgXpXpXdXsXP Q | ' ' ] +.q A.mXuXpXiXnXd.XX",
"dXXXhXkXkXkXjXnXQ [.9X2X3X2X5XOXwXzXvXBXgXAXM.].",
"zXxXzXlXlXlXkXNX| 9XCXbXnXnXmXcXlXbX1X[.jXAXG.{.",
"kXaXlXlXlXlXjXMX[ 2XbXjXkXkXlXsXgXbX}.X..XIXA.{.",
"lXsXlXlXlXlXjXMX{ 3XnXkXlXlXzXdXgXMX|.+.rXFXS.{.",
"kXpXhXgXgXgXfXvX] 2XnXkXlXlXzXaXaXnXH.e.6XAXZ.{.",
"zXnXGXFXFXFXDXUX+.5XmXlXzXzXxXcXAXFXUXUXDXUXR.[.",
"9X( -.>.;.;.:.1.q OXxXaXdXaXcX5.[ 2.-.=.-.4.P XX",
"qXZ ^.(.].`./.).S.qXlXhXhXdXAX[ r..X^././.).U.wX",
"uX..CXSXiXkXSXZXmXxXlXzXzXhXGX1..XUXVXZXZXAXNXkX",
"yX' hXkXa.*.zXjXiXjXlXlXlXgXFX:.^.VXgXjXjXkXdXgX",
"yX] kXlX(.) aXcXsXkXlXlXlXgXFX:./.ZXjXlXlXzXgXgX",
"yX' hXgXg.j.gXjXiXgXhXhXhXdXAX;./.ZXjXlXlXzXdXgX",
"uX| ZXDXPXIXDXAXMXZXAXAXAXBXUX4.).AXkXzXzXzXvXlX",
"0XF A.D.A.A.S.F.w.B.F.D.D.A.R.P U.NXsXgXdXvX!.<X",
"dX$X[.{.{.{.{.{.[.{.{.{.{.{.[.XXwXjXgXgXgXkX<XuX"
};

/* XPM */
static const char *const xpm_icon_8[] = {
/* columns rows colors chars-per-pixel */
"16 16 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"hXwXlXgXzXtXaXlXjXkXeX0XqXqX9XhX",
"wXD.zXzXmXR.+XMXcXtX#.x.x.x.h.rX",
"lXzXkXOXaXxXvX5X>XdXk.AXmXNXbXkX",
"hXkX7XO.XXxXxXQ.f.jXf.cXhXkXdXhX",
"zXnXjX].uXbXHX1X,XBXx.bXlXlXcXlX",
"tXE.zXcXnX{.b.R.R.n.:.cXdXzX&XwX",
"aX@XzXaXCXn.M.*X X X@XlXfXBX!.-X",
"zXcXlXjXZXK.=XJXMXBXvXlXm.qX+X>X",
"kXhXcXlXDXK.XXMXhXjXhXzXe.tXOX:X",
"kXyXiXrXcXM. XBXjXkXuXuX3XhX~.:X",
"eX#.l.v.v.:.@XcXgXyX#.j.z.l.t.0X",
"0Xc.CX;XzXvXlXlXcXiXj.BXcXbXcXkX",
"0Xx.bX;.+XxXdXjXlXyXj.vXhXkXfXhX",
"0Xn.DX`.fXBXmXNXCXkXk.bXkXzXzXlX",
"wX0.XXOXXXQ.(.}.|.~.t.cXfXzX2XyX",
"kX5X:X;X>X:X:X>X>X:X0XkXhXlXyXfX"
};

const char *const *const xpm_icons[] = {
    xpm_icon_0,
    xpm_icon_1,
    xpm_icon_2,
    xpm_icon_3,
    xpm_icon_4,
    xpm_icon_5,
    xpm_icon_6,
    xpm_icon_7,
    xpm_icon_8,
};
const int n_xpm_icons = 9;
