#include "gtk.h"
/* XPM */
static const char *const xpm_icon_0[] = {
/* columns rows colors chars-per-pixel */
"128 128 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXeXeXeXeXeXeXeXeXeXeXeXeXeXeXwXqXqXqXqXqXqXqXqXwXeXeXeXeXeXeXwXqXqXqXqXqXqXqXwXeXeXeXeXeXeXwXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X3X-X&X&X&X&X&X&X&X&X&X&X&X&X-X4X0XqXqXqXqXqXqXqX3X=X&X&X&X%X-X2X8XwXrXrXeXwX7X1X=X%X&X&X&X-X5XqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX<Xc.D e y 5 3 2 1 1 2 3 5 y e F b.<XwXqXqXqXqXqX3Xd.g 8 2 3 8 G t.oXhXBXFXMXsX'.5.N 6 3 2 8 x C.6XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X;X`.F.t.D e @     @ e F y.F.`.;X9XqXqXqXqXqX6X7.q @   . n D.>XaXBXJXPXUXPXHXMXuX-Xz.d     # i l.9XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXeXrXhX2X&XV     V &X2XhXrXeXeXqXqXqXqXqXwXu.p .   > <.XXNXIXUXUXUXUXUXUXUXUXUXIXcX_.O.#   X x m.tXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XaXUXUXUX~     ~ UXUXUXaX0XqXqXqXqXqX0XF.m +   0 i.9XLXUXUXUXUXUXUXUXUXUXUXUXUXUXGX,X4.>   @ I W.eXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqX0X>XQ ;   4 5.jXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXHXuXo.:   1 +.5XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXwX5XN.w   O } 8XKXUXUXUXUXDXyX#X}. X;XkXKXUXUXUXUXUXGX:XK X   l P.9XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXwX=XF &   b R.IXUXUXUXUXDXoX#.e + & z u.qXKXUXUXUXUXUXJXm.p   ; ^ 5XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqX9X).6     t.BXUXUXUXUXUX.XF . ; 7 4 % O :.tXUXUXUXUXUXUXnX+.    g +X9XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXeX>Xu.3   p Y.PXUXUXUXUXUX*.# e u.sX<XR % 5 R.UXUXUXUXUXUXKXz.9   t c.4XwXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXrXOXI $   } 2XUXUXUXUXUXSX=.t *.qXUXSXH.9   M.UXUXUXUXUXUXUXOXI   ; ' ,XwXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXtX{.s o   l.mXUXUXUXUXUXKX9X$XrXHXUXFXS.8 + D.UXUXUXUXUXUXUXkX<.  @ M -XeXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXtX`.4 .   I.PXUXUXUXUXUXUXUXUXUXUXIXyXR @ A @XUXUXUXUXUXUXUXJXa.  X p *XeXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXtX).=     ^.UXUXUXUXUXUXUXUXUXUXIXiX6.* 4 M.bXUXUXUXUXUXUXUXUXz.    8 &XeXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXtX).*     (.UXUXUXUXUXUXUXUXUXPXrXq.* 2 3.sXUXUXUXUXUXUXUXUXUXc.    6 &XeXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXtX`.2 .   Y.IXUXUXUXUXUXUXUXPXeX6.% 2 &.fXUXUXUXUXUXUXUXUXUXLXf.  . u *XeXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXtX[.y o   N.ZXUXUXUXUXUXUXPXwX-.# 2 #.lXIXUXUXUXUXUXUXUXUXUXMX8.  + v -XeXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXrX.XZ @   =.wXUXUXUXUXUXUX9X..# ; [ rXNXVXVXGXIXUXUXUXUXUXUX:X/   & R :XwXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXeX-X2.1   v ).UXUXUXUXUXPX*.&   o 1 k l l d z.mXUXUXUXUXUXLXZ.p   q u.3XwXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXwX6XY.4   @ k.GXUXUXUXUXLXL                 ;.xXUXUXUXUXUXAX<.X   a ].8XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXwX$Xx @   U @XUXUXUXUXPX`.c.m.m.m.m.m.m.v.&XSXUXUXUXUXIX/.m   $ K 2XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXwX3Xq.5   - ,.hXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXKXtX_ $   i j.9XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqX0X$XA #   s m.mXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXLXkXy.8   % / ,XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXeXeXeXeXeXeXeXwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqX8Xa.d o   l !.jXLXUXUXUXUXUXUXUXUXUXUXUXUXUXJXaXF.0   O M M.wXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X7X1X;X=X*X&X*X=X;X<X6X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqX6X%.4 .   r b.pXGXIXUXUXUXUXUXUXUXUXIXIXDX0Xi.1 . o q w.qXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X8X:XU.>.F h w 5 q f Z $.G.-X7X0XqXqXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqX=X .& X X 1 ~ @XcXSXLXIXUXUXUXIXJXZXzX(.K ; X O : 3.,XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX0X+X>.M u 4 # o       o # 2 u v @.{.9XwXqXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXOX-.3 +   & u ! j.*XvXFXIXSXlXoXy.I w $   + 9 e.&XwXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrX;Xw.H 7 o                       . 3 C 4.@XtXqXqXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXwX#Xi.x X     $ w m L T W Y J v 8 +     @ C c.-XwXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXwX{.#.p X     & h ^ -.3.8.4.:._ c ;       w { !.eXwXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXrX,XZ.%.p X                       @ z <.K.5XrXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX/.' : X   > S &.x.].3XqXyXwX5X XM.:.L 2     & T I.wXqXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXeXeX=Xh.O.~ z , @       # 2 M / %.m.<XrXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrX{.' O . . u | !.5X0XwXwXqXqXqXwXwXqX6X[.$.g O   X L ^.eXqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXwXtX8X>XK.0.Y     o _ r.W.1X0XtXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX;X@.; . O f h.=XqXqXqXqXqXqXqXqXqXqXqXqXqX<Xm.n #   * ~ %XqXqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXZXUXUX[.    # 8XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0Xw.i X . f K.:X9XqXqXqXqXqXqXqXqXqXqXqXqXqX0X<XQ.V . . q >.5XqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX8XXX5XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X+XH X   i j.:X9XqXqXqXqXqXqX8X6X5X6X9XqXqXqXqX0X2XM.c   X l {.8XqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXwX,X$.OXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X>.7   : | -X9XqXqXqXqXqXeX<Xh.M s b V.yXqXqXqXqX0X<X,.2   , X.<XwXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX*Xh Q.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X:XM o   S !.qXqXqXqXqXqXqX'.+.t     & f.iXqXqXqXqXqXwX[./   o p OXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&Xw d.,XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XI.i   & &.5XqXqXqXqXqXwX,Xu., , i # ; g.iXqXqXqXqXqXwX9Xt.;   3 Z.3XwXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X9 / {.rXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX1X:.5   h x.0XqXqXqXqXqXeX#X4.c 1.' : - g.iXqXqXqXqXqXqXwXG.M   ; O.%XeXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX%X3 p T.yXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX;XD #   ^ ].wXqXqXqXqXqXwX-XK.V.<X>., = g.iXqXqXqXqXqXqXeXoX@.  O b |.rXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X2 < J.uXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX=Xg o   -.3XwXqXqXqXqXqXqX0XqXeX1X:., = g.iXqXqXqXqXqXqXwX5Xp.  X w ].tXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X2 @ S.uXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX*Xw     4.wXqXqXqXqXqXqXqXqXqXwX1X:., = g.iXqXqXqXqXqXqXqXwXx.  . , _.tXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X2   C.iXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX&X5     8.yXqXqXqXqXqXqXqXqXqXwX1X:., = g.iXqXqXqXqXqXqXqXyXM.    & (.tXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X2 + A.uXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX*Xq     4.wXqXqXqXqXqXqXqXqXqXwX1X:., = g.iXqXqXqXqXqXqXqXwXx.  . > _.tXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X2 , H.uXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX=Xf o   :.5XwXqXqXqXqXqXqXqXqXwX1X:., = g.iXqXqXqXqXqXqXwX6Xs.  X q ].tXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX%X2 u Y.yXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX;XZ #   _  XwXqXqXqXqXqXqXqXqXwX1X:., = g.iXqXqXqXqXqXqXwX+X&.  O x |.rXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X8 R ].tXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX<X$.2   c M.qXqXqXqXqXqXqXqXqXwX1X:., = g.iXqXqXqXqXqXqXwXK.Z   - { $XeXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&Xe u.:XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX6XG.u   ; ;.5XqXqXqXqXqXqXqXqXwX1X=.- $ f.iXqXqXqXqXqXwX9Xi.<   4 m.2XwXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX*Xs T.qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X-Xv .   K ].qXqXqXqXqXqXqXqXwX3Xy.m v V.yXqXqXqXqXqXwXoX{   X r .XqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXwX>X} XXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X@.3   2 $.<X0XqXqXqXqXqXqXqXqX8X6X5X9XqXqXqXqXqXqX4X6.7   > _ >XwXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X].3XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X{.C     f m.<X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX4XD.N     d /.8XqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX7X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X4.w   O n W.2X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX4X_.K @   6 #.3XqXqXqXqXqX0XpXIXUXUXQ     Q UXUXIXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX@X{ *   # V m.,XwXwXqXqXqXqXqXqXqXqXqXwXwX4XD.L *   % Y  X0XqXqXqXqXqX0XaXUXUXUX~     ~ UXUXUXaX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX!.T X   . x >.'.8XwXeXwXqXqXqXwXwXwX8XoX7.B @   . Z K.rXqXqXqXqXwXrXrX8X<XJ.k.s     s k.J.<X8XrXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXI.P = .   2 ~ r.G.XX5XwXyXeX7X@XK.u.] 7     % C B.qXqXqXqXqXrX0X{.k.<.[ s > o     o > d { <.l.}.0XrXqXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXwX/.^ q X     ; M +.p.x.M.c.s.*.A <       5 U K.qXwXqXqXqXeX+XZ.{ 6   X                 X   6 } Z.@XeXqXqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX%X>.l , o                       X : d #. XrXqXqXqXqX8X X*.q .     * 8 a h l l h a 8 *     . w =..X8XqXqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX5X{.o.p 3 : O X .   . X O - 4 r ` /.3X0XqXqXqXqX7X'.K , o X & r _ C.3XlXCXCXlX3XV.) e & X X > K ].7XqXqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X<XOXZ.O.v w , & : 0 z [ n..X>X8XqXqXqXqXqX8X'.c = X   0 <.4XhXCXJXPXUXUXPXJXCXhX4X,.0   X * v ].8XqXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX3X%X|.]._.(._.'.|.$X2XqXwXqXqXqXqXqXeX XJ = .   L R.hXKXIXUXUXUXUXUXUXUXUXIXKXgXT.K   . * K .XeXqXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXwXeXrXrXrXrXrXeXwXqX0X0X0X0X0X0XeXOX*.> X   { .XMXUXUXUXUXUXUXUXUXUXUXUXUXUXUXMX X{   X , -.@XrXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXeXbXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXbXR.w o   L .XDXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXDX.XK   o w A.qXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXtXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXuX>.. X 0 R.MXUXUXUXUXUXmX@XL.V.F.{.sXUXUXUXUXUXUXMXT.9 X . } }.rXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXtXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXR.0   & <.gXUXUXUXUXLXhXn.D =   + l r.rXLXUXUXUXUXUXgX,.&   6 l.tXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X3X<.    r 4XKXUXUXUXUXMXb.- @ t x i % : >.NXUXUXUXUXUXLX5Xe     >.3XwXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X8 + # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # O   * _ hXIXUXUXUXUXuXA 1 F |.AXOXT   q iXIXUXUXUXUXIXhX( * . R !.rXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X1                                                                                                           8 Z.CXUXUXUXUXUXbXC.v.$XBXUXzXp.  4 6XLXUXUXUXUXUXCXB.7   y f.yXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX%X,                                                                                                           a 4XJXUXUXUXUXUXLXMXmXFXUXKX}.P o B kXUXUXUXUXUXUXJX2Xp   ; 4.uXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX8X{.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.|.)     h zXIXUXUXUXUXUXUXUXUXmXw.L l r [ XXHXUXUXUXUXUXUXPXkXh   o *.uXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXyXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX8.    l CXUXUXUXUXUXUXUXUXIXlXW     + P ].JXUXUXUXUXUXUXUXVXl     O.uXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXyXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX9.    l CXUXUXUXUXUXUXUXUXUXSX~.s.=.z $ Y 1XHXUXUXUXUXUXUXVXl     O.uXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX8X{.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.|._ .   h lXPXUXUXUXUXIXAXfXfXDXUXUXAXv.  & g.nXUXUXUXUXUXPXkXh   O =.uXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX%X,                                                                                                           a 3XJXUXUXUXUXLXqX9.7.|.mXUXLXR.+ & 1.zXUXUXUXUXUXJX1Xp   : 5.yXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X1                                                                                                           8 V.CXUXUXUXUXUX,X8 # Z Y.rX'.(   < N.NXUXUXUXUXUXVXM.7   y g.tXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X8 + # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # O   * ) hXIXUXUXUXUXzX7.: # , 4 < % & J pXPXUXUXUXUXIXhX/ * X W ~.rXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X4X1.    e 2XKXUXUXUXUXLXgXn.G , . * v q.7XJXUXUXUXUXUXKX4Xe     >.3XwXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXtXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXE.q   & >.gXUXUXUXUXUXUXvX2XoX{. X:XgXIXUXUXUXUXUXUXgX;.& . 7 z.tXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXtXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXuX,.. X 0 Y.MXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXmXI.8 X . | |.rXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXeXbXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXnXR.w o   J  XDXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXSX}.G   o e S.qXqXqXqXqXqXqXCXUXUX'.    # 6XUXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XeX+X=., X   {  XmXUXUXUXUXUXUXUXUXUXUXUXUXUXUXmX}.]   X , :.@XrXqXqXqXqXqXqXAXUXUX{.    # 9XUXUXMX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX.XL * .   J U.gXKXIXUXUXUXUXUXUXUXUXIXKXgXI.H   . = P XXeXqXqXqXqXqXwX0XzXlXfXC.    @ ).dXcXjX0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X].v * X   9 :.3XhXVXJXPXUXUXPXJXVXhX2X;.8   X = b ].9XqXqXqXqXqX7X>X|.L.-.N 8     . e A 5.T.oX<X8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X].L , o X & e ( N.1XkXVXVXkX1XM./ w % X X , P ].7XqXqXqXqX0X3X!.( e 6 : O X       X + < 6 i X.[.5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X.X-.e .     * 7 p h l l h p 7 *     . e ;.XX9XqXqXqXqXeX{.O.s ; X                       o , l >.%XeXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX@XA.} 7   X                 X   8  .S.@XeXqXqXqXwXqXJ.I 5       5 R t.J.Q.).E.F.8.H ,     . 0 ~ ^.wXwXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXqX}.l.>.E t ; o     o ; y E >.z.|.qXrXqXqXqXqXqXN.V %     t :.D.%XjXJXUXUXUXGXdXXXV.o.6     * L I.eXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXtX3X!.d.4.*.O.O.*.4.f.!.3XtXrXqXqXqXqXqXeXH.V .   # P b.gXLXUXUXUXUXUXUXUXUXUXLXwXs.Z X   X T !.tXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXeXtXyXyXuXuXyXyXtXeXqX0X0X0X0X0X0X0X{.P $   ; ) OXFXUXUXUXUXUXUXUXUXUXUXUXUXUXVX).P $   & { @XwXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXeXiXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXiX0X+.5   # ) qXFXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXZX<XF +   e 5.9XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX0X6XSXUXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXJX1Xl     I +XFXIXUXUXUXPXSXmXcXcXnXAXLXUXUXUXUXIXZX).v     C }.0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXwX,X&.aXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXNX@.,   t b.FXIXUXUXUXUXgXZ.U i y G z.dXUXUXUXUXUXIXVX9.5   3 @.7XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX*Xg :XUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXaXa X   :.fXUXUXUXUXUXuX;.4 #     @ 4 :.9XUXUXUXUXUXUXtX(   . v -X0XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X4 B Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q W G #   6 D.LXUXUXUXUXUXj.6 + Q T.U.Q % , H.UXUXUXUXUXUXHXu.>   u G.6XwXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X1                                                                                                                                                                   E %XUXUXUXUXUXUXN.( ,..XUXKX!.w   p.JXUXUXUXUXUXUX(.V   2 %.<XwXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X1                                                                                                                                                                   t.kXUXUXUXUXUXUXyX:XyXHXUXZXK.7   c.PXUXUXUXUXUXUXuX$.  # A ;XeXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X2 . N ~ Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q ~ G @     J.JXUXUXUXUXUXUXUXUXUXdX(.V.D r ^ &XUXUXUXUXUXUXUXFXu.  o f =XeXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X2 #  XUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX0X< .   !.UXUXUXUXUXUXUXUXUXFXD..     A /.CXUXUXUXUXUXUXUXUXk.    q *XeXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X2   [.UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX7X*     ).UXUXUXUXUXUXUXUXUXHX[.F j < # U OXUXUXUXUXUXUXUXUXc.    5 &XeXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X2 # '.UXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXPX3X1 .   E.UXUXUXUXUXUXUXPXLXLXFXmXjXP.q   @.0XUXUXUXUXUXUXUXh.  . e *XeXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X2   y c z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z z x s O     D.GXUXUXUXUXUXZXoXP.&XZXUXUXmXm   D |.UXUXUXUXUXUXAXt.  O j =XeXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X1                                                                                                                                                                   8.sXUXUXUXUXUXZX%.r T P.CXVX).u   Y $XUXUXUXUXUXUX9X..  $ H ;XwXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X1                                                                                                                                                                   H XXUXUXUXUXUXLXy.,   x O.O.M + # t.vXUXUXUXUXUXIXT.v   5 1.1XwXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX&X5 / <.1.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<.<...&   , B.KXUXUXUXUXUXwXw.k $     @ f :.3XUXUXUXUXUXUXGXq.*   i T.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXeX=Xx 4XUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXkXj o   X.wXUXUXUXUXUXUXkXR.,.' ` ;.H.kXUXUXUXUXUXUXUX<XY   o N >X0XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXwX1X8.gXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXCX3.3   6 s.VXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXvX-.>   8 1.8XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXrXSXLXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXLXGX0Xn .   Z (.ZXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXMXH.d   X J #X0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXwXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtX8X>.0 . X P ,XCXIXUXUXUXUXUXUXUXUXUXUXUXUXUXPXMXoXz . . a y.0XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X%X^ *   $ D /.VXUXUXUXUXUXUXUXUXUXUXUXUXUXvXJ.z + . > %.>XwXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX/.P X   + c 8.eXHXUXUXUXUXUXUXUXUXIXGX,X;.f . . + [ }.rXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXI.T *     5 / u./.yXFXUXUXUXAX0XT.0.U ,   . : ] (.eXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXeX!.} e       > V #.u.k.v.h.t...v *     X a %.}.eXwXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX@X5.Z 3 .                       o 8 J y.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX9X}.#.v u 2 # o       o $ 5 i N 2.#X0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X7X-XH.$.A f q 5 w j G <.T.>X8X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X6X<X;X=X*X&X*X=X;X1X7X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXwXeXeXeXeXeXeXwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_1[] = {
/* columns rows colors chars-per-pixel */
"96 96 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X8X7X7X7X7X7X7X7X7X7X8X0XqXqXqXqXqXqX0X8X7X7X7X8X0XqXwXwXqXqX9X8X7X7X7X8XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXpXcXvXcXxXzXzXzXxXcXvXcXpXqXqXqXqXqX9XrXlXlXxXnXzXpXqX9X8X0XrXgXbXvXzXzXfX8XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX{.u.r.f.l.k.j.k.l.f.r.u.[.qXqXqXqX9XuX<Xx.c.z.9.j.[.qXsXgXuX:XH.0.p.c.l.Y.fX7XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXW.{ F -           = F { W.qXqXqX9XrX1XK       V t.+XMXLXUXFXyXS.o.$     ; n.dX6XwXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXsXxXMX3X X9   7 }.3XMXxXsXqXqX0XrX9XI     t A.PXUXUXUXUXIXUXUXUXUX2X|     % D.gX7XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X2XMXUXUXi   r UXUXNX2X0XqXqXwXwX,..   f &XUXUXJXKXKXKXKXLXPXLXKXUXmXw.    a [.yX9XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XnXLXUXt   q UXLXmX7XwXqXqXtX.X&   6 XXUXKXPXIXUXUXUXUXUXUXIXUXLXUXSX+.    ..dX8XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXrXE     t.UXJXIXIXUX&X^ i x y.nXUXPXUXIXUXiXt   2 J.fX8XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwXqXeX;X*   c GXUXIXPXUX'.*   o .   Q bXUXPXUXUXUXR.    :.hX8XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwX0XyXf.$   7.UXKXUXKXUXz   3.UX<Xy   W.UXKXUXIXUX0Xt   H %XyX0XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwX9XpXK   1 >XUXPXUXIXPX5.X.jXUXHX,.  F.UXJXUXIXUXFX;.  > P.fX8XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwX9XaXh   r KXPXIXUXUXUXUXUXPXUXiXp O OXUXKXUXUXPXUXS.    B.jX8XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwX9XsX0   i UXKXUXUXUXUXKXGXUXCX_   :.KXUXIXUXUXLXUXR.    x.lX7XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwX9XsX0   i UXKXUXUXUXUXPXUXBX/   Q vXUXLXUXUXUXLXUXE.    x.lX7XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwX9XaXh   r LXPXIXUXUXIXUXbXP   L lXUXGXUXUXUXUXPXUXF.    B.jX8XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwX9XpXH   2 2XUXPXUXIXUXlXD   ` KXUXUXUXUXUXUXIXUXFX<.  ; L.gX8XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwX0XyXp.#   w.UXLXUXKXUXH   . 6 n j 9 x.UXUXIXIXUXqXa   D #XyX9XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwXqXeX=X&   n LXIXIXKXUXq   X         ,.UXUXIXIXUX^.    -.fX8XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXeXK   . f.UXKXIXUXKXHXHXJXKXKXJXPXUXUXIXUXfXs   , S.gX8XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X8X8X8X8X8X9X0XqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXtX[.#   q %XUXLXPXUXUXUXUXUXUXUXUXUXPXUXDX=.    [ aX8XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X7X9XiXgXjXlXkXhXdXrX7X9XqXqXqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXwXqX%..   v 3XUXUXKXPXIXUXUXIXIXLXKXUXVXg.    y ).iX8XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X0XjXiX XK.C.b.m.S.W.<XkXpX7X0XqXqXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwXqXqX0XtX6XA     y L.KXUXUXPXJXHXKXUXUXUX7X+.    # m.hX6XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XfX0Xv.%.n $       . 5 ! 7.).hXwX7XwXqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXqX9XtX;XN     . X.Y.uXUXUXUXUXZXoXj.a     & d.fX6XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XhXG.W @                     u <.:XaX7XqXqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqX8XyX1X} +     + B o.:.1.&.! q       d A.fX7XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XsX9.4       h %.i.k.g.q.W ,       N }.pX8XqXqXqXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqX8XuX8XC.S                     1 3.|.gX7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XsX$.      N S.{.0XfXkXhXaX>X~.6.-     7 `.pX9XqXqXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqX8XwXfX-XJ.u.v : + + % 9 ..N.).yXpX8XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XhX9.      3.=XaXyXwX9X7X8X9XrXpXrXW.m     y %XuX0XqXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqX0X8XuXpXDXCX2X    ' KXVXxXiXqX9XqXqXqXqXqXqXqXqXqX",
"qXqXqXwXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XfXG.3     c.7XyX0X0X0XqXyXiXyX0X9XwXrX*XK     R 8XrXqXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqXqXqXqX6XJXUXNX    %.UXUXfX6XqXqXqXqXqXqXqXqXqXqXqX",
"qX0XtX*XOXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0X0XW     4.8XeX0XqX0XtXtX:X$X*X0XqXqX9XrX|.d     P.tXqXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqXqXqXwX8XGXUXkX    +.UXKXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX8XzXj.;.bX7XqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XjXv.@   M -XyX0XqX0XqX9Xn.:   > 5XrXqXqX8XsXj.%   k qXwXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XbXq.G fX9X0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX%.    A.aX0XqXqX9XyX3.  = # 6 5XrXqXqXqX7XpXZ   # .XeXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XcXf.1 }.pX9XqXqXqXqXqXqXqXqXqXqXqXqX9XiX Xb   j {.yX0XqX0XrX1XP H 5.= 1 6XrXqXqXqX8XyX7.O o 1.yX0XwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XxXj.  Z.jX7XqXqXqXqXqXqXqXqXqXqXqXqX8XgXK.$   %.0XwXqXqX0XeX4XI.9XS.$ 2 6XrXqXqXqX0XeX|.4   V iX0XwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XzXx.  z.lX7XqXqXqXqXqXqXqXqXqXqXqXqX8XjXV.    p.fX9XqXqXqXqXwXuXdXz.& 2 6XrXqXqXqXqX9XeXe   p pX9XwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XzXv.  s.xX7XqXqXqXqXqXqXqXqXqXqXqXqX8XlXb.    k.kX8XqXqXqXqXqX5XpXv.% 2 6XrXqXqXqXqX7XfXy   4 pX9XwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XzXc.  g.xX7XqXqXqXqXqXqXqXqXqXqXqXqX8XkXm.    g.hX8XqXqXqXqXqX8XaXv.& 2 6XrXqXqXqXqX8XpXt   0 pX9XwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XxXl.  n.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XhXS..   q.aX0XqXqXqXqXqX8XaXv.& 2 6XrXqXqXqX0X0X4X0   k pX9XwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XxXh.  I.fX8XqXqXqXqXqXqXqXqXqXqXqXqX9XdXW.5   W >XrX0XqXqXqXqX8XaXb.* 4 6XrXqXqXqX9XtXS.;   W iX0XwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XcXi.j 2XrX0XqXqXqXqXqXqXqXqXqXqXqXqX0XrX<X!   , !.pX9XqXqXqXqX8XsXl.  % 6XrXqXqXqX7XsX]   $ V.tX0XwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XbXq.~ nX7XqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XkX7.    5.rXwX0XqXqXqX9XiXH.M D 7XeXqXqX9XrX%Xy   - 2XeXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX9XfXU.n.hX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XpX).u   - W.rX0XqXqXqXqX0XuXzXlXqXqXqX0X7XeX`     @.eXwXqXwX7XmXUXUXy   w UXUXMX7XwXqXqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXwX8X7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XhX<.    n *XrX8X0XqXqXqX9X6X6XqXqX9X7XiXf.%   3 &XtXqXqXwX7XnXJXUXt   q UXKXnX7XwXqXqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwX;XB     K {.sX8X8X0XqXqXqX0X9X7XrXeXf.,   o h.eXqXqXqX0X2XMXUXUXp   r UXUXNX2X0XqXqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XaX}.7     d g.iXuXeX9X7X8X0XtXsX$X` %     @.yXeX0XqX7XyXfXyXT.u.:   - y.Y.yXfXyX7XqXqXqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XpX`.y .   $ C 7.|.eXfXpX4XD.] t     X @.wXwX0X0XqXfX!.U c <   . o .   < v U Q.fXqX0XqXqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XpX%XE       O 4 e y t 0 ;       3 h.yXeX9X0XuX<X_ 2                       1 _ <XuX0XqXqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XuX8XI.l $ O             $ - #.&XrXeX0X0XuX.Xf X     r 5.$XpXzXpX%X5.t     . s |.iX0XqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXtXqXXX2.V p 5 0 j E V.2XeXyXwXqXqXuX.X3     7 A.AXHXUXUXUXUXUXJXAXD.8     1 |.uX0XqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX0X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X9XqXeXtXyXuXuXyXtXwX0X9X8X8X8X8X<Xf     H eXUXUXUXLXJXJXKXIXUXUXUXrXK     s <XqXqXqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XtXfXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXdXsXaXaXaXaXaXsXsXdXdXdXdXiXmX' .   H BXPXIXIXPXUXUXUXUXUXLXUXIXPXCXL   X ) dX8XqXqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXwX8XhXUXKXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXJXUX4X7   8 rXPXIXIXUXUX0X`./.#XCXUXPXUXIXPXtX8   < Q.tX9XqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXwX5XmXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX|     A.UXIXIXPXUXS.e       O.kXUXPXUXPXUXF.    P kX6XqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX9XaX).+.1.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.>.:.1.7   r ZXUXUXJXUX~.    u.C.d   >.UXJXUXUXUXFXy   g -XwX0XwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XcXp.                                                                                      4.HXUXUXKXUXW.~ D.UXUX^.  S UXLXIXUXUXJX6.  ; f.fX7XwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XxXd.                                                                                  O   $XUXUXUXUXUXHXCXUXUXDX=.  +.UXJXUXUXUXUX&X    @.zX6XwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XrXaXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXrXZX{     aXUXUXUXUXUXUXUXZXn ; o <.BXUXIXUXUXUXUXdX    ) cX5XwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXwX6XbXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX&.    zXUXIXUXUXUXUXUXAXL 1   7 Q.UXLXUXUXUXUXvX    E bX5XwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XtXhXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXiXFX .    pXUXUXUXUXIXLXJXUXUXUXQ.  p 0XUXLXUXUXUXdX    _ cX5XwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XxXf.                                                                                  O   $XUXUXIXIXUXH.} }.UXUXpX  4 [.UXKXUXUXUX*X    #.zX6XwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XcXp.                                                                                      4.JXUXUXJXUXe.  2 N.^.Y   x lXUXPXUXUXJX5.  ; f.fX7XwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qX9XaX/.| ;.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.*.&.%.;.5   r ZXUXUXPXUXbX;.$       v `.UXKXUXUXUXFXu   g -XwX0XwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXwX5XmXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX}     S.UXIXUXPXUXGX}.m.c.L.aXUXKXUXUXPXUXF.    P jX6XqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXwX8XhXUXKXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXJXUX3X6   8 eXPXIXUXPXUXUXUXUXUXUXKXUXUXIXPXrX8   < W.tX9XqXwX8XGXUXlX    +.UXLXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XtXgXdXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXpXMX' .   J VXPXPXUXIXKXHXHXJXPXUXUXPXPXVXK   . ( dX8XqXqXwX8XFXUXjX    O.UXKXfX8XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX0X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X,Xd     K eXUXUXUXUXUXUXUXUXUXUXUXrXL     d ,XqXqXqXqXwX9XPXUXBX    %.UXUXjX8XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwX0XiX|.1     8 S.DXJXUXUXUXUXUXJXDXF.9     1 {.iX0XqXqXrXrX1X).o.b     4 T k.*X0XtXeX0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XiX|.d X     y 5.&XdXbXdX*X6.y     . s {.iX0X0XeXrX=X$.> &             + * l I.9XuX0XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX<X) <                       , ) ,XiX0X0XwXuXk.5       : w u i t 6 O     . R &XpX9XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXdXQ.P g ;           ; g P Q.dXqX0X0XwXeX&.O     p %.|.SXUXUXUXuXl.K &     y '.pX7XwXqXqXqXqX",
"qXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwX9XyXlX;Xg.#._ W _ #.g.-XkXyX9XwXwXeXuX&.    % $.kXUXUXIXLXKXPXUXUXUXP.z     6 }.pX8XqXqXqXqX",
"qXqXqX9X6X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X5X2X7XiXgXkXlXkXgXiX7X2X5X7X7X7X0Xj.o   , H.UXUXJXKXHXGXGXJXLXLXKXUXyX~     M :XwX0XqXqXqX",
"qXqX0XyXvXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXbXcXzXzXlXzXzXcXbXnXnXnXnXBXtX5   % H.UXJXPXPXUXUXUXUXUXUXPXUXKXUXvXG     >.jX7XqXqXqX",
"qX0XsX~.{.UXKXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXUXUXe.    $.UXJXIXIXUXnXx.$.#.V.BXUXLXUXPXUX<X;   u (.pX8XqXqX",
"qX8XnX7.=.UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX<   p kXUXPXIXUXpXM         ` kXUXPXIXUXUXv.    4.kX7XqXqX",
"qX7XxXf.  k p a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a y     %.UXJXUXKXUX]   A tXoXh   F.UXKXUXPXUX6X1   W <XrX0XqX",
"qX7XzXk.                                                                                                                                : |.UXPXUXPXUXP.c.uXUXUXw.  g.UXLXUXIXUXNX .  5 T.dX8XqX",
"qX7XzXk.  5 0 7 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 7 9 #   w DXIXIXUXUXUXUXUXLXcX`.d 2 OXUXLXUXUXIXUXB.  X V.jX8XqX",
"qX7XzXv.  Y.UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXp   u UXLXUXUXUXUXPXUX&X    $ L.UXUXIXUXUXLXUXT.    c.lX7XqX",
"qX7XzXb.  W.UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXy   p UXKXUXUXUXUXUXUXsX:._ ; * (.UXPXUXUXLXUXQ.    l.zX7XqX",
"qX7XzXz.  *.~.F.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.J.G.T.9   t UXPXIXUXUXJXzXhXUXUXUXM.  P xXUXPXUXPXUXJ.    M.kX8XqX",
"qX7XzXh.                                                                                                                            o   5 yXUXPXIXIXLX! V Q.UXUXg.  F fXUXPXIXUXPXe.  % G.gX8XqX",
"qX7XlXz.                                                                                                                                + z.UXLXUXLXUX{   2 @.' ,   a.UXIXIXIXUXyXn   b |.iX9XqX",
"qX7XbXq.M =X_.].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].].{.R.@   H UXKXIXIXUXVXw.,     7 d.VXUXPXUXIXUXOX    @.iX9X0XqX",
"qX8XzXh.l.UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXB   & P.UXKXUXIXUXKXaX9X0XsXUXUXPXUXIXUXMXJ   # z.jX7XqXqX",
"qXqXeX5XkXUXKXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXUX&X.   z tXUXPXUXPXUXUXUXUXUXUXPXUXIXUXKXj.    T 0X0X0XqXqX",
"qXqXqXeXtXwXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXwXtXqXR .   ~ vXUXUXPXIXPXPXPXPXIXIXIXUXKX/..   3 D.dX7XqXqXqX",
"qXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XyX&Xy     D ,XUXUXUXIXLXLXPXUXUXUXMXk..     6.hX7XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX'.7     ; z.5XNXUXUXUXUXPXuXXXK       o.aX8XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XpX}.N       1 | B.T.!.J.e.m       2 6.sX8XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XpX:X>.u                     @ T D.hX8XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX8XwXjX(.4.Q 5 X       % v @.z.0XdX7XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X7XpXkX<XT.V.c.l.m.G.}.iXjX0X7XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X7XrXdXjXlXzXkXgXiX9X7X0XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X8X8X7X7X8X8X9X0XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_2[] = {
/* columns rows colors chars-per-pixel */
"88 88 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X7X6X6X6X6X6X6X6X6X6X9XqXqXqXqXqXqX8X6X6X6X7X0XqXwXwXqX9X7X6X6X6X8XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXlXcXcXcXcXxXcXcXcXcXaXqXqXqXqXqX0XhXxXcXvXlXuX0X8X8X0XiXlXvXcXcXgX0XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrX X/ N N b C Z N b C C x.qX0XqX0XwXrX*.S N z ) T.aXMXmXuXJ.! z B A 2.iXqX0XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX5X{.J.} t     & Z g.~.-XqXqX0XqXtX(     # j.eXnXPXUXUXLXcXqXe..     O.iX0X0XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXrXCXUXPX;   m.UXIXpXtXqXqX9XuX,.    F eXUXUXUXJXGXHXPXUXUXUX3Xn     r.iX9XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX6XxXUXDX*   z.UXDX9XqXqX9XsXY.%   B vXUXIXLXUXUXUXUXUXPXIXIXUXdXh   > '.pX9XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqX0XeX6XY   ; >XUXIXIXUXiXa.*.y.0XUXIXUXPXUX.X+   ` qXwXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqX0XiX`.X   a.UXIXPXUX1Xi       6 $XUXIXIXIXUX3.  % #XyX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqXqXtX9.  > ;XUXLXKXUXY   d.UXf.  F UXLXUXKXUX|.$ . d.uX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqXqXwXM   ~ SXUXIXIXUXx.8.BXUXuX< e UXLXUXIXUXMXF   L iX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqXwXqXr   4.UXKXUXUXUXUXUXPXUXl.  -.UXPXUXUXKXUXX.  k pX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqXwXqX2   e.UXHXUXUXUXJXGXUX^.  u zXUXUXUXUXHXUX&.  r pX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqXwXqX8   9.UXJXUXUXUXUXUXR.  4 4XUXPXUXUXUXJXUX#.  a pX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqXwXwXf   %.UXPXIXUXIXUXH.  x hXUXUXUXUXUXIXPXUX(   N iX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqXqXeX)   z uXUXPXLXUXG.. O 1.'.G.^.FXUXIXPXUX4Xu   O.uX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqX0XuXF.    /.UXLXJXUXe             hXUXPXKXUXG.  O Q.yX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqX0XyX%X7   H VXUXPXUX7XOX#X&X-X*X1XJXUXPXUXzXb   i 1XrXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9X7X7X7X7X7X9XqXqXqXqXqXqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqXqX0XiXt.    a.UXUXLXUXUXUXUXUXUXUXUXLXUXIX2.    x.iX0XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X8XiXjXlXxXlXjXiX8X7XqXqXqXqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqXqX9XtX;Xn   O H.UXUXFXJXKXLXLXKXJXJXUXKXc.    G 3XrX0XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XhXpX).c.a.e.a.v.`.aXgX9X0XqXqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqXqXqX8XsX`.2     1.lXUXUXIXLXLXUXUXUXsX$.    q oXpX8XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuXrXa.U t           u T f.yXuX9XqXqXqXqXqXwX7XcXUXDX*   x.UXDX0XqXqXqXqX7XsXR.8     l e.-XLXUXUXJX@X7.s     y _.aX7XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9XdXU.M o       + @ +       O V W.dX9XqXqXqXqXwX7XcXUXDX*   x.UXDX0XqXqXqXqXqX8XsX.XD X     2 g l l f <     O Y %XpX8XwXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XaXd.>     y X.Y.%X>X$XI. .r     > k.aX9XqXqXqXwX7XcXUXDX*   x.UXDX0XqXqXqXqXqXqX9XyXqXM.z 0             q n F.wXyX9XqXqXqXqXqXqXqX",
"qXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XdXd.    - -.4XtXyXtXrXtXyXtX2X%.*     l.dX9XqXqXwX7XcXUXDX*   x.UXDX0XqXqXqXqXqXqXqX0XwXiXyX<Xn.~     ] V.4XyXuXwX0XqXqXqXqXqXqXqXqX",
"qXqXqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuXU.:   6 F.iXwX0X0X0X9X8X9X0XwXiXZ.4   < Q.uX0XqXwX7XcXUXDX*   x.UXDX0XqXqXqXqXqXqXqXqXqXqX6XFXUX4X    fXUXAX5XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXwX<X8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XrXM   ; F.pX7X0XqXwXuXdXdXyX0X0X7XpXC.*   A uX9XqXwX7XcXUXDX*   x.UXDX0XqXqXqXqXqXqXqXqXqXwX7XBXUXXX    2XUXMX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX8XhX-.@XuX0XqXqXqXqXqXqXqXqXqXqXqXqX7XhXa.o   -.pX7XqXqX0XuXZ.u < i.jX7X0X7XiX#.  O h.gX7XwX7XcXUXDX*   x.UXDX0XqXqXqXqXqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX6XcXm K.aX0XqXqXqXqXqXqXqXqXqXqXqX0X8XpXU   y 3XwX0XqX9XuX1.  O   1.zX6XqX0XwX2Xw   Q sX8XwX7XcXUXDX*   x.UXDX0XqXqXqXqXqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX6XcXV [ iXqXqXqXqXqXqXqXqXqXqXqXqX9XiX).r   X.rX0XqX0XeX1XZ / |   6.zX6XqXqX0XrX{   i ].iX0X7XcXUXDX*   x.UXDX0XqXqXqXqXqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX6XcXn t uXqXqXqXqXqXqXqXqXqXqXqXqX7XjXc.    Y.tX0XqXqXeX5X/.uX;.  6.zX6XqXqX0XyXK.    N.hX8X7XcXUXDX*   x.UXDX0XqXqXqXqXqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX6XcXm : uXqXqXqXqXqXqXqXqXqXqXqXqX7XlXp.  + %XtXqXqXqXqXwXiXtX#.  6.zX6XqXqX0XyXOX.   j.kX8X7XcXUXDX*   x.UXDX0XqXqXqXqXqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX6XcXm O uXqXqXqXqXqXqXqXqXqXqXqXqX7XxXe.  @ >XrXqXqXqXqXqX8XyX$.  6.zX6XqXqXqXtX*Xo   a.zX8X7XcXUXDX*   x.UXDX0XqXqXqXqXqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX6XcXm > uXqXqXqXqXqXqXqXqXqXqXqXqX7XlXa.  + $XtXqXqXqXqXqXqXyX$.  6.zX6XqXqX0XyXOX.   k.kX8X7XcXUXDX*   x.UXDX0XqXqXqXqXqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX6XcXn u uXqXqXqXqXqXqXqXqXqXqXqXqX7XjXv.    I.yX0XqXqXqXqXqXyX$.  6.zX6XqXqX0XyXG.    B.hX9X7XcXUXDX*   x.UXDX0XqXqXqXqXqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX6XcXC  .iXqXqXqXqXqXqXqXqXqXqXqXqX9XiX`.u    .tX0XqXqXqXqXqXyX#.  5.zX6XqXqX0XeX`   s {.uX0X7XcXUXDX*   x.UXDX0XqXqXqXqXqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX6XcXM L.aX0XqXqXqXqXqXqXqXqXqXqXqXqX8XaXT   e <XwX0XqXqXqXqXyX&.  8.lX6XqX0XeX,X0   ~ dX8XwX7XcXUXDX*   x.UXDX0XqXqXqXqXqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX8XgX<.#XuX0XqXqXqXqXqXqXqXqXqXqXqXqX7XgXf.O   %.iX7X0XqXqXqXqX8X4X8XwXqX0X7XiX..  # z.gX7XwX7XcXUXDX*   x.UXDX0XqXqXqXqXqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXwX2X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XyXV   * C.pX7X0XqXqXqXwXrXwXqX0X7XpXn.$   F iX9XqXwX7XxXUXVX*   k.UXAX0XqXqXqXqXqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuXW.,   4 V.iXeX0X0X0X0X0X0X0XeXiXn.1   3 /.yX0XqXqX6XnXUXUX=   Z.UXPX0X0XqXqXqXqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XdXk.    * @.<XrXyXyXtXyXyXeX:XX.%   o b.dX9X0XwXyXtX5XZ.@.#   M i.oXwXiXrX0XqXqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XaXl.1     q [ J.OX*XOXG.` 9     2 b.aX9X8XtXeXE.Z 8 .           3 u 9.2XuX9XqXqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XdXQ.A +       . o .       @ F /.dX9X8XuX1X] #     : j n m c e       a K.sX8XqXqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XuXuXj.! i           a ~ x.iXyX9X8XuX%Xc     w 3.|.FXUXUXUXrXv.P     % f.dX8XqXqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9XgXsX].N.j.p.j.B.{.dXgX9X0X0XtX1Xx     _ 7XUXUXIXKXJXLXUXUXHXv.>   o n.gX7XqXqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X7X8XuXhXkXlXkXgXuX8X7X0XqXqXrX]     6.AXUXJXKXUXUXUXUXLXJXUXUX#X6   9 _.pX8XqXwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXwX8XmXHXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXAXZXZXCXZXZXAXDXDXDXDXPX-X#   _ SXUXLXUXUXAXcXcXSXUXUXPXLXUX^.@   ] tX0X0XwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXeX5XUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXR   w 7XUXLXPXUXXXb     Z >XUXIXIXUXUX-.  @ U.fX8XwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX9XiXG.p.h.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.h.s.:   2.UXJXKXUX*X    7.+.  9 mXUXIXIXUXtX%   0.uX0XwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX6XbXp                                                                             , |.UXLXKXUX|.~ E.UXUX~   7XUXIXIXUXBX^   M >XtXqX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XzXL . , ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; ; : *   g FXUXIXUXUXPXJXUXGX#Xp 0 vXUXIXUXPXUXM.  , @XuXqX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXwX8XbXDXZXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXMXUX~   N UXKXUXUXUXIXUXrX.   3 @XUXIXUXUXLXUXQ.  o XXiXqX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXeX5XUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX[   V UXKXUXUXUXUXUXnXr.| & a uXUXIXUXLXUX/.    .XpXqX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX9XaXv.<.7.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.4.y.g   c UXPXIXUXUXgX,XHXUXUX5.  r.UXIXUXLXUXP.  * OXiXqX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX6XbXp                                                                             w rXUXPXKXUXj.  ,.NXmXT   g.UXIXUXUXLX3.  r *XyXqX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qX7XlXW - 0 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 8 3 X   v.UXJXLXUX2Xw   & O   s hXUXIXIXUXkXu   ..9XwXqX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXwX7XDXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXs   K GXUXPXPXUXdX6.D H w.xXUXIXUXIXUXR.    n.gX8XwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXwX7XDXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXM.    v.UXLXUXIXUXUXUXUXUXUXIXUXIXUXdXr   c *XtX9XwX7XCXUX@X    6XUXBX7XwXqXqXqXqXqXqXqXqXqXqX",
"qXqX0XfXlXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXxXuXs   > #XUXUXIXPXKXJXJXKXIXIXIXUXGX^   X h.hX7XqXqX6XVXUX#X    7XUXNX6XwXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX8X7X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X7X6XrXK.&   6 !.UXUXUXPXLXLXLXUXUXUXfX/     _ tX9X0X9XwXrXFXUX#X    7XUXSXeXwX9XqXqXqXqXqXqXqXqXqX",
"qXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqX0XsXf.o   + *.eXBXUXUXUXUXLXjXT.t     Z 6XrX9X8XpXwX|.D.^ u o O a ' J..XrXiX8XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XdXn.0     % ~ M.Q.(.P.4.u     . ) 6XrX0X9XsX].X.$                 = *..XdX8XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XgX_.] @                   c h.tXrX0X9XdXz.6     O J <.w.q.>.S       w C.fX8XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX8XpXtXU.0.M , o   * r ..m.=XjX9X0X9XdX8.X   . =.+XvXUXUXUXUXkX.X..    + d.fX8XqXqXqXqXqX",
"qXqXqX9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X6X8XaXtX;XOX X}.XX$X7XdXeX5X8X6XaXz.X   q ).UXUXUXGXSXSXJXUXUXLXU.3   @ A.aX8XqXqXqXqX",
"qXqX0XpXaXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXiXyXiXfXhXjXkXjXgXaXyXuXpXuXlX.X5   q 6XUXUXJXIXUXUXUXUXKXLXUXUX=X2   t @XuX9XqXqXqX",
"qX0XtX{.ZXUXLXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXLXLXLXLXLXLXLXLXPXPXLXPXUX6.  X ).UXIXIXUXJX@XQ._.fXUXIXIXIXUXJ.    ,.rXqXqXqXqX",
"qX7XxXH hXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXvX$   =.UXUXIXUXmX{       k *XUXIXIXUXKX[   > -XtXqXqXqX",
"qX6XcXN p I H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H H K v   O +XUXLXKXUX4.  R 2Xy.  v UXPXUXKXUX/.  o B.yX0XqXqX",
"qX6XxXA                                                                                                                     J vXUXIXPXUX/.N.bXUXMX: - LXUXIXPXUXfXn   ! iX0XqXqX",
"qX6XcXB   U K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K K L L :   <.UXLXUXUXIXUXUXcXXX-.6 t.UXPXUXIXLXUX}   x pX0XqXqX",
"qX6XcXn ; UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX9   e.UXJXUXUXUXKXUXG.    z eXUXPXUXUXHXUX%.  t pX0XqXqX",
"qX6XcXn ; UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX0   w.UXJXUXUXUXUXUXcXXXc.  f kXUXIXUXHXUX%.  u pX0XqXqX",
"qX6XcXV   M v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v v ;   :.UXLXIXUXPX).U.HXUXUXM   T.UXKXIXLXUX[   b iX0XqXqX",
"qX6XxXA                                                                                                                     A kXUXPXLXUXI   %.qXH.+ % <XUXPXPXUXuXz   ) iX0XqXqX",
"qX6XcXM m { ) _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ ) ' J      XUXLXPXUX,Xc       > f.UXPXUXKXUXW.  O F.yX0XqXqX",
"qX7XzXI xXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXNX-    .LXUXIXIXUXlXL.l.N.:XUXUXUXIXUXDXQ   3 :XtXqXqXqX",
"qX0XrX@XAXPXJXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXJXJXUXt.    I.UXIXIXLXUXUXUXUXUXLXUXIXUXUXn.    8.tXqXqXqXqX",
"qXqX0XiXuXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXrXgX+Xw   3 *XUXUXKXKXJXHXHXKXPXKXUXUX{.=   f *XyX0XqXqXqX",
"qXqXqX9X9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X6XsXC.+   1 H.JXUXUXLXHXHXLXUXUXDXm.=   = Y.sX8XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXd.@     ' ^.fXUXUXUXUXuXE.W     = m.fX8XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXS.t       n } &.%.[ z       f U.fX8XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XaX@X,., O             O 3 9.*XsX8XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XuXrX-XV.Q x t y v ( F.:XtXyX8XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XqXtXyXiXpXpXpXiXiXyXtXqX0XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0X0X0X0X0X0X0XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_3[] = {
/* columns rows colors chars-per-pixel */
"64 64 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0X0X0X0X0XqXqXqXqXqXqX0X0XqXqXqXqXqX0X0X0XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXuXtXrXrXtXuXtXqXqXqX0XwXrXuXyXwXqXqXrXiXtXeX9XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX=X+X:X<X<X:X+X-XqXqX9XyXqX3X$X$X7XwXqX>XOX<X5XiX0XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXN._ t & & t _ N.wX9XuX*Xv   S H.hXGXCX#X=.* $ l.dX9XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXuXMXmX3 3 mXNXuXqXeX7Xn   1.DXUXUXUXUXUXUX{.<   J.sX9XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5XnXUXw w UXnX5X0XdX5.  <.UXUXUXnX0XZXUXKXUX>X  r 3XeX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XnXUX0 0 UXnX7XtX2X; r AXUXUX~.2 # y <XUXGXUXd.  g.hX8XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XnXUX0 0 UXnX5XdXV.  y.UXKXUXh %.PXz H UXHXUXuX+ H tXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XnXUX0 0 UXnX4XlX=.  >XUXPXIXjXPXUXM ] UXKXIXUXM * 7XeX0XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XnXUX0 0 UXnX3XxXo.  tXUXPXIXUXUX@.= dXUXPXKXUXJ   5XrX0XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XnXUX0 0 UXnX4XzX&.  3XUXLXUXUX;.r nXUXPXUXPXUXC $ 6XeX0XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XnXUX0 0 UXnX5XgXx.  z.UXHXUX..  ` h.C.UXPXUXfX- N rXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0X0XqXqXqXqXqXqXqXwX7XnXUX0 0 UXnX6XyX:X@ j PXUXUX[ K D a ..IXJXUXV.  t.hX8XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X8XwXrXrXeX0X8XqXqXqXqXqXwX7XnXUX0 0 UXnX7X0XsX#.  f.UXPXUXUXUXUXUXJXUXyX- 1 -XtX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XuXgXqX5X4X7XuXgXqX9XqXqXqXwX7XnXUX0 0 UXnX7XqXtX,Xt   M.UXUXPXJXKXUXUX0Xg   x.dX8XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XfX.X4.k @ . > E b.0XpX8XqXqXwX7XnXUX0 0 UXnX7XwX9XpX{.7   Q }.VXUXPX8Xp.#   5.sX9XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXuX8.#   < v M g     z (.dX8XqXwX7XnXUX0 0 UXnX7XwXqX8XaX%XE     < s w     q N.sX0XqXqXqXqXqX",
"qXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqX0XuX[   9 x.*XaXaX0X`.X.  + I.sX8XwX7XnXUX0 0 UXnX7XwXqXqX9XyXiX!.e.z   ; X.Z.<XfX9XqXqXqXqXqXqX",
"qXqXqXwXtX0XqXqXqXqXqXqXqXqXqX8XfX8.  C -XgXeXwXjXsXiXfXB.  , @XuX0X7XnXUX0 0 UXnX7XwXqXqXqX0X0XiXUXzX  4.UXxXqX9XqXqXqXqXqXqXqX",
"qXqXrX1X/.uX0XqXqXqXqXqXqXqX9XuX.X# 9 ;XiX6XiX=X6.D.tX6XjXd.  ) aX0X7XnXUX0 0 UXnX7XwXqXqXqXqXqX6XUXpX  >.UXtX7XwXqXqXqXqXqXqXqX",
"qX0XuXOX..aX9XqXqXqXqXqXqXqX8XgX4.  z.gX6XuX].e   G jX7X0XyXc . $XuX6XnXUX0 0 UXnX7XwXqXqXqXqXqX8XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qX0XtX;Xd *XyX0XqXqXqXqXqXqXwXqXj 1 *XrX8XpXF.i.C L kX8X8XdXy.  d.hX4XnXUX0 0 UXnX7XwXqXqXqXqXqX8XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qX0XrX<X= {.pX9XqXqXqXqXqX0XrX5X@ b sX9XqX0XrXkXb J kX8X9XpX).  $.zX3XnXUX0 0 UXnX7XwXqXqXqXqXqX8XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qX0XrX1X% ].pX9XqXqXqXqXqX0XrX4X. M gX8XqXqXqXrXv J kX8X9XpX{.  o.zX3XnXUX0 0 UXnX7XwXqXqXqXqXqX8XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qX0XtX,X2 .XiX0XqXqXqXqXqX0XeX7X> g wX0XqXqXqXyXM P kX8X9XpXL.  ,.lX4XnXUX0 0 UXnX7XwXqXqXqXqXqX8XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qX0XyX%XF 5XeX0XqXqXqXqXqXqX0XuXE   `.pX9XqXqXyXp M lX8X8XgX..  K.sX5XnXUX0 0 UXnX7XwXqXqXqXqXqX8XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qX0XuX#Xe.fX8XqXqXqXqXqXqXqX8XgXb.  X.dX8X0XqXeXn.Z.iX6XyX%X; 9 7XeX7XbXUX0 0 UXbX7XwXqXqXqXqXqX8XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qXqXwX9X1XwXqXqXqXqXqXqXqXqXqXqX0Xz   B.hX0X8X9XdXsX6XuXeXP   f.dX8X4XmXUXw w UXmX4XqXqXqXqXqXqX8XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qXqXqXqXeXqXqXqXqXqXqXqXqXqXqX9XpX).+   s.yXdXpXyXuXgX%XI   E tX8XrXsXxXrX8 8 rXxXsXtX9XqXqXqXqX8XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XdXI.,   x y.).{.L...-   E 6XwXiX<X,.s         d ,.1XaX8XqXqXqX8XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwX9XdX@X) X           0 g.yXwXpXE.0   3 ;.v.v.;.3   0 E.sX8XqXqX8XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qXqXqXqX9X8X8X8X8X8X8X8X8X8X8X8X8X9X6XrXuX@Xp.+. .;.G.5XaX6XyXE.X   l.ZXUXUXUXUXZXk.  X W.aX9XqX8XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qXqXqXqXuXiXiXiXiXiXiXiXiXiXiXiXiXiXiXyXyXfXcXmXMXnXkXaXrXsX9X0   {.UXUXUXUXUXUXUXUX{.  0 1XrX0X8XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qXqX0XrXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXk.  z.UXJXUXH.B ^ 0XUXHXUXk.  <.hX8X8XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qXqXeX6X4.:.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.>.4., 7 CXUXUXH.  w.H 5 SXUXUXZX3 u 8XwX7XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qX0XtX,X                                                      >.UXKXUX7X;XUXR.$ nXUXLXUX-.  (.sX6XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qXqXqXwXlXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXjXAX).  m.UXKXPXUXUXC   h.UXPXKXUXn.  I.gX5XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qXqXqXwXlXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXjXAX_.  n.UXKXUXGXPXwXd.: 4XUXHXUXn.  I.gX5XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qX0XtX,X                                                      :.UXJXUXu.F FX[.  `.UXJXUX=.  (.aX6XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qXqXeX6X4.:.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.,.>.4., 6 CXUXUX1Xs X O *.LXIXUXZX2 i 9XwX7XUXaX  ,.UXiX8XwXqXqXqXqXqXqXqX",
"qXqX0XrXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXk.  k.UXHXUXJXeXdXUXUXJXUXh.  1.hX8X7XUXuX  :.UXuX8XwXqXqXqXqXqXqXqX",
"qXqXqXqXuXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXuXdX0X0   [.UXUXUXUXUXLXUXUX].  w 1XeX8X9XUXNX  7.UXfX7X0XqXqXqXqXqXqXqX",
"qXqXqXqX9X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X6XiXW.X   j.ZXUXUXUXUXCXh.  o Q.iX9XdX6XOX4.  m !.>XiXuX8XqXqXqXqXqXqX",
"qXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwX9XsXW.q   3 -.n.n.=.2   q W.pXwXrXd.7             ( @XsX8XqXqXqXqXqX",
"qXqXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwX9XsX1X1.p         p 2.2XpXwX6XT   1 3.$X0X5XA.C   > I.dX8XqXqXqXqX",
"qXqXqXqX9X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X5X0XsX4X!.J.J.!.5XsX0X5XwXT   ` xXUXUXUXUXUXUXH.. O ).pX9XqXqXqX",
"qXqXqX9XdXMXbXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXbXcXmXSXGXGXSXmXcXxXJXn.  ' UXUXUXUXVXPXUXPXUX`.  z 0XqXqXqXqX",
"qXqXuX+XB.UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXr , xXUXUX,Xx + c 5XUXHXUX2.  n.gX8XqXqX",
"qX0XtX:X3 q 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 q e   3.UXHXUX( R bXv l UXHXUX9X  E iX0XqXqX",
"qX0XrX1X$ > 8 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 9 %   #XUXPXUXbXAXyXL -.UXLXIXUXv : 7XeX0XqX",
"qX0XrX1X& VXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXi.  yXUXPXPXUX7X$ O ].UXLXLXUXJ   5XrX0XqX",
"qX0XrX<X= y.F.B.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.C.N.I.U   6XUXIXLX6XBXUX9.% nXUXKXUXS + 6XrX0XqX",
"qX0XrX2Xo                                                                                 C.UXKXUXn z ].m 0 AXUXUXxX4 l wXwXqXqX",
"qX0XuX+X' ;XoX+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+XOX%X).  C UXLXUX9XU 3 K 5XUXGXUXT.  6.gX8XqXqX",
"qXqXrX<X7XUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX@.  H.UXJXUXUXUXUXUXHXUXvXw # oXuX9XqXqX",
"qXqXqXeXyX9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X8XyX$X> . _.UXUXLXJXJXUXUXbXK   q.fX8XqXqXqX",
"qXqXqXqX9XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX7XsXI.O   1.9XUXUXUXxXY.w   } uX0XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XdX).x     v J S 4   # q.uXwX0XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XpX0Xn.E :   O k 6.oXfX0X0XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XqXgXuX7X5X6XwXgXuX8XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X0XeXrXrXwX8X9XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0X0XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_4[] = {
/* columns rows colors chars-per-pixel */
"48 48 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X8X8X8X8X9XqXqXqX8X8X0XqXqX0X8X9XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuXkXjXhXkXpXqXqX0XgXlXuX9X9XpXlXdX9XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX}.k.c.n.g.~.wXqXyXF.f.].iXuX(.s.I.pX0XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX#X~.l = P..X9XfX*.  j.cXUXUXfX9.  t.gX8XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XUXw.Z UXeXsXI.  `.UXUX_.wXUXUXA.  {.pX9XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4XUX,.b UX6XgXZ E UXUXQ L v #XUXUXv ) kX8XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX4XUX1.n UXqX6X= W.UXFX].UXP B.UXUXz.7 uXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX4XUX1.n UXwX1Xo $XUXLXUXx.' UXKXUXY.= rXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX4XUX1.n UXqX9X< G.UXUX8.& [.xXPXUXu.t pX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X8X8X8X8X9XqXqXqXqX4XUX1.n UX6XjXR m UXKX .' U W.UXUXe $.kX8XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqX8XtXjXlXlXkXdX0X0XqXqX4XUX1.n UX7XiX{.  l.UXUXUXUXUXUX1.+ -XyX0XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqX9XfX:X0.S b ) L.aXtX9XwX4XUX1.n UX9X8XfXk.  o.0XJXGX4XQ   G.dX8XqXqXqX",
"qXqXqX0XqXqXqXqXqXqXqXqX9XdXx.@ 1 T / b   Z 1XyX0X4XUX1.n UX9XqX9XfXP.e + r w o f (.dX8XqXqXqXqX",
"qXqX0XrXwXqXqXqXqXqXqX8XfXx.  >.0XvXHXjX].7 p 9XeX3XUX1.n UX9XqXqX9XsXyX0Xu V yXuXaX8XqXqXqXqXqX",
"qX0XyX@X,XrXqXqXqXqX0XtX:X@ >.cXyXOX7.+XkX1X* $.kX1XUX1.n UX9XqXqXqX9X6XUXY *.UX3X9XqXqXqXqXqXqX",
"qX8XhXb.B.gX9XqXqXqX8XjX9.1 qXwX6X_   D.iXhXr., 6X6XUX1.n UX9XqXqXqXqX7XUXH X.UX5XqXqXqXqXqXqXqX",
"qX8XlXf.$.xX7XqXqXqX8XlXA R kX9X5X2XC Z.fXrXXXO /.rXUX1.n UX9XqXqXqXqX7XUXJ O.UX5XqXqXqXqXqXqXqX",
"qX8XlXd. .cX7XqXqXqX8XlXb _ xX7XwXaXn A.fXeX-Xo E.tXUX1.n UX9XqXqXqXqX7XUXJ O.UX5XqXqXqXqXqXqXqX",
"qX8XkXj.9.kX8XqXqXqX8XkX) b fX8X0XuXp M.fXiXU.# OXqXUX1.n UX9XqXqXqXqX7XUXJ O.UX5XqXqXqXqXqXqXqX",
"qX9XsXI.`.pX0XqXqXqX9XdXL.  ].fX5XeXi.'.rXzXP k dX<XUX1.n UX6X0XqXqXqX7XUXJ O.UX5XqXqXqXqXqXqXqX",
"qXqXqX9X0XqXqXqXqXqXqX0XaXZ 7 2XlXyXlXjXzXy.  U.pXeXUX8.C UXfX0X0XqXqX7XUXJ O.UX5XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqX0XtX1Xp * r.XX&XI.P   y.gXaXK.' @   R c.uXtX9XqX7XUXJ O.UX5XqXqXqXqXqXqXqX",
"qXqXqXqX0X0X0X0X0X0X0XqX9XtX9X#., O X # k I.fXeX/   J x.B.'   x <XyX9X7XUXJ O.UX5XqXqXqXqXqXqXqX",
"qXqXqXqX0X0X0X0X0X0X0X0X0X8XqXhX5X~.T.XXsXiXdX/ + -XUXUXUXUXgXi s eXqX7XUXJ O.UX5XqXqXqXqXqXqXqX",
"qXwX7XlXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX&X  -XUXaX-.8.MXUXMX% 6.kX4XUXJ O.UX5XqXqXqXqXqXqXqX",
"qX9XdXP.r l h h h h h h h h h h g f f g g b ; ~ UXUX7.h.[ &.UXUXr.e yX7XUXJ O.UX5XqXqXqXqXqXqXqX",
"qX0XyXOXi.g.f.f.f.f.f.f.f.f.f.f.f.f.f.f.f.z.o n.UXHXUX-Xz ~.UXUX&XO -XwXUXJ O.UX5XqXqXqXqXqXqXqX",
"qXqXwX7X:X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X6X1 M.UXJXjXfX=.o.UXUX:XX =XwXUXJ O.UX5XqXqXqXqXqXqXqX",
"qX8XgXV.  # O O O O O O O O O O O O O O o %   $.UXUX%.] W ..UXUXb.4 qX8XUXJ o.UX5XqXqXqXqXqXqXqX",
"qXwX9XfXUXLXLXLXLXLXLXLXLXLXLXLXLXLXLXLXHXUXT.  gXUXBXF.P.HXUXUXq ..kX1XUXK +.UX2XqXqXqXqXqXqXqX",
"qXqX0XrXdXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXsXkXc i mXUXUXUXUXUX! - >XyXfXUXH X.UXsXwX9XqXqXqXqXqX",
"qXqXqX0X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X7XrX<Xd % e.#X-Xb.q - _.jX6X<.f     l q.rXyX9XqXqXqXqX",
"qXqXqX0X9X0X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X0X7XrXqX4.q X   2 | ;XhXOX9 > w.XX}.3.# f <XtX0XqXqXqX",
"qXqX0XtXaXuXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXyXiXnXfX2X1XuXnXfXwX8 C FXUXUXUXUXbXp l yXqXqXqXqX",
"qX0XiX_.gXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXLXUXs.- FXUX`.H ;.vXUXkXX g.hX8XqXqX",
"qX8XhXM.. ; * * * * * * * * * * * * * * * * * * * * * & & * & <   f.UXPX,.@XL k.UXUX@.v fX9XqXqX",
"qX8XlXg.! <X{. X X X X X X X X X X X X X X X X X X X X X X XoX_.. OXUXJXUXZ.u 9XUXUXJ.: tXqXqXqX",
"qX8XkXj.Z R.B.Z.Z.Z.Z.Z.Z.Z.Z.Z.Z.Z.Z.Z.Z.Z.Z.Z.Z.Z.Z.Z.Z.C.S.n.  .XUXFX%XnX| -.UXUXF., tXqXqXqX",
"qX8XhXM.< u e e e e e e e e e e e e e e e e e e e e e e e e w d   q.UXPX_ U v D.UXUX] V gX9XqXqX",
"qX0XuX|.MXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXM.O bXUXHX.X6XUXUXeX  M.gX8XqXqX",
"qXqX0XtXtXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXeXuXf i kXUXUXUXUXeX< F aX0XqXqXqX",
"qXqXqX0X9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X8XtX<Xl X +.G.S.]   F 9XeX0XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX9XtXyXh.v : > V M.aXrX9XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXhXfXtXtXgXgX0X0XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X9XqXqX9X8XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_5[] = {
/* columns rows colors chars-per-pixel */
"44 44 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X8X8X8X9XqXqXqX8X8X0XwXqX9X8XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXiXkXhXjXfXwX0X0XgXjXrX8XwXhXjXqX0XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX_.k.M.l.F.4XeXyXA.z.*XsX1XM.b.qXeX0XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX%XR.. -.#X2XhX$.% ~.UXUXPX+X7 U sX0XqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XiXUXy -XPXtXW.o 6XUX$Xw.dXUXnX= l.hX8XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXIX0 [.NXsXE } UXaXR p.F LXUXp.z fX9XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXUXq |.ZXuXu J.UXCXUX_./ UXUX*X1 9XeX0XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXUXq |.ZXuXp F.UXUX^.G DXUXUX$X3 9XwX0XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqX0X8X7X7X9XqXqXqX0XwXUXq |.VXdX) ~ UXjX3 B ,.MXUX6.m gX9XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqX8XtXjXcXxXfXqX9XqX0XwXUXq |.CXeX].. @XUXAXlXlXUXiX+ Z.fX8XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqX9XgX>Xy.` ..A.yXiX9XqXwXUXq |.AX2XhX0.. A.HXUXUX^.* | fX9XqXqXqX",
"qXqXqX0XqXqXqXqXqXqXqX9XfXf.& p U G < p  XaX8XeXUXq |.AX5X0XfXx.3 8 n e > 9.sXqXqXqXqXqX",
"qXqX0XrXwXqXqXqXqXqX8XgXf.  c.sXDXVX6X^ - &XtXqXUXq |.AX5XwX9XdXaX|.  H.dXdX0X0XqXqXqXqX",
"qX0XuX|.=XtX0XqXqX0XtX>X& c.nX8X2.8.sXxXA K hX0XUXq |.AX5XwXqX6XsXIX3 gXnX4XqXqXqXqXqXqX",
"qX8XjXc.v.hX9XqXqX8XjXt.p iXtX XB C pXiX^., &XuXUXq |.AX5XwXqX9XaXNX> 0XxX7XqXqXqXqXqXqX",
"qX8XlXf.,.xX7XqXqX7XcX` R zX5XfXT.L sX0X6X1 R.fXUXq |.AX5XwXqX9XaXCX> wXcX7XqXqXqXqXqXqX",
"qX8XkXh.6.zX8XqXqX7XxX..K kX5XgXC.C dXqX>X< ^.dXUXq |.AX5XwXqX9XaXCX> wXcX7XqXqXqXqXqXqX",
"qX9XfXZ.U.dX9XqXqX9XfXA., 7XwXiXU.] rXjXv.5 8XwXPX0 [.VX4XwXqX9XaXCX> wXcX7XqXqXqXqXqXqX",
"qXqXeX,X6XwXqXqXqXqXqXyXp ~ cXaXeXtXlX2X> 2.fXpXUXu :XUX6X0XwX9XaXCX> wXcX7XqXqXqXqXqXqX",
"qXqXqXeXwXqXqXqXqXqX9XpX.X- A ^.6X,Xv.> H iXfX!.:.  Z B.9XpX9X9XaXCX> wXcX7XqXqXqXqXqXqX",
"qXqXqX0X9X9X9X9X9X9X9X7XiX$XH : , > 3 <.yXuX( @ ! k.,.0 0 }.aX6XaXCX> wXcX7XqXqXqXqXqXqX",
"qXqXqXeXpXpXpXpXpXpXpXpXuXgXnX1X).[.yXvXcX) 7 pXUXUXUXUX&.- :XeXpXCX> wXcX7XqXqXqXqXqXqX",
"qXqX9XaXDXAXAXAXAXAXAXAXAXZXCXHXUXIXZXLX$X@ iXUXF.T B.UXUXG ^ jXuXCX> wXcX7XqXqXqXqXqXqX",
"qX8XfXS.o < : : : : : : : : : ; - - ; 2   X.UXSXh.(.x lXUX.X4 3XdXVX> wXcX7XqXqXqXqXqXqX",
"qXqX0XtXgXfXfXfXfXfXfXfXfXfXfXfXfXfXaXAXA i.UXLXUX9.! JXUXhX> oXjXBX> wXcX7XqXqXqXqXqXqX",
"qX9XsXR.x A V V V V V V V V V V V V B J   8.UXBXi.;Xu *XUX6X< *XhXVX> wXcX7XqXqXqXqXqXqX",
"qX0XtX&XN.Z.C.C.C.C.C.C.C.C.C.C.C.C.V.P.;.y UXUXx.I h.DXUX0.l aXtXNX> 0XlX6XwXqXqXqXqXqX",
"qXwX8XfXUXPXPXPXPXPXPXPXPXPXPXPXPXPXLXUXZX0 &.UXUXUXUXUX=X. D.aXbXUX2 jXKX9X9XqXqXqXqXqX",
"qXqXqX9X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X2XrX|.; G  XgX6X0.X 2.zXqXf.G   b e.5XpX9XqXqXqXqX",
"qXqXqXqX8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X5XyX*XW < = ; g C.jX-Xa 7 w.(.f.a 4 {.pX9XqXqXqX",
"qXqX0XrXlXzXlXzXzXzXzXzXzXzXzXzXzXzXzXlXzXjXbXGXfX5XeXCXZXhXa Y UXUXUXUXUX=., 1XrX0XqXqX",
"qX9XpX/.=XhXiXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXiXyXsXhXgXrXvXx.8 UXUX4.J c.UXUXn O.lX8XqXqX",
"qX8XhXn.  2 : : : : : : : : : : : : : : : : : : ; ; ; : 2   j.UXvXK.R.Z KXUXU.e yXqXqXqX",
"qX8XlXf.u.UXnXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXMXLXg U.UXUXUX%.o.IXUX,X, 7XeX0XqX",
"qX8XjXb.  p w e e e e e e e e e e e e e e e e e e e e e i   m.UXdX;.`.w aXUX^.0 rXqXqXqX",
"qX9XaXT.I.$XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX X,X1.f UXUXV.' _.UXUXI ( lX8XqXqX",
"qXqXqXwXNXAXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXCXDXjX3 =.UXUXUXUXUXl.@ $XuX0XqXqX",
"qXqXqXqX6X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X3XtX}., n P.-X!.P @ L.dX8XqXqXqX",
"qXqXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwX9XpX2XO.e , 0 ( $XdX8XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XrXlXyX7XrXlXuX8XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X8XqXeXqX8X0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_6[] = {
/* columns rows colors chars-per-pixel */
"32 32 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXuXuXuXuXqX9XtXiXqXwXpXwX9XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX#X@X@X$X0XpX>XXX8X3X|.9XyX0XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XXXE E  XsXG.g $XUXIXx.l >XtX0X",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4XUXh.g.UX2Xd ,XSX5.XXUXw.! kX8X",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4XDXi.t.UXb.! UX=XW.Q UXVXc <XrX",
"qXqXqXqXqXqXqXqXqXqXqX0X0XqXqXwX4XGXp.y.UXy.$.UXUXd.(.UXKXV @XuX",
"qXqXqXqXqXqXqXqXqX9X9XtXyX0X8XwX4XGXp.u.UXW.b UX_.u l.UX9Xx rXqX",
"qXqXqXqXqXqXqXqX9XiXhX:X%XpXfX9X4XGXp.p.DXiXT >.UXUXUXHXm a.jX8X",
"qXqX0XqXqXqXqX9XaXoXL l x m Z.gX2XGXp.p.FX7X8XK A l.e.d ,.dX9XqX",
"qXwXrXqXqXqX9XiXXXe w.wXuXL.e z.iXSXp.p.GX3XtXuX|.s ' ;XdX0XqXqX",
"tX-X].uX0XqX9XhXL q.JXM.} nX|.p -XKXi.p.GX5XqX2XUX,.`.GX3XqXqXqX",
"pX X5.fX8X0XtX;Xz 2XyXL.G 3XlXU x.UXy.p.GX4XwX4XJX+.L.mX6XqXqXqX",
"pX X:.fX8X0XyX%Xv 7XeX7XR 2XzX~ d.UXt.i.FX4XwX4XPX@.I.NX6XqXqXqX",
"uX#XJ.pX9XqX0XpXn K.bX;X2.tX0Xf `.UXh.j.JX1XwX4XPX@.I.NX6XqXqXqX",
"qX0XqXwXqXqX9XgXA.e |.lXmX0XD ..bXVX@.@.ZXiX0X4XPX@.I.NX6XqXqXqX",
"qXqX7X6X6X6X6X4XiXf.e J Y u ] lXv.h I P j v.gX1XPX@.I.NX6XqXqXqX",
"qXeXmXBXNXNXNXNXmXLXlX/.T.8XUXN.p rXUXUXqXs b.uXJX@.I.NX6XqXqXqX",
"wX8XI.J.K.K.K.K.K.G.I./.(.Y.U.h tXFX&.4.UX0Xj 3XIX@.I.NX6XqXqXqX",
"eX5Xe.5.7.7.7.7.7.7.6.5.3.r.( F UXDX%X! bXUXL P.UX+.I.NX6XqXqXqX",
"eX5Xe.5.7.7.7.7.7.7.7.7.6.t.( S UXuXI.I 6XUXK P.UXo.P.MX6XqXqXqX",
"wX8XI.J.K.K.K.K.K.K.K.K.K.J.R.j wXUXy.Z.UX8Xj 1XUX:.(.ZX3XwXqXqX",
"qXeXMXBXBXBXBXBXBXBXBXBXBXmXLXB.a wXUXUX9Xa n.nX0Xn ,.sXsX9XqXqX",
"qXqX2X,X,X,X,X,X,X,X,X,X,X,X:XyXj.i C C i k.zX} k ,.| p k.gX9XqX",
"wX9XnXKXGXGXGXGXGXGXGXGXGXGXHXDXUXGX2X1XDXUX .! UXGXUXfXa C.fX8X",
"tX-X5.p.y.u.u.u.u.u.u.u.u.u.u.u.r.u.k.k.a.5.z UX5X{ <.UX&Xb pX0X",
"uX+XP B.z.x.x.x.x.x.x.x.x.x.x.x.x.x.z.k.S.! +.UXPX X[ UXLXN $XuX",
"uX$Xn #.X.o.o.o.o.o.o.o.o.o.o.o.o.o.o.X.=.v ` UX%Xf.U nXJXb ;XtX",
"tX>X/.*X@X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X$X+Xg fXUXM.&XUXB.J hX8X",
"qXwXhXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXgXVXl.a ;XUXUXB.t oXiX9X",
"qXqX8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X6XfXC.b n c J XXaX9XqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XfXpX$X;XhXiX9XqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X0XuXtX8X9XqXqXqX"
};

/* XPM */
static const char *const xpm_icon_7[] = {
/* columns rows colors chars-per-pixel */
"24 24 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXwX8X].)..XuX3X^.pX>X(.uXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXwXQ.P :X9XW ^.mXSX;.t.jX",
"qXqXqXqXqXqXqXqXqXqXqXqXtXaX-.UXy.h.UX4.L.UXE =X",
"qXqXqXqXqXqXqX0X8X8X0XqXrX0X+.PX@.@XUXC.T.UX2.I.",
"qXqXqXqXqXqX9XyXkXkXtX8XtXqX#.LXy.l.UXW :XUXQ *X",
"qX0XqXqXqX9XfX#Xp.f.>XsXwXwX%.MX5XY F.UXkXX.e.jX",
"eXeXqXqX9XfXB.A d.t.C T.hX9X&.bXwX<X>.D W x.aX0X",
";X-XrX9XyX#XZ 0X~._.-XA 7XrX%.mX6X9XUX7.+XVX7XqX",
"!.N.fX6XkXs.q.CXd.&.GX%.G.hX#.nX7X5XMXq.|.dX9XqX",
"^.Z.fX6XkXh.5.AX^.2.FXO.H.gX*.bX5X6XVX0. XhX9XqX",
"<X1XrX0XtX>XA @XdXrX^.D kXeX` SXuX3XCX0. XhX9XqX",
"wX7X5X6X3XyXL.B $.+.B @XN.^ ) ( Q.yXNXq. XhX9XqX",
"rXnXmXmXmXbXFXlX_._.KXD.' sXxX0XJ (.HX9. XhX9XqX",
",Xu.e.r.r.r.w.y.k.k.t.V JX5X_ DX=X[ UX8. XhX9XqX",
",Xy.e.r.r.r.r.e.q.y.;.I UXaXO.%XBXQ HX7.{.fX9XqX",
"2XZ.V.V.V.V.V.V.V.B.H.J aXwXh.UXP.#.UXi.:XcX6XqX",
"rXvXbXvXvXvXvXvXvXxXSX(.S $XbXH.G 8X_.m #.;XaX8X",
"wXsXdXdXdXdXdXdXdXdXaXMX-X<.#.e.dX1...,X^.F R.dX",
"-XQ.'._._._._._._._.`.(. X$X.X<Xc.#.UXm.qXcXU 0X",
"^.Q u.8.9.9.9.9.9.9.9.0.8.7.5.i.z #XUXC.3.UX:.R.",
"].{ 0.4.5.5.5.5.5.5.5.5.5.5.4.y.B /.UX,.p.UX .].",
"5XuXhXfXfXfXfXfXfXfXfXfXfXfXdXzX,XJ sXVXUXD.{ gX",
"wXqX9X9X9X9X9X9X9X9X9X9X9X9X0X7XaXR.T 7.o.{ 5XrX",
"qXqXwXqXqXqXqXqXqXqXqXqXqXqXqXwX9XdX0XY.].gXrX0X"
};

/* XPM */
static const char *const xpm_icon_8[] = {
/* columns rows colors chars-per-pixel */
"16 16 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXwX3X+X8X<X9X3X3XtX",
"qXqXqXqXqX0X0XwX4Xy.;Xd.1X=Xg.wX",
"qXqXqXqX9XtXuX8XuXY.W.+X'.:X~.).",
"qXqXqX0XaX=X#XiXtXS.*Xh.@X2Xd.1X",
"wXeX9XaXD.8.r.z.fXm.gX~.y.r. XiX",
"&X|.sX%Xu.}.U.l.[.G.9XcXW..XkX8X",
"#X).dXoXu.7X'.v.oXJ.fXaX~.|.iX0X",
"qXiXyXjXM.h.m.f.L.<.D.jX!.|.aX9X",
"0X-X*X*X<XL.Y.A.G.;XS.R.].{.aX0X",
"3X8.>.<.,.8.8.^ PXs.dXM. X+XpX9X",
"0X*X%X&X&X#X,XM.S.4XD.R.B.Z.gX0X",
"wXjXkXkXkXkXkXkXU.M.T.j.A.V.b.pX",
"&X7.f.s.s.s.s.d.v.m.,.(.&X+XL.|.",
"-Xh.V.N.N.N.N.N.n.A.7.L.1X;XB.+X",
"wXdXfXfXfXfXfXfXfXdXlXb.B.n.F.aX",
"qX8X8X8X8X8X8X8X8X8X8XpX.X@XaX0X"
};

const char *const *const xpm_icons[] = {
    xpm_icon_0,
    xpm_icon_1,
    xpm_icon_2,
    xpm_icon_3,
    xpm_icon_4,
    xpm_icon_5,
    xpm_icon_6,
    xpm_icon_7,
    xpm_icon_8,
};
const int n_xpm_icons = 9;
