#include "gtk.h"
/* XPM */
static const char *const xpm_icon_0[] = {
/* columns rows colors chars-per-pixel */
"128 128 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXjX9X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X9XjXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXxX9Xd 1 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 1 d 9XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXxX7X1                                                                                                                                                                                                                             1 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXxX7X<                                                                                                                                                                                                                             3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXxXeX( 7.d.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.f.| | f.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.h.|     o ^ $.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.R ) #.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.@.%.N     3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXiXf.eXVXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXCXR.R.CXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXDXR.    + m.}.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.].h.Z.[.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'.'..X%.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^.^.^./.(.(.(./.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^./.T.p.R N G &.G.(.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^./.e.k , r 2 q O.F.(.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^./.P.( = A 3. .; d i._.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^./.G.-.( j./.L.w 0 q._.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^.^./.(./.R.0.4 B l.).^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^.^.^.^.R.g.p 4 3.W./.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^.^.^.E.u.p 2 -.K./.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^./.R.8.u 4 ;.K././.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^.W.3.t # a | =.>.M.(.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^./.F.T     - 4 4 9 4.`.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^./.K.3.Y R T Y Y E f.).^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^./.).`.`.`.`.`.`.(.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.a.M.(.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + m.{.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.'.h.C.].`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`. X%.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.9XmXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXMXI.I.MXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXVXU.    X S W R R R R R R R R R R R R R R R R R R R R R R E B F E R R R R R R R R R R R R R R R R R R R R R R Q g     3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXu.Y.'.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.].z.z.].`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.}.e.                                                                                                                  3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXy.z.N.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.N.u.u.N.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.M.Z.&.                                                                                                                  3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.7XvXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXbXP.P.bXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXNXP.    . 8 t r r r r r r r r r r r r r r r r r r r r r e @     ; t r r r r r r r r r r r r r r r r r r r r r t <     3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + j.~.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.U.e     Y ^.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E./.o.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R '.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXxXbXmXmXbXxXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^./.).`.`._.(.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXkXrX{.U.U.}.wXzXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.T.v.r.q.s.A.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXxX0Xv.n 5 2 N m.rXzXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.Q.e.M 1 ; w ) D.~.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXxXR.w 0 ..( 4 z F.vXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.~.F.f = M L w , ..Y./.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXzXaXU.` z.7XeXY , t.nXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.W.s.h I j.T.-.# Z P./.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXjXtX2XqX8X(.z i B.bXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.~.U.N.D.Q.E.$.% T I./.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXxXzX:XI d u 0.7XzXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^./.).).^.0.p 0 p.W.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXzXcXxX3XX.F u Q ^.cXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^./.!.9.j 9 #.T.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXzXaX>XXX3X,X XY % %.dXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.R.7.h w O.Y.).^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXxXeXC.~ y. X+X) @ [ yXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.T.6.i % H d.m.B.!.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXzXdXA.0 2 F C 4 e d.xXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.!.n.t   > j V V Q I./.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXzXwXZ.J t t J B.eXzXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.W.s.p - > , , 1 j K./.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXkXsXwX8X8XwXaXkXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.W.J.G.G.G.G.G.K.~.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXzXxXxXxXxXzXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^./.(.(.(.(.(./.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.7XcXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXbXP.P.bXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXMXP.    + c._./././././././././././././././././././././.W.r     E './././././././././././././././././././././.].@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXi.:XaXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXsXG.G.sXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXgXS.    + j.!.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.I.e     Y ^.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E./.o.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXy.t.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.0.0.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.f.}     o #.r.w.w.w.w.w.w.w.w.w.w.w.w.w.w.w.w.w.w.w.w.w.8.5     m y.w.w.w.w.w.w.w.w.w.w.w.w.w.w.w.w.w.w.w.w.w.y.P     3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXi.+X8X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X9XA.A.9X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6XwXm.    O s.Y.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.H.w     I R.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.E.|     3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.8XbXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXnXI.I.nXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXBXI.    + c.`./././././././././././././././././././././.W.r     E ]./././././././././././././././././././././.].@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXvXP.P.vXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXmXP.    + x._.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXI.I.nXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXBXI.    + v.'.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.Q.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.9XnXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXbX0.h x U P P P P P P P P P P P P P P P P P P P P P P Y x     . g m n n n n n n n n n n n n n n n n n n n n n b &     E ].(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.[.#.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXu.Z.I.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.I.h                                                                                                                   A v.l.l.l.l.l.l.l.l.l.l.l.l.l.l.l.l.l.l.l.l.l.b.!     3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXu.Z.I.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.I.x                                                                                                                   A v.l.l.l.l.l.l.l.l.l.l.l.l.l.l.l.l.l.l.l.l.l.b.!     3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.9XnXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXnXU     > m n n n n n n n n n n n n n n n n n n n n n n n c p v n n n n n n n n n n n n n n n n n n n n n n n b &     E ].(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.[.#.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     m _.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.).U.e.Q.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.Q.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^.^.^.(.`.'.'._./.^.^.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.(.`.'.'._./.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^./.!.A.5.#.*.t.J.(.^.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.!.Z.6.$.&.r.L./.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^./.V.( 7 < : d :.P./.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.(.m./ 9 < > s ,.L./.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^./.Q.<.* e _ J ; d e.).^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.*.< u ` U - a y._.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^./.H.#.l <.U.U.z < =._.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^./.Y.4.Q q.R.J.a < 3.`.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^.!.T.K.R.!.N.s r 7.).^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.~.R.I.I.N.1.3 g p._.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^.^././.^.J.K ; ^ H./.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^./.~.d.q 1 w 2.Y./.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^.^./.^.A.U , L c.^.^.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^./.(./.B.~ Z 5 n s._.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^./.~.m.Y 1 P c.(.).^.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^./.I.z.s.A.G.v.z # | W.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^.~.v.U % 7 #.p.f.Z./.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^./.H.o.z ` i.w.f X [ E./.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^./.U.@.    6 g j l :._.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^./.E.=.7 > s r > f e.(.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^./.K.&.e r e e e y $._.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^./.K.4.V i d P s.E./.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^.^.!.E.E.E.E.E.E.!.^.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.!.W.E.W.Q.~.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     n (.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.P.w.E./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.E.r     R `.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.'.@.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.6XxXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXcXP     M ].`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.].R.y.^.'.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.'./.t     Q |.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`. X%.    3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXp.8XbXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXbXI     i q.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.q.2.` 6.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.7.4     n r.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.r.L     3 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXuXu.+X8X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X8XS                                                                                                                                                                         1 7XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXzXiXh.u.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.p.a.m < 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 1 d 9XxXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXkXiXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuX0X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X9XjXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX"
};

/* XPM */
static const char *const xpm_icon_1[] = {
/* columns rows colors chars-per-pixel */
"96 96 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXkXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXkXlXlXlX",
"lXlXjXmXLXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXLXmXjXlXlX",
"lXkXmX>Xz.C.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.C.z.>XmXkXlX",
"lXgXKXz.                                                                                                                                                                                z.LXgXlX",
"lXgXJXM.                                                                                    o $ +                                                                                   :   C.HXgXlX",
"lXhXSXW.#.Q.F.H.H.H.H.H.H.H.H.H.H.H.H.H.H.H.H.F.E.@.d.U.G.H.H.H.H.H.H.H.H.H.H.H.H.H.H.H.D.Q.8   & w.6.7.7.7.7.7.7.7.7.7.7.7.7.7.7.7.7.9.,.` w.6.7.7.7.7.7.7.7.7.7.7.7.7.7.7.7.6.y.^     N.HXgXlX",
"lXjXCX[.~.UXSXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXAXUXD.6XUXFXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXZXUXf   > %X X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X+X/.g.$X X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X|.>Xy.    N.HXgXlX",
"lXjXZX`.D.CXdXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXsXVXc.@XbXfXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXaXAXp   : _.W.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.^.J.e._.W.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.E.|.6.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.).L.t.'.~.^.^.^.^.^.^.~.~.~.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.).L.t.'.~.^.^.^.^.^./.[.}.]./.~.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.).L.t.'.~.^.^.^.^./.^.M.j.B.!.).~.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.).L.t.'.~.^.^.^./.(.} @   # &.^.~.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.).L.t.'.~.^.^.^.).l.  V a.l # l.[.!.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.).L.t.'.~.^.^.^./.M.-.A.-X,.  s.}.!.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.).L.t.'.~.^.^.^.^.'. X{.Z.< c E./.~.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.).L.t.'.~.^.^.^.^.!./.A.8 5 B._.!.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.).L.t.'.~.^.^.^.^._.B.4 0 F.}.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.).L.t.'.~.^.^.^.(.C.:   n ` Y z.].!.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.).L.t.'.~.^.^.^.).p.# = - = X 4..XQ.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.).L.t.'.~.^.^.^.^./.'.`.`.`.`.(.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.).L.t.'.~.^.^.^.^.^.~.~.~.~.~.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.).L.t.'.~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : (.R.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.!.G.q./.R.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.T.[.4.    N.HXgXlX",
"lXjXCX'.F.CXdXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXdXCXv.#XnXgXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXaXAXp   , ;XOX@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X%X`.l.-XOX@X@X@X@X@X@X@X@X@X@X@X@X@X@X@XoX3Xa.    N.HXgXlX",
"lXjXZX_.U.UXBXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXNXUXM.<XPXCXZXZXZXZXZXZXZXZXZXZXZXZXZXZXZXMXUXd   @ S C C C C C C C C C C C C C C C C C Z M g A C C C C C C C C C C C C C C C C V F d     N.HXgXlX",
"lXjXBXoX2.z.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.k.0.i.j.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.d.n.4                                                                                       $   N.HXgXlX",
"lXjXCX[.m.uX2X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X2XyXk.`.qX3X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X<XaXy   .                                       o X .                                       #   N.HXgXlX",
"lXjXZX`.K.LXcXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXcXLXm.-XDXbXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXzXIXs   ; E.K.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.Y.C.    M '.G.P.P.P.P.P.P.P.P.P.P.P.P.P.P.J./.<.    N.HXgXlX",
"lXjXCX`.G.AXgXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXgXAXb.%XMXjXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXfXFXa   : |.(._._._._._._._._._._._._._._._.[.Y.    A &X~._._._._._._._._._._._._._._./.@X0.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].!.^.^.^.^.^.^.^.^.^.^.^.^.^.^.~._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXhXfXfXgXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.~.Q.W.W.!.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXjXmXPXIXCXjXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.~.|.#XOX_.!.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXjXMX&Xe.9.Q.nXzXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.9.` $.B.[.!.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXhXVX~.2 = $ > Z.vXlXlXlXlXfXGXa   : ].~.^.^.Q.[.[ . ;   g V.(.~.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXjXNX5.g `.qX5 g bXkXlXlXlXfXGXa   : ].~.^.~./.J., v G.6.  ..|.Q.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXkX0XqXlX,X> Y bXlXlXlXlXfXGXa   : ].~.^.^./.R.M.P..Xc.  $.}.Q.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXvXCXY   h 3XcXlXlXlXlXfXGXa   : ].~.^.^.^./.].[.F.2 y T.^.~.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXkXgXcX'.s.6 N zXlXlXlXlXfXGXa   : ].~.^.^.^.^.).Z.9 7 S.`.~.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXkXxXdXy.,.7XbXS   5XxXkXlXlXfXGXa   : ].~.^.^.~./.M.8 , n.!.W./.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXjXmXw.  e p   P xXlXlXlXlXfXGXa   : ].~.^.~.^.Y.1   3 d w ]  XQ.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXkXzXjXS.! E r.aXzXlXlXlXlXfXGXa   : ].~.^.~./.H.U b n v j o.|.Q.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXkXVXIXUXDXkXlXlXlXlXlXfXGXa   : ].~.^.^.^.(.OX@X#X#X$X X!.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXkXgXdXdXfXkXlXlXlXlXlXfXGXa   : ].~.^.^.^.~.W.E.E.E.E.Q.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.DXjXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXjXDXb.&XBXlXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXgXHXa   : [.~././././././././././././././.^._.P.    C #XW././././././././././././././.!.XX8.    N.HXgXlX",
"lXjXCX'.F.BXsXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXsXBXv.@XvXfXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXpXZXp   : `.W.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.(.K.    C OXT.!.!.!.!.!.!.!.!.!.!.!.!.!.!.E.|.6.    N.HXgXlX",
"lXjXBXOX>.a.t.y.y.y.y.y.y.y.y.y.y.y.y.y.y.y.y.t.u.8.w.y.t.y.y.y.y.y.y.y.y.y.y.y.y.y.y.y.e.h.3   & t.9.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.e.2.    f f.7.0.0.0.0.0.0.0.0.0.0.0.0.0.0.8.a.(     N.HXgXlX",
"lXjXCX`.J.HXzXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXlXHXn.=XZXcXvXvXvXvXvXvXvXvXvXvXvXvXvXvXvXkXLXa   : {.^.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.'.U.    Z $XQ.(.(.(.(.(.(.(.(.(.(.(.(.(.(.~.oX9.    N.HXgXlX",
"lXjXCX`.G.AXgXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXgXAXb.%XMXjXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXdXFXa   : '.!.^.^.^.^.^.^.^.^.^.^.^.^.^.^.~.).L.    C +XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXSXb.%XNXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXfXGXa   : ].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXAXv.$XMXjXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXdXDXa   : '.!.~.~.~.~.~.~.~.~.~.~.~.~.~.~.~.).L.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.G.AXgXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXdXJXC.,XHXNXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXbXUXs   > oX'.].].].].].].].].].].].].].].].|.W.    C +XE.~.~.~.~.~.~.~.~.~.~.~.~.~.~.Q..X7.    N.HXgXlX",
"lXjXZX_.P.UXnXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXMXVXjX8.R.2X=X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X%X0Xr   - S.M.N.N.N.N.N.N.N.N.N.N.N.N.N.N.N.C.k.    A *X^.`.`.`.`.`.`.`.`.`.`.`.`.`.`.(.@X0.    N.HXgXlX",
"lXjXCX{.x.3X#X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%XOXeXW                                           . O o                                       o   n !.V.D.D.D.D.D.D.D.D.D.D.D.D.D.D.A.R.-.    N.HXgXlX",
"lXjXBXXX6.V.v.b.b.b.b.b.b.b.b.b.b.b.b.b.b.b.z.J.J   $                                                                                   o   j v.t.p.p.p.p.p.p.p.p.p.p.p.p.p.p.u.z.[     N.HXgXlX",
"lXjXZX_.Y.UXCXSXSXSXSXSXSXSXSXSXSXSXSXSXSXSXMXUX,.    J ^ W Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q W Q M W Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q Q ! Y .   S ;X(.].].].].].].].].].].].].].].`.%Xw.    N.HXgXlX",
"lXjXCX'.F.CXdXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXpXJX%.    I.;X+X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X+X$Xf.OX#X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X&X'.    C OXR.!.!.!.!.!.!.!.!.!.!.!.!.!.!.W. X6.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    b.).E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.Q.7.T.W.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.~.H.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    N.].~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.(.0.!./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.^.^.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    N.].~.^.^.^.^./.(.(./.^.^.^.^.^.^.^.^.(.0.!./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^./.(.(.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    N.].~.^.^.^.(./.E.W.^.(.^.^.^.^.^.^.^.(.0.!./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.).!.E.Q.(./.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    N.].~.^.~.).K.K # # K Y./.^.^.^.^.^.^.(.0.!./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.(.y.e o 3 4.).~.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    N.].~.^.!.[.Y   ! T   Q ).^.^.^.^.^.^.(.0.!./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.(.L.% 0 +.s $ f.{.!.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    N.].~.~./.E. . .W.@X1 g _.^.^.^.^.^.^.(.0.!./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^./.D.o.m.4X,.  0..XQ.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    N.].~.^.^./.|.].[.7.  $.).^.^.^.^.^.^.(.0.!./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.(.|.C.I q S I.(.~.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    N.].~.^.^.^.R._.y.  N W./.^.^.^.^.^.^.(.0.!./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.).[.h.6 , A J._.~.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    N.].~.^.^.!.].9.+ J ).{.^.^.^.^.^.^.^.(.0.!./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^./.Z.s.I..X6.  } {.!.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    N.].~.^.~._.3.@ < &.3.8./.^.^.^.^.^.^.(.0.!./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.).l.$ b 6.L   O.}.!.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    N.].~.~./.Q.b   o     3 _.^.^.^.^.^.^.(.0.!./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^./.Q.{ %   O E P.).~.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    N.].~.^.^.~.K.S.Z.C.C.S.^.^.^.^.^.^.^.(.0.!./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.(.(.J.Z.G.!.).~.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    N.].~.^.^.^._.'.'.'.].'.^.^.^.^.^.^.^.(.0.!./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^./.`.'.`./.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    N.].~.^.^.^.^.~.~.~.~.~.^.^.^.^.^.^.^.(.0.!./.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.    C @XE.^.^.^.^.^.^.~.~.^.^.^.^.^.^.Q..X7.    N.HXgXlX",
"lXjXCX`.H.SXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXdXIX*.    n._.W.W.W.W.W.W.W.W.W.W.W.W.W.W.W.W.W.!.8.R.Q.W.W.W.W.W.W.W.W.W.W.W.W.W.W.W.W.^.H.    V XXY.W.W.W.W.W.W.W.W.W.W.W.W.W.W.R.}.5.    N.HXgXlX",
"lXjXCX'.F.CXdXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXpXKX%.    P.=XoXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOX@Xs.XX+XOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOX$X_.    G 6X|.OXOXOXOXOXOXOXOXOXOXOXOXOXOX.X1Xi.    N.HXgXlX",
"lXjXCX`.R.UXCXSXSXSXSXSXSXSXSXSXSXSXSXSXSXSXNXUX3.  + | <.-.;.;.;.;.;.;.;.;.;.;.;.;.;.;.;.-.:.W =.;.;.;.;.;.;.;.;.;.;.;.;.;.;.;.;.-.>.#.+   d 6.*.;.;.;.;.;.;.;.;.;.;.;.;.;.;.=.3.R o   C.HXgXlX",
"lXjXCX].5.R.F.H.H.H.H.H.H.H.H.H.H.H.H.H.H.H.A._.B                                                                                                                                       z.LXgXlX",
"lXkXcXrX].`.'.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.[.P.n.B.Z.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.C.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.N.N.V.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.A.C.V.z.>XmXkXlX",
"lXlXkXcXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXZXCXDXJXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXLXmXjXlXlX",
"lXlXlXkXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXhXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXkXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX"
};

/* XPM */
static const char *const xpm_icon_2[] = {
/* columns rows colors chars-per-pixel */
"88 88 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXgXdXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXlXlXlXlX",
"lXlXjXSXLXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXLXHXkXlXlXlX",
"lXgXSX,.M K H H H H H H H H H H H H H H H H H G H H H H H H H H H H H H H H H H H H F D G G G G G G G G G G G G G G G G G G G G G G G G G G G G G G G G G G G G G G V ^ nXkXlXlX",
"lXdXPXz                                                                                                                                                               * NXjXlXlX",
"lXfXHX) s C N N N N N N N N N N N N N N N N M y B N N N N N N N N N N N N N N N B b O   8 k f g g g g g g g g g g g g g f h y i h g g g g g g g g g g g g g g f l 2   0 MXkXlXlX",
"lXjXBXV.kXIXGXHXHXHXHXHXHXHXHXHXHXHXHXHXHXJXZXN.IXGXHXHXHXHXHXHXHXHXHXHXHXHXHXGXIXcXo   1.<X}.XXXXXXXXXXXXXXXXXXXXXXXXXX.X&XC.P.$X.XXXXXXXXXXXXXXXXXXXXXXXXXXX{.3X!   0 NXkXlXlX",
"lXhXVXx.4XlXfXgXgXgXgXgXgXgXgXgXgXgXgXgXfXhXiXf.lXdXgXgXgXgXgXgXgXgXgXgXgXgXgXfXlX9XX   %.|.T.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.E.).j.M.(.W.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Y.XXI   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.^.~.].l.V.`.~.^.^.^.^.~.!.!.~.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.^.~.].l.V.`.~.^.^.^.~./.|..X`.!.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.^.~.].l.V.`.~.^.^.^.(.R.y.w.Z.`.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.^.~.].l.V.`.~.^.^.(.R.b o X 8 z.).~.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.^.~.].l.V.`.~.^.~.`.w.o 1.D.= Z {.!.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.^.~.].l.V.`.~.^.^./.I.D.`.).1 R [.~.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.^.~.].l.V.`.~.^.^.^./.].~.l 1 G.(.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.^.~.].l.V.`.~.^.^.~._.R.v 7 F.|.~.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.^.~.].l.V.`.~.^.^.(.L.z   X.i.f.).~.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.^.~.].l.V.`.~.^.~.`.9.  O $   g {.!.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.^.~.].l.V.`.~.^.^.^.R.J.F.S.S.J./.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.^.~.].l.V.`.~.^.^.^.(.`.`.`.'._.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXW.^.^.^.^.^.^.^.^.^.^.^.^.^.~.].l.V.`.~.^.^.^.^.~.~.~.~.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XvXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXzXdXj.vXjXkXkXkXkXkXkXkXkXkXkXkXkXkXkXvXeXX   =.OX!./././././././././././././.^.[.z.C.].^./././././././././././././.Q.@XY   0 NXkXlXlX",
"lXhXVXv.rXVXnXnXnXnXnXnXnXnXnXnXnXnXnXnXnXmXzXk.VXbXnXnXnXnXnXnXnXnXnXnXnXnXnXnXVXpXX   &..XE.~.~.~.~.~.~.~.~.~.~.~.~.~.Q.'.k.B._.!.~.~.~.~.~.~.~.~.~.~.~.~.~.R.OXU   0 NXkXlXlX",
"lXhXVXb.+X8X3X4X4X4X4X4X4X4X4X4X4X4X4X4X4X5X<Xd.7X3X4X4X4X4X4X4X4X4X4X4X4X4X4X3X8X%X    # - * = = = = = = = = = = = = = = * @ @ * = = = = = = = = = = = = = = * - @   7 NXkXlXlX",
"lXhXVXm.g.N.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.m.b.0.M.n.m.m.m.m.m.m.m.m.m.m.m.m.m.n.B.h.  X                                                                               6 NXkXlXlX",
"lXhXVXv.pXGXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXAXmXl.FXVXCXCXCXCXCXCXCXCXCXCXCXCXCXCXFXhX    n o.` ] ] ] ] ] ] ] ] ] ] ] ] ' ..S     E } ] ] ] ] ] ] ] ] ] ] ] ] ] ` O.p   8 NXkXlXlX",
"lXhXVXv.6XxXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXjXaXh.zXgXhXhXhXhXhXhXhXhXhXhXhXhXhXhXxXqXo   3.3X.XOXOXOXOXOXOXOXOXOXOXOXOXXX<Xd.    I.=XoXOXOXOXOXOXOXOXOXOXOXOXOX X5X!   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   %.}.T.W.W.W.W.W.W.W.W.W.W.W.W.T.[.7.    m.(.E.W.W.W.W.W.W.W.W.W.W.W.W.Y..XI   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXhXdXfXjXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.~.W.W.Q.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXkXVXIXKXbXjXlXlXlXlXbXeXX   *.oXQ.^.^.^./.XX@X X^.~.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXkXbXP...<.:XMXjXlXlXlXbXeXX   *.oXQ.^.^.(.U.X.L *.Q.).~.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXgXAXd.  l < v 2XbXjXlXlXbXeXX   *.oXQ.^.~.(.c O i . ( ^./.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXkXvXt.3.IX/.  R.SXgXlXlXbXeXX   *.oXQ.~.).Z.f -.@XP = T.).^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXkXcXlXB.J A 0XvXjXlXlXbXeXX   *.oXQ.^.^.^._.]./.y M /./.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXzXBXsXV , ^ rXvXjXlXlXbXeXX   *.oXQ.^.^.~.(.~.v : C._.~.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXkXzXsXQ.OXGX[.  i.AXhXlXlXbXeXX   *.oXQ.^.^.).E.c 6 F..X~.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXkXkXvX` $ -.R   A.CXhXlXlXbXeXX   *.oXQ.~.^.E.g   m ] ( Y.(.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXjXvX3X| y k v.lXlXkXlXlXbXeXX   *.oXQ.~.(.A.i 7 q 5 w L._.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXcXVXAXZXMXzXlXlXlXlXbXeXX   *.oXQ.^.^./.}.|.}.}.}./.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXjXhXhXkXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.!.!.!.!.!.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.6XcXhXjXjXjXjXjXjXjXjXjXjXjXjXjXjXkXsXh.xXhXjXjXjXjXjXjXjXjXjXjXjXjXjXhXxXqXX   &..XE.~.~.~.~.~.~.~.~.~.~.~.~.W.|.9.    N.`.!.~.~.~.~.~.~.~.~.~.~.~.~.R.OXU   0 NXkXlXlX",
"lXhXVXv.aXGXCXZXZXZXZXZXZXZXZXZXZXZXZXZXZXAXmXl.GXCXZXZXZXZXZXZXZXZXZXZXZXZXZXCXGXhXX   :.&X).'.'.'.'.'.'.'.'.'.'.'.'._.#Xt.    D..X`.'.'.'.'.'.'.'.'.'.'.'.'.(.=XR   0 NXkXlXlX",
"lXhXVXn.K.).^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.!.t.(.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.X   ' G.c.b.b.b.b.b.b.b.b.b.b.b.b.c.S.$.    r.V.b.b.b.b.b.b.b.b.b.b.b.b.b.x.J.A   0 NXkXlXlX",
"lXhXVXn.K.).^.^.^.^.^.^.^.^.^.^.^.^.^.^.^./.!.t.(.^.^.^.^.^.^.^.^.^.^.^.^.^.^.^._.P.X   ' G.c.b.b.b.b.b.b.b.b.b.b.b.b.c.S.$.    r.V.b.b.b.b.b.b.b.b.b.b.b.b.b.x.J.A   0 NXkXlXlX",
"lXhXVXv.aXGXCXZXZXZXZXZXZXZXZXZXZXZXZXZXZXAXmXl.GXCXZXZXZXZXZXZXZXZXZXZXZXZXZXCXGXhXX   :.&X).'.'.'.'.'.'.'.'.'.'.'.'._.#Xt.    D..X`.'.'.'.'.'.'.'.'.'.'.'.'.(.=XR   0 NXkXlXlX",
"lXhXVXv.6XcXhXjXjXjXjXjXjXjXjXjXjXjXjXjXjXkXsXh.xXhXjXjXjXjXjXjXjXjXjXjXjXjXjXhXxXqXX   &..XE.~.~.~.~.~.~.~.~.~.~.~.~.W.|.9.    N.`.!.~.~.~.~.~.~.~.~.~.~.~.~.R.OXU   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXfXj.vXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXbXeXX   *.oXQ.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXxXpXd.kXsXfXfXfXfXfXfXfXfXfXfXfXfXfXdXkX8XX   %.}.T.W.W.W.W.W.W.W.W.W.W.W.W.T.[.7.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.6XxXhXhXhXhXhXhXhXhXhXhXhXhXhXhXjXhXCXC.UXKXLXLXLXLXLXLXLXLXLXLXLXLXLXLXUXnXo   3.3X.XOXOXOXOXOXOXOXOXOXOXOXOXXX<Xd.    N.`.!.~.~.~.~.~.~.~.~.~.~.~.~.R.OXU   0 NXkXlXlX",
"lXhXVXv.pXGXCXCXCXCXCXCXCXCXCXCXCXCXCXCXBXLXk.F 0.6.6.6.6.6.6.6.6.6.6.6.6.6.6.6.9.,..   n o.` ] ] ] ] ] ] ] ] ] ] ] ] ' ..S     S..X`.'.'.'.'.'.'.'.'.'.'.'.'.(.=XR   0 NXkXlXlX",
"lXhXVXm.g.N.m.m.m.m.m.m.m.m.m.m.m.m.m.m.v.H.&                                                                               o   2.h.p.a.a.a.a.a.a.a.a.a.a.a.a.y.b.M   9 NXkXlXlX",
"lXhXVXb.+X8X3X4X4X4X4X4X4X4X4X4X4X4X4X4X1XiXh   & ; - - - - - - - - - - - - - - - - # + & = = = = = = = = = = = = = = * = $ .   l.R.K.L.L.L.L.L.L.L.L.L.L.L.L.G._.K   0 NXkXlXlX",
"lXhXVXv.rXVXnXnXnXnXnXnXnXnXnXnXnXnXnXnXxXPXz   N oXE.~.~.~.~.~.~.~.~.~.~.~.~.~.!.).D.d.'.Q.~.~.~.~.~.~.~.~.~.~.~.~.~.W.|.9.    Z.}.(.).).).).).).).).).).).).~.%XT   0 NXkXlXlX",
"lXhXVXv.8XvXkXkXkXkXkXkXkXkXkXkXkXkXkXkXgXDXl   B @XQ./././././././././././././.^.`.G.g.[.^./././././././././././././.!..Xq.    B.'.~.^.^.^.^.^.^.^.^.^.^.^.^.E.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXl   N +XW.^.^.^.^.^.~.~.^.^.^.^.^.^.~._.F.f.].!.^.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^.^.~.^.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXl   N +XW.^.^.^.~.(.`._.^.^.^.^.^.^.~._.F.f.].~.^.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.^._.`.(.~.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXl   N +XW.^.^.~._.E.F.P./.(.^.^.^.^.~._.F.f.].~.^.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^./._.L.F.R._.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXl   N +XW.^.!.'.v.u O * -.(.~.^.^.^.~._.F.f.].~.^.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^./.^.| = . d B.(.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXl   N +XW.~.~._.k : u.H O v.].!.^.^.~._.F.f.].~.^.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^._.m.% Q p.- m {.!.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXl   N +XW.^./.W.p.b.&Xr.  z.[.!.^.^.~._.F.f.].~.^.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^./.L.z.!.^.5 V {.!.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXl   N +XW.^.^./.[..XV.= G (.~.~.^.^.~._.F.f.].~.^.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.~.(.+X+.  q V._.~.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXl   N +XW.^.^.~.`.m.5 m ^._.!.^.^.^.~._.F.f.].~.^.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.R.!.V.2.0 b _.~.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXl   N +XW.^.~.).x.5 w M.S.T.(.^.^.^.~._.F.f.].~.^.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^._.p.j 0.T.s $ E.^.~.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXl   N +XW.~.~.^.a   & - . w.|.Q.^.^.~._.F.f.].~.^.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.(.G.a   + # +.'.~.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXl   N +XW.^.^.!.j.0.8.8.5.F._.~.^.^.~._.F.f.].~.^.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.).R.i.0.v._.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXl   N +XW.^.^./.{. X.X.XXX_.~.^.^.^.~._.F.f.].~.^.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.~.(.|..X].~.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXl   N +XW.^.^.^.!.Q.!.!.Q.~.^.^.^.^.~._.F.f.].~.^.^.^.^.^.^.^.^.^.^.^.^.^.Q. X0.    B.'.~.^.^.^.~.!.!.!.^.^.^.^.^.W.+XY   0 NXkXlXlX",
"lXhXVXv.8XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXgXFXl   M .XT.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.W./.A.s.).E.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.R.[.8.    m.).W.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Y.XXI   0 NXkXlXlX",
"lXhXVXv.9XbXlXlXlXlXlXlXlXlXlXlXlXlXlXlXhXFXz   S 3X|.oXoXoXoXoXoXoXoXoXoXoXoXoXXX$XW.m.&X.XoXoXoXoXoXoXoXoXoXoXoXoXoX X>Xs.    P.&XXXoXoXoXoXoXoXoXoXoXoXoXoX}.4X!   q MXkXlXlX",
"lXhXCXz.6XcXjXjXjXjXjXjXjXjXjXjXjXjXjXjXdXSXa   # u e r r r r r r r r r r r r r r t 0 5 t e r r r r r r r r r r r r r e y 2     8 t r r r r r r r r r r r r r e i &   1 NXkXlXlX",
"lXjXNXF.i.k.h.h.h.h.h.h.h.h.h.h.h.h.h.h.g.b.k 2 5 ; : : : : : : : : : : : : : : : : > , : : : : : : : : : : : : : : : : ; < 7 7 > : : : : : : : : : : : : : : : ; 3 1 l MXkXlXlX",
"lXlXlXcXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXzXmXNXNXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXMXMXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXBXNXNXMXlXlXlXlX",
"lXlXlXkXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX"
};

/* XPM */
static const char *const xpm_icon_3[] = {
/* columns rows colors chars-per-pixel */
"64 64 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXlXlXlXlX",
"lXlXlXlXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXlXlXlXlX",
"lXlXlXkXpXaXaXaXaXaXaXaXaXaXaXaXaXpXpXaXaXaXaXaXaXaXaXaXaXaXpXyXiXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXuXuXkXlXlXlX",
"lXkXcXuX=                                                     - #   . . . . . . . . . . . o . . . . . . . . . . . . + 5 uXcXkXlX",
"lXkXvXyXn S S S S S S S S S S S G v v G S S S S S S S S S F n   , c j k k k k k k k k k j i k k k k k k k k k k j c # + uXcXkXlX",
"lXlXvXeX{.UXAXDXDXDXDXDXDXDXDXSXUX2X2XUXSXDXDXDXDXDXDXDXAXUX0X  ^ >X{..X.X.X.X.X.X.X X.X}.C.XX X.X.X.X X X.X.X.X}.;Xc . pXcXkXlX",
"lXlXvXeXE.vXfXgXgXgXgXgXgXgXgXfXbXoXoXbXfXgXgXgXgXgXgXgXfXbX*X  T  XR.Q.Q.Q.Q.Q.Q.Q.Q.!.E.k.~.Q.Q.Q.W.Q.W.W.Q.Q.E.}.j . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.^./.!.x.(.^.^.^.]._.`._.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.^./.!.x.(.^.~.'.1.t v B._.~.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.^./.!.x.(.~.^.~.Z r.I A XX!.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.^./.!.x.(.^.^.!.}.`.f 1.{.!.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.^./.!.x.(.^.~.).T.x O.OX!.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.^./.!.x.(.~.~.).d   B X.].!.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.^./.!.x.(.~.^.!.a.w.1.g._.~.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  T  XR.!.!.!.!.!.!.!.!.!.W.l.~.Q.!.!.].{.|.'.Q.!.E.}.j . pXcXkXlX",
"lXlXvXeXQ.mXhXkXkXkXkXkXkXkXkXjXMX+X+XMXjXkXkXkXkXkXkXkXhXMX;X  ~ ;X].|.|.|.|.|.|.|.|.|.[.B. X}.|.|.{.[.[.{.|.|.[.=Xx   pXcXkXlX",
"lXlXvXeX~.CXxXvXvXvXvXvXvXvXvXxXZX$X$XZXxXvXvXvXvXvXvXvXxXCX>X  N l.t.u.u.u.u.u.u.u.u.i.r.$.i.u.u.u.u.u.u.u.u.u.t.k.r o iXcXkXlX",
"lXlXcXiXv.E.Y.T.T.T.T.T.T.T.T.Y.Q.N.N.Q.Y.T.T.T.T.T.T.T.U.~.z.                                                        - yXcXkXlX",
"lXlXvXrXW.cXdXgXgXgXgXgXgXgXgXdXvXoXoXvXdXgXgXgXgXgXgXgXdXvX%X  j <.$.&.&.&.&.&.&.&.$.<.S   W >.$.&.&.&.&.&.&.&.$.,.7 + iXcXkXlX",
"lXlXvXeX!.NXkXzXzXzXzXzXzXzXzXkXBX#X#XBXkXzXzXlXkXlXzXzXkXBX:X  ^ <X}.XXXX.X X.XXXXX}.>X<.  k.-X|.XXXXXXXXXXXXXX|.:Xc   pXcXkXlX",
"lXlXvXeXQ.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXkXcXVXxXkXlXjXNX;X  T |.R.Q.W.(._.Q.Q.Q.R.}.$.  t.].E.Q.Q.Q.Q.Q.Q.Q.R.{.j . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXkXkXnXsX XpXnXkXjXNX;X  R oXQ.~.'.S.b.(./.^.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXhXbX7Xm x F 9XvXhXNX;X  R oXE.[.4.4 q ` _.~.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXhXnX:XV.5Xt [.VXgXNX;X  R oXW.'.<.j.k.3 ^./.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXkXjXCXZXF F iXcXhXNX;X  R oXQ.!.oX`.j -.[.!.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXhXmXOXf.,X3 b.DXfXNX;X  R oXW.(.J.p T ~.~.^.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXhXbX2XC h J 2XbXhXNX;X  R oXE.{._   f v E.(.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXkXkXbXcX9XzXbXkXjXNX;X  R oXQ.^.R.Y.H.K.^.^.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeXQ.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXkXlXbXlXjXlXjXMX;X  R oXQ.^.(.)._.).^.^.Q.XX&.  u.}.Q.^.^.^.^.^.^.^.Q. Xk . pXcXkXlX",
"lXlXvXeX~.CXxXvXvXvXvXvXvXvXvXxXZX$X$XZXxXvXvXcXcXcXvXvXxXZX,X  E +X!.(.(./././.(.(.~.OX*.  p..X~.(.(.(.(.(.(.(.~.oXk . pXcXkXlX",
"lXlXcXrXU.aXwXeXeXeXeXeXeXeXeXwXsX[.[.sXwXeXeXeXeXeXeXeXqXsX.X  U ].P.T.T.T.T.T.T.T.I.'.+.  w.).I.T.T.T.T.T.T.T.I.`.h . pXcXkXlX",
"lXlXcXiXn.^.Q.!.!.!.!.!.!.!.!.Q.(.Z.Z.(.Q.!.!.!.!.!.!.!.W._.M.  S D.z.v.v.v.v.v.v.v.z.S._   -.Z.x.v.v.v.v.v.v.v.x.A.i o pXcXkXlX",
"lXlXvXeX/.FXnXMXMXMXMXMXMXMXMXnXGX*X*XGXnXMXMXMXMXMXMXMXnXFX3X  W #X/._._._._._._._./.@X=.  a.oX(._._._._._._._./.+Xl   pXcXkXlX",
"lXlXvXeXQ.nXgXjXjXjXjXjXjXjXjXhXmX+X+XmXhXjXjXjXjXjXjXjXgXmX-X  R XXW.~.~.~.~.~.~.~.W..X%.  y.}.Q.~.~.~.~.~.~.~.W. Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXjXNX@X@XNXjXlXlXlXlXlXlXlXjXNX;X  R oXQ.^.^.^.^.^.^.^.Q.XX&.  u.|.!.^.^.^.^.^.^.^.Q..Xk . pXcXkXlX",
"lXlXvXeXQ.nXhXjXjXjXjXjXjXjXjXhXmX+X+XMXhXkXkXkXkXkXkXkXhXMX-X  R XXW.~.~.~.~.~.~.~.W..X%.  u.}.Q.~.~.~.~.~.~.~.Q. Xk . pXcXkXlX",
"lXlXvXwX/.GXmXNXNXNXNXNXNXNXNXmXGX*X*XDXbXmXmXmXmXmXmXmXbXDX2X  Q %X(.'.'.'.'.'.'.'.(.$X-.  a.oX(._._._._._._._.(.+Xl   pXcXkXlX",
"lXlXcXyXA.*X+X@X@X@X@X@X@X@X@X+X*Xm 4 e 0 0 0 0 0 0 0 0 0 w 5   + 4 1 2 2 2 2 2 2 2 1 3 ;   <.I.C.A.A.A.A.A.A.A.C.U.s X pXcXkXlX",
"lXlXcXyXA.*X+X@X@X@X@X@X@X@X@X+X*X4   1 , , , , , , , , , , 1 4 4 1 2 2 2 2 2 2 2 2 1 3 ;   <.I.C.A.A.A.A.A.A.A.C.U.s X pXcXkXlX",
"lXlXvXwX/.GXmXNXNXNXNXNXNXNXNXMXDXe 1  X`.'.'.'.'.'.'.'.'.`.}.m.Q.[.`.'.'.'.'.'.'.'.(.$X-.  a.oX(._._._._._._._.(.+Xl   pXcXkXlX",
"lXlXvXeXQ.nXhXjXjXjXjXjXjXjXjXhXbX0 , `.!.~.~.Q.Q.Q.~.~.~.!.).c.Y./.~.~.~.~.~.~.~.~.W..X%.  u.}.Q.~.!.Q.Q.!.~.~.Q. Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXkXmX0 , '.~.^./.}. X[.^.^.^.~._.v.T.(.^.^.^.^.^.^.^.^.Q.XX&.  u.|.!.~.'. X|.).^.^.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXkXmX0 , '.~./.!.[ g :._.~.^.~._.v.T.(.^.^.^.^.^.^.^.^.Q.XX&.  u.|.W.`.g.b F D.).~.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXkXmX0 , '.!.[.i.s 8.2 x.].!.~._.v.T.(.^.^.^.^.^.^.^.^.Q.XX&.  u.}.Q.`.L -.F B .X!.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXkXmX0 , '.~./.T..Xb.w G._.~.~._.v.T.(.^.^.^.^.^.^.^.^.Q.XX&.  u.|.!.!.[.i.i 1.{.!.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXkXmX0 , '.~.!.}.v.i v. XQ.^.~._.v.T.(.^.^.^.^.^.^.^.^.Q.XX&.  u.}.!.W.H.c.C P [.~.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXkXmX0 , '.!.`.v.  r ( M.`.~.~._.v.T.(.^.^.^.^.^.^.^.^.Q.XX&.  u.}.Q./.C D k n [.!.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXkXmX0 , '.!._.N.X.#.' A._.~.~._.v.T.(.^.^.^.^.^.^.^.^.Q.XX&.  u.|.Q.).D.X.&.T.(.~.Q..Xk . pXcXkXlX",
"lXlXvXeX!.MXjXlXlXlXlXlXlXlXlXkXmX0 , `.!.!.).XX XoX(.!.~.!.).c.U.^.!.~.~.~.~.~.~.~.W. X%.  y.{.W.!.(..X X^.!.~.W.|.j . pXcXkXlX",
"lXlXvXeXQ.nXgXjXjXjXjXjXjXjXjXhXbX0 1 XX'.].'._._._.].].].'.|.B.^.{.].].].].].].].]._.&X;.  f.@X`.].'._.`.].].]._.$Xz . pXcXkXlX",
"lXlXvXwX^.FXnXMXMXMXMXMXMXMXMXmXSX0 & B.v.v.b.b.b.b.b.b.b.v.M.7.j.n.v.b.b.b.b.b.b.b.x.D.)   =.A.x.b.b.b.b.b.b.b.x.S.y   pXcXkXlX",
"lXlXxXpXN.^.Q.!.!.!.!.!.!.!.Q.Q./.i + $ # # # # # # # # # # # $ # # # # # # # # # # # # % = $ # # # # # # # # # # # # 4 uXcXkXlX",
"lXlXlXlXpXwXeXeXeXeXeXeXeXeXeXeXeXyXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXiXuXuXuXiXuXuXuXuXuXuXuXuXuXiXuXuXkXlXlXlX",
"lXlXlXlXxXvXvXvXvXvXvXvXvXvXvXvXvXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX"
};

/* XPM */
static const char *const xpm_icon_4[] = {
/* columns rows colors chars-per-pixel */
"48 48 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXkXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXlXlXlXlX",
"lXlXkXbXFXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXSXSXDXDXDXDXDXDXDXDXSXDXDXDXDXDXDXDXDXDXZXlXlXlXlX",
"lXkXbX5XV.D.S.S.S.S.S.S.S.D.S.S.S.S.S.S.S.S.D.K.G.D.F.F.F.F.F.F.F.F.F.F.F.F.F.F.F.F.D.Q.lXlXlXlX",
"lXhXSXK.  ; * * * * * * * @ % * * * * * * * %     + O O O O O O O X O O O O O O O #   > lXzXlXlX",
"lXkXmX=X6XbXkXlXlXlXlXjXnXOXyXcXkXlXlXlXkXvXwX# g.}.~./././././.~.C.(.^.^.!.!.^.!..X,.3 xXzXlXlX",
"lXkXmX*X7XmXlXzXzXzXzXlXMX+XuXvXzXzXzXzXlXnXeXo f.}.~./././././.^.C.(.^.)..X|.).!.XX>.< xXzXlXlX",
"lXkXmX*X5XbXkXkXkXkXkXjXnXoXtXcXkXkXkXkXjXvXwXo d.{.!.^.^.^.^.^.!.B.^./.R._ [ Y.~. X:.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXkXnXOXyXcXlXlXlXlXkXbXwXo f.{.!.^.^.^.^.^.~.V.^.`.m.6.S 4.{.|.>.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXkXnXOXyXcXlXlXlXlXkXbXwXo f.{.!.^.^.^.^.^.~.V./.!.}.v./ !.!..X>.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXkXnXOXyXcXlXlXlXlXkXbXwXo f.{.!.^.^.^.^.^.~.V.^._.Z.: V m.). X>.< xXzXlXlX",
"lXkXmX*X5XbXkXkXkXkXkXjXnXoXtXxXkXkXkXkXjXvXqXO a.'.E.Q.Q.Q.Q.Q.W.M.!.!.I.B.x.H.Q.{.;.< xXzXlXlX",
"lXkXmX*X0XBXvXbXbXbXbXcXVX#XaXMXvXbXbXbXcXNXyXo m.-X.XoXoXoXoXOX.XL.OXoX+X%X&X@X X,X8., xXzXlXlX",
"lXkXmX*X#XrX8X9X9X9X9X8XrX_.,XwX9X9X9X9X8XeX=X@ F { ) _ _ _ _ _ ` ^ ` _ _ ) ) _ ) | c 6 zXzXlXlX",
"lXkXmX*X).=X$X$X$X$X$X#X=XY. X&X$X$X$X$X#X=X'.# 1 q 9 9 9 9 8 q 1   5 0 9 9 9 9 8 w * 0 lXzXlXlX",
"lXkXmX*XwXCXnXmXmXmXmXbXAX$XdXBXmXbXnXbXnXCXpXo c.%X{.}.}.}.{.#XN.o E.oX}.|.|.|.[.-X5., xXzXlXlX",
"lXkXmX*X5XvXjXkXkXkXkXhXbXoXrXxXjXMXvXmXhXcXqXo s.].!.'._.).W.'.j.  G.).Q.!.!.!.W.|.:.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXkXnXOXyXxXnXl.) R.mXcXwXo d.}.W.^ G m.).].l.  H._.~.^.^.^.Q..X>.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXkXnXOXyXcXkX}.=.,.AXxXwXo d. XH.j.)  . X'.l.  H._.~.^.^.^.Q..X>.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXkXnXOXtXvXhX7X:.#.BXxXwXo f.].'.j.P E.^.[.l.  H._.~.^.^.^.Q..X>.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXkXnXOXtXcXxX6.E y.MXcXwXo d.|.I.y x y.`.'.l.  H._.~.^.^.^.Q..X>.< xXzXlXlX",
"lXkXmX*X5XvXjXkXkXkXkXhXbXoXtXxXlXzXrXmXjXcXqXo d.{.W.W.P.Y.~.].k.  H._.!.~.~.~.Q. X:.< xXzXlXlX",
"lXkXmX*XqXCXnXnXnXnXnXbXZX$XsXBXnXnXNXbXbXVXiXO h. X(.`.'.`./.|.x.  L.].(.).).)./.OX<.< xXzXlXlX",
"lXkXmX*X|.4X,X,X,X,X,X>X4XQ.$X2X,X,X,X,X>X3XoXo w.R.G.H.H.J.G.T.y.  n.I.H.J.J.J.F.!.@.1 xXzXlXlX",
"lXkXmX*X|.4X,X,X,X,X,X>X4XQ.$X2X,X,X,X,X>X3XoXo w.R.G.J.J.J.G.T.y.  n.I.H.J.J.J.F.!.@.1 xXzXlXlX",
"lXkXmX*XqXCXnXnXnXnXnXbXZX$XsXBXnXnXnXnXbXVXiXO h. X/.).).).(.|.x.  L.].(.).).)./.OX<.< xXzXlXlX",
"lXkXmX*X5XvXjXkXkXkXkXhXbXoXtXxXkXkXkXkXjXcXqXo d.[.!.~.~.~.!.].k.  H._.!.~.~.~.Q. X:.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXkXnXOXyXcXlXlXlXlXkXbXwXo f.{.!.^.^.^.!.[.l.  H._.~.^.^.^.Q..X>.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXkXnXOXyXcXlXlXlXlXkXbXwXo f.{.!.^.^.^.!.[.l.  H._.~.^.^.^.Q..X>.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXkXnXOXyXcXlXlXlXlXkXbXwXo f.{.!.^.^.^.!.[.l.  H._.~.^.^.^.Q..X>.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXkXnXOXyXcXlXlXlXlXkXbXwXo f.{.!.^.^.^.!.[.l.  H._.~.^.^.^.Q..X>.< xXzXlXlX",
"lXkXmX*X5XvXjXkXkXkXkXhXbXXXeXzXhXjXjXjXhXxX0Xo s.].Q.!.!.!.Q.'.j.  G.).!.~.~.~.Q. X:.< xXzXlXlX",
"lXkXmX*XwXCXnXmXmXmXmXbXZX*XlXFXZXAXAXAXCXGXhX@ v.$X{.|.|.|.{.#XN.  L.[.(.).).)./.+X<.< xXzXlXlX",
"lXkXmX*X).=X$X$X$X$X$X$X*XJ p f f d d d d g p   < q 8 9 9 9 9 0 2 O z.H.A.S.S.S.Z.T.o.1 xXzXlXlX",
"lXkXmX*X#XrX8X9X9X9X9X9XwX, d { ^ ^ ^ / / ^ ) ^ ` _ _ _ _ _ ) [ J X B.E.I.I.I.U.L.).%.1 xXzXlXlX",
"lXkXmX*X0XBXvXbXbXbXvXbXBX5 3.1X.X+X&XOXOXOX+XP..XOXoXoXoXoXXX=XC.  K.]./.'.]./.^.oX,.< xXzXlXlX",
"lXkXmX*X5XbXkXkXkXkXkXkXvX4 %.}.Q.Y.b.^.Q.Q.!.M.E.Q.Q.Q.Q.Q.E.`.h.  H.).(.G.A.`.W..X:.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXlXbX4 &..X!.Q n %.{.!.(.V.!.^.^.^.^.^.!.[.l.  G.[.F.P N d.`.|.>.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXlXbX4 &.XXT.^.>. .|.!.(.V.!.^.^.^.^.^.!.[.l.  H._.!.F.V a.'.|.>.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXlXbX4 &. X`.*.d F.(.~.(.V.!.^.^.^.^.^.!.[.l.  G.].D.<.x -.{.|.>.< xXzXlXlX",
"lXkXmX*X6XbXkXlXlXlXlXlXbX4 &.XXQ.! L 3.`.!./.V.!.^.^.^.^.^.!.[.k.  G.'.U.{ _ A./. X:.< xXzXlXlX",
"lXkXmX*X5XvXjXkXkXkXjXkXvX5 -.OX^.XXXX[.(.(._.Z./.).).).).)./.|.x.X L.'._.XX.X'.~.OX<.1 xXzXlXlX",
"lXkXmX&X0XVXbXbXbXbXbXbXBX, .._.K.L.K.P.I.I.Y.z.L.U.I.I.I.I.L.!.i.  N.R.P.L.L.P.L.).@.= xXzXlXlX",
"lXkXnX:X^.+X.XXXXXXXXXXXOXn i b x c c c c c c l x c c c c c x v k w z c c c c c x b i M lXzXlXlX",
"lXlXlXlXaXiXpXpXpXpXpXpXiXjXhXdXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXdXgXkXfXdXfXfXfXfXfXdXgXlXlXlXlXlX",
"lXlXlXlXxXxXxXxXxXxXxXxXxXzXzXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXxXzXzXxXxXxXxXxXxXxXxXzXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX"
};

/* XPM */
static const char *const xpm_icon_5[] = {
/* columns rows colors chars-per-pixel */
"44 44 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX",
"lXlXlXkXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXlXlXlXlX",
"lXlXkXbXFXDXFXFXFXFXFXFXDXFXFXFXFXFXFXFXFXSXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXZXlXlXlXlX",
"lXkXbX5XM.Z.C.C.C.C.C.C.A.C.C.C.C.C.C.C.C.K.D.A.S.A.A.A.A.A.S.A.A.A.A.A.S.S.A.W.jXzXlXlX",
"lXhXAXY., i t t t t r y 7 w t t t t t t r   - 5 3 3 3 3 3 3 , 3 3 3 2 3 3 6 $ * fXxXkXlX",
"lXkXnX>XpXSXBXBXBXBXNXDX<XjXAXBXBXBXBXCXmX6 z.@X].{.{.{.{.{.K.{.{.[. X{.].@Xc.4 gXxXkXlX",
"lXkXnX;X8XcXjXkXkXkXhXvX#XtXxXjXkXkXjXlXfX2 s.[.Q.~.~.~.~.~.C.~.~.(.J.^.~.].f.2 gXxXkXlX",
"lXkXnX;X9XvXkXlXlXlXkXnX$XyXcXlXlXlXkXzXhX2 d.{.!.^.^.^.^.^.Z.!.[.y.V { (.[.f.2 gXxXkXlX",
"lXkXnX;X9XbXkXlXlXlXkXnX$XyXcXlXlXlXkXxXhX3 d.{.!.^.^.^.^.^.Z.^.^.!.d.' _.[.f.2 gXxXkXlX",
"lXkXnX;X9XbXkXlXlXlXkXnX$XyXcXlXlXlXkXxXhX3 d.{.!.^.^.^.^.^.Z.~.'.l.7 9.).[.f.2 gXxXkXlX",
"lXkXnX;X8XvXkXkXkXkXjXbX$XyXxXkXkXkXkXzXgX3 a.'.E.Q.Q.Q.Q.Q.B.W.(.k.%.q.W.'.s.2 gXxXkXlX",
"lXkXnX;XeXNXvXvXvXvXcXBX*XaXMXvXvXvXvXnXzX3 n.-XXXoXoXoXoXoXI.oXXX&X,X=XXX-XM.1 gXxXkXlX",
"lXkXnX-X&XeX9X9X9X9X8XrX[.,XwX9X9X9X9XqX6X3 H ..] [ [ [ ] { ' { ] ' ` ' ] ..H 4 fXxXkXlX",
"lXkXmX-X}.,X-X-X-X-X=X,X!.+X:X-X=X=X-X:X&X3 0 k f f f f h i   i h f g g f k 0 5 fXxXkXlX",
"lXkXnX;XtXVXnXnXnXnXbXCX=XdXBXbXZXVXbXMXcX3 b.=X X$XOX X#XE.- R.#X.XXXXX X=Xm.1 gXxXkXlX",
"lXkXnX;X8XcXjXkXkXkXhXvX#XtXzXxX:X6XcXkXgX2 a.'.!.b.L.^./.A.& Z.(.Q.!.!.W.'.s.2 gXxXkXlX",
"lXkXnX;X9XbXkXlXlXlXkXnX$XtXMX5X@.( 9XnXgX3 a.OXg.J P E.`.D.& S.`.~.^.^.!.{.f.2 gXxXkXlX",
"lXkXnX;X9XbXkXlXlXlXkXnX$XyXxXzX;XW 6XmXgX3 d.[.(.v.[ /._.D.& S.`.~.^.^.!.{.f.2 gXxXkXlX",
"lXkXnX;X9XbXkXlXlXlXkXnX$XtXMX1X;.D &XNXfX3 s.oXb.2 @.^._.D.& S.`.~.^.^.!.{.f.2 gXxXkXlX",
"lXkXnX;X8XvXjXkXkXkXjXbX$XtXxXjX`.+XxXlXgX2 s. XA.t.s.W.`.D.& S._.!.~.~.!.[.f.2 gXxXkXlX",
"lXkXnX;XrXBXbXnXnXnXvXCX=XsXNXnXFXSXvXMXxX3 g.|.`..X X(.].H.& G.[.(.).)./. Xj.2 gXxXkXlX",
"lXkXnX-XOX6X2X2X2X2X1X7X).*X4X2X<X<X2X3X,X< w.W.H.H.J.J.U.b.% v.Y.J.K.K.J.E.e.2 gXxXkXlX",
"lXkXnX-XOX6X2X2X2X2X1X7X).*X4X2X2X2X2X3X,X< w.W.J.K.K.J.U.b.% v.Y.J.K.K.J.E.e.2 gXxXkXlX",
"lXkXnX;XrXBXbXnXnXnXvXCX=XsXNXbXnXnXbXmXxX3 g. X/.).).(.].H.& G.[.(.).)./. Xj.2 gXxXkXlX",
"lXkXnX;X8XvXjXkXkXkXjXbX$XtXxXkXkXkXjXzXgX2 s.{.!.~.~.~._.D.& S._.!.~.~.!.[.f.2 gXxXkXlX",
"lXkXnX;X9XbXkXlXlXlXkXnX$XyXcXlXlXlXkXxXhX3 d.{.!.^.^.~.`.F.& S.`.~.^.^.!.{.f.2 gXxXkXlX",
"lXkXnX;X9XbXkXlXlXlXkXnX$XyXcXlXlXlXkXxXhX3 d.{.!.^.^.~.`.F.& S.`.~.^.^.!.{.f.2 gXxXkXlX",
"lXkXnX;X9XbXkXlXlXlXkXnX$XyXcXlXlXlXkXxXhX3 d.{.!.^.^.~.`.F.& S.`.~.^.^.!.{.f.2 gXxXkXlX",
"lXkXnX;X8XcXjXkXkXkXhXvX@XeXlXgXhXhXgXjXdX2 a.].W.!.!.Q.(.A.& S._.!.~.~.!.[.f.2 gXxXkXlX",
"lXkXnX;XtXVXnXnXnXnXbXCX1XbXKXFXGXGXFXHXAX6 b.=X XXXXX.X#XE.& G.[.(.).)./. Xj.2 gXxXkXlX",
"lXkXmX-X}.,X-X-X-X-X=X1X[ N D A A A A S Z   q j f g g f h i + x.L.F.G.G.D.Y.0.2 gXxXkXlX",
"lXkXnX-X&XeX9X9X9X9X8XiXq h  .( ( ( _ ) ` ' { ] [ [ [ ] } ! # M.E.I.L.I.P.~.y.2 gXxXkXlX",
"lXkXnX;XeXNXvXvXvXvXcXSXd 0.<XXX:X=XOXOX@XU. XOXoXoXoXXX$XW.& F.].).XX].^.|.h.2 gXxXkXlX",
"lXkXnX;X8XvXkXkXkXkXjXBXa -.|.R.2.s./.W.~.V.R.!.Q.Q.Q.W.(.A.& S.].P.2.M./.[.f.2 gXxXkXlX",
"lXkXnX;X9XbXkXlXlXlXjXBXa :.@Xn...N S._./.A.Q./.^.^.^.~.`.F.& A.|.l.T P (.[.f.2 gXxXkXlX",
"lXkXnX;X9XbXkXlXlXlXjXBXa >. X`.d...(.~.(.A.Q./.^.^.^.~.`.F.& S._.~.8.F ^.{.f.2 gXxXkXlX",
"lXkXnX;X9XbXkXlXlXlXjXBXa :.@XB.6 ! Y././.Z.Q./.^.^.^.~._.D.& A. Xi.S R Q.{.f.2 gXxXkXlX",
"lXkXnX;X8XcXjXkXkXkXhXNXs >.XXR.I.F.Q./.(.A.!./.^.^.^.~.`.F.* D.`./.H.^./.{.g.3 gXxXkXlX",
"lXkXnX-XeXBXbXnXnXnXvXFXy -..X!.)._.^.~./.C.Q.^.~.~.~.!._.S.O A._.!.).^.Q.{.d.; gXxXkXlX",
"lXkXnX>X].*X$X$X$X$X#X;XV d F Z C Z Z Z A N Z Z Z Z Z Z A B 8 N A Z Z Z Z D z k gXxXkXlX",
"lXlXlXkXuXyXuXuXuXuXuXyXdXiXrXtXtXtXtXtXtXyXtXtXtXtXtXtXtXyXdXyXtXtXtXtXtXtXyXgXlXlXlXlX",
"lXlXlXlXcXcXcXcXcXcXcXcXxXcXvXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXxXcXcXcXcXcXcXvXcXzXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXlXlXlXlX",
"lXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlX"
};

/* XPM */
static const char *const xpm_icon_6[] = {
/* columns rows colors chars-per-pixel */
"32 32 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"lXlXnXnXnXnXnXnXnXnXnXnXnXnXnXbXbXnXnXnXnXnXnXnXnXnXnXnXnXvXlXlX",
"lXkX7X4X5X5X5X5X6X6X5X5X5X5X4X0X0X7X7X7X7X7X8X7X7X7X7X7X7XeXkXlX",
"vXtXC v n n n n x x n n n b N 0 e j g g g g d g g d f h s e eXvX",
"vXrX<XcXhXjXjXcX6X6XcXjXjXdXGX' *. XW.!.Q.^.F.R.~.].`.!.W.s 7XnX",
"vXtX3XnXlXzXlXbX9X9XbXlXzXhXJX} ;.oX~./.^.).J.Q.).9.d.`.!.h 7XnX",
"vXtX2XvXkXlXkXvX8X8XvXkXlXgXHX{ -.XXQ.^.~.(.H.Q.^.0._ ).Q.g 7XnX",
"vXtX2XbXkXlXkXvX8X8XvXkXlXgXHX{ =. XW.~.!./.F.R._.R O.'.E.g 7XnX",
"vXtX3XmXzXxXzXnX0X0XnXzXxXjXKX{ 2.*X'.{.[.|.U.`.[.N.D.}.].j 7XnX",
"cXtX:XhXaXsXaXhX2X2XhXaXsXuXVX[ ~ m.d.g.f.h.i.f.g.c.z.g.s.i 9XbX",
"cXyX}.1X:X>X:X1X.X.X1X;X-X=XeX) 1 M l l z b = p v l z c k 4 wXbX",
"vXrX5XNXcXvXcXNXqXqXMXBXDXxXPX} 2.=XXX@X[.#Xb c.#X[.{.}.].j 7XnX",
"vXtX1XvXjXkXjXcX7X6XVX_.g.uXJX[ =.OX3.;.W.[.z d.].Q.!.~.W.g 7XnX",
"vXtX2XbXkXlXkXvX8X7XmX:X_ %XUX[ -.OXh.( T.|.x g.{.~.^./.!.g 7XnX",
"vXtX2XvXkXlXkXvX8X6XBXW.) =XUX] =.%X) / /.{.x f.{.!.~./.!.g 7XnX",
"vXrX4XMXxXcXxXMXqX0XBXaXtXxXLX} ;.OXH.H.!. Xx h.|.^./.).~.g 7XnX",
"cXtX&XiXrXrXrXiX:X:XiXyXpX0XbX) &.].E.Q.Y.(.l i.(.Y.T.R.U.f 7XnX",
"cXyXXX5X1X2X2X6X+X+X6X1X1X,XpX~ @.^.H.J.J.W.j r.E.J.K.L.H.d 8XnX",
"vXrX6XBXvXbXvXNXwXwXNXvXbXzXIX| :.+X^.(./. Xc h.|./.(.).^.h 7XnX",
"vXtX1XvXjXkXjXcX7X7XcXjXkXfXGX{ -..XQ.~.!.{.x f.[.!.~.^.Q.g 7XnX",
"vXtX2XbXkXlXkXvX8X8XvXkXlXgXHX{ ;.XX!.^.!.}.x g.{.~.^./.!.g 7XnX",
"vXtX1XvXjXkXjXvX6X4XzXfXgXaXSX] =.|.E.Q.E.'.z f.[.!.~.^.Q.g 7XnX",
"vXrX5XBXcXvXvXMXyXjXUXHXJXDXUX@.6.1X.XoXXX;Xn g.|./.(.).^.g 7XnX",
"cXyX@XwX7X8X6XyXR.:.w.9.9.7.a.s x X._ ' ` } 5 a.Q.P.I.U.L.d 8XnX",
"cXyX@XwX7X8X4XjX:.6  .^ ( / / _ ] ' ' ' ` } 5 a.Q.P.I.U.L.d 8XnX",
"vXrX5XBXcXvXzXUXw. .5X XoX#X#X^.].@XoXoXXX;Xn g. X!.)._.^.h 7XnX",
"vXtX1XvXjXkXfXHX9.~ @X7.E G.).G.U.~.Q.Q.E.'.z f.}.@.X._.Q.g 7XnX",
"vXtX2XbXkXlXgXJX9./ $Xb.( I.`.K.T.(.^.^.!.}.x g.}.t.G ^.!.g 7XnX",
"vXtX1XvXjXkXfXHX9.~ *X@.E T.(.H.U.^.!.!.W.].x f.].@.o.^.W.g 7XnX",
"vXrX5XBXvXbXzXUX9._ ;X}. X]. XE.(.|.{.{.].#Xx z.#X|.oX|.].f 6XnX",
"cXuXoX5X1X2X,XuX7.g O.' ] ] { ( ` [ ] [ ]  .y U | ' ' [ ` s wXbX",
"lXlXuXrXtXtXtXrXtX6X<X1X1X1X1X2X2X1X1X1X1X1X0X4X1X1X1X1X1XwXlXlX",
"lXlXcXvXvXvXvXvXvXnXmXmXmXmXmXmXmXmXmXmXmXmXbXmXmXmXmXmXmXvXlXlX"
};

/* XPM */
static const char *const xpm_icon_7[] = {
/* columns rows colors chars-per-pixel */
"24 24 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"dX|.[.{.{.{.|.{.{.{.].+X+X X.X.X.XXX.X X.X X&XhX",
"4X@.%.%.%.$.} %.%.#.>.B M ( W Q W Y Q / ! ^ g &X",
"eXhXJXSXDXAXtXDXSXVXUXy.q.-X{. X[.!.XXR./.=X^  X",
"wX9XlXgXhXgX,XjXhXsXSX8.3.}.W.!.E.H.).,.:.+XE .X",
"wXqXcXkXlXjX1XlXkXgXGX9.3.|.W.~.W.H.'. .5.+XE .X",
"wXrXmXxXcXxX4XvXcXkXKXq.q.%X'.{.'.Q.}.Y.Q.%X^  X",
"wX%X6X3X4X3X.X4X2X>XpX<.p R H L J C P U I U u %X",
"wX8XkXfXfXdX>XgXlXkXZX8.O.L.Z.F.t.n F.m.m.H.K oX",
"wXeXbXzXzXlX2XNX_.K.UX8.w.'.<.^.G.U XX_._.+X~ .X",
"wXqXcXkXlXjX<XMXY.O.IX7.6.!.Y P.Z.K ].Q.Q.}.Q .X",
"wXrXmXxXcXxX3XNXOX`.UX0.9.Q.<.(.C.L }.^.~..X! .X",
"wX2XpXyXyXtX*XyXgXdXbX3.<.{.(.^.v.J (.Y.Y.`.E XX",
"wX;XqX8X9X8X+X9X6X3XhX,.,./.H.W.z.G !.P.L./.T XX",
"wXtXMXcXvXcX5XbXvXlXLXw.6.oX~.[.V.P }.^.^..X! .X",
"wXqXcXkXkXjX1XlXkXfXFX9.4. XQ.`.N.L [.!.Q.|.Q .X",
"wXwXcXkXlXkX1XzXlXgXGX0.5..XQ.'.B.L [.!.!.|.Q .X",
"wXeXbXlXzXlX6XcXzXjXJX8.2..XQ.'.N.L {.~.!. XQ .X",
"wX$X5X1X3X<XY U W R ~ c c H D J z C W.H.G.Q.Y XX",
"wXqXxXjXzXiXD J.P.U.P.G.L.Y.I.Q.z.K [.!.^.}.Q .X",
"wXwXvXkXcXpXY `.s.6.}.U.W._./.[.C.K .X5.9.@XW .X",
"wXqXcXjXxXiXP ).<.o.].K.Y.~.Q.).m.K  X=._ @XE .X",
"wXtXmXxXnXdXU  Xm.Z.oX~._. X{.OXG.I %XV.H.-X~  X",
"wX,XtXqXrX6XA 1.r.q.6.,.2.6.5.9.$.v 0.q.9.q.C #X",
"jXwXwXwXwXwX;X.X.X.XXXoXoXXXXXXXOX%XXX.X.X.X&XhX"
};

/* XPM */
static const char *const xpm_icon_8[] = {
/* columns rows colors chars-per-pixel */
"16 16 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"jX>X*X=X-X=X*X,X1X>X>X,X>X;X5XlX",
"uXV.B.B.z.N.G.<...q.4.1.7.q.| 5X",
"pXkXSXCXdXNXUXQ.h.+X)._.n.A.r.;X",
"pXyXzXjXqXgXVXI.h. X/._.c.D.r.:X",
"pX4XrXwX:XwXxXZ.R 4.$.] ;.>.W 4X",
"pXtXlXgXwX9XeXY.q.C.m. .Y.R.,.,X",
"pXpXbXzXdX$XI.(.s.9.D.>.].|.8.>X",
"pX0XdXpX5XwXhXH.u.Q.F.$.~.(.3.,X",
"pX7XiXyX1XtXcXD.u.[.D.#.!./.3.,X",
"pXpXbXxXeXkXSXY.f..XP.*.`.[.6.>X",
"pXyXxXkXyXjXAXP.i.|.J.&.).'.6.>X",
"pX3XwXyX9.2.u.o.] -.[ ' /.(.<.,X",
"pXyXxXzX4.M.c.W.E./.J.;.D.D.0.>X",
"pXiXbXxXt.b.p.].~.}.T.1.b.k.t.-X",
"aX8XiXaX2.s.z.h.j.v.s.| c.b.=.4X",
"kXaXpXaXqX2X3X2X2X2X3X6X2X1X7XlX"
};

const char *const *const xpm_icons[] = {
    xpm_icon_0,
    xpm_icon_1,
    xpm_icon_2,
    xpm_icon_3,
    xpm_icon_4,
    xpm_icon_5,
    xpm_icon_6,
    xpm_icon_7,
    xpm_icon_8,
};
const int n_xpm_icons = 9;
