#include "gtk.h"
/* XPM */
static const char *const xpm_icon_0[] = {
/* columns rows colors chars-per-pixel */
"128 128 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX9X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X8XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX9XaXbXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXcXbXgX9XwXqXqXqXqXqXqX",
"qXqXqXqXqX9XaXE.=.6.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.3.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.4.3.4.5.;.V.sX8XqXqXqXqXqXqX",
"qXqXqXqXqX6XbX=.                                                                                                                                                                                                                              S mX5XqXqXqXqXqXqX",
"qXqXqXqXqX6XcX6.  * + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + %   ! vX6XqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + %X4X<X1X1X1X1X1X1X1X1X1X1X1X1X1X1X1X1X1X1X1X1X<X2XW.,X1X1X1X1X1X1X1X1X1X1X1X1X1X1X1X1X1X1X1X1X1X1X<X5X2   $.UXnXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXVXJX8X%XLXBXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXCXMXUXA.    wXPXzX8XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 4XiXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXyX[.eXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXpX3   :.UXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXfX4XUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXU.    kXUXCX6XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + <XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUXpX<XUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXUXIXKXHXHXHXKXIXUXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXLXUXUXUXUXUXUXUXLXUXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXPXUXSXJ.1.X.3.P.GXUXPXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXUXUXUXH.G 6.A.U.Z.3.H Y.UXUXUXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXUXIXUXt.W 0XLXUXUXUXKX6XU s.UXPXUXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   -.UXGXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpX1XUXPXUXUXPXUXD.E JXUXUXIXLXIXUXUXDXU U.UXLXUXJXUXL.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX3   &.UXCXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXHXGXUXrX-XUXPXUXIXUXCXZ iXUXIXUXUXUXUXUXPXUX0XD FXUXIXJXUXL.    iXUXbX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + <XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX3   5.UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXVXdXUXJXUXJXUXC.q.UXUXUXUXUXUXUXUXUXUXPX5.J.UXJXJXUXK.    DXUXUX4XeXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 2XyXwXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXwXrX`.qXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXwXuX1   r / Y R R R R R R R R R R R R R R R R R R R R T W K s ;XUXUXLXUX#.E.UXIXUXUXUXUXUXUXUXIXUXK.:.UXLXPXUXI.  + L H e.hX7XqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + W.[.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.~.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.`.{.O                                                         5.>X{.].4Xn OXUXPXUXUXUXUXUXUXUXPXUX].V 5X[.[.>X1.        B mX5XqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + ,XeX9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9XqX`.8X0X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9XeXc 9 s d d d d d d d d d d d d d d d d d d d d d d g u #   W.UXFXDXUXX.~.UXIXUXUXUXUXUXUXUXIXUXI.&.UXDXGXUX_.; j b y ;.lX7XqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXiX$XxXhXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXgXmX*X    Q.UXLXLXUXn.i.UXUXUXUXUXUXUXUXUXUXIXq.Z.UXLXUXLXUXeXNXUXUXLX5XeXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX2X_.wX7X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X6XyX[.    Q.UXKXIXUXvXS kXUXIXUXUXUXUXUXIXUXpXF NXUXIXUXPXUX=XpXUXPXMX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXLXUXz.[ UXUXUXIXIXIXUXUXUX( B.UXLXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXIXUX;.[ cXUXUXUXUXUXkX( 4.UXIXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXIXUXUXl.Z a.'.%X`.u.C m.UXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXIXUXUXxXc.} U | b.bXUXUXIXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXPXUXUXUXUXUXUXUXPXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXIXKXHXHXHXKXIXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + <XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXJXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 2XyXwXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXwXyX{.eXyXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXuX9X|.iXqXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXqXsX X    ^.UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX2XkXUXUXZX6XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + -X8X4X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X<XfXI   1 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 1   X.gX<X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X4XeX'.  O , 6 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 4 ; > 9 % } xX6XqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + !.}.].].].].].].].].].].].].].].].].].].].].(.=Xz                                                         I ;X(.].].].].].].].].].].].].].].].].].].].'.oXC.                                                              V mX5XqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 4XpXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtX0XmXD   3 2.*.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.:.-   { NX9XtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXaX>XT ! } ] [ [ [ [ [ [ [ [ [ [ [ [ [ [ [ [ [ [ [ [ ' o.N     o.$.b.dX8XqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + <XrX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X5XxXD   b UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX7   ] vX5X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XiXtX3XBXcXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXbXlXIXN.    FXUXUX4XeXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   l UXAXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXGXPX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrX@X}.qX4X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X5X2XfXa.    uXUXbX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   z UXKXUXUXUXUXUXUXUXPXKXJXJXKXPXUXUXUXUXUXUXUXUXUX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   z UXKXUXUXUXUXUXIXKXUXUXUXUXUXUXJXIXUXUXUXUXUXUXUX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   z UXKXUXUXUXUXPXUXUXuXU.l.x.W.kXUXUXIXUXUXUXUXUXUX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   z UXKXUXUXUXLXUXCX5.' <.9.8.>.) a.LXUXPXUXUXUXUXUX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   z UXKXUXUXLXUXgXR 9.yXUXUXUXIX8X&.' BXUXPXUXUXUXUX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   z UXKXUXPXUXAX/ v.UXUXUXKXKXUXUXUX7.%.IXUXIXUXUXUX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   z UXKXUXPXUXM.+.UXIXLXIXUXUXIXLXUXUXW _.UXIXUXUXUX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   z UXKXIXUXnXX.&XUXKXUXUXUXUXUXUXKXUX^.=.HXUXUXUXUX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   z UXKXIXUX8XQ IXIXIXUXUXUXUXUXUXIXUXBXP bXUXUXUXUX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   z UXKXPXUX:XL UXJXUXUXUXUXUXUXUXUXLXUXm kXUXUXUXUX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   z UXKXIXUX2XT UXKXUXUXUXUXUXUXUXIXPXUXD zXUXUXUXUX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   z UXKXIXUXaX' dXUXPXUXUXUXUXUXUXPXUX4X] CXUXUXUXUX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   z UXKXUXUXUX7.x.UXGXUXUXUXUXUXUXFXUX3.c.UXUXUXUXUX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   z UXKXUXLXUX5XD yXUXKXIXUXUXIXJXUX&XH kXUXIXUXUXUX5   [ bX5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqX0XqXrXtXtXrXwX0XqXqXqXqXqXqX6XcXD   z UXKXUXUXLXUXT.V 6XUXUXUXUXUXUXoXm @XUXLXUXUXUXUX5   [ bX5XqXqXqXqXqX0X0XeXrXtXtXeXqX0XqXqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqX9XpXtX2X;X-X,X0XsXwX0XqXqXqXqX6XcXD   z UXKXUXUXUXIXUX[.C 1.#XhXfX X#.G :XUXPXUXUXUXUXUX5   [ bX5XqXqXqXqX9XeXaX8X,X=X;X3XuXiX9XqXqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqX0X0XaX.X4.R M c J +.P.yXrX9XqXqXqX6XcXD   z UXKXUXUXUXIXUXUXVXS.' M B  .T.GXUXIXUXUXUXUXUXUX5   [ bX5XqXqXqX9XtXwXZ.| G c B ! w.&XpX9XqXqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqX0XpXM.N w.9XUXUXlXF.G &.9XeX0XqXqX6XcXD   z UXKXUXUXUXUXUXIXUXUXUXUXUXUXUXUXIXUXUXUXUXUXUXUX5   [ bX5XqXqX0XtX2X[ Y W.nXUXLX<X,.V I.pX0XqXqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqX0XpXB.m 9XUXUXPXPXUXUXKX|  .uX9X0XqX6XcXD   z UXKXUXUXUXUXUXUXUXPXLXLXLXLXIXUXUXUXUXUXUXUXUXUX5   [ bX5XqX0X0XqXR 3.UXUXUXLXIXUXUX+Xx W.uX0XqXqX0XtX=XoXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XzXg.    fXUXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqX0XuX$XM 6XUXFXPXUXUXIXJXUXUX~ f.lX8XqX6XcXD   z UXJXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXPXUX5   [ bX5XqX7XzX<.$.UXUXKXIXUXUXPXFXUX XV 6XeXqXqX0XtX*XXXtX9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X6XkXf.    dXUXMX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + <XrX0X0XtX0.6.UXGXUXUXUXUXUXUXJXUX X~ 1XrX9X5XxXD   c UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX6   ] vX5X9XtX$XT 8XUXKXUXUXUXUXUXUXJXUXX.z.uX9X0X9XyX2X&XhXiXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXtXBXz.    vXUXDX6XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 4XpXtXyXrXY 3XUXPXUXUXUXUXUXUXIXUXSX} T.lXeX0XMXF   f vXeXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXsX2   { NX9XwXvXv.;.PXUXIXUXUXUXUXUXUXIXUX|.^ hXeXtXyXrXR.L.oX}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.].;X4.    %XgXrX0XqXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + XX:X=X-X$Xz GXUXUXUXUXUXUXUXUXUXJXUXo.i.0X&X#XwXN   .                                                 $   ) eX#X$XtX*.e.UXKXUXUXUXUXUXUXUXUXUXgXl 4X%X*X2XR.                                                      @ @     I bX5XqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + `.@XXXoX}.j LXUXUXUXUXUXUXUXUXUXJXUXO.0.1X X}.2Xm                                                         F 6X{.|.6Xo.t.UXKXUXUXUXUXUXUXUXUXUXlXg =X XXX&XL.  #                                                           A mX5XqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 6XaXyXuXtXK rXUXIXUXUXUXUXUXUXIXUXLX .P.cXrXqXNXF   9 Z.z.c.c.c.c.c.c.c.c.c.c.c.c.c.c.c.c.c.c.c.c.c.v.l.2.).gXrXeXmXj.1.UXUXIXUXUXUXUXUXUXIXUX&XR kXrXtXhXoX    2.}.R.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.E._.d.V.).T. XtX0XqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + <XeX0X0XrX>.h.UXHXUXUXUXUXUXUXKXUX;X~ %XyX9X5XxXD   h SXgXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXzXaX&XgX7X0X8XiX[.~ pXUXLXUXUXUXUXUXUXKXUX3.r.uX9X9XiX}.    }.UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXeXCXUXUXPX5XeXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqX0XiX[.N vXUXGXPXUXUXIXJXUXUX$.6.lX8XqX6XcXD   s dX4X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X9X1X_.eX0XqXqX7XlX| t.UXIXKXIXUXUXPXHXUXqXn :XrXqX0XpX}.    R.UXFXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXHXUX-XpXUXPXmX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXiXr.I BXUXUXKXJXUXUXUXy.Y 9XwX0XqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqX0XeX>XC Z.UXUXIXJXKXUXUXdXC M.iXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqX0XwXtX4.H I.vXUXUXJX@X..E ;XyX9XqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqX9XiX.XG :.>XUXUXUXhXC.Z d.iXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqX0XqXaXD.[ ) ~ Q ( ) e.9XyX9XqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqX9XiX1X3.) ( Q ^ ) | E.sX0X0XqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqX0X9XhX9X_.J.F.Q.;XjXeX9XqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqX8XtXhX%XR.F.L.].rXfX8XqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqX8XeXsXhXhXdXyX9X0XqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqX0X9XuXfXhXgXaXwX8XqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqX0X0X9X9X0X0XqXqXqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqX0X0X9X9X0XqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   d jX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    Q.UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXUX>XdXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 1XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XcXD   s sX4X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X9X1X).wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX}.    R.UXDXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXGXUX=XiXUXUXBX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + <XeX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X5XxXD   h DXhXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXlXxXsX*XhX7X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XiX}.    |.UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXrXAXUXIXNX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + 6XaXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXqXNXF   8 b.s.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.g.f.h.s.-.~.gXrXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXtXhXoX    ;.^.H.L.L.L.L.L.L.L.L.L.L.L.L.L.L.L.L.L.L.L.L.J.R.y.j.UXUXFX6XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + _.OX.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X{.<Xm   .                                                     F 4X[..X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X X$XK.  @                                                       /.4X2XwXqXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX4.  + OX,X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X%XtXc                                                     %   _ tX$X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X=X4XW.                                                    . $   :XxXuX0XqXqXqXqXqXqXqX",
"qXqXqXqXqX6XcX5.  % 4XiXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXuX;XF.oX+XOX+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+XOX$X4   | MX9XtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrX^.R.%XOX+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X+X.X2X0.    zXUXZX6XwXqXqXqXqXqXqX",
"qXqXqXqXqX6XbX;.    ,XrX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX9XiX{.yXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXfX+   Q nX4X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XyX,X@XgXuXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXrXNXd.    fXUXVX7XwXqXqXqXqXqXqX",
"qXqXqXqXqX8XgXV.S ! 4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX+X9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XeX[ A d.hX7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX,X*XrX9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X8XsXP.D U pXGXlX8XwXqXqXqXqXqXqX",
"qXqXqXqXqXqX9XsXmXvXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXxXmXgX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XpXnXbX0X6X8XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXwX8X5X6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6X5X8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X5X5XqXwXwXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_1[] = {
/* columns rows colors chars-per-pixel */
"96 96 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXqXqXqX",
"qXqXqXqXtXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXtXqXqXqXqX",
"qXqXqX0X=X=X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X=X=X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X-X-X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X-X=X-X;X0XqXqXqX",
"qX0XtX=XO .                                                                                 X X                                                                                 . X     =XyX0XqX",
"qX0XyX=X.   2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 < < 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2     5 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 7 3 2 7 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 7 X   3 9 ;XtX0XqX",
"qX0XyX-X  2 rXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXyX=X;XyXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXrX8X& # FXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXgXyXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXe   cXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXPXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXPXUXUXUXUXUXPXKXJXKXPXUXUXUXUXUXIXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXPXUXUXUXPXUXUXUXUXUXUXUXPXUXUXUXIXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXPXUXUXPXUXUX.Xr.-.0./.LXUXPXUXUXIXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # SXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXfXtXUXPXUXPXUXvX&.0.$XeX;Xf.| rXUXLXUXIXUXw   xXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX6X% # AXUXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXPXUXdXrXUXPXIXUXSX .[.UXUXUXUXUX2X_ jXUXPXIXUXw   zXUXeX0XqXqX",
"qX0XyX-X  2 wX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XrX&X=XrX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XwX6X% # HXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXjXiXUXLXKXUXv.V.UXJXPXIXIXJXUX].7.UXKXPXUXw   bXUXeX0XqXqX",
"qX0XyX-X  2 yXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXpX-X:XpXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXuX0X* # vXDXCXZXZXZXZXZXZXZXZXZXZXZXZXZXZXCXJXwX4XUXUXUXJX:.bXUXPXUXUXUXIXUXKX<.kXUXUXUXr   pXHXwX0XqXqX",
"qX0XyX-X  < =X*X*X*X*X*X*X*X*X*X*X*X*X*X*X*X*X&X-X X.X-X&X*X*X*X*X*X*X*X*X*X*X*X*X*X*X&X=X@X$   + O O O O O O O O O O O O O O O O O + @ * 3XaXfX$X#.UXLXUXUXUXUXUXKXUXq.R.xXtXsX4   o @ -XtX0XqX",
"qX0XyX-X  < ;X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X=X:X.XoX:X=X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X;X%X6 & - = = = = = = = = = = = = = = = = = =   . 6XhXkX=X$.UXLXUXUXUXUXUXKXUXq.!.nXpXhXd $ > , ;XtX0XqX",
"qX0XyX-X  2 yXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXpX-X:XpXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXiX%X$XwX7X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X,   SXUXUXLX>.xXUXPXUXUXUXIXUXHX,.zXUXUXUXgX8XUXHXrX0XqXqX",
"qX0XyX-X  2 wXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XrX&X=XrX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XrX-X-XyXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXeXwX,   mXUXJXUXB.c.UXKXPXIXIXJXUX~.w.UXKXPXUXkXeXUXUXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX,   MXUXPXUXJX+.E.UXUXUXUXUX$X' bXUXIXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXPXUXVX,.3.{.1XoXr.+.gXUXPXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXPXUXUX=Xf.2.i. XUXUXPXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXIXIXUXUXUXUXUXUXPXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXPXKXJXKXPXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*X-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX&X=XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XrX=X;XyXwXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXwXuX;X-XrX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX0X,   BXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXlXrXUXUXrX0XqXqX",
"qX0XyX-X  2 yXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXiX@X$X0X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X0X#X#XiXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtX<   cXIXGXHXHXHXHXHXHXHXHXHXHXHXHXHXHXGXUXsX7XUXFXrX0XqXqX",
"qX0XyX-X  < -X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X:XoX< & & % % % % % % % % % % % % % % % % % & % 1 +X;X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X-X=X:   = - = = = = = = = = = = = = = = = = - - - = ; -XtX0XqX",
"qX0XyX-X  < =X*X*X*X*X*X*X*X*X*X*X*X*X*X*X*X*X;X.Xo   # # # # # # # # # # # # # # # # # # #   + oX;X*X*X*X*X*X*X*X*X*X*X*X*X*X*X*X=X*X7 O @ O + + + + + + + + + + + + + + + + +     O % -XtX0XqX",
"qX0XyX-X  2 yXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXiX5Xo : VXFXSXSXSXSXSXSXSXSXSXSXSXSXSXSXSXFXBX= + 7XiXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXiX#X.X7X2X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X2X7X7   sXKXwX0XqXqX",
"qX0XyX-X  2 wX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XeX1XX > IXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXLX- O 2XeX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XrX>X=XiXeXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXeXiX7   vXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XX : KXUXIXIXIXIXKXUXUXUXUXUXKXUXIXIXIXUXHX= + 3XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX-X%XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XrX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XX : KXUXUXUXUXPXUXUXfXrXgXUXUXLXUXUXUXUXHX= + 3XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XX : KXUXUXUXIXUXOX1.:.-.:.2.#XUXPXUXUXUXHX= + 3XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XX : KXUXUXPXUXS.@.6XUXUXUX4XO.G.UXLXUXUXHX= + 3XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XX : KXUXLXUX.X#.UXUXIXKXIXUXUX+.+XUXLXUXHX= + 3XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XX : KXUXUXUX>.0XUXJXUXUXIXJXUX6X<.UXUXUXHX= + 3XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XX : KXUXUXyX<.UXIXUXUXUXUXIXIXUX,.aXUXUXHX= + 3XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XX : KXUXUX6X<.UXJXUXUXUXUXUXKXUX>.9XUXUXHX= + 3XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XX : KXUXUXtX<.UXPXUXUXUXUXUXIXUX,.iXUXUXHX= + 3XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1XX : KXUXUXLX:.tXUXKXUXUXUXKXUXqX,.IXUXUXHX= + 3XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqX9X7X7X7X9XqXqXqXqX0XrX1XX : KXUXKXUX'.=.UXUXPXJXPXUXUX&.|.UXKXUXHX= + 3XrXqXqXqXqXqX9X7X7X7X9XqXqXqXqXqX0XtX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqX0XqXsXlXxXlXaX0X0XqXqX0XrX1XX : KXUXUXPXUXv.-.uXUXUXUXtX*.M.UXPXUXUXHX= + 3XrXqXqXqX0X0XaXlXxXlXsXqX0XqXqXqX0XtX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqXqX9XyXtXH.<.+.3.Y.iXtX9XqX0XrX1XX : KXUXUXUXIXUX`.;.3.4.3.;.].UXIXUXUXUXHX= + 3XrX0XqX9XtXuXI.3.+.1.J.yXyX9XqXqX0XtX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wXqXqX9XiX$X' 9. X,X[.<. .,XyX0XqXrX1XX : KXUXUXUXUXPXUXKXqX<XwXLXUXPXUXUXUXUXHX= + 3XrXqX9XuX:X| 2.{.,X|.7.] %XiX9XqX0XtX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wXqX0XtX;XQ OXUXUXUXUXUX!.( 6XeX0XrX1XX : KXUXIXUXUXUXLXUXUXUXUXUXLXUXUXUXIXUXHX= + 3XrX0XrX4X/ /.UXUXUXUXUX X! ,XtX0X0XtX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  2 wX0X8XfX-.E.UXKXPXPXPXKXUXm.0.gX7XeX1XX : LXUXUXUXUXUXUXUXIXIXIXUXUXUXUXUXUXUXJX- + 2XeX8XgX7.C.UXKXPXPXPXKXUXI.>.fX8X0XrX;X&XyXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXqXyX7   xXUXeX0XqXqX",
"qX0XyX-X  2 yXrXsX%X%.FXUXIXUXUXUXPXUXbX@.3XiXuX5XX : KXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXHX- + 6XuXiX1X@.MXUXPXUXUXUXIXUXSX$.*XaXeXiX;X%XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX7   zXUXeX0XqXqX",
"qX0XyX-X  < :X-X7XS.6.UXKXUXUXUXUXUXLXUX&.W.3X>XOX@   3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3   # @X>X3XR.=.UXKXUXUXUXUXUXKXUX4.F.6X-X;Xt - < , , , , , , , , , , , , , , , , , , .   < 5 ;XtX0XqX",
"qX0XyX-X  < *X%X3XB.7.UXKXUXUXUXUXUXLXUX&.Y.,X=X X@                                         . * XX*X,XI.=.UXKXUXUXUXUXUXKXUX5.Z.2X%X%X-                                         o O     =XyX0XqX",
"qX0XyX-X  2 yXrXdX+X=.KXUXIXUXUXUXIXUXNX#.<XpXiX5Xo * -X,X:X>X>X>X>X>X>X>X>X>X>X>X>X>X>X:X1X|.|.iXrXaX>X#.CXUXIXUXUXUXIXUXHX&.$XsXtXtX<   sXZXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXAXqX:XDXmXeX0XqXqX",
"qX0XyX-X  2 qX0X8XsX$.[.UXHXIXIXIXHXUXJ.3.fX8XeX<Xo - wXyXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXpX>X:XrX9X8XdX<.I.UXHXIXIXIXHXUX).&.sX8XqX0X,   ZXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXcXyXUXUXtX0XqXqX",
"qX0XyX-X  2 wXqX9XuX@XW 5XUXUXUXUXUX+X~ ,XtX9XrX1Xo = 7XwX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XrX=X*XtX0X0XtX:XQ %XUXUXUXUXUX1XQ %XyX9XwXqX,   MXUXPXIXIXIXIXIXIXIXIXIXIXIXIXIXIXPXUXhXqXUXLXrX0XqXqX",
"qX0XyX-X  2 wXqXqX9XaX].^ h.2XsX:Xt._ #XiX9XqXrX1Xo = 8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqX9XiXOX) u.>XsX1Xf./ }.pX9XqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqX9XiX6Xv.$.' &.B.qXuX9XqX0XrX1Xo = 8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqX9XuX0XN.&.' $.b.8XiX9XqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqX9XeXdXhXjXhXsXwX9XqXqX0XrX1Xo = 8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqX9XwXsXhXjXhXdXeX9XqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqX9X8X7X8X9XqXqXqXqX0XrX1Xo = 8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqX9X8X7X8X9XqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1Xo = 8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1Xo = 8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1Xo = 8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1Xo = 8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1Xo = 8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1Xo = 8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1Xo = 8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1Xo = 8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1Xo = 8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1Xo = 8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X*XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX,   MXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXjXwXUXPXrX0XqXqX",
"qX0XyX-X  2 wXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XrX1Xo = 7XwX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XrX=X*XrX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX0X,   MXUXPXIXIXIXIXIXIXIXIXIXIXIXIXIXIXPXUXjXwXUXLXrX0XqXqX",
"qX0XyX-X  2 tXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXuX4Xo - wXyXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXpX>X:XiXeXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXtXrX<   ZXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXxXyXUXUXtX0XqXqX",
"qX0XyX-X  1 ,X>X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X<X@X@   7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 8 3 w %X<X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X>X:X,   q e w w w w w w w w w w w w w w w w e 9 7 8XcXqXqXqXqX",
"qX0XyX-X  < %X$X$X$X$X$X$X$X$X$X$X$X$X$X$X$X$X*X|.@                                           X  X*X$X$X$X$X$X$X$X$X$X$X$X$X$X$X$X%X$X>                                             $XiX0XqXqXqX",
"qX0XyX-X  2 yXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXuX_.`.-X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X*X+X% @ 7XiXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXiX{.(.-X%X&X&X&X&X&X&X&X&X&X&X&X&X&X&X&X%X-X5   NXUXrX0XqXqX",
"qX0XtX;X  4 qX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XtX>X<XaXyXuXuXuXuXuXuXuXuXuXuXuXuXuXuXyXiXwX= # 2XeX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XrX1X:XsXyXuXuXuXuXuXuXuXuXuXuXuXuXuXuXyXyXsX0   lXUXeX0XqXqX",
"qXqXqX0X=X;XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X-X-X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X;X=XwXtXqXqXqXqX",
"qXqXqXqXyXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXyXqX0XqXqXqXqX",
"qXqXqXqX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_2[] = {
/* columns rows colors chars-per-pixel */
"88 88 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX9X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8XqXqXqXqX",
"qXqX0XiXzXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXzXhX0XqXqXqX",
"qX9XiX].0.s.d.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.s.u.p.d.d.d.d.d.d.d.d.d.d.d.d.d.d.d.d.d.s.d.d.d.d.d.d.d.d.d.d.d.d.d.d.d.d.d.d.i.i.i.B.tX0XqXqX",
"qX7XzX0.                                                                                                                                                              $ dX9XqXqX",
"qX7XkXs.  } M.g.j.j.j.j.j.j.j.j.j.j.j.j.j.j.k.q.g.k.j.j.j.j.j.j.j.j.j.j.j.j.j.h.x.e.  v (.K.I.I.I.I.I.I.I.I.I.I.I.I.I.P.R.c.n.R.P.I.I.I.I.I.I.I.I.I.I.I.I.I.L.~.i o A.(.eXqXqXqX",
"qX7XlXd.  M.FXfXkXkXkXkXkXkXkXkXkXkXkXkXkXkXxX1XfXlXkXkXkXkXkXkXkXkXkXkXkXkXkXjXnX5X  ^ UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXMXAXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXB   UXUX7XwXqXqX",
"qX7XlXs.  g.fX4X7X7X7X7X7X7X7X7X7X7X7X7X7X7X0X.X4X8X7X7X7X7X7X7X7X7X7X7X7X7X7X6XrX+X  T UXDXKXKXKXKXKXKXKXKXKXKXKXKXKXHXUXtXsXUXHXKXKXKXKXKXKXKXKXKXKXKXKXKXGXUXb   VXFX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX%X  R UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpXhXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXLXUXn   DXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX%X  R UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpXhXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXLXUXn   DXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX%X  R UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpXhXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXLXUXn   DXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX%X  R UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpXhXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXLXUXn   DXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX%X  R UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpXhXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXLXUXn   DXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX%X  R UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpXhXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXLXUXn   DXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX%X  R UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpXhXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXLXUXn   DXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX%X  R UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpXhXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXLXUXn   DXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX%X  R UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpXhXUXIXUXUXUXUXIXKXJXKXIXUXUXUXUXLXUXn   DXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX%X  R UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpXhXUXIXUXUXUXPXUXUXUXUXUXPXIXUXUXLXUXn   DXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX%X  R UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpXhXUXIXUXIXUXUXpXJ.z.F.eXUXUXIXUXLXUXn   DXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX%X  R UXKXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXpXhXUXIXIXUXUXM.,.~.#X(.3.k.UXUXIXLXUXn   DXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX%X  T UXGXLXLXLXLXLXLXLXLXLXLXLXLXLXJXUXyXdXUXIXPXUXv.j.UXUXUXUXUXM.s.UXIXLXUXn   CXGX8XqXqXqX",
"qX7XlXs.  j.kX7X0X0X0X0X0X0X0X0X0X0X0X0X0X0XeXOX7XqX0X0X0X0X0X0X0X0X0X0X0X0X0X9XuX$X  ~ UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXvXmXUXLXUX9X1.UXUXLXPXLXPXUX8.:XUXHXUXb   UXUX8XwXqXqX",
"qX7XlXs.  k.xX0XeXeXeXeXeXeXeXeXeXeXeXeXeXeXyX+X0XrXeXeXeXeXeXeXeXeXeXeXeXeXeXwXaX&X  A 8X%X=X=X=X=X=X=X=X=X=X=X=X=X=X*X<XT.{.UXUXUXN..XUXLXUXUXUXPXUX-Xk.UXIXUXV   +X1XqXqXqXqX",
"qX7XlXs.  q.1X.XOXOXOXOXOXOXOXOXOXOXOXOXOXOX+X'.oX+XOXOXOXOXOXOXOXOXOXOXOXOXOXoX&XY.                                        2 9X,XpX$.rXUXIXUXUXUXIXUXjXO.yX>XyXw     & dX9XqXqX",
"qX7XlXs.  g.fX4X7X7X7X7X7X7X7X7X7X7X7X7X7X7X0XoX4X8X7X7X7X7X7X7X7X7X7X7X7X7X7X7X0X;X' X.%.$.$.$.$.$.$.$.$.$.$.$.$.$.$.%.%.X a UXZXUXs.;XUXPXUXUXUXIXUX8Xr.UXCXUXf. .9.j.tX0XqXqX",
"qX7XlXs.  k.lX8XwXwXwXwXwXwXwXwXwXwXwXwXwXqXrX+X8XeXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXiX6XaXvXzXzXzXzXzXzXzXzXzXzXzXzXzXzXxXzX* t UXIXUX+Xf.UXHXIXUXIXHXUXb.`.UXIXUXGXnXUXUX8XwXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXoX>X9X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X6X% t UXPXUXUX1. XUXUXUXUXUX=X-.IXUXIXUXkX0XUXVX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXPXUXNX:.n.pXZXdXC.&.xXUXPXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXPXUXIX!.q.:.9.T.JXUXPXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXPXUXUXUXUXUXUXUXPXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXIXPXPXPXPXPXIXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXoX6XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XeX@X1XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXLXPXPXPXPXPXPXPXPXPXPXPXPXPXPXUXzXeXUXAX9XqXqXqX",
"qX7XlXs.  h.jX6X0X0X0X0X0X0X0X0X0X0X0X0X0X9XtX&XeXiXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXsX-X6XwX0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX0X% y UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXNXsXUXLX9XqXqXqX",
"qX7XlXs.  l.cXqXtXtXtXtXtXtXtXtXtXtXtXtXrXyX0XT.OX&X$X%X%X%X%X%X%X%X%X%X%X%X%X%X$X=XW.+XpXrXtXtXtXtXtXtXtXtXtXtXtXtXrXyXrX- r nXjXkXkXkXkXkXkXkXkXkXkXkXkXkXjXvX4X$XmXfX0XqXqXqX",
"qX7XlXs.  w.4XOX#X#X#X#X#X#X#X#X#X#X#X#X+X<Xn.                                        N 3XOX#X#X#X#X#X#X#X#X#X#X#X#X#X$X+X+                                       .   8 sX9XqXqX",
"qX7XlXs.  a.iX,X2X2X2X2X2X2X2X2X2X2X2X2X<XrXF.  p / T R R R R R R R R R R R R R Q P   A uX,X2X2X2X2X2X2X2X2X2X2X2X2X2X3X2XI N G D D D D D D D D D D D D D D S J = . G &.uX0XqXqX",
"qX7XlXs.  k.xX9XeXeXeXeXeXeXeXeXeXeXeXeX0XhXE.  u.UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXDX  F zX0XeXeXeXeXeXeXeXeXeXeXeXeXeXwXuX7XwXxXjXkXkXkXkXkXkXkXkXkXkXkXkXkXgXCXl   UXUX7XwXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  7.UXDXKXKXKXJXHXGXGXGXHXJXKXKXHXUXfX  D jX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX%X*XqX7X7X7X7X7X7X7X7X7X7X7X7X7X7X5XpXg   VXGX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  8.UXJXUXUXPXUXUXUXUXUXUXIXIXUXPXUXkX  D kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X:XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXh   SXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  8.UXJXUXLXUXbXA.5.,.a.=XUXUXIXIXUXkX  D kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X:XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXh   SXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  8.UXJXLXUX<XO.!.kXMX2X7.h.UXUXPXUXkX  D kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X:XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXh   SXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  8.UXGXUXfXO.cXUXUXUXUXUXl.V.UXKXUXkX  D kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X:XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXh   SXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  8.UXGXUXk..XUXGXIXUXLXUXUX,.kXUXUXkX  D kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X:XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXh   SXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  8.UXKXKX2.SXUXIXUXUXUXLXUXD._.UXUXkX  D kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X:XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXh   SXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  8.UXLXFX;.UXPXUXUXUXUXLXUXY.U.UXUXkX  D kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X:XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXh   SXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  8.UXKXIX5.nXUXIXUXUXUXLXUXm..XUXUXkX  D kX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX=X:XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXh   SXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqX0X0X0X0XqXqXqXqX8XfXT.  8.UXGXUXS.K.UXJXPXPXKXUXVX=.SXUXUXkX  D kX7XqXqXqXqX0X0X9X0XqXqXqXqXqX0XtX=X:XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXh   SXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqX8XyXdXfXuX9X0XqXqX8XfXT.  8.UXHXUXDX%.#XUXUXUXUXJX:.}.UXKXUXkX  D kX7XqXqXqX9XwXsXfXpX0X9XqXqXqX0XtX=X:XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXh   SXKX8XqXqXqX",
"qX7XlXs.  j.kX7XqX0XwXgX*XE.T.@XfXrX9XqX8XfXT.  8.UXJXPXUXbX1.d.XX&XI.$._.UXPXPXUXkX  D kX7XqXqX9XfX6X^.U.].yXaX9XqXqX0XtX=X:XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXh   SXKX8XqXqXqX",
"qX7XlXs.  j.kX7X0XrX0X0.;.a.d.1.1.1XuX9X8XfXT.  8.UXJXUXPXUXUX%Xn.h.T.nXUXPXUXPXUXkX  D kX7XqX0XsXV.#.y.f.q.#.(.sX9XqX0XtX=X:XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXh   SXKX8XqXqXqX",
"qX7XlXs.  j.kX7XwXqX+.G.KXUXUXUX(.{ 1XrX8XfXT.  8.UXJXUXUXPXIXUXUXUXUXUXLXUXUXPXUXkX  D kX7X9XsXf.3.bXUXUXUXqX| R.aX9X0XtX=X;XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXh   SXKX8XqXqXqX",
"qX7XlXs.  h.jX4XgXu.C.UXUXKXKXUXUX).,.dX5XdXT.  9.UXKXUXUXUXUXIXLXLXPXUXUXUXUXUXUXlX  D jX6XiX(.&.UXUXLXKXLXUXcX{ 3XwX9XrX-X>XtXqXwXwXwXwXwXwXwXwXwXwXwXwXwX9XgXh   DXLX8XqXqXqX",
"qX7XlXs.  l.cXrX4X-.ZXUXPXUXUXIXIXUX1.#XiXjXW.  8.UXHXIXIXIXIXIXIXIXIXIXIXIXIXLXUXjX  G cX9XhX7.&XUXKXUXUXUXJXUXC.N.kXwXiX&X-XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XfXj   AXJX8XqXqXqX",
"qX7XlXs.  t.8X>XE.5.UXKXUXUXUXUXKXUXj.N.3X4XB.  % > ; ; ; ; ; ; ; ; ; ; ; ; ; ; : :   n 9X#X<X..NXUXUXUXUXUXPXUX-X;.0X%X&Xt = & % % % % % % % % % % % % % % % &     X b aX9XqXqX",
"qX7XlXs.  u.qX<X!.5.UXLXUXUXUXUXKXUXh.Z.6X6XC.  : 0 7 8 8 8 8 8 8 8 8 8 8 8 8 8 8 w < ~ 9X&X3Xo.mXUXUXUXUXUXPXUX*X,.eX=X-X=   u t t t t t t t t t t t t t t t y y y 7 J pX9XqXqX",
"qX7XlXs.  l.xXeX7X=.NXUXLXUXUXIXUXIX>.&XuXjXE.  *.NXrXiXiXiXiXiXiXiXiXiXiXiXiXiXuXdX-X7XyXwXhX0.oXUXJXUXUXUXJXUXn.Z.jXeXrX& u UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXNXaXUXLX9XqXqXqX",
"qX7XlXs.  j.kX4XgXg.l.UXUXLXKXUXUXI.6.fX5XdXY.  #.lX6X0X0X0X0X0X0X0X0X0X0X0X0X0X9XrX@X1XeX9XiX{.O.UXUXPXKXIXUXpX} 7XqXqX0X% t UXLXIXIXIXIXIXIXIXIXIXIXIXIXIXPXUXzXeXUXAX9XqXqXqX",
"qX7XlXs.  j.kX7XqXtX-.z.VXUXUXDXH.X.7XeX8XfXT.  #.zX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqX9XdXb.&.pXUXUXUX=X[ `.pX9XwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7X0XeXtXd.=.8.9.;.w.8XtX0X8XfXT.  #.zX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqX9XdXI.&.6.9.2.;.XXaX9XqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqX0XqXgX2X[.].>XgXeX0XqX8XfXT.  #.zX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqX9XdXeX.X`.@XaXiX9XqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqX8XrXpXaXtX9XqXqXqX8XfXT.  #.zX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqX9XqXiXaXuX0X0XqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqX0X0X0XqXqXqXqX8XfXT.  #.zX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqX0X0X0XqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  #.zX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  #.zX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  #.zX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  #.zX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  #.zX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  #.zX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  #.zX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXT.  #.zX7XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX% t UXIXUXUXUXUXUXUXUXUXUXUXUXUXUXIXUXxXrXUXSX9XqXqXqX",
"qX7XlXs.  j.kX7XqXqXqXqXqXqXqXqXqXqXqXqX8XfXY.  @.kX4X8X8X8X8X8X8X8X8X8X8X8X8X8X8XwX+X<XeX0XqXqXqXqXqXqXqXqXqXqXqXqX0XwX0X% t UXKXLXLXLXLXLXLXLXLXLXLXLXLXLXKXUXkXwXUXSX9XqXqXqX",
"qX7XlXs.  l.xX0XrXrXrXrXrXrXrXrXrXrXrXrXqXjXE.  -.AXpXfXfXfXfXfXfXfXfXfXfXfXfXfXdXkX<XqXtXrXrXrXrXrXrXrXrXrXrXrXrXrXeXtXeX% i UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXSXhXUXJX9XqXqXqX",
"qX7XlXs.  p.tX;X,X,X,X,X,X,X,X,X,X,X,X,X:XqXA.  3 z g h h h h h h h h h h h h h h z e { qX:X,X,X,X,X,X,X,X,X,X,X,X,X,X<X:X; . N n n n n n n n n n n n n n n n M z m fXzX0XqXqXqX",
"qX7XlXd.  t.6X@X&X&X&X&X&X&X&X&X&X&X&X&X#X2XB.  @                                     b 6X@X&X&X&X&X&X&X&X&X&X&X&X&X%X&X#X7 .                                       1XuX0XqXqXqX",
"qX7XzXi.  h.vXqXtXtXtXtXtXtXtXtXtXtXtXtXtXtXeX).:X3X1X2X2X2X2X2X2X2X2X2X2X2X2X<X8X|.  C cXqXtXtXtXtXtXtXtXtXtXtXtXtXtXrXiX.XoX5X1X2X2X2X2X2X2X2X2X2X2X2X2X2X,XrXu   IXUX8XqXqXqX",
"qX8XhXN.= C.fX7X0X0X0X0X0X0X0X0X0X0X0X0X0X9XeX=XqXtXrXrXrXrXrXrXrXrXrXrXrXrXrXeXaX:Xq | fX7X0X0X0X0X0X0X0X0X0X0X0X0X0X0XeX<X3XuXrXrXrXrXrXrXrXrXrXrXrXrXrXrXwXgXW s MXZX9XqXqXqX",
"qXqX0XtXdXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXsXiX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XpXsX0X9XqXqXqXqX",
"qXqXqX0X9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X9XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_3[] = {
/* columns rows colors chars-per-pixel */
"64 64 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX0X8X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X8X9XqXqXqXqX",
"qXqX0XrXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXdXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXfXdXfXsXqXqXqXqX",
"qX0XrX,XL.L.K.K.K.K.K.K.K.K.K.K.K.L.K.K.K.K.K.K.K.K.K.K.K.K.P.I.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.J.I.J.Q.eXqXqXqX",
"qX8XfXL.  .                                               .     X X X X X X X X X X X X   X X X X X X X X X X X o     j uX0XqXqX",
"qX9XfXL.. OX2X:X>X>X>X>X>X>X>X>X<XXX*X,X>X>X>X>X>X>X>X>X-XyX+.a LXmXNXNXNXNXNXNXNXNXNXBX4XMXBXNXNXNXNXNXNXNXNXnXUXB  .UX6XqXqXqX",
"qX9XfXK.  2XpXrXtXtXtXtXtXtXtXtXiX,X9XyXtXtXtXtXtXtXtXtXqXmX=.s UXUXUXUXUXUXUXUXUXUXUXUXdXUXUXUXUXUXUXUXUXUXUXUXUXZ #.UX5XqXqXqX",
"qX9XfXK.  :XrX9X0X0X0X0X0X0X0X9XwX=X5XqX0X0X0X0X0X0X0X0X6XzX&.a UXLXIXIXIXIXIXIXIXIXPXUXyXLXIXIXIXIXIXIXIXIXIXJXUXC +.UX5XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XrX-X5XwXqXqXqXqXqXqXqXqX7XxX&.a UXPXUXUXUXUXUXUXUXUXUXUXuXPXUXUXUXUXUXUXUXUXUXKXUXC +.UX5XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XrX-X5XwXqXqXqXqXqXqXqXqX7XxX&.a UXPXUXUXUXUXUXUXUXUXUXUXuXPXUXUXUXUXUXUXUXUXUXKXUXC +.UX5XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XrX-X5XwXqXqXqXqXqXqXqXqX7XxX&.a UXPXUXUXUXUXUXUXUXUXUXUXuXPXUXUXUXUXUXUXUXUXUXKXUXC +.UX5XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XrX-X5XwXqXqXqXqXqXqXqXqX7XxX&.a UXPXUXUXUXUXUXUXUXUXUXUXuXPXUXUXUXUXUXUXUXUXUXKXUXC +.UX5XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XrX-X5XwXqXqXqXqXqXqXqXqX7XxX&.a UXPXUXUXUXUXUXUXUXUXUXUXuXPXUXUXUXPXUXUXPXIXUXKXUXC +.UX5XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XrX-X5XwXqXqXqXqXqXqXqXqX7XxX&.a UXPXUXUXUXUXUXUXUXUXUXUXuXPXUXIXUXUXUXHXUXUXPXLXUXC +.UX5XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XrX-X5XwXqXqXqXqXqXqXqXqX7XxX&.a UXPXUXUXUXUXUXUXUXUXIXUXuXPXUXUXUX].V.Z.S.fXUXHXUXC +.UX5XqXqXqX",
"qX9XfXK.  >XtX9X0X0X0X0X0X0X0X0XeX-X5XwX0X0X0X0X0X0X0X0X7XxX&.a UXLXIXIXIXIXIXIXIXIXIXUXuXLXPXUXJ.`.PXUXhXj.fXUXUXC +.UX5XqXqXqX",
"qX9XfXK.  <XiXwXrXrXrXrXrXrXrXeXyX:X8XtXeXrXrXrXrXrXrXrX0XbX-.h UXUXUXUXUXUXUXUXUXUXUXUXdXUXUXeXL.UXUXLXUXjXS.UXUXD =.UX5XqXqXqX",
"qX9XfXL.  XX,X=X-X-X-X-X-X-X-X-X:XoX&X;X-X-X-X-X-X-X-X-X&XwX+.  e 8 8 8 8 8 8 8 8 8 8 q e oXSXF.5XUXPXUXLXUXm.0XCXN   J yX0XqXqX",
"qX9XfXK.  *X9X5X5X5X5X5X5X5X5X5X8X&X1X7X5X5X5X5X5X5X5X5X4XeXE.<.0.7.7.7.7.7.7.7.7.6.q.=.  %XUX`.&XUXJXPXIXUXN.BXIXE.i.K.eXqXqXqX",
"qX9XfXK.  ,XyXqXwXwXwXwXwXwXwXwXtX;X7XrXwXwXwXwXwXwXwXwXwXeXqXeXvXlXzXzXzXzXzXzXzXlXMX5X  6XUXBXl.LXUXUXUX{._.UXUXCXUXUX8XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XeX-X5XwXqXqXqXqXqXqXqXqXqXeX2X*X0X7X7X7X7X7X7X7X7X6XeXOX  4XUXUX7Xl.*X6XU.F.UXIXUXhXvXGX9XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XrX-X5XwXqXqXqXqXqXqXqXqXqXeX4X;XrX0XqXqXqXqXqXqXqX0XuX#X  4XUXKXUXNX.X_.7XUXUXPXUXlXMXKX9XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XrX-X5XwXqXqXqXqXqXqXqXqXqXeX4X;XrX0XqXqXqXqXqXqXqX0XuX#X  4XUXPXPXUXUXUXUXIXIXIXUXlXMXKX9XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XrX-X5XwXqXqXqXqXqXqXqXqXqXeX4X;XrX0XqXqXqXqXqXqXqX0XuX#X  4XUXPXUXIXLXLXPXUXUXIXUXlXMXKX9XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XrX-X5XwXqXqXqXqXqXqXqXqXqXeX4X;XrX0XqXqXqXqXqXqXqX0XuX#X  4XUXPXUXUXUXUXUXUXUXIXUXlXMXKX9XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XrX-X5XwXqXqXqXqXqXqXqXqXqXeX4X;XrX0XqXqXqXqXqXqXqX0XuX#X  4XUXPXUXUXUXUXUXUXUXIXUXlXMXKX9XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XrX-X5XwXqXqXqXqXqXqXqXqXqXeX4X;XrX0XqXqXqXqXqXqXqX0XuX#X  4XUXPXUXUXUXUXUXUXUXIXUXlXMXKX9XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XrX-X5XwXqXqXqXqXqXqXqXqXqXeX4X;XrX0XqXqXqXqXqXqXqX0XuX#X  4XUXPXUXUXUXUXUXUXUXIXUXlXMXKX9XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XeX&X2X9X7X7X7X7X7X7X7X7X7X9X<X=XeX0XqXqXqXqXqXqXqX9XuX#X  <XUXGXJXJXJXJXJXJXJXHXUXgXvXFX9XqXqXqX",
"qX9XfXK.  <XuXwXeXeXeXeXeXeXeXwXtX8XgXvXxXxXxXxXxXxXxXxXxXvXdX9XtXwXeXeXeXeXeXeXeXqXpX%X  dXUXUXUXUXUXUXUXUXUXUXUXGXUXUX8XqXqXqX",
"qX9XfXK.  &X8X3X4X4X4X4X4X4X4X4X7Xw.$.*.&.&.&.&.&.&.&.&.&.*.$.u.9X3X4X4X4X4X4X4X4X3X0X|.  $.f.t.u.u.u.u.u.u.u.y.i.w.8.b.rX0XqXqX",
"qX9XfXL.  oX<X;X;X;X;X;X;X;X;X;X1X> $ j p a a a a a a a p k o 8 6X-X;X;X;X;X;X;X;X-X1X}.k q 7 8 8 8 8 8 8 8 8 7 q     R yX0XqXqX",
"qX9XfXK.  1XiXeXrXrXrXrXrXrXeXeXdX< D.UXUXUXUXUXUXUXUXUXUXUXj.0 kXwXrXrXrXrXrXrXrXrXeXtX1XdXgXgXgXgXgXgXgXgXgXsXSXb >.UX4XqXqXqX",
"qX9XfXK.  >XrX9X0X0X0X0X0X0X0X0XpX> b.UXHXKXUXUXPXUXUXKXHXUXy.8 fX9X0X0X0X0X0X0X0X0X0XqX%X9X9X9X9X9X9X9X9X9X9X6XkXk #.UX5XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XpX> n.UXHXUX7XS.Z.D.eXUXGXUXu.8 gX9XqXqXqXqXqXqXqXqXqXwX*X0XqXqXqXqXqXqXqXqXqX8XzXk $.UX5XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XpX> n.UXUX*XM.nXPXxXv.1XUXUXu.8 gX9XqXqXqXqXqXqXqXqXqXwX*X0XqXqXqXqXqXqXqXqXqX8XzXk $.UX5XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XpX> b.UXBXb.UXUXIXUXLXb.DXUXy.8 gX9XqXqXqXqXqXqXqXqXqXwX*X0XqXqXqXqXqXqXqXqXqX8XzXk $.UX5XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XpX> v.UX4X/.UXKXUXKXUXE.eXUXt.8 gX9XqXqXqXqXqXqXqXqXqXwX*X0XqXqXqXqXqXqXqXqXqX8XzXk $.UX5XqXqXqX",
"qX9XfXK.  >XtX0XqXqX0X0XqXqXqX0XpX> b.UX0XE.UXJXIXJXUXP.pXUXy.8 gX9XqXqXqX0X0XqXqXqXqXwX*X0XqXqXqXqXqXqXqXqXqX8XzXk $.UX5XqXqXqX",
"qX9XfXK.  >XtX0X0XqXiXiXwX9XqX0XpX> n.UXLXv.cXUXUXUXdXM.UXUXy.8 gX9XqX9XwXpXuX0X0XqXqXwX*X0XqXqXqXqXqXqXqXqXqX8XzXk $.UX5XqXqXqX",
"qX9XfXK.  >XtX8XuXeXXX|.9XpX9XqXpX> n.UXUXhXj.&XiX+Xk.vXUXUXu.8 gX0X9XaX7X}.oXtXyX9XqXwX*X0XqXqXqXqXqXqXqXqXqX8XzXk $.UX5XqXqXqX",
"qX9XfXK.  >XrXuX@Xp.H.P.p.).pX9XpX> b.UXFXUXZX{.Y. XFXUXFXUXy.8 gX8XpX!.s.I.F.p.&XuX0XwX&X9X0X0X0X0X0X0X0X0X0X7XlXk #.UX5XqXqXqX",
"qX9XfXK.  :XiX>Xt.cXUXUXGXd..XuXiX: B.UXUXUXUXUXUXUXUXUXUXUXs.7 fXuX].k.PXUXUXgXr.4XwXwX>XyXuXuXuXuXuXuXuXuXuXrXMXz *.UX5XqXqXqX",
"qX9XfXK.  :XkXV.>XUXJXJXUXfXk.aXiX< i.FXdXjXhXfXfXfXhXjXdXGX3.q fXyXh.vXUXJXJXUX@XF.fX8X(.@X#X#X#X#X#X#X#X#X#XOX5Xg { DX6XqXqXqX",
"qX9XfXL.  {.6Xi.cXUXUXUXIXKXa.*X;Xt                           p 2X@Xf.UXPXUXIXUXdXa.9X~.@                             a iX0XqXqX",
"qX9XfXK.  -XhXm.4XUXJXJXUXzXj.uXyX4 X.XXQ.^.^.^.^.^.^.^.^./.T.!.eXtXg.BXUXKXHXUX=XC.lXOX  R.eX2X4X4X4X4X4X4X4X3X8XoX#X9XqXqXqXqX",
"qX9XfXK.  >XpX*Xi.SXUXUXUXz.].iXiX< t.VXuXsXsXsXsXsXsXsXsXfXrX3XwXpX^.m.UXUXUXNXy.,XaX#X  eXUXUXUXUXUXUXUXUXUXUXUXBXHXUX9XqXqXqX",
"qX9XfXK.  >XeXiX{.i.Q./.d.R.pX9XpX1 8.jX5X9X9X9X9X9X9X9X9XqX2X-XrX9XpXL.g.).E.u.XXyXyX#X  2XUXJXPXPXPXPXPXPXPXLXUXjXnXHX9XqXqXqX",
"qX9XfXK.  >XtX8XiX8X).^.1XaX9XqXpX1 9.zX7XqXqXqXqXqXqXqXqXeX4X;XrXqX0XaX,X~._.0XuX8XuX#X  4XUXPXUXUXUXUXUXUXUXIXUXlXMXKX9XqXqXqX",
"qX9XfXK.  >XtX0X0XwXaXsXeX9XqX0XpX1 9.zX7XqXqXqXqXqXqXqXqXeX4X;XrX0XqX9XrXsXaXwX0X0XuX#X  4XUXPXUXUXUXUXUXUXUXIXUXlXMXKX9XqXqXqX",
"qX9XfXK.  >XtX0XqXqX0X0XqXqXqX0XpX1 9.zX7XqXqXqXqXqXqXqXqXeX4X;XrX0XqXqXqX0X0XqXqX0XuX#X  4XUXPXUXUXUXUXUXUXUXIXUXlXMXKX9XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XpX1 9.zX7XqXqXqXqXqXqXqXqXeX4X;XrX0XqXqXqXqXqXqXqX0XuX#X  4XUXPXUXUXUXUXUXUXUXIXUXlXMXKX9XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XpX1 9.zX7XqXqXqXqXqXqXqXqXeX4X;XrX0XqXqXqXqXqXqXqX0XuX#X  4XUXPXUXUXUXUXUXUXUXIXUXlXMXKX9XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XpX1 9.zX7XqXqXqXqXqXqXqXqXeX4X;XrX0XqXqXqXqXqXqXqX0XuX#X  4XUXPXUXUXUXUXUXUXUXIXUXlXMXKX9XqXqXqX",
"qX9XfXK.  >XtX0XqXqXqXqXqXqXqX0XpX1 9.zX7XqXqXqXqXqXqXqXqXeX4X;XrX0XqXqXqXqXqXqXqX0XuX#X  4XUXPXUXUXUXUXUXUXUXIXUXlXMXKX9XqXqXqX",
"qX9XfXK.  >XtX0X0X0X0X0X0X0X0X0XiX1 7.hX4X8X8X8X8X8X8X8X8X0X1X=XeX0X0X0X0X0X0X0X0X9XuX@X  1XUXHXKXKXKXKXKXKXKXJXUXgXnXKX9XqXqXqX",
"qX9XfXK.  1XiXeXrXrXrXrXrXrXeXeXdX3 s.HXhXlXlXlXlXlXlXlXlXxXaX6XyXeXrXrXrXrXrXrXrXwXaX&X  sXUXUXUXUXUXUXUXUXUXUXUXFXHXUX9XqXqXqX",
"qX9XfXL.. +X2X>X,X,X,X,X,X,X>X>X3X> : N x c c c c c c c c c c ..4X:X>X,X,X,X,X,X,X:X5X`.  l J S D D D D D D D D H N v.NX8XqXqXqX",
"qX8XgXH.  #X6X1X2X2X2X2X2X2X2X1X5X-.! ^ ^ ~ ~ ~ ~ ~ ~ ~ ! } 4 1 wX<X2X2X2X2X2X2X2X1X5X$X' ~ W Q Q Q Q Q Q Q Q E [   ` UX6XqXqXqX",
"qX9XsX!.f 3XuXwXeXeXeXeXeXeXeXeXtXqXhXvXcXcXcXcXcXcXcXcXkXHXk.J fXwXeXeXeXeXeXeXeXeXeXtXwXxXvXcXcXcXcXcXcXcXcXzXHX' r.UX6XqXqXqX",
"qXqXqXeXiXqXqXqXqXqXqXqXqXqXqXqXqX9X7X7X7X7X7X7X7X7X7X7X7X6XwXuX0XqXqXqXqXqXqXqXqXqXqX0X8X7X7X7X7X7X7X7X7X7X7X7X6XtXrX9XqXqXqXqX",
"qXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_4[] = {
/* columns rows colors chars-per-pixel */
"48 48 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXqXqXqX",
"qXqXqXwXiXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXiXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXsXiXaXwXqXqXqX",
"qXqXwX8X|.).).).).).).).).).).).).).).).).).).|.(./././././././.(.(./././././././.^.}.`.5XwXqXqX",
"qX0XiX|.- z z z z z z z z k k z z z z z z z l * b V B B B B B B M b V B B B B B N A > g =XtX0XqX",
"qX0XaX).z fXyXuXuXuXuXuXpX8X0XpXuXuXuXuXyXpXqXk NXUXUXUXUXUXUXUXLXVXUXUXUXUXUXUXUXUXI @XbX7XqXqX",
"qX0XaX).z yX9X0X0X0X0X0XwX2X3XwX0X0X0X0X0XwX4Xj zXUXPXIXIXIXIXUXAXvXUXPXIXIXIXIXKXUXP .XvX7XqXqX",
"qX0XaX).z uX0XqXqXqXqX0XeX2X4XwXqXqXqXqX0XeX5Xj xXUXIXUXUXUXIXUXAXbXUXIXUXUXUXUXLXUXP .XvX7XqXqX",
"qX0XaX).z uX0XqXqXqXqXqXeX2X4XeXqXqXqXqXqXeX5Xj xXUXIXUXUXUXUXUXSXbXUXIXUXUXUXUXLXUXP .XvX7XqXqX",
"qX0XaX).z uX0XqXqXqXqXqXeX2X4XeXqXqXqXqXqXeX5Xj xXUXIXUXUXUXUXUXSXbXUXIXPXIXIXIXLXUXP .XvX7XqXqX",
"qX0XaX).z uX0XqXqXqXqXqXeX2X4XeXqXqXqXqXqXeX5Xj xXUXIXUXUXUXUXUXSXbXUXPXUXUXUXUXKXUXP .XvX7XqXqX",
"qX0XaX).z uX0X0XqXqXqX0XwX2X4XwX0XqXqXqX0XeX5Xj jXUXJXKXKXKXJXPXBXxXUXUX0X'.'.uXUXUXP }.cX7XqXqX",
"qX0XaX).z pXwXeXeXeXeXwXtX4X6XtXeXeXeXeXwXtX6Xx IXUXUXUXUXUXUXUXUXAXUXqX].KXFX/.aXUXU 2XMX6XqXqX",
"qX0XaX).k 8X2X2X2X2X2X2X4X-X:X4X2X2X2X2X2X4X>X4 +.6.2.2.2.2.2.4.=.P.LXW.LXUXUXCX!.UXM L 7XeX0XqX",
"qX0XaX).k 0X3X4X4X4X4X4X6X:X>X5X4X4X4X4X4X4X4Xp.:.,.,.,.,.,.:.0.n Q UXW.GXUXUXMX).UXT.p.2XrX0XqX",
"qX0XaX).z pXwXwXeXeXeXwXtX4X5XrXwXeXeXeXeXwXeXqXxXcXcXcXcXcXlXKXM.| UXtX/.VXmXQ.zXUXZXUXsX9XqXqX",
"qX0XaX).z uX0XqXqXqXqX0XeX2X4XwX0XqXqXqXqXqXqX-X7X8X8X8X8X8X5XgXi.{ UXUXsX}.|.lXUXUXlXAXiX0XqXqX",
"qX0XaX).z uX0XqXqXqXqXqXeX2X4XeXqXqXqXqXqXqXqX>X0XqXqXqXqXqX8XlXs.{ UXHXUXUXUXUXPXUXxXGXiX9XqXqX",
"qX0XaX).z uX0XqXqXqXqXqXeX2X4XeXqXqXqXqXqXqXqX>X0XqXqXqXqXqX8XlXs.{ UXKXIXPXPXIXIXUXxXGXiX9XqXqX",
"qX0XaX).z uX0XqXqXqXqXqXeX2X4XeXqXqXqXqXqXqXqX>X0XqXqXqXqXqX8XlXs.{ UXKXUXUXUXUXIXUXxXGXiX9XqXqX",
"qX0XaX).z uX0XqXqXqXqXqXeX2X4XwX0XqXqXqXqX0XqX>X0XqXqXqXqXqX8XlXs.{ UXJXIXIXIXIXIXUXxXFXiX0XqXqX",
"qX0XaX).z uX0XqXqXqXqXqXeX3X6XtXeXeXeXeXeXeXeX,X0XqXqXqXqXqX8XlXs.| UXPXUXUXUXUXUXUXvXHXpX9XqXqX",
"qX0XaX).z uXqXqXqXqXqXqXeX2X,X6X5X5X5X5X5X5X5X;XqXwXqXqXqXqX9XzXa.` UXSXHXHXHXHXGXLXkXCXuX0XqXqX",
"qX0XaX_.j 4X:X>X>X>X>X-XwX2.6 x j h g h j k f s +X1X:X>X>X>X;X0Xc.: n l z z z z z c t l %XyX0XqX",
"qX0XaX).z yX0X0X0X0X0X7XcX' 0.PXhXnXZXVXzXvXgXs #XyX9X0X0X0X0XwX2XXX%X#X$X$X$X#X+X2Xk W.kX8XqXqX",
"qX0XaX).z uX0XqXqXqXqX7XvX[ c.UXUXDX7XsXUXUXUXc $XuX0XqXqXqXqXwX7X5XiXyXyXyXyXyXrXxXZ %XvX7XqXqX",
"qX0XaX).z uX0XqXqXqXqX7XcX[ h.UXtXW.wX$X`.UXGXz $XyX0XqXqXqXqXwX6X<XwX0X0X0X0X0X8XgXB oXcX7XqXqX",
"qX0XaX).z uX0XqXqXqXqX7XcX[ k.UXQ.KXUXUX=X0XUXk $XyX0XqXqXqXqXwX6X1XeXqXqXqXqXqX8XhXB OXcX7XqXqX",
"qX0XaX).z uX0X0X9X9X0X7XcX[ x.UX'.UXFXUXhX*XUXk #XyX0X0X9X0XqXwX6X1XeXqXqXqXqXqX8XhXB OXcX7XqXqX",
"qX0XaX).z uX9XrXpXpXuX6XvX[ j.UX/.fXUXUX`.hXUXl $XyXqXpXpXiX0XwX6X1XeXqXqXqXqXqX8XhXB OXcX7XqXqX",
"qX0XaX).z yXrX,XY.P.OXeXxX[ g.UXmX~.|.~.,XUXAXz #XpX7XW.P.'.tXwX5X>X0X8X8X8X8X8X6XdXB .XcX7XqXqX",
"qX0XaX).z sX<XD.dXmXW.`.CX' V.UXUXUXIXUXUXUXUXn %XuXS.8XVX XR.aX9XtXxXlXlXlXlXlXhXSXG 1XmX6XqXqX",
"qX0XaX).j pXL.dXUXUXUXS.gX .( '.K.I.W.T.P.T.J.4 =X'.;XUXUXUXR.>X#Xs.s.s.s.s.s.s.p.x.4 4.rXqXqXqX",
"qX0XaX_.h eXD.MXUXLXUXH.eX+.9 ~ P I I I I U L _ 2XT.qXUXGXUX(.:Xh.> #.[ { { { { { } { { =XtX0XqX",
"qX0XaX).l hXOXW.UXUX-XI.CX{ e.LXjXxXxXxXxXxXxXwXyX2XH.SXUXyXS.nXs.#.UXUXUXUXUXUXUXUXJXUXsX9XqXqX",
"qX0XaX).z tXyX[.K.Y.R.9XxX| <.jX4X8X8X8X8X7X8X-X9XyX$XH.R.K.3XxXa.[ UXFXKXKXKXKXJXIXkXAXiX0XqXqX",
"qX0XaX).z uX9XuXqX7XiX8XcX| 2.xX8XqXqXqXqXqXqX>XqX0XyXrX5XuX0XkXs.{ UXKXUXUXUXUXIXUXxXGXiX9XqXqX",
"qX0XaX).z uX0X0XqXwX0X7XcX| 2.xX8XqXqXqXqXqXqX>X0XqX0XqXwX0X8XlXs.{ UXKXUXUXUXUXIXUXxXGXiX9XqXqX",
"qX0XaX).z uX0XqXqXqXqX7XcX| 2.xX8XqXqXqXqXqXqX>X0XqXqXqXqXqX8XlXs.{ UXKXUXUXUXUXIXUXxXGXiX9XqXqX",
"qX0XaX).z uX0XqXqXqXqX7XcX| 2.xX8XqXqXqXqXqXqX>X0XqXqXqXqXqX8XlXs.{ UXKXUXUXUXUXIXUXxXGXiX9XqXqX",
"qX0XaX).z yX0X0XqXqXqX7XxX| <.kX5X8X8X8X8X8X9X;X0XqXqXqXqXqX8XlXs.[ UXGXLXLXLXLXKXUXkXFXiX9XqXqX",
"qX0XaX).x pXwXeXeXeXeX9XbX .0.FXsXhXhXhXhXhXhX0XeXeXeXeXeXeX0XxXs.O.UXUXUXUXUXUXUXUXFXPXiX9XqXqX",
"qX0XaX).j 5X,X,X,X,X,X;XyX..> J B C C C C C N K %X2X,X,X,X,X:XeXf.= ! P I I I I I Y F -XsX9XqXqX",
"qX0XaX_.x eX7X7X8X8X8X7XwX|.A.G.F.F.F.F.F.H.C.7 @XwX7X8X8X8X7XqX&XZ.S.A.A.A.A.A.Z.U.w }.xX7XqXqX",
"qXqXwX6X#XeXwXwXwXwXwXwXwXtXdXgXgXgXgXgXgXgXfX&X9XeXwXwXwXwXwXwXeXsXhXgXgXgXgXgXgXkX:X9XrX0XqXqX",
"qXqXqXwXyXqXqXqXqXqXqXqXqX0X9X9X9X9X9X9X9X9X9XyXwXqXqXqXqXqXqXqX0X9X9X9X9X9X9X9X9X8XtXwX0XqXqXqX",
"qXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_5[] = {
/* columns rows colors chars-per-pixel */
"44 44 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX0X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9XqXqXqXqX",
"qXqXqXeXfXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXfXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXfXfXwXqXqXqX",
"qX0XeX1XG.B.V.V.V.V.V.V.V.V.V.V.V.V.V.B.C.L.M.M.M.M.M.M.m.M.N.m.M.M.M.M.M.m.G.G.8XwXqXqX",
"qX9XfXG.2 ] ! ~ ~ ~ ~ ^ W ! ^ ~ ~ ~ ~ / T , o.+.O.+.+.+.@. .} @.O.+.+.+.o.=.5 _ wXqXqXqX",
"qX9XhXB.] GXgXlXlXlXlXxXiXfXzXlXlXlXkXbXqXL UXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUX .5XlX8XqXqX",
"qX9XhXV.! gX4X8X8X8X7X9X>X3X9X7X8X8X7XwX&XS FXLXKXKXKXKXPXBXvXIXJXKXKXKXGXUX( @XhX8XqXqX",
"qX9XhXV.~ lX8XqXqXqXqXeX1X7XwXqXqXqX0XtX-XD KXUXUXUXUXUXUXAXMXUXIXUXUXUXLXUX) %XjX8XqXqX",
"qX9XhXV.~ lX8XqXqXqXqXeX1X7XwXqXqXqX0XtX-XD KXUXUXUXUXUXUXAXMXUXIXPXPXPXLXUX) %XjX8XqXqX",
"qX9XhXV.~ lX8XqXqXqXqXeX1X7XwXqXqXqX0XtX-XD KXUXUXUXUXUXUXAXMXUXUXUXUXUXKXUX) %XjX8XqXqX",
"qX9XhXV.~ lX7XqXqXqX0XwX1X6XwXqXqXqX0XtX=XS FXLXJXKXKXKXIXBXnXUXIX<XoX8XUXUX) @XjX8XqXqX",
"qX9XhXV.^ xX9XeXeXeXwXrX3X8XrXeXeXeXwXuX;XJ UXUXUXUXUXUXUXUXFXUX+X7XKX=X>XUX_ 6XkX8XqXqX",
"qX9XhXV.W iX>X1X1X1X1X3X=X>X2X1X1X1X1X4X%Xw ..#.@.@.@.@.%.' W.zX.XUXIXUX).GXI R rXqXqXqX",
"qX9XhXV.! fX3X7X7X7X6X8X>X3X8X7X7X7X7X7X6XG.b.m.n.n.n.v.K.E :.UX_.UXUXHX'.IX*X).qXqXqXqX",
"qX9XhXV.^ zX9XwXwXwXwXrX2X8XeXwXwXwXwXwXwX9XkXjXjXjXjXfXDXx.5.UX6X}.wX`.hXUXZXUXtX0XqXqX",
"qX9XhXV.~ lX7XqXqXqXqXeX1X7XwXqXqXqXqXqX0X:X9X9X9X9X9X6XjXy.4.UXUXnXqXZXUXUXvXCXrX0XqXqX",
"qX9XhXV.~ lX8XqXqXqXqXeX1X7XwXqXqXqXqXqXqX,XqXqXqXqXqX8XlXi.4.UXHXUXUXUXPXUXbXSXrX0XqXqX",
"qX9XhXV.~ lX8XqXqXqXqXeX1X7XwXqXqXqXqXqXqX,XqXqXqXqXqX8XlXi.4.UXKXIXPXUXIXUXbXSXrX0XqXqX",
"qX9XhXV.~ lX8XqXqXqXqXeX1X6XqX0X0X0X0X0X0X,XqXqXqXqXqX8XlXi.4.UXJXIXIXIXIXUXbXAXrX0XqXqX",
"qX9XhXV.~ lX8XqXqXqXqXeX2X0XyXtXtXtXtXtXtX1XqXqXqXqXqX8XzXi.7.UXUXUXUXUXUXUXMXHXrX0XqXqX",
"qX9XhXV.~ lX7XqXqXqX0XeX>X&X-X-X-X-X-X-X=X*XqX0XqXqXqX8XlXy.&.UXbXNXNXNXMXVXfXzXeXqXqXqX",
"qX9XhXC.E yX;X,X,X,X;XeXX.s K S A A A G M m ,X<X,X,X,X:X0XN.u G Z Z Z Z Z F e U 8XwXqXqX",
"qX9XhXV.^ lX8XqXqXqX8XzXW ).UXLXUXUXUXIXbXN 6XeXqXqXqXqXeX7X1X7X6X6X6X6X3XpXG #XhX8XqXqX",
"qX9XhXV.~ lX8XqXqXqX8XlX~ [.UXMX#XOXsXUXNXA 6XwXqXqXqXqXwX5X3XrXwXeXeXeX0XjXU -XhX8XqXqX",
"qX9XhXV.~ lX8XqXqXqX8XzX! {.UX/.cXFX[.kXSXC 6XeXqXqXqXqXwX5X2XwX0XqXqXqX8XgXP *XhX8XqXqX",
"qX9XhXV.~ lX8X0X9X0X8XzXQ +XjX9XUXUXCX&XGXV 6XeX0X0X0XqXwX5X2XeXqXqXqXqX8XgXI =XhX8XqXqX",
"qX9XhXV.~ lX7XyXpXyX7XzX! XXVX#XUXUXuX1XGXV 6XwXeXpXiXqXwX6X2XeXqXqXqXqX8XgXI =XhX8XqXqX",
"qX9XhXV.~ kXeX$X/.$XeXlX~ _.UXoXOX>X).JXmXZ 5XuX<X(.|.eXeX4X<X0X8X8X8X8X6XsXP %XhX8XqXqX",
"qX9XhXV.~ mX_.'.fX'.).MXQ &XUXUXHXVXUXUXUXF qX:XY.iX=XT.yX9XiXxXlXlXlXlXjXSX! 8XkX8XqXqX",
"qX9XhXV.~ yXT.UXUXUXT.yX( :.^.H.R.W.L.U.M.p 7XY.BXUXUX@XOX$Xi.p.i.i.i.i.y.k.0 t.iX0XqXqX",
"qX9XhXV.! 9XQ.UXUXUX!.8X_ N #.} } } |  .[ 3.7XP.JXUXUX;X$Xh.p y.2.4.4.4.4.5.<.w.7XwXqXqX",
"qX9XhXV.~ MXE.*XFX=XE.MX~ I.GXlXxXxXxXxXxXwXpXOX(.CXyXJ.xXp.y.UXUXUXUXUXUXUXIXUXtX0XqXqX",
"qX9XhXV.~ lX0X].R.].0XlX^ m.sX5X7X7X7X8X7X:X0XyX@XR.~.3XxXi.2.UXDXKXKXKXJXPXzXVXrX0XqXqX",
"qX9XhXV.~ lX7XiXiXiX7XlX^ N.hX9XqXqXqXqXqX,XwX0XyXiXiX0XlXi.4.UXKXUXUXUXIXUXbXSXrX0XqXqX",
"qX9XhXV.~ lX8X0X0X0X8XlX^ N.hX9XqXqXqXqXqX,XqXqX0X0X0X8XzXi.4.UXKXUXUXUXIXUXbXSXrX0XqXqX",
"qX9XhXV.~ lX8XqXqXqX8XlX^ N.hX9XqXqXqXqXqX,XqXqXqXqXqX8XlXi.4.UXKXUXUXUXIXUXbXSXrX0XqXqX",
"qX9XhXV.~ lX7XqXqXqX7XlX/ M.dX6X8X8X8X8X8X:XqXqXqXqXqX8XlXi.3.UXFXLXLXLXKXIXxXAXrX0XqXqX",
"qX9XhXV./ xX9XeXeXeX9XxX/ J.VXdXgXgXgXgXhXqXeXwXeXeXeX0XxXi.w.UXUXUXUXUXUXUXKXJXrX0XqXqX",
"qX9XhXB.Y uX:X1X1X1X:XyX^ p W L P P P I G _ 1X1X<X1X1X>XtXf.0  ./ ) ) ) ) _ T 4XiX9XqXqX",
"qX9XgXF.| fX5X8X8X8X7XwX.XI.T.Y.Y.Y.Y.W.A.N 5X9X8X8X8X7XqX-XI.I.I.I.I.I.L.~.B %XfX9XqXqX",
"qXqXwX9X7XwXwXwXwXwXwXqXtXdXdXdXdXdXdXdXsX9XqXwXwXwXwXwXqXeXsXfXfXfXfXfXfXgXwXqXwXqXqXqX",
"qXqXqXwXwXqXqXqXqXqXqXqX0X9X9X9X9X9X9X9X9XwXqXqXqXqXqXqXqX0X9X9X9X9X9X9X9X9XwXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_6[] = {
/* columns rows colors chars-per-pixel */
"32 32 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXwXsXfXfXfXfXfXfXfXfXfXfXfXfXsXfXgXfXfXfXfXfXfXfXfXfXgXdXaXwXqX",
"wX8X).T.E.E.E.E.E.E.E.E.E.E.E.).T.U.Y.Y.Y.Y.T.U.Y.Y.Y.U.!._.9XwX",
"sX).b 2.*.-.-.-.%.*.-.-.-.=.;.l 8.u.t.t.y.e.7.y.t.t.r.j.T  .pX0X",
"fXT.2.SXfXjXjXlXpXdXkXjXjXjXzX$.UXUXUXUXUXUXUXUXUXUXUXUXR.!.cX7X",
"fXE.*.fX5X8X8X9X1X4X9X8X8X7X0X .mXUXKXKXLXFXbXPXKXKXFXUXA.J.zX7X",
"fXE.-.jX8XqXqXwX4X7XwXqXqXqXeX..VXUXIXUXUXKXNXUXIXUXLXUXF.L.xX7X",
"fXE.-.jX8XqX0XwX4X7XwXqXqX0XeX..mXUXJXKXLXFXnXUXUXNXJXUXD.J.zX7X",
"fXE.-.lX9XwXwXrX5X8XeXwXwXwXrX+.UXUXUXUXUXUXJXUX,X3X$XUXK.^.cX7X",
"fXE.%.pX1X4X4X5X,X1X5X4X4X3X0XY u.n.x.x.v.s.|.uXrXUXvXrXA.{ dX9X",
"fXE.*.dX4X7X7X8X1X3X8X7X7X6X0XQ.M.V.V.B.H.4.-.NX1XUXaX9XjX'.wXqX",
"fXE.-.kX9XwXwXeX5X8XeXwXwXwXwX0XhXhXhXgXnX+X7.UX9X-X2XUXKXPXrX0X",
"fXE.-.jX8XqXqXwX4X7XwXqXqXqXqX1X8X9X9X8XiX/.4.UXUXIXUXUXZXmXrX0X",
"fXE.-.jX8XqXqXwX4X7XwXqXqXqXqX2X0XqXqX0XaX_.4.UXJXUXPXUXSXBXrX0X",
"fXE.-.jX8XqXqXwX3X6XwX0XqXqXqX1X0XqXqX0XaX_.4.UXKXIXPXUXZXNXrX0X",
"fXE.-.kX8XwXqXwX9XqXrXeXeXeXrX8XqXwXwX0XsX).3.UXUXUXUXUXPXDXrX0X",
"fXE.$.uX,X2X<XeXz.R O.X.O...} ` ,X3X2X1X7X|.G &.@.#.@.*.^ *.wXqX",
"fXE.=.jX7X0X7XlX<.Q.UXbXkXSXDX[ =XrX0X0XqX9X-X-X-X-X&XqX2.C.zX7X",
"fXE.-.kX8XqX8XlX5.$XUX-X3X0XUX%.-XtXqXqXwX9X5XyXtXtXwXvXs.R.zX7X",
"fXE.-.jX7XqX7XlX4.+XwXCXUX2XNX=.=XrX0X0XqX9X2XwX0XqX7XkXu.U.zX7X",
"fXE.-.jXwXeXwXlX4.+XrXxXUX:XCX*.*XiXrXtXrX8X2XqX0X0X7XjXy.I.zX7X",
"fXE.=.zX|.'.{.xX4.&XUX3X4XhXUX&.>X2X`.'.4XwX8XsXaXaXuXNXg.!.xX7X",
"fXR.>.3X-XUX1X<X0.S.cXtXwXaXiX/ $X|.DXmX'.7X].)._.)./.#X+.h.kX8X",
"fXR.:.+XlXUXMXOX0.D ;.+.#.@.o.;.+X*XUXUX+X/.D 9.3.4.4.6.*.0.qXqX",
"fXR.:.pX).yX`.uX9.P.NXsXfXfXgXeXrX[.4X-X&X].8.UXUXUXUXUXUXPXrX0X",
"fXE.=.zX2XoX<XcX5.C.dX6X8X8X9X,X0XeX$X&XfX).3.UXGXLXKXIXVXmXrX0X",
"fXE.-.jX9XuX0XkX6.A.hX9XqXqXwX2X0XqXyXrXpX_.4.UXLXUXUXUXSXBXrX0X",
"fXE.-.jX7X0X7XkX6.V.sX6X8X8X8X<X0XqX0X9XaX_.3.UXGXKXKXIXVXNXrX0X",
"fXE.;.lX9XwX9XzX8.T.ZXhXkXkXlXtXwXwXwXqXsX_.e.UXUXUXUXUXUXDXeXqX",
"fXR.@.aX2X5X2XdX1.( l.u.i.p.y.j.4X5X5X4XeX_.K T.D.F.F.J.M.8XiX9X",
"sX/.5.pX3X6X5XqX{.h.z.l.l.l.l.R :X8X5X6X7X1Xb.g.j.j.g.M.( U.jX8X",
"qX0X0XwXwXwXwXqXyXjXjXjXjXjXkXtXqXwXwXwXwXeXgXkXkXkXjXlXsXwXqXqX",
"qXqXwXqXqXqXqXqX0X8X8X8X8X8X8XqXqXqXqXqXqXqX8X8X8X8X8X8X0XqXqXqX"
};

/* XPM */
static const char *const xpm_icon_7[] = {
/* columns rows colors chars-per-pixel */
"24 24 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXOX`.].].].[.].].].'.XX'.)._._._._.)._.).`.oX7X",
"OX' a.r.t.t.q.t.t.r.s.^ a.C.M.N.v.v.N.M.Z.p.o.eX",
"`.a.NXsXfXfXyXgXfXsXMXs.bXUXUXUXUXUXUXUXUXMXB.cX",
"].r.sX6X8X8X1X8X8X6XpXr.uXUXJXPXZXZXPXIXUXpXc.lX",
"].t.fX8XqXqX4XqXqX9XdXy.sXUXIXUXDXDXUXBXUXgXn.zX",
"].t.fX8XqXqX4XqXqX8XfX0.iXUXPXUXGXLXpX8XuXuXj.zX",
"[.q.yX1X4X4X>X4X4X2XrXh.;.y.q.u.#. XqXUXdX1Xw.5X",
"].t.gX8XqXqX4XqXqXqXwX8X7X7X5XsXb.'.mX6X0XKXAXtX",
"].t.fX8XqXqX4XqXqXqXwX4X0XeX0XhXZ.[.UXNXUXIXMXuX",
"].t.fX8XqXqX2X8X9X8X0X2X8XqX8XgXC.`.UXLXPXHXnXyX",
"].t.hX9XwXwXrXfXdXdXfXuXqXwX0XjXV.+XUXUXUXUXIXuX",
"].w.uX2X4X8Xg.0.p.p.r.0.;X7X2XyXD.%.A.v.m.l.k.8X",
"].r.sX6X0X6X-.fX9X<XNX0.'.uX8X0X3X[.[.].oXA.4.jX",
"].t.gX8XwX5Xa.FXuXuXJXS.].pX0XwX8XeXaXiXjX$Xm.cX",
"].r.hXwXyX5Xi.qXnXIXtXM.].dXeXrX5X3X9X7XtX{.k.zX",
"].u.pX X$X6Xp.UXfXuXUXF.{.2X X1XwXpXhXfXvX=XN.vX",
"].r.;XZXfXoX;.|.|.`.-X;.E.qXUX*X@XV.C.V.F.u.#.aX",
"].r.-XSXhXOXO.c.B.B.n.S.#X6XUX3Xv.w.#X`.].`.).9X",
"].u.iX|.#X8X5.gXhXgXjXyXeX=X}.eXS.@XUXUXUXUXPXiX",
"].r.hXqXtX9X<.6X8X7X9X2X8XtXwXkXV._.UXGXLXGXnXyX",
"].y.gX8XwX8X5.yXiXuXpX0X0XqX8XgXZ..XUXUXUXUXCXyX",
"'.w.dX6X0X8XO. XoXXXXXoX8X0X7XfXm.S.CXdXhXdXlXuX",
"}.i.yX2X4X8Xd.7.e.q.t.{ OX9X3XqX].4.u.r.a.-.x.kX",
"8X4XeXwXwXwXsXgXfXfXhX8X8XeXwXqXtXgXdXdXfXiX5XeX"
};

/* XPM */
static const char *const xpm_icon_8[] = {
/* columns rows colors chars-per-pixel */
"16 16 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"9X|.'.].].].].{.'.`.`.`.`._..X9X",
"|.f.~.Y.I.Y.~.u.'.>X$X@X=X%Xc.qX",
"'.~.cXiXyXpXzXW.ZXUXUXUXUXUX[.rX",
"].Y.iX7X4X7XuXF.gXUXHXFXkXnXE.wX",
"].I.yX4X2X5XeXQ.B.R.M.|.vXaX).5X",
"].T.sX9X5X0X0X9X9XiX{.XXZXMXDXeX",
"].R.sX9XrXpXpXtX0XpX}.:XUXUXKXeX",
"].I.yX7XY.P.T.H.1XrX.Xm.+X[.).0X",
"].Y.sX2XZ.wXtXU.oXuX7X+XOX Xz.tX",
"].E.uX1XE.VXHX[.$XqX0XiXjXsXQ.uX",
"{.H.sX:XZ.6X9XP.[.iX>X.X}.'.j.rX",
"{.A.cX2Xe.R.E.I.*XbX(.C.;XOX X0X",
"].E.0X<XS.dXpXyX8X6X}.:XUXUXLXeX",
"'.U.gX6Xc.0X7X6X9XsX{.XXUXLXCXeX",
"|.E.rX7Xb.m.C.l.:XwX%Xb.R.G.W.iX",
"9X9XqXqXtXqXwX5X9XqXeXwX8X8X3XqX"
};

const char *const *const xpm_icons[] = {
    xpm_icon_0,
    xpm_icon_1,
    xpm_icon_2,
    xpm_icon_3,
    xpm_icon_4,
    xpm_icon_5,
    xpm_icon_6,
    xpm_icon_7,
    xpm_icon_8,
};
const int n_xpm_icons = 9;
