#include "gtk.h"
/* XPM */
static const char *const xpm_icon_0[] = {
/* columns rows colors chars-per-pixel */
"128 128 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9X7XqXqXqXqXqXqX8X7X6X7X0XqXqXqXqX7X9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X7X7X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9X8XqXqXqXqXqXqX0X7X7X0XqXqXqXqXqX7XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX9XjXaXyXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXiXkX0XqXqXqXqXqXfXxXcXkXrX0XqXqX0XjXpXuXuXuXuXuXuXuXuXuXuXuXuXuXuXyXlXxXyXuXuXuXuXuXuXuXuXuXuXuXuXuXuXyXaXhX9XqXqXqXqX0XrXzXxXyX0XqXqXqXqXjXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX7XjXX.b.!.Y.T.T.T.T.T.T.T.T.T.T.T.T.T.T.T.T.T.T.T.T.T.Y.R.T.T.T.T.T.T.T.T.T.T.T.T.T.T.T.R.K.] tX0XqXqXqXeX8.F v ( #XuX9X9XsXX.C.W.Y.T.T.T.T.T.T.T.T.T.T.T.I.(.E A ~.Y.T.T.T.T.T.T.T.T.T.T.T.T.Y.!.b.#.gX8XqXqX9XuX%XK A (.pX9XqXqX8X} wXqXqXqXqXqXqXqX",
"qXqXqXqXqX9XaXb.OX4X<X<X<X<X<X<X<X<X<X<X<X<X<X<X<X<X<X<X<X<X<X<X4XH.-X2X<X<X<X<X<X<X<X<X<X<X<X<X<X2X*Xf 4XqX0X0XqXeX$X{.@.  [ 9XqX9XeXl {.6X<X<X<X<X<X<X<X<X<X<X<X<X<X3X}.f.3X<X<X<X<X<X<X<X<X<X<X<X<X<X<X9XQ.b pX9XqX9XuX+Xp   &.7X9XqXqXeX*Xj eXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XyX!.4XyXeXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXeXiXY.8XtXrXrXrXrXrXrXrXrXrXrXrXrXrXtXqXQ ).tXrXoXT 0.eXDXT.  n >XeXrX XW <XiXeXrXrXrXrXrXrXrXrXrXrXrXrXeXuXW.tXeXrXrXrXrXrXrXrXrXrXrXrXrXeXaX&XW $XrX8XyX|.y O v.AXxXuXqX0XyXR./ rXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXY.<XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXyX8.-.lXqX*X=   ! >.i   ;.tX5XxX8.;.rXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XP.wX0XqXqXqXqXqXqXqXqXqXqXqXqX0XwXwXO.a.lXwX+Xr   a 9.3.-._.rX6XlXX.p.uX0XqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX_.S :XtXwX~.B     - %.<XwXeX5XK P.dX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XgXV.T 8XjXV.            e.yXyX#XZ oXuX0XqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XdX3./ fXtXwX1X@X+X%XqXqXwXjXO.+.dX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9XaX` >.zX6X].}. XOXOX.X:XiXtXU y.sX0XqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwX0XX.  [ 2XkXpXyXrXtXtXkX7X%.  Q 3XrX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XrX1XP   1.eXNXdXuXtXiXjXbX:X^ O.rXqX0XqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqXqXqXqXqXqX0XeX0X .  y R R p.OXuXzXpX$Xh.Q E h   W 3XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX<XL   z T ~ l.&XaXzXyX Xy.J ;.rXwX0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqXqXqXqXqX9XeXqX .  8 T.jXF.} ~ Y U U ! [ N.gX[.i   W 4XtX9XqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqX9XtX1XP   h  XsXc.] ! U U Y ~  .U.fXqX0XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqXqXqXqX9XwX0X .  8 R.pX7XfXyX|.E.6.U.}.eXhX7XiX[.p   W 3XrX9XqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqX9XtX1XP   j  XuX8XhXqX{.L.7.!..XuXfX8X0XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqXqXqX9XwX9X .  8 E.pX9X0X8XqXiXhX^.sXaXwX8X0X9XiX{.p   W 2XrX9XqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqX9XrX<XP   j .XiX9X0X8XwXsXiX/.kXuXqX8XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqXqX0XeX0X .  7 R.pX9XqXqXqXqX9XqXL.7X0X0XqXqXqX9XiX[.p   W 3XrX0XqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqX0XtX<XP   h  XiX9XqXqXqX0XqX4XP.eX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqX0XeXqX .  7 T.pX9XqXqXqXqXqXqXeXP.9XwXqXqXqXqXqX9XiX[.i   W 4XtX0XqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqX0XtX1XP   h |.iX9XqXqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqX9XwX0X .  8 R.pX8XqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqX8XiX[.p   W 4XrX9XqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqX9XtX1XP   h |.uX8XqXqXqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqX9XwX9X .  8 E.pX9XqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqX8XiX{.p   W 3XrX9XqXqXqXqXqXqX0XP.wXqXqXqXqXqXqX8XrX<XP   j  XiX8XqXqXqXqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqX0XeX0X .  8 R.pX9XqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqX9XiX{.p   W 2XrX9XqXqXqXqXqX0XP.wXqXqXqXqXqX9XtX<XP   h  XiX9XqXqXqXqXqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9XtXI.4XwX0XqXqX0XeXqX .  7 T.pX9XqXqXqXqXqXqXqXqX0X0XwXP.8XqX0XqXqXqXqXqXqXqXqX9XiX[.i   W 4XtX0XqXqXqXqX0XP.wXqXqXqXqX0XtX1XP   h  XiX9XqXqXqXqXqXqXqXqX0XwX5XI.rX9X0XqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXiXtXpXU.8XuXiXyX9XwXqX .  8 R.pX8XqXqXqXqXqXqXqX0XtXiXtXyXI.wXyXiXyX0XqXqXqXqXqXqXqX8XiX[.p   W 4XrX9XqXqXqX0XP.wXqXqXqX9XtX1XP   h |.uX8XqXqXqXqXqXqX0X0XyXiXuX9XU.iXtXiXrX0XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XaX2XB.| Z s M / f.$XdX9X .  8 R.pX8XqXqXqXqXqXqXqX0XsX,Xb.[ C s B _ l.=XsX0X0XqXqXqXqXqXqX8XiX{.p   W 3XrX9XqXqX0XP.wXqXqX8XtX<XP   j  XiX8XqXqXqXqXqXqX0XqXsX%Xg.( M s Z | N.2XaX9XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XaXP.P R m. X<X@XL.[ S f.+.  8 R.pX9XqXqXqXqXqXqXqX0XaXA.H ! V.XX<X+XG.) D n.pX0XqXqXqXqXqXqXqX9XiX{.p   W 2XrX9XwX0XP.wXqX9XtX<XP   j .XiX9XqXqXqXqXqXqX0XqXuXh.A [ K.+X<X.XN.R P L.aX9XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XaXv.v F.jXjXvXbXzXaXkX XA   u Y.pX9XqXqXqXqXqXqXqX9XaXs.m Y.kXlXbXvXkXiXlX(.C 9.pX9XqXqXqXqXqXqXqX9XiX[.i   W 3XtX0X0XP.wX9XtX<XP   h  XiX9XqXqXqXqXqXqXqX0XuX>.D }.kXfXcXvXzXaXjXG.b v.aX9XqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XaXP.v .XhXwXI.[ T =.XXtXyXtXF } zX7XqXqXqXqXqXqXqX9XdXB.M &XhX8Xb.^ W e.,XqXaX2XS g.fX8XqXqXqXqXqXqXqX8XiX[.p   W 4XtX8XP.qXtX1XP   h |.iX9XqXqXqXqXqXqXqX8XfXq.L 8XpXrXR.} E 2.;XwXfXXXb L.aX9XqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrX2XP F.hX9Xp.        a '.rXyX@XP /.sX9XqXqXqXqXqX0XyX=XH E.gX7X>.        L ,X9XsX].F XXpX0XqXqXqXqXqXqXqX8XiX[.p   W 4XeXH.yX<XP   j  XiX8XqXqXqXqXqXqXqX9XsX_.F +XuXyXv.#       l $XqXgXF.P 2XtXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqX0XiXB.R kX9X7Xd 4 S.wX%.  ,.hX3XxX,.;.pXqXqXqXqXqXqX0XpXg.` zXqX#X  8 }.2XN   M.jX2XcX+.9.pXqXqXqXqXqXqXqXqXqX9XiX{.p   Q ,XE.<XL   j .XiX9XqXqXqXqXqXqXqXqXqXpX:.<.xXwX$Xr   M.7X)   -.rX6XkXR N.iX0XqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XeX0X0X0X0X0X0X0X0X0X0X0X0X0X9XtX| B.pX0X0XE.}.lXbXt.  >.hX6XiX`.~ &XtX9X0X0X0X0XqX9X` J.iXwX-X=.D.tXuX8.  k.kX5XpXW./ 1XeX0X0X0X0X0X0X0X0X0X8XuX].u   Q V.U   g |.uX8X0X0X0X0X0X0X0X0X9XtX*X~ `.yXpXK.  W 7XkX^.* q 6XwXpXB.| tXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXY.4XiXtXtXtXtXtXtXtXtXtXtXtXtXtXtXpXZ +XrX0XqXyXpXr.D 1 ! &XyX9X0XwXG .XgXrXtXtXtXtXuX9XV :XwX0XrXzXpXyX-X8 : _.pX8XqX6XA ;XsXtXtXtXtXtXtXtXtXtXtXeXdX Xa   p   l oXdXeXtXtXtXtXtXtXtXtXtXrXfXXXF eX9XuXf.  >.rXrX@Xd   <XrXeX+XZ eXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXR.H.Y.I.I.I.I.I.I.I.I.I.I.I.I.I.I.U.s 4XwXqXqX9XyX(     b |.uX9X7XkXA a.!.I.I.I.I.I.Y.D.g rX0XqX0X4XeX,XN   e.aX8XqX8XdXb v.E.I.I.I.I.I.I.I.I.I.I.I.L.R.Z.r   9 z.!.J.I.I.I.I.I.I.I.I.I.I.I.!.d.V kX6XuXi.  2.tXeX$Xh   ,XtXwX4Xl wXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.-X8X5X5X5X5X5X5X5X5X5X5X5X5X5X4X8XM &XeXqX9X6X0XqX0Xj.  J 3XwX8XuXF (.rX4X5X5X5X5X7X>Xm 2XwXqX9XyX=Xz   ;.eX8XqX0X0X0XV oX0X5X5X5X5X5X5X5X5X5X5X5X5X3XwX8X0   W =X8X3X5X5X5X5X5X5X5X5X5X4XeX_.S uX8XpXm.  O.0XyX|.0 $ 3XrXeX&XN eXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.2XtXwXeXeXeXeXeXeXeXeXeXeXeXeXwXuX/ P.yXeX%XH $.1XNX!.  g $XtXeX+XW $XpXwXeXeXeXwXrX0XQ ~.yX8XtX#Xj   a.NXxXaXqX0XyX{.W <XyXwXeXeXeXeXeXeXeXeXeXeXeXeXqXyXh.p   ! 4XyXqXeXwXeXeXeXeXeXeXwXiX%XE OXeXpX_.  e oXNXs.  K 9XqXyXI.( rXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqX0XiXg.[ kX9X4Xr   V | r   @.tX5XxXy.+.rXwXqXqXqXqXqXqXuXw.#.zX9X<Xf   3 o.O.) A.tX5XcX2.,.yXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XW.{.i   W 4XtX0XqXqXqXqXqXqXqXqXwXrX@.t.zX6XqX]   b ] :   G.iX5XzX{ f.iX0XqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXyX#XS  XiXqXoXW .   < :.1XwXrX6XI F.fX9XqXqXqXqXqX0XiX[.D $XgX[.  .         ..tXrX,XJ E.sX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XJ.pX].p   W 4XrX9XqXqXqXqXqXqXqX9XfXG.P 5XeXwX:X+.@   t b.rX7XpX|.A #XyX0XqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XdXf.A tXiXqX8X,X,X1XwX0XeXfXO.X.aX0XqXqXqXqXqXqXqX8XfX7.I aXtX%X&X=X>X>X;X<XyXgX` $.gX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.0XiX{.p   W 3XrX9XqXqXqXqXqXqXqX0XaXX.o.fXrX0XqX<X,X3XrX7XpX0XG g.sX0XqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwX9X+.  D @XzXpXtXwXeXtXzX6XO.E 5XeX0XqXqXqXqXqXqX0XeX5X`   T -XMXdXtXwXrXsXbX1X]   I :XtX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wX9XiX{.p   W 2XrX9XqXqXqXqXqXqX0XeX6XW o.5XzXtXeXwXeXuXxX@XI ;.yXqX0XqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqX0XeXqX .  u } P ,._.wXkXyXOXy.I X.5XtX0XqXqXqXqXqXqX0XrX6X^   g { P 3.[.eXkXtX X0.U { m   H >XyX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqX9XiX[.i   W 3XtX0XqXqXqXqXqXqX0XtX6XX.P t.OXyXkXwX`.1.G 0.uXqX0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqX9XeXqX .  8 Y.zX/.-.~ H Z F E +.F.aXeX0XqXqXqXqXqXqX9XrX7X^   y (.lXE.&.! G Z F W #.L.jXOXl   H >XyX9XqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqX8XiX[.p   W 4XrX9XqXqXqXqXqXqX0XeXaXG.+.W F Z G ~ -.(.fX0X0XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqX9XwX0X .  8 R.pX7XsXpX*X.Xa.(.$XrXgX0X0XqXqXqXqXqXqX9XeX6X^   u _.pX7XdXuX%X}.p.'.%XtXfX7XuX+Xz   H :XtX8XqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqX8XiX[.p   W 3XrX9XqXqXqXqXqXqX0X0XfXrX$X(.a. X&XpXsX8XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqX0XwX9X .  8 R.pX9XqX9XqXtXgX~.rXpXwX9XqXqXqXqXqXqXqX9XeX5X^   u _.pX9XqX9XqXyXsXQ.iXuXqX9XqX9XuX+Xz   H :XtX9XqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqX9XiX{.p   W 2XrX9XqXqXqXqXqXqXqX9XwXpXrX!.fXtXqX9XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqX0XeX0X .  7 R.pX9XqXqXqXqX9XrXI.4XwXqXqXqXqXqXqXqXqX0XrX6X^   y ).pX9XqXqXqXqX0XwXP.8XqXqXqXqXqX0XuXOXl   H :XyX0XqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqX9XiX[.i   W 3XtX0XqXqXqXqXqXqXqXqXwX5XP.rX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqX9XeXqX .  8 T.pX8XqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqX9XrX7X^   y ).iX9XqXqXqXqXqXqXeXP.9XwXqXqXqXqXqX9XuXOXl   H >XyX9XqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqX9XiX[.p   W 4XtX9XqXqXqXqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqX9XwX0X .  8 R.pX8XqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqX9XeX6X^   u ).iX8XqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqX8XuXOXz   H >XtX8XqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqX8XiX[.p   W 3XrX9XqXqXqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqX9XwX9X .  8 E.pX9XqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqX9XeX5X^   u _.pX9XqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqX9XuX+Xz   H :XtX9XqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqX9XiX{.p   W 2XrX9XqXqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqX0XeX0X .  7 R.pX9XqXqXqXqXqXqXqXqX0XtXI.5XwXqXqXqXqX0XrX6X^   y ).pX9XqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqX9XuX+Xl   H :XyX0XqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqX9XiX[.p   W 3XrX0XqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXR.2XtXqX0XqX0XeXqX .  7 T.pX9XqXqXqXqXqXqXqX0X0XqXuXY.7XrXqX0XqX0XrX7X^   y ).iX9XqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqX9XuXOXl   H >XyX0XqX0XqXwXqXU.rXwXqX0XqXqXqXqXqXqXqX9XiX[.i   W 4XtX9XqX0XqXrX8XY.uXqX0X0XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX9XiXK.*X0XyXpX9XwX0X .  8 R.pX8XqXqXqXqXqXqXqX9XyXpX9X9XD.>X0XuXpX8XeX6X^   u ).iX8XqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqX8XuXOXz   H >XtX8XiXiX9X3XS.6X9XiXuX9XqXqXqXqXqXqXqX8XiX[.p   W 4XrX8XiXuX0X,XD.9X9XpXuX9XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX7XkX] f Q 7._.dX0X .  8 E.pX9XqXqXqXqXqXqXqX9XdX=Xg.` C h m Q q.[.fX5X^   u _.pX9XqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqX8XuX+Xz   H ;XhX.Xy.^ N h B ( p.#XdX0X0XqXqXqXqXqXqX8XiX{.p   W 2XgX{.w.Q M h V ` f.=XdX9XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX<X(.-.S 3.X.  8 R.pX9XqXqXqXqXqXqXqX0XaXB.H _ H.;X0X:XQ.#.D 8.`   y _.pX9XqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqX9XuX+Xz   Y 0.G  .Y.;X0X;XI.{ F z.pX0X0XqXqXqXqXqXqX9XiX{.p   ( 7.D @.~.<X0X&XH.` H N.aX0XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX9XgXkXiXzX>X(   y T.pX9XqXqXqXqXqXqXqX0XaXs.M E.xXiXyXkXvXjXlX&XU   g (.pX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqX9XuXoXN   Z .XcXlXxXlXgXpXxX`.A 8.pX0XqXqXqXqXqXqXqX9XiX].l   P $XxXtXiXlXvXhXzXW.N a.aX0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqX8XhX#.` oXeXrXfX[ E jX7X0XqXqXqXqXqXqX0XsXA.m *XgX4XiX.XT k r.pXrXaXR { lX7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XkX>.Z rXsX1Xe.A Z 5.:X0XaX2XA l.dX9XqXqXqXqXqXqXqX7XlX .Y aXuX5XaXL.Z c J.uXsX=XM A.sX9XqXqXqXqXqXqXqX",
"qXqXqXqXqX6XxXJ   a |.eXeX2XR F.fX8XqXqXqXqXqXqXtX,XH Y.fX4XuXS.;     ' xX3XuX=XP E.dX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XpX].J oXaX3X_   % &   L 1X8XsX_.F #XuX0XqXqXqXqXqX9XdXW.P =XyX6XyX7..     j.kX2XgXT.H >XtXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXtX}   >.xX2XzXp.} yXqXqXqXqXqXqX0XiXb.~ lX6XqXwX( j o.  X.zX7X6XxX4.&.uXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXiX7.$.zXrX`.  l >X3Xn   K.gX3XcX{ p.iXqXqXqXqXqXqXqXuX=.4.xX5XeX>Xz P W   b.jX7X6XlX! v.iX0XqXqXqXqXqXqX",
"qXqXqXqXqXqX0XpXv.  &.cX4XyX+X~ |.iX9XqXqXqXqX0XtX[ S.iX9XwX9XY.6Xf.  ..zX6X9XuX{.! %XyX0XqXqXqXqXqXqXqXqXqXqXqXqX0XwXP.9XqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX1X~ ^.uXwX&X8.~.tXrX&.  G.hX6XiXU.( 8XwX0XqXqXqX0XyX&X! {.uX9XeX1XU.eX}   v.jX7X9XiXA.] tXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXwX3Xg O K.sX7X9XpXU E.hXqXeXeXeXeXwXyXV #XrXqXqXqXuXdX0.  ..zX6X0X0XrXG }.sXwXeXeXeXeXeXeXeXeXeXeXeXeXwXtXI.qXeXeXeXeXeXeXeXeXeXeXeXeXeXwXiX-XC 7XqX0XtXzXuXiX{.$ e +XuX9XwX,XV 5XtXeXeXeXeXwXsX XD rX0X0XqXwXaXwX_   v.jX7X0XrX#XC eXqXqXqXqXqXqXqX",
"qXqXqXqXqX7XkX)   :.pX9XqX7XxXY 6.^.L.P.P.P.P.P.I.s 4XwXqXqXqX7XpXe.  ..zX6XqX7XkXA p.Q.P.P.P.P.P.P.P.P.P.P.P.P.P.P.I.Z.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.R.b.x sX8XqX0X4XyXXXr   M.aX8XqX0XtXj A.Y.P.P.P.P.P.Q.s.V kX7XqXqXqX9XwX`   v.jX7XqXwX4Xl wXqXqXqXqXqXqXqX",
"qXqXqXqXqX6XvXl W rX9XqXqX9XaXU U.sX7X9X9X9X9X8XwXB $XrXqXqXqX9XaXe.  ..zX6X0X0XtXG '.iX8X9X9X9X9X9X9X9X9X9X9X9X9X9XqXP.7X9X9X9X9X9X9X9X9X9X9X9X9X9X9XrX%XV 9X0X0X9XiX[.7   a.uX7XqX0XwX1XB 1XqX9X9X9X9X8XuX[.D tX9X0XqXqXwXwX`   v.jX7X0XeX%XV eXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXI.fXbXdXqX0XtX%X~ }.aX0XwXwXwXwXqXyX_ H.uX0XqXqX9XaXe.  ..zX6X9XyX.XQ %XuXqXwXwXwXwXwXwXwXwXwXwXwXwXqXeXP.9XwXqXwXwXwXwXwXwXwXwXwXwXwXqXtX1XQ _.yX8XyX(.5 % D.ZXlXpXqX9XuXE.^ 8XeXqXwXwXwXqXuX&XW .XyX9XqXqXwXwX`   v.jX7X0XuXH.` rXqXqXqXqXqXqXqX",
"qXqXqXqXqX8XfX0.J G w.eX6XlXg.[ eXwX0XqXqXqXqX0XiXl.) lX7XqXqX9XaXw.  | zX6X6XcXq.$.yXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXuX3.;.xXqX@X<   3 ~ R D A.yX5XxXX.t.iXqXqXqXqXqXqXqXtX%.0.cX6XqXqXwXwX)   v.jX7X6XzX) l.iX0XqXqXqXqXqXqX",
"qXqXqXqXqX6XvXx     L qXeX6XQ N.gX8XqXqXqXqXqXqXyX=XD (.dX6XqX9XaXt.  O.zX4XyX<XL P.fX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XaX).F $XkXE.  = .       5.rXuXXXG  XiX0XqXqXqXqXqX9XdXI.K <XyX8XqXwXwX[   n.jX3XdX).F *XyX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqX9X9X8X5XrXjX%.R gX7X0XqXqXqXqXqXqX0XsXb.V 2XdX6X9XqX8X5X6X7XyXpX] &.dX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XdX5.R tXiX2X3X6X9X9X7X5XiXyXS 9.jX8XqXqXqXqXqXqXqX9XsX&.' pXuX6X9X0X7X4X7X6XsX3XC b.sX0XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX0XwXuXcX5XO.  g _.iX9XqXqXqXqXqXqXqX0XiX8.A '.cXpXqX9XwXuXcX,X_ ` 0XwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXrXo.R &XnXpXeX0XeXaXvX+XD   Y ;XtX9XqXqXqXqXqXqX0XwX0X` ( >XcXyX0X9XeXpXcX[.D 8.iX0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX0XsXeX|.8.K @.^   u `.iX9XqXqXqXqXqXqX0XqXpXf.F +.E.6XdXqX[.2.H :.0XeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXrX5.F ;._.9XdX7X^.%.K :.M   H >XyX0XqXqXqXqXqXqX0XeX0X:.G 1.[.qXdX5XW.+.F f.pX0X0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqX7XjX..l E -.I.dX6X!   i `.iX9XqXqXqXqXqXqXqX9XfXXX9./ A v V E >.W.sXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXdX(.2.Q C v Z ~ 5.'.lXoXz   H >XyX9XqXqXqXqXqXqX0XwXsXQ.,.E V v A / 8..XfX9XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX9XpXC.{.<XrXdX9XeX5X~   i `.iX8XqXqXqXqXqXqXqX8XpXpX1X;Xv.oX<XyXsX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9XaXuX<X$Xc.*X<XiXaX7XuX+Xz   H :XtX8XqXqXqXqXqXqX0X0XsXyX<XOXv.;X1XpXpX8XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXW.6XiXqX9XqX9XrX6X~   u _.iX8XqXqXqXqXqXqXqX0XqXeXaXE.0XyXqX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XqXtXtXR.iXrXqX0XqX9XuX+Xz   H :XtX9XqXqXqXqXqXqXqX9XqXyXqXE.aXeXqX0XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXY.<XeXqXqXqXqX0XrX6X~   u _.iX9XqXqXqXqXqXqXqXqX0XtXI.5XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XP.wX0XqXqXqXqX9XuXOXl   H >XyX0XqXqXqXqXqXqXqXqXwX5XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqX9XrX5X~   i `.pX9XqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqX8XuXOXl   H >XyX9XqXqXqXqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqX9XeX5X~   i `.iX8XqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqX8XuX+Xz   H :XtX8XqXqXqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqX9XrX6X~   i _.iX8XqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqX9XuX+Xz   H :XtX9XqXqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqX0XrX6X~   u _.iX9XqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqX9XuXOXl   H :XyX0XqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqX0XrX5X~   i `.pX9XqXqXqX0XtXI.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqX9XuXOXl   H >XyX9XqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqX9XeX5X~   i `.iX9XqXqX0XtXI.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqX8XuXOXz   H >XtX8XqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqX9XrX6X~   i `.iX8XqX0XtXI.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqX9XuX+Xz   H :XtX9XqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqX0XrX6X~   u _.iX9X0XtXI.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX+Xl   H :XyX0XeX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX5X~   u `.iX8XtXI.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XuXOXl   H >XyXqX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XeX5X~   i `.uXeXI.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XuXOXz   H >XuX3XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XrX6X~   i _.fXL.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XuX+Xz   H >X0XP.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XrX6X~   y {.T.3XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XuX+Xl   K %XI.rX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.1XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX0XrX6X~   p Z.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX0XP.eXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XiXOXl   D /.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXE.3XyX8X9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X8XwX3X!   e 9XyX8X9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XqXP.8X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X8XP.qX9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X7XtX.Xv   U gX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX6XxXZ % i.W.K.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.L.L.!.!   9 f.W.K.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.I.Z.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.Z.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.P.H.^.p.O 5 lX7XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX7XlX^   w W.aX0XwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqX0XbXU   a '.pX0XwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXrXI.0XeXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXeXqXP.rXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwX0XpX{.g   ! fX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX9XuXR./   8 E.pX9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XtXK.R   u _.iX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXP.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XP.wX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XiX[.p   U (.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXR.,X..  9 W.pX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXT.,X^   i `.pX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XiX{.a   W ;XI.rX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXU.2XwX}   0 W.pX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXL.7X6X~   i `.iX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XiX{.s   R 3X9XP.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.,XtX0X}   0 E.pX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXU.3XtX6X~   i _.iX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XiX{.a   R 3XyX3XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.1XwXeXqX}   9 E.pX9XqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XqXrX6X~   u _.iX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XiX[.a   R 3XtXqX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrX0XeX0X|   9 W.pX9XqXqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XeX0XrX5X~   i `.pX9XqXqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqXqX9XiX{.a   R 2XtX0XeX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqX9XwX9X}   0 W.pX9XqXqXqXqXqXqXqXqXqXqXqX0XtXI.5XwXqX9XeX5X~   i `.iX9XqXqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqXqX8XiX}.s   R 1XrX9XqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqX9XwX0X}   0 W.pX8XqXqXqXqXqXqXqXqXqXqX0XtXI.5XeXqXqX9XrX6X~   i `.iX8XqXqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqXqX8XiX{.a   R 3XrX9XqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqX0XeXqX}   9 E.pX9XqXqXqXqXqXqXqXqXqX0XtXI.5XeXqXqXqX0XrX6X~   u _.iX9XqXqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqXqX9XiX[.a   R 3XtX9XqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqX0XeX0X|   9 E.pX9XqXqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqX0XrX5X~   u `.iX9XqXqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqXqX9XiX{.a   R 2XtX0XqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqX0XwX9X|   0 W.pX9XqXqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqX9XeX5X~   i `.iX9XqXqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqXqX9XiX}.s   R 1XrX9XqXqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqX9XwX0X}   0 W.pX8XqXqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqX9XeX6X~   i `.iX8XqXqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqXqX8XiX{.s   R 2XrX9XqXqXqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXT.<XrXqXqXqXqXqXqXqX9XeXqX}   9 E.pX8XqXqXqXqXqX0XtXI.5XeXqXqXqXqXqXqXqX9XrX6X~   u _.iX9XqXqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqXqX8XiX[.a   R 3XtX9XqXqXqXqXqXqXqXwX6XI.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XuXY.<XeX0XqXqXqXqXqXqXqX0XeX0X}   9 E.pX9XqXqXqXqX9XrXI.4XwXqXqXqXqXqXqXqXqX0XrX6X~   u `.iX9XqXqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqXqX9XiX{.a   R 2XtX0XqXqXqXqXqXqXqXqXwX5XI.rX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX0XyX!.9XaXwX9X0XqXqXqXqXqXqX0XeX9X|   9 W.pX9XqX9XqXtXfX!.eXiXwX9XqXqXqXqXqXqXqX9XrX5X~   i `.pX9XqXqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqXqX9XiX}.a   R 1XrX9XqXqXqXqXqXqXqX9XwXiXrX!.fXtXqX9XqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX9XaXb.Q.%XwXgXqX9XqXqXqXqXqXqX9XwX9X}   0 W.pX7XsXpX=XoXd._.%XrXfX0X0XqXqXqXqXqXqX9XeX5X~   i `.iX8XqXqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqXqX8XiX{.s   R 2XrX9XqXqXqXqXqXqX0X0XfXrX%X_.d.XX*XiXsX8XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX8XhX$.b W O.C.uXrX9XqXqXqXqXqXqX9XeXqX}   9 R.lX(.:.~ F C S E @.G.pXeX0XqXqXqXqXqXqX9XrX6X~   i _.iX8XqXqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqXqX8XiX[.a   R 3XtX9XqXqXqXqXqXqX0XeXaXH.#.E S C F ~ ;.).fX0X0XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX9XfXiX#Xa.Y ] <XyX0XqXqXqXqXqXqX0XeXqX}   i | P >._.wXjXtXOXt.P X.6XtX0XqXqXqXqXqXqX0XrX6X~   u _.iX9XqXeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wXqX9XiX{.a   R 3XtX0XqXqXqXqXqXqX0XtX6XX.P r.oXtXhXqX`.<.F q.uXqX0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX0XyXpXpXxX0X-.L <XrX0XqXqXqXqXqXqX0XeX9XO.  S @XxXiXtXyXiXaXxX6Xo.W 6XeX0XqXqXqXqXqXqX0XrX5X~   i `.pX9XeXP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.wX9XiX{.a   R 1XrX9XqXqXqXqXqXqX0XeX6XW X.4XcXsXpXiXpXdXcX+XP :.uXqX0XqXqXqXqXqXqXqX",
"qXqXqXqXqX7XzXY t n.0XrXgX,.( uX0X0XqXqXqXqXqXqX9XdXf.Z tXiX5XtXT.g - :.dXqXdXO.X.aX0XqXqXqXqXqXqXqX9XeX5X~   i `.iX0XP.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XP.0XiX{.s   R 2XrX9XqXqXqXqXqXqXqX0XaXX.X.dXyX6Xd.g 3 L /.qXiX9XF h.sX0XqXqXqXqXqXqXqX",
"qXqXqXqXqX7XjX{ X   t.pXqXqXQ v.hX8XqXqXqXqXqXqXyX$XS |.iX5XiXj.  +   { zX4XtX5XP G.fX9XqXqXqXqXqXqXqX9XrX6X~   i _.aXK.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XJ.pX[.a   R 3XtX9XqXqXqXqXqXqXqX9XfXH.P 5XeXrX4.  3 d   - ).eXpX}.A $XyX0XqXqXqXqXqXqX",
"qXqXqXqXqXqX9XhXT.  u @XwXkXx.] qXwX0XqXqXqXqX0XiXg.[ lX6XwXqX_ T >.  ..zX7X6XxXt.+.rXwXqXqXqXqXqXqXqXqX0XrX6X~   u '.Y.7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XW.}.p   R 3XtX0XqXqXqXqXqXqXqXqXwXrX#.t.lXtX X; = ^.xX<.  ` wX7XlX[ f.iX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXG   G.fXqX=XQ {.sXqXwXwXwXwXwXuX/ P.yX0XwX9X{.uXs.  ..zX6X0XtX+XE $XpXwXwXwXwXwXwXwXwXeXqXtX6X^   i F.tXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXtXj.d   W 3XyXqXeXwXwXwXwXwXwXwXwXiX%XE +XeXpXA.  [ 9XaX].2 , 4XeXyXP./ rXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX9XO.  b.hX5XsXU L.iX4X6X6X6X6X5X9XM &XeXqXqXqXtXsX0.  ..zX6X0X9XuXF (.rX5X6X6X6X6X6X6X6X6X6X4X8X,XW   6 8X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X9X1   Y ;X9X3X6X6X6X6X6X6X6X6X6X5XrX_.S uX8XuXa.  <.rXeX#Xg   ,XrXeX*XN eXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX9XO.  n.jX3XxXY 7./.P.I.I.I.I.I.U.s 4XwXqXqXqX7XpXe.  ..zX6XqX7XkXA a.!.P.I.I.I.I.I.I.I.I.I.I.P.P.).W y Y.U.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.T.a R ).I.P.I.I.I.I.I.I.I.I.I.I.I.!.d.C kX6XuXa.  <.rXeX@Xg   <XtXwX4Xl wXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwXqXF   G.gX7XuXY !.kXeXtXtXtXtXrXiXZ +XrX0XqXqX9XaXe.  ..zX6X0X0XwXH  XfXrXtXtXtXtXtXtXtXtXtXtXtXrXrXfXkXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXjXfXtXrXtXtXtXtXtXtXtXtXtXtXtXrXfXXXF wX8XaXA.  ' 9XsX'.1 , 5XeXrX+XZ eXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX9XfXP.  u +XrXtXXX^ .XuX9X0X0X0X0X9XtX| B.pX9XqXqX9XaXe.  ..zX7X9XiX`.^ *XtX9X0X0X0X0X0X0X0X0X0X0X0X0X0X8X8X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X8X8X0X0X0X0X0X0X0X0X0X0X0X0X0X9XtX=X~ `.yXuX X; & E.jX;.  ` wX0XpXB. .tXqXqXqXqXqXqXqX",
"qXqXqXqXqX7XkX_     t.sX4XzXy. .iXqXqXqXqXqXqX0XiXN.R kX6XqXqX9XsX9.  [ xX6X6XcX1.-.pXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXpX:.<.cX4XyX5.  > t   ; ).uX5XkXR M.iX0XqXqXqXqXqXqX",
"qXqXqXqXqX7XlXE a M.yX8XuX:XJ U.fX8XqXqXqXqXqXqXrX2XP G.hX5XqX9XpXh.3 <.jX3XpX@XG (.sX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XsX).F +XpX8XqXj.l 8 Y _.uX4XhXG.P 1XtXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX0XaXfXyX7XuXtX^ ;.fX8XqXqXqXqXqXqXqX0XaXL.v XXhX6X9X0XyXdXyX6XaX9XI 0.fX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XfX0.P 9XpX7XyXdXgXaXeX6XhXoXb L.sX9XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX0XqXtXlX#XU O.rXqX0XqXqXqXqXqXqXqXqX9XaXc.v F.kXpXrXqXqXtXzX|.G ;.uX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XuX:.D }.lXtXqX0XwXiXkXG.b c.aX9XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX8X=XT.X.Z y.rXwX0XqXqXqXqXqXqXqXqXqXqX9XaXL.P R N.@X4X*XP.{ A g.yXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXyXh.A [ P.*X4X+XB.R P L.aX9XqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqX7XjX} h ^ i.oXsXqX0XqXqXqXqXqXqXqXqXqXqXqXqX9XsX2XN.| Z k N / d.$XsXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXsX$Xf.( N k Z } M.1XsX9XqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwXeXrXuXuX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XtXiXtXeXwXeXrXiXyX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XyXiXrXeXwXeXtXiXtX9XqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XqXqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_1[] = {
/* columns rows colors chars-per-pixel */
"96 96 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX0XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXqXqXqXqX9X9X0XqXqXqXqX0XqX0X0X0X0X0X0X0X0X0X0X0X9X0XqX0X0X0X0X0X0X0X0X0X0X0X0X0XqXqXqXqXqX9X9XqXqXqXqX0X0XqXqXqX",
"qXqX0XrXaXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXtXtXtXtXtXtXtXtXtXtXtXtXyXiXqXqXqXqXeXfXgXuX0XqXqXeXiXtXtXtXtXtXtXtXtXtXtXtXtXgXaXrXtXtXtXtXtXtXtXtXtXtXtXuXyX0XqXqXqX9XiXdXwXqXqX0XyXrX0XqXqX",
"qXqXrX>X'.:X*X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X:X=X=X=X=X=X=X=X=X=X=X=X=X&X.XwXqXqXqX<XI.L.#XyX0XeX3X X;X*X=X=X=X=X=X=X=X=X*X;XP.`.:X*X=X=X=X=X=X=X=X=X*X-X+X&XrX0XqX9XpXXXQ.7XwX0XtX&X<XrXqXqX",
"qX0XaX'.y.$X].{.{.{.{.{.{.{.{.{.{.{.{.{.{.{.}.{.K.}.{.{.{.{.{.{.{.{.{.{.|.`.I qXqXqXfX>Xy.C - '.tXsXK.] ;X'.{.{.{.{.{.{.{.{.[.+X>.r.$X].{.{.{.{.{.{.{.{.[.%Xg.%.lX7X9XfXM.  ^ 0XqX7XlX*.H.hX9XqX",
"qXqXrX:X$XdXyXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXyX).iXuXuXuXuXuXuXuXuXuXuXuXiX+.OXpX*X` K.UXQ.  r.sXiXz.f.vXrXuXuXuXuXuXuXuXuXuXaXwX#XdXyXuXuXuXuXuXuXuXuXyXjX|.&.rX9XaXm.O <.ZXlX0X0XrX=._.aX0XqX",
"qX0XtX*X].yX8X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9X^.0X0X0X0X0X0X0X0X0X0X0X8XsXr.s.zX<Xa   m ; < |.rXuX[ ].yX9X0X0X0X0X0X0X0X0X9XrX&X'.yX8X0X0X0X0X0X0X0X0X9X0XqX[ [.xXS.    Z v _ 7XiXXX[ qXqX0XqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X^.wXqXqXqXqXqXqXqXqXqXqX0XtX%X{ 1XgX;X<.G ..[.eXzXS.,.iX0XqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqXqXqXqX9XsXA.-.NXK.S ~ F M ] 8XjX#.M.fX9XqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X^.wXqXqXqXqXqXqXqXqXqXqX9XqXsXS c ,XzXcXbXxXhXjXL.% f.jX8XqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqXqXqXqX8XiX%X, *.jXDXvXvXMXZXpX2.;.pX0XqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X^.wXqXqXqXqXqXqXqXqXqX0XwXyX[   U { y.{.2X:XR.-.} j   B.dX9XqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqXqXqX8XiX&Xp = ` | b.#X4X#XN.[ 5.tXwX0XqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X^.wXqXqXqXqXqXqXqXqX9XwXtX}   &.gX*Xp.@.E ( ;.V.uX<Xj   V.dX8XqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqXqX8XiX*Xa . F.lX^.6.} T } 5.^.dXqX0XqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X^.wXqXqXqXqXqXqXqX0XwXyX}   -.iX0XtXsXdX].>XdXdX0XyX>Xk   V.dX8XqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqX8XiX*Xa . G.dX8XaXaXrX^.rXaXaX9XqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X^.wXqXqXqXqXqXqX9XwXyX}   -.iXqX0X0X8XyX].;XrX9XqX8XuX>Xl   V.dX8XqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqX8XiX*Xa . G.dX8XqX0X0XqX^.0X0X0XqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X^.wXqXqXqXqXqX9XwXyX}   -.iXqX0XqXqX0XuX].;XtXqXqXqX8XuX>Xk   V.dX8XqXqXqXqX0XtX-X[.uX0XqXqXqXqX8XiX*Xa . G.dX8XqXqXqXqXqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X^.wXqXqXqXqX0XwXyX}   -.iXqX0XqXqXqX0XuX].;XtXqXqXqXqX9XuX>Xl   V.dX9XqXqXqX0XtX-X[.uX0XqXqXqX8XiX*Xa . G.dX8XqXqXqXqXqXqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X8X~.0X0XqXqX9XwXyX}   -.iXqX0XqXqXqXqX9XtX'.=XeX0XqXqXqXqX8XuX>Xl   V.dX8XqXqX0XtX-X[.uX0XqXqX8XiX*Xa . G.dX8XqXqXqXqXqX0X9X~.9X0XqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0X8XyXdX XgXtX8X9XwXyX}   -.iXqX0XqXqXqXqX8XqXkX#X8XdX9X9XqXqXqXqX9XuX>Xk   V.dX9XqXqXtX-X[.uX0XqX8XiX*Xa . G.dX8XqXqXqXqX0X8XtXfX XfXtX8X0XqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqXqXqXqX9XtXgX*XU.e.T.-XfXuXtX}   -.iXqX0XqXqXqX0X0XgX6X_.p.N.XXpXpX9XqXqXqXqX8XuX>Xl   V.dX8X0XtX-X[.uX0X8XiX*Xa . G.dX8XqXqXqXqX9XrXgX=XT.e.Y.=XgXrX9XqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqXqXqX9XpX:X,.| =.>.-.} 3.;X .  -.iXqX0XqXqXqX0XeXyXl.] %.:.:.+...(.dX9XqXqXqXqX8XuX>Xk   V.dX8XtX-X[.uX7XiX*Xa . G.dX8XqXqXqXqX9XiX<X1.| -.>.-.| <.,XiX9XqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqXqX9XpX}.R c.8XVXKXbX<Xn.j   ;.iXqX0XqXqXqX0XwXqX+.;.&XcXKXCXwXQ.! n.sX9XqXqXqXqX8XuX>Xl   V.sXwX-X[.rXuX*Xa . G.dX8XqXqXqXqX9XiXXXE l.4XMXKXmX3Xx.E .XiX9XqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqX0XtX:XR /.nX X( N e.eXlX^.v 1XtX9XqXqXqXqX0XaX-.q.cX5X*.m &.,XjX7X~ P.sX9XqXqXqXqX8XuX>Xk   V.jX*X`.gX%Xa . G.dX8XqXqXqXqX0XrX1XE Q.vX<X&.N 3.8XxX^.E <XrX0XqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqX8XgX,.c.nX).@ 6 S   %.iXhXz.6.gX8XqXqXqX9XsXU...fX9Xl O P . A 8XaXOX/ 8XwXqXqXqXqXqX9XuX>Xl   S.4XXX>Xp . G.dX8XqXqXqXqXqX8XhX3.j.zX4Xn o D   T wXkXz.1.gX8XqXqX",
"qX0XtX=X}.uX0XqXqXqXqXqXqXqXqX0XyX*X .1XdX^.o.7XPXf w dXwX-X| ,XtX0XqXqX0XpX,.K.jX&Xl m.GX6.  ;XeXdX<.H.sX9XqXqXqXqXqX8XuX>Xk   l.R.f   H.dX8XqXqXqXqXqX0XtX:X .:XkXV.  k.JX<.  W.fX,X| -XtX0XqX",
"qX0XtX=X{.yX9X0X0X0X0X0X0X0X0X8XdXU.>.dX8XwXhXH.L q m.aX7XaX&.~.aX9X0X0XqX9X[ <XwXqXyXyXyXV t 9XqXsXZ.6.kX7X0X0X0X0X0X0X8XyX:Xh O u & S.sX7X0X0X0X0X0X0X9XsXW.=.pXsX2.  +XdXW.  g.hXiX;.R.dX9XqX",
"qX0XrX:XK.).^.^.^.^.^.^.^.^.^.~. Xe.1.lX7XrXkXr.5 ; y.yX6XlX=.a.|.~.^.^./.E.R eXqX0XqXuX_   P.pX7XdXY.` OXQ.^.^.^.^.^.^.^.!._.~.n   @.OXW.^.^.^.^.^.^.^.~.|.u.;.jXiX:.  #XpX~.  s.gXhX>.J.gX9XqX",
"qX0XtX=X}.iX0XwXwXwXwXwXwXwXwX0XfXT.:.aXeX|.H.sXUX~   OXtXuX&./.dX0XwXwXwX0X{ >XwXeXtX) $ W.NXwX9XsXV.9.lX8XwXwXwXwXwXwXwXwX9XpX4Xe # B.fX9XwXwXwXwXwXwX0XdX~.=.uXgXf.  T.GXx.  S.dXuX;.E.dX9XqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqX0XtX-X} :XjXx.  [ c.2 0 <XuX=X| 1XrX0XqXqX0XpX1.D.fX8XW   _ H.x.>X0XsX;.L.aX9XqXqXqXqXqXqXqXqX0XwX>XT.v   V.dX8X0XqXqXqXqX0XrX,X| ;XaX#X= a m.4 e 3XyX>X| :XrX0XqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqX8XfX3.b.lX5X2., . b Q.rXkXp.0.hX8XqXqXqX9XsXT.O.jX%X# ; @     E.kXoX] 8XwXqXqXqXqXqXqXqXqXqX0XtX*XXX4Xj   V.dX8XqXqXqXqXqX8XhX7.d.jXtXP.s   v ).rXjXg.5.hX8XqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqX9XuX;Xj ~.zXdX9X5XeXdXkXL.~ 6XeXqXqXqXqXqX8XkX! <.jXrX5X3X6X0XdX8XN M.kX7XqXqXqXqXqXqXqXqXqX0XtX;X'.fX;Xk   V.dX9XqXqXqXqXqXeX4XQ U.kXdXwX5XeXsXkXT.Q 3XeXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqX9XwXtX .  c z.=XsXxXpX=Xp.~ &XuX9XqXqXqXqX8XdXB.$ 2 <. XyXzXjX7XP.!   k %XiX8XqXqXqXqXqXqXqXqX0XtX-X[.rXyX>Xl   V.dX8XqXqXqXqX9XuX$XQ a.=XaXxXaX-Xd.Q @XuX9XqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqX0XwXyX}   ;.1X6.| &.=.&.| 0.6XuX9XqXqXqXqX8XdXV.  k =XV.[ $.=.*.o.+.XXL.  a &XiX8XqXqXqXqXqXqXqX0XtX-X[.uX7XuX>Xk   V.dX9XqXqXqXqX9XuX5X8.| &.=.&.| 7.4XuX9XqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqX9XwXyX}   -.iXtXgX,X~.p./.<XhXeX9XqXqXqXqX8XdXV.  l :XiXdX0X Xj.F.%XsXuXsXH.. s &XiX8XqXqXqXqXqXqX0XtX-X[.uX0X8XuX>Xl   V.dX8XqXqXqXqX9XeXhX<X/.p.^.,XhXeX9XqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqX9XwXyX}   -.iXqX9X8XtXaX|.dXrX8XqXqXqXqXqX8XfXV.  k :XuX8X9X0XjX@X7XsX9X9X8XdXH.. a &XiX8XqXqXqXqXqX0XtX-X[.uX0XqX8XuX>Xk   V.dX8XqXqXqXqXqX8XrXsX|.sXrX8X0XqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqX0XwXyX}   -.iXqX0XqXqX0X9X~.0X0XqXqXqXqXqX8XdXV.  l :XuX8XqXqX9XyX'.-XrX0XqXqX8XdXH.. s &XiX8XqXqXqXqX0XtX-X[.uX0XqXqX8XuX>Xl   V.dX8XqXqXqXqXqX0X0X~.9X0XqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqX9XwXyX}   -.iXqX0XqXqXqXqX0X^.wXqXqXqXqXqX8XfXV.  l :XuX8XqXqXqX0XuX].;XtXqXqXqXqX8XdXH.. s &XiX8XqXqXqX0XtX-X[.uX0XqXqXqX8XuX>Xl   V.dX8XqXqXqXqXqXqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqX0XwXyX}   -.iXqX0XqXqXqXqXqX0X^.wXqXqXqXqX8XdXV.  l :XuX9XqXqXqXqX0XuX].;XtXqXqXqXqXqX8XdXH.. a &XiX8XqXqX0XtX-X[.uX0XqXqXqXqX9XuX>Xk   V.dX9XqXqXqXqXqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X|.uX8X0X9XwXtX}   -.iXqX0XqXqXqXqX9X0XqX/.wX0X9XqX8XdXV.  l :XuX8XqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqX8XdXH.. s &XiX8X0X9XrX;X{.uX8X0XqXqXqXqX8XuX>Xl   V.dX8X0X9X0XwX/.qX0X9XqXqXqXqXqXqXqX",
"qX0XyX&X`.iXsXyXqXyX}   -.iXqX0XqXqXqX0X0XsXpX9XE.0XpXsX8XdXV.  k :XuX9XqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqX8XdXH.. a &XuXqXfXpX&X).iXsXuX9XqXqXqXqX9XuX>Xk   V.dX8XsXpX0XE.9XpXsX9XqXqXqXqXqXqX",
"qX0XiX.XL O.e.%XsX[   -.iXqX0XqXqXqX0XwXaXU.,.] U ' <.T.kXB.  l :XuX8XqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqX8XdXH.. p -XeXx.$.~ T X.w.#XdX9XqXqXqXqX8XuX>Xl   B.jXY.<.[ U ` ,.Y.sXwX0XqXqXqXqX",
"qXqX0XtXgX@Xs.{ S   &.iXqX0XqXqXqX0XeXqX-.X.K.,XjXwXG.+.! $ k :XuX8XqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqX8XdXG.@ i ` 2.[.sXgX#Xj.~ B.dX9XqXqXqXqX8XuX:Xk $ Q @.F.>XjXeXL. .:.wXeX0XqXqXqX",
"qX0XrX:XG.qXcX1Xv Y gX0X0XqXqXqXqX0XyX+.w.sXlXyXS.D.pXfX<.3 =XiX8XqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqX7XkXh.> !.mX>XB.J.9XvX,X( V.dX9XqXqXqXqX8XiX=X2 4.sXkXuXF.S.iXfX9.$.uX0XqXqXqX",
"qX9XsX_.  H :XdX,X{ *XtX0XqXqXqX8XgXl.;.jXwX*XF   X *XrXjX<.V.dX9XqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqX0XyX*.L.mXJ.1     Z =XdX>X~ $XuX0XqXqXqX9XdXM.3.kXwX;XL     #XtXhX=.c.gX8XqXqX",
"qXqX9XpXn.  v.dXgXu.p.sX8X0X0X9XqX6X[ $XtXpXn.Z E : >XeXuX|.{ 0X0X9X0X0X0X0X0X0X0X0X8XyX].-XeX9X0X0X0X0X0X0X0X0X8XuX/. .uXfXD 8 4XA.  h.fXgXh.r.sX8X0X0X9X0X9X[ XXyXpXC.V ! * =XrXyX@X[ 9XwX0XqX",
"qXqX8XfXT.  V.gXeX X+.sXyXuXuXtXkX_.-.uX0XwX;XpX&.@ ,XtX0XrX$. XjXyXuXuXuXuXuXuXuXuXyXfX}.1XsXyXuXuXuXuXuXuXuXuXrXbXu.b.sXqX X+XzX!.  c.hXwXOXO.pXuXuXuXtXjX{.&.tX0XwX;XpX,.o -XyX9XtX*.(.sX9XqX",
"qX0XrX<X3 m 0XqXqX6XE ].].].].'.$Xp.1.lX8XqXwXdX..$ ,XtX7XkX=.j.@X'.].].].].].].].].].}.U.^.{.].].].].].].].].]._.*X] T.dX9XtXfX;X7 z 8XwX0X0XW _.[.].].'.@Xg.;.lX8XqXwXfX#.O -XyX6XlX>.K.gX9XqX",
"qX9XfXY.e 1XuX9XeX<X) >X;X;X;X=X8XN.,.hX8XqX8XaXX.$ ,XtX7XgX=.F.7X-X;X;X;X;X;X;X;X;X-X1X^.oX,X;X;X;X;X;X;X;X;X;X*XeX%.K.sX8XiX$Xw e ;XiX9XwX4X) =X:X;X;X-X7XS.-.hX8XqX8XsX%.O -XyX7XhX:.I.fX9XqX",
"qX0XtX=X.XyX7X9XaXE.;.dXrXtXtXeXdXXX@.0XwXqX9XsXO.* ,XrXwX6XO.%XsXrXtXtXtXtXtXtXtXtXeXsX{.,XpXrXtXtXtXtXtXtXtXtXwXlXc.u.sXwXOXt q %XtX7X9XpX^.=.aXrXtXtXeXsX#XO.7XwXqX9XsX*.$ -XtXqX8X+.XXiX0XqX",
"qX9XaX_.    2.iXgX-.V.dX9XqXqX0X9XpX..Q.dX8X9XdX`   ,XqXfXP.#.sX8X0XqXqXqXqXqXqXqXqX9XyX].;XrX0XqXqXqXqXqXqXqXqX0XrX&X/ 7XkXn     @   -.uXhX<.n.dX9XqXqX0X9XaX+.Y.fX7X9XdX .  =XeXdXR.o.aX9XqXqX",
"qX0XrX:Xl.k.'.bXK.{ uX0XqXqXqXqX9XpX(.! 8XpX7XeX).B.4XdX2XW }.iX0XqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqX9XfXy.:.vXR.v.n.c.k.).bXT.' tX0XqXqXqXqX0XiX].W 4XaX7XeX`.N.3XsX6XQ `.pX0XqXqX",
"qXqX0XyXBXNXhXA.& g :XuX8XqXqXqXqX9XsXn.~ +XfXsXxXmXfX XQ Z.sX9XqXqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXtX&.;.aXBXMXNXNXjXG.- p =XuX8XqXqXqXqX9XsXB.Q .XfXaXxXMXgXoX! N.dX9XqXqXqX",
"qXqXeX3XG.h.' ,.h.. g ;XuX8XqXqXqXqX9XsXP./ ,.Z.I.M.;.) E.sX9XqXqXqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqX0XwXtXi.! w.D.K.j.[ :.l.+ p &XiX8XqXqXqXqX9XsXT.( >.C.I.N.>.( U.sX9XqXqXqXqX",
"qX0XiX X' f.'.iXkXZ.  h ;XuX8XqXqXqXqX9XsX8XH.7.` 9.K.0XaX9XqXqXqXqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqX0XqXfX*Xb.*.} s.`.uXkXG.. a &XiX8XqXqXqXqX9XaX0XK.8.` 7.J.9XsX9XqXqXqXqXqX",
"qXqXtX;X;XvXyX0X8XfXA.  j -XuX8XqXqXqXqX9XwXsXkXOXzXaXwX9XqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XrXlXwX=XbXyX0X7XdXH.. s &XiX8XqXqXqXqX9XwXsXlXOXkXsXwX9XqXqXqXqXqXqX",
"qX0XtX*X'.rX9XqXqX8XdXA.  j ;XuX8XqXqXqXqXqX9X7XQ.8X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwX&X`.rX9XqXqX8XdXH.. a &XiX8XqXqXqXqXqX9X8XQ.7X9XqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqX8XfXA.  j -XuX8XqXqXqXqXqX0X^.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqX8XdXH.. s &XiX8XqXqXqXqXqXqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqX8XdXA.  j ;XuX8XqXqXqXqX0X^.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqX8XdXH.. s &XiX8XqXqXqXqXqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqX8XfXA.  h ;XuX8XqXqXqX0X^.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqX8XdXH.. a &XiX8XqXqXqXqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqX8XdXA.  j ;XuX8XqXqX0X^.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqX8XdXH.. s &XiX8XqXqXqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqX8XdXA.  h ;XuX8XqX0X^.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqXqX8XdXH.. a &XiX8XqXqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqX8XdXA.  j -XuX8X0X^.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqXqXqX8XdXH.. s &XiX8XwX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqX8XdXA.  j ;XuX8X^.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqXqXqXqX8XdXH.. a &XiX9X^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqXqXqXqXqXqXqXqX8XfXA.  j ;XuX!.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqXqXqXqXqX8XdXH.. s &XiX!.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX*X[.yX9X0X0X0X0X0X0X0X0X0XqX0X8XdXZ.  j -X_.9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XyX].;XrX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XrX-X[.yX9X0X0X0X0X0X0X0X0X0XqX0X8XsXG.  a &X`.9XqXqXqXqXqXqXqXqXqX",
"qX0XtX=X+XaXeXtXtXtXtXtXtXtXtXtXtXtXtXwXjXG.. j ^.pXwXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXeXsX{.,XpXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXpX>X}.sXeXtXtXtXtXtXtXtXtXtXtXtXtXwXjXP.O p T.rX0XqXqXqXqXqXqXqXqX",
"qX0XyX$XS.6X-X=X;X-X-X-X-X-X-X-X-X-X-X-X*X6Xz.  x 5X>X*X;X-X-X-X-X-X-X-X-X-X-X-X-X-X=X<X^.XX>X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X>XXX^.<X=X-X-X-X-X-X-X-X-X-X-X-X-X-X*X5Xm.@ M sX9XqXqXqXqXqXqXqXqX",
"qX9XsX/.  ! XX{.].[.[.[.[.[.[.[.[.[.[.[.[.`.$XJ.  7 Y.XX`.[.[.[.[.[.[.[.[.[.[.[.[.[.].|.Y./.}.[.[.[.[.[.[.[.[.[.[.[.[.[.[.[.[.}.^.Y.|.].[.[.[.[.[.[.[.[.[.[.[.[.[.`..X~.9 6 sX9XqXqXqXqXqXqXqXqX",
"qX0XrX>XI   >.gXuXyXuXuXuXuXuXuXuXuXuXuXuXuXeXzXr.. v 1XfXrXuXuXuXuXuXuXuXuXuXuXuXuXyXfX}.1XsXyXuXuXuXuXuXuXuXuXuXuXuXuXuXuXyXsX<X|.fXyXuXuXuXuXuXuXuXuXuXuXuXuXrXfX2Xx   h.uX9XqXqXqXqXqXqXqXqX",
"qX0XtX-XXX@.  %.uX0X8X0X0X0X0X0X0X0X0X0X0X0X0X6X{.A.  g =XyX7X0X0X0X0X0X0X0X0X0X0X0X8XyX].-XeX9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XeX=X].yX8X0X0X0X0X0X0X0X0X0X0X0X7XyX-Xg   Z.{.9XqXqXqXqXqXqXqXqXqX",
"qX0XtX&X|.dX .  &.iXqX0XqXqXqXqXqXqXqXqXqXqXqX0X!.fXA.  j -XuX8XqXqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqXqXqXqXqX8XuX;Xh   Z.fX!.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X[.uXtX..  &.iXqX0XqXqXqXqXqXqXqXqXqXqX0X^.9XdXA.  j ;XuX8XqXqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqXqXqXqX8XuX;Xh   Z.dX9X^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X}.yXqXyX..  &.iXqX0XqXqXqXqXqXqXqXqXqX0X^.wX8XfXA.  h ;XuX8XqXqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqXqXqX8XuX;Xh   Z.dX8XqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX8XwXyX..  &.iXqX0XqXqXqXqXqXqXqXqX0X^.wXqX8XdXA.  j ;XuX8XqXqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqXqX8XuX;Xh   Z.dX8XqXqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0X0XwXyX..  &.iXqX0XqXqXqXqXqXqXqX0X^.wXqXqX8XdXA.  h ;XuX8XqXqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqXqX8XuX;Xh   Z.dX8XqXqXqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqX9XwXyX..  &.iXqX0XqXqXqXqXqXqX0X^.wXqXqXqX8XdXA.  j -XuX8XqXqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqXqX8XuX;Xh   Z.dX8XqXqXqXqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqX0XwXyX..  &.iXqX0XqXqXqXqXqX0X^.wXqXqXqXqX8XdXA.  j ;XuX8XqXqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqXqX8XuX;Xh   Z.dX8XqXqXqXqXqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX=X{.uX0XqXqXqX9XwXyX..  &.iXqX0XqXqXqXqX0X^.wXqXqXqXqXqX8XfXA.  j -XuX8XqXqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqXqX8XuX;Xh   Z.dX8XqXqXqXqXqXqX^.qXqXqXqXqXqXqXqXqXqX",
"qX0XtX*X[.yX9XqXqXqXqX0XqXyX..  &.iXqX0XqXqX0X9X~.0X0XqXqXqXqXqX8XdXA.  j ;XuX8XqXqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqXqX8XuX;Xh   Z.dX8XqXqXqXqXqX0X0X~.9X0XqXqXqXqXqXqXqXqX",
"qX0XtX-X%XjX0X9XqXqXqXqX0XwXyX..  &.iXqX9X8XtXsX|.dXrX8XqXqXqXqXqX8XfXA.  h ;XuX8XqX0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0XqX8XuX;Xh   Z.dX8XqXqXqXqX0X8XrXdX|.sXtX8X0XqXqXqXqXqXqX",
"qX0XuXOXg.|.qXfX0X0XqXqXqX9XwXyX..  &.uXyXgX:XW.u.~.>XhXeX9XqXqXqXqX8XdXA.  j ;XuX8X0XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX0X8XuX;Xh   Z.dX8XqXqXqXqX9XeXhX>X!.u.Q.:XhXrX9XqXqXqXqXqX",
"qX0XyX%X+.$.{ B.pXwX0XqXqXqX0XwXyX..  =.>X4.| *.*.$.| 7.4XuX9XqXqXqXqX8XdXA.  h ;XuX7XuX].;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.uX7XuX;Xh   Z.dX8XqXqXqXqX9XuX3X6.| %.&.%.| 4.2XiX9XqXqXqXqX",
"qXqX9XpXKXdX X@.>.tXqXqXqXqXqX9XwXtXo.  c b.:XhXGXNX,Xd.! $XuX9XqXqXqXqX8XdXA.  j -XyXrX[.;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-X[.rXyX;Xh   Z.dX8XqXqXqXqX9XuX@XQ f.<XmXKXnX,Xh.Q OXiX9XqXqXqX",
"qX0XiX.XK P.jXgX<.6.dX9XqXqXqXqX9XuX-Xk ).hXsXXX~ ' 6XcXI.! 5XeXqXqXqXqXqX8XdXA.  j =XfX`.;XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX;X`.fX-Xh   Z.dX8XqXqXqXqX0XeX3XW Y.vX4X,.F 0.qXxXE.W 2XrX0XqXqX",
"qX9XiXoX9   B.sXyX' (.aX0XqXqXqX8XfX2.m.jXeX_.2   X ;XqXkXa.8.hX8XqXqXqXqXqX8XfXA.  g 2X.X=XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX*XXX2Xg   Z.dX8XqXqXqXqXqX8XhX6.f.zX5XB . M   ! eXkXh.4.hX8XqXqX",
"qXqX9XuXuXw l 0XgXV.6.aX0XqXqX0XtX=X| <XwXpXA.0.@.* ,XrXrX;X| <XrX0XqXqXqXqXqX8XdXA.  c W.3XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX<XT.c   Z.dX8XqXqXqXqXqX0XrX>X .;XkXZ.  f.HX:.  !.fX>X| ;XtX0XqX",
"qXqXqX8XaXX.X 3XuX&X} rXqXqXqX9XfXY.>.dX9XqX0XxXO.# ,XtX8XaX&.^.sX0XqXqXqXqXqXqX9XfXA.  w ,XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX&X< X Z.fX9XqXqXqXqXqXqX0XdX!.=.iXdX2.  OXfXE.  g.hXuX;.E.dX9XqX",
"qXqXqX9XuX$.  2XrX7XT ^.^.^.^.~. Xe.1.lX7XqX9XpXX.$ ,XtX7XlX=.a.|.~.^.^.^.^.^.^.^.!.}.g.4 C.'.~.^.^.^.^.^.^.^.^.^.^.^.^.^.^.~.'.V.6 h.{.!.^.^.^.^.^.^.^.~.|.u.;.jXiX;.  #XpX~.  a.gXhX>.J.gX9XqX",
"qXqX0XeXzXG 3 8XtX*X} rX0XqXqX9XdXY.:.dX9XqX9XsXo.$ ,XtX8XaX*.^.sX9XqXqXqXqXqXqXqXqX9XuXdXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXdXuX9XqXqXqXqXqXqXqXqX9XsXQ.=.iXgXs.  E.FXb.  Z.fXiX;.E.dX9XqX",
"qX0XeX4X6.  *.uXdXC.6.aX0XqXqX0XtX=X .2XeXqX9XsX..+ ,XrXrX:X .,XrX0XqXqXqXqXqXqXqXqXqX9X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X9XqXqXqXqXqXqXqXqXqX0XtX:X| :XaX+X% h S.8 9 1XyX,X .;XtX0XqX",
"qX9XsX/.  } 2XrXiX[ ^.aX0XqXqXqX8XgX<.x.lX7X9XaX#.> ,XqXkXd.7.hX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XhX4.h.jXtXG.y   j !.tXjXl.2.gX8XqXqX",
"qXqXwX9X2XyXsXgX3.5.dX9XqXqXqXqX0XrX,XR ^.lXeX9X2X,XeXlXT.! 4XeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX2XW W.kXsX9X<X0XaXkX!.W 1XrX0XqXqX",
"qXqX0XtXcXeXXX#.;.tXqXqXqXqXqXqXqX9XiX XR z.1XsXxXfX<Xg.Q @XuX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XiXOXE j.>XdXvXsX,Xl.W oXiX9XqXqXqX",
"qX0XyX%X&.*.[ M.iXwX0XqXqXqXqXqXqXqX9XiX<X1.| ;.:.-.} 5.3XuX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XiX2X4.| ;.:.-.| 2.1XiX9XqXqXqXqX",
"qX0XrX<XH._.qXfX0X0XqXqXqXqXqXqXqXqXqX9XrXgX-XR.J.E.:XhXeX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XrXhX:XE.J.E.;XgXrX9XqXqXqXqXqX",
"qXqX0XrXhXaXqX9XqXqXqXqXqXqXqXqXqXqXqXqX0X8XtXdXgXdXtX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X8XtXdXgXdXtX8X0XqXqXqXqXqXqX",
"qXqXqXqX9X0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9X9X9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9X9X9X0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_2[] = {
/* columns rows colors chars-per-pixel */
"88 88 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX9X9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9X0XqXqXqXqX8X7X9XqXqXqX9X0X0X0X0X0X0X0X0X0X0X0X8X8X0X0X0X0X0X0X0X0X0X0X0X9X9XqXqXqXqX9X7X0XqXqXqX9XqXqXqXqX",
"qXqX0XaXpXtXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXtXsXuX0XqX9XwXhXxXdX0XqX0XdXuXyXyXyXyXyXyXyXyXyXtXfXhXtXyXyXyXyXyXyXyXyXyXtXaXiX0XqXqX8XsXcXrX0XqX0XsXeXqXqXqX",
"qX9XaXG.T. X[.[.[.[.[.[.[.[.[.[.[.[.[.[.[.[.[.[.[.[.[.[.[.[.[.[.].oXA./.pX9XaX<X0...B.uX9XyXM.(.}.[.[.[.[.[.[.[.].OXh.t.OX].[.[.[.[.[.[.[.[.XXF.W.aX9X8XgXF.} $XtX9XuXZ.1XrXqXqX",
"qX9XpXT.OX4X<X<X<X<X<X<X<X<X<X<X<X<X<X<X2X*X~.4X,X<X<X<X<X<X<X<X>XtXk.e.jX7XXX:X7X&.  }.uX9X' $X3X<X<X<X<X<X<X<X<X4X}.A.6X,X<X<X<X<X<X<X<X>XrXM.5.hX6XhXC.@ <.fXqX9XtX{ =XtXqXqX",
"qX0XtX X4XyXeXrXrXrXrXrXrXrXrXrXrXrXrXeXyX5X{.iXeXrXrXrXrXrXrXrXwXaX.X-.rXuXB d ^.~   |.gX'.>.iXeXrXrXrXrXrXrXrXrXtXqX XuXeXrXrXrXrXrXrXrXeXpX$X*.0XpXx.  G _.T.4XiXOX-.wXqXqXqX",
"qX0XyX[.<XeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXeX3X].tX0XqXqXqXqXqXqXqXqX9XuX=.).xXW.c # e b.tXkXw.Z.dX9XqXqXqXqXqXqXqXqXwX4X`.tX0XqXqXqXqXqXqXqXqX9XpX,.R.NXN O 3 . 3 #XvX9.b.gX9XqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3X].yX0XqXqXqXqXqXqXqXqX7XhXE.k %XnXsX9XrXkXlXh.M uXwX0XqXqXqXqXqXqXqXqXwX5X'.tX0XqXqXqXqXqXqXqXqX8XfX_.g oXjXwX7X9XuXlXM.+.tXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3X].yX0XqXqXqXqXqXqXqX8XiX:Xh : ] k.#XrX8X/.8.H   %.aXqX0XqXqXqXqXqXqXqXwX5X'.tX0XqXqXqXqXqXqXqX8XuX2Xb = _ k.#XrX0X{.5.$.5XrX0XqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3X].yX0XqXqXqXqXqXqX8XuX>Xa = }.5Xp.*.] ..,.D.dXe.  &.aXqX0XqXqXqXqXqXqXwX5X'.tX0XqXqXqXqXqXqX8XyX3Xz @ ).8Xd.=.[  .>.Z.yXrX0XqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3X].yX0XqXqXqXqXqX8XuX>Xa ; [.aXwXdXiX_.=XaXsX0XfXr.  &.aXqX0XqXqXqXqXqXwX5X'.tX0XqXqXqXqXqX8XyX3Xz $ /.sXwXdXiX].%XaXdXqX0XqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3X].yX0XqXqXqXqX8XuX>Xa ; [.sX8X0X8XyX X,XeX9X0X9XfXr.  &.aXqX0XqXqXqXqXwX5X'.tX0XqXqXqXqX8XyX3Xz $ /.dX8X0X8XyXXX;XrX8XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3X].yX0XqXqXqX8XuX>Xa ; [.sX8XqXqX0XyX}.>XrXqXqX0X9XfXr.  &.aXqX0XqXqXqXwX5X'.tX0XqXqXqX8XyX3Xl $ /.dX8XqXqX0XyX X-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX3X].yX0XqXqX8XuX>Xa ; [.sX8XqXqXqX0XyX}.>XrXqXqXqX0X9XfXr.  &.aXqX0XqXqXwX5X'.tX0XqXqX8XyX3Xl $ /.dX8XqXqXqX0XyX X-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqXqXqXqXqXqX9XwX3X].tX8X0X8XuX>Xa ; [.sX8XqXqXqX0X8XyX}.>XeX9XqXqXqX0X9XfXr.  &.aXqX0XqXwX5X'.tX0XqX8XyX3Xz $ /.dX8XqXqXqX0X8XyX.X-XrX9XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqXqXqXqXqX9XaXdX4X'.aXdXeXuX>Xa ; [.sX8XqXqXqX9XeXdXsX{.>XfXsX0X0XqXqX0X9XfXr.  &.aXqX0XeX5X'.tX0X8XyX3Xz $ /.dX8XqXqXqX9XeXdXdX X-XfXsX0X0XqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqXqXqX0XqXsX!.5. ./ &.s.1X1Xi ; [.sX8XqXqXqX9XiX5Xk.*.( } 3.I.pXwX0XqXqX0X9XfXr.  &.aXqXqX5X`.tX8XyX3Xl $ /.dX8XqXqXqX9XiX7Xc.-.) [ 1.J.iXeX0XqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqXqXqXqXyX9.@.Q.iXzX=Xj.X.e 3 '.sX8XqXqXqX9XpX}.] s.:XzXtX!.-.<.wXwX0XqXqX0X9XfXr.  &.aXwX4X'.wXtX3Xz $ /.dX8XqXqXqX9XiXoX[ y.#XkXaX(.>.-.9XeX0XqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqXqX9XsX9.7.zX6X8.;..XnX;Xj n.lX7XqXqXqX0XyX$X/ |.nXD.#.N.dXkXs.:.pX0XqXqXqX0X9XfXr.  &.dX5X(.aX2Xz $ /.dX8XqXqXqX0XtX;X( ).MX X;.u.eXcXl.%.yX0XqXqXqX",
"qX0XyX[.<XeXqXqXqXqXqXqXqXqX9XaX!.+.lX@X1 4 u > ^.xX.X$.eX0XqXqXqX9XaX*.I.NXO.  c   &.iXhX,.F.dX9XqXqXqX0X9XfXr.  *.wX X6Xl $ /.dX8XqXqXqXqX9XdX,.D.mXI.# i = M 7XzX5.N.gX9XqXqX",
"qX0XyX[.2XyXeXeXeXeXeXeXeXeXwXdX6.Y.zX'.@.pXiX  y.gXuX,.^.dXqXeXwXaXXX=.yXyXQ 0.HX ., iXiX].:.iXeXeXeXeXeXwXqXhXy.  $.~.x $ ).gX0XeXeXeXeXwXiX$X%.0XhXZ t zXQ.  m.xX X*.eXqXqXqX",
"qX0XyX[.*X5X3X3X3X3X3X3X3X3X3X4Xo.1XwXwXkXs.q M ,XeXsXn.s.iX<X3X<XyXm.a.gX8XyXdX8Xt ! dX9X0XO.=X5X3X3X3X3X3X2X1XtX5.  6 4 I.eX<X3X3X3X3X3X1XtXS.w.gX0X7  .kX,X> =.zX0X#.;XtXqXqX",
"qX0XyX[.~.{.].].].].].].].].].'.` 9X0XeXkXx.G 8 F.yXdXZ.>.%X_.].`.$X0.g.jX7XtX8Xc 9 :XrX8XuX} E.}.].].].].].].'._.$Xi.  E $X`.'.].].].].].`.#Xu.t.kX9X7  .kX,X> -.lXyX+.*XyXqXqX",
"qX0XyX[.4XiXtXyXyXyXyXyXyXyXtXaX=.+XfXT.o.uXCX* J aXaXf.C.zXeXyXeXkXT.q.aXqX5XV k tXnX9XeX,X&.qXuXtXyXyXyXyXyXyXtXeXjX$.  =.hXtXrXyXyXyXyXrXjX/.6.uXfXC t lXQ.  m.hX1X%.1XrXqXqX",
"qX0XyX[.,XeX0X0X0X0X0X0X0X0X8XdXs.j.nX!.$ 4 t O B.iXeXX.$XrX9X0X9XwX1X| 2XfXm   t x D >XjXS.7.sX8X0X0X0X0X0X0X0X0XqX1XOXy.  %.pX0X9X0X0X0X9XqX6X..:XdXY.# u * N 5XgXI.2.sX9XqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqX0XeX1X..;XxXoX2.=.Q.uXxXs.w.fX9XqXqXqX8XdXv.e.nXu.>.&. .<.eX0X..%XtX0XqXqXqXqXqXqXqXqXeX5X_.kXe.  &.aXqX0XqXqXqX9XfXn.<.jXaX).*.r.1XfX9X[ |.iX9XqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqX8XuX1Xe j .XhXZXDXxXrXs.@.qXwX0XqXqX0X9XgXs.  q.kXCXSXFXVX=XS > +XpX8XqXqXqXqXqXqXqXqXwX5X`.eXdXr.  &.aXqX0XqXqXqX0XiX4.2.7XzXSXVXfX%X} C.dX9XqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqX8XuX>Xi 3 n.$.:.c.N.s.X.w.qXrX0XqXqXqX9XfXd.  ..h.[ t.N.v.<.O.B.p 1 oXaX8XqXqXqXqXqXqXqXwX5X'.rX8XfXr.  &.aXqX0XqXqXqXwXuXj.} y.M.n.6.| Y.dX9XqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqX8XuX>Xa ; '.lXeX/.s.:.C.$XfXwX0XqXqXqX9XfXf.   .iXgX=XD.<.y.W.0XlX%Xw 3 oXaX8XqXqXqXqXqXqXwX5X'.tX9X9XfXr.  &.aXqX0XqXqX0X0XfX>XG.2.e.R.0XsX9XqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqX8XuX>Xa ; [.sX7X0XdXiX%XzXrX9X0XqXqX0X9XfXf.   .iXqX8XeXzX-XeXgXqX7XpX*Xw 3 oXaX8XqXqXqXqXqXwX5X'.tX0X0X9XfXr.  &.aXqX0XqXqXqX9XwXzX:X0XhXqX9XqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqX8XuX>Xa ; [.sX8XqXqXqX<X_.eX9XqXqXqXqX9XgXf.   .iXqX0XqX9XrX].-XwX0XqX8XiX*Xw 3 oXaX8XqXqXqXqXwX5X'.tX0XqX0X9XfXr.  &.aXqX0XqXqXqX9XrX{.&XeX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqX8XuX>Xa ; [.sX8XqXqXqXeX3X].yX0XqXqXqX9XfXf.   .iXwX0XqXqX0XyX}.>XrXqXqXqX8XiX*Xw 3 oXaX8XqXqXqXwX5X'.tX0XqXqX0X9XfXr.  &.aXqX0XqXqX0XyX X-XtX0XqXqXqXqXqXqXqX",
"qX0XyX].>XwXqXqX8XuX>Xa ; [.sX8XqXqXqXqXwX<X`.eX9XqX0X9XfXf.   .iXqX0XqXqXqX0XyX}.>XrXqXqXqXqX8XpX*Xw 3 oXaX8XqXqXqX3X).eX9XqXqXqX0X9XfXr.  &.aXqX0XqX9XrX}.*XeX0XqXqXqXqXqXqXqX",
"qX0XtXoXtXaX9X7XiX>Xa ; [.sX8XqXqXqX0X9XaXyX$XkXwX8X9XfXf.   .iXqX9XqXqXqXqX0XyX}.>XrXqXqXqXqXqX8XiX*Xw 3 oXaX7X9XiXpX#XkXeX9XqXqXqX0X9XfXr.  &.aXqX8XqXkX;X9XfX9X0XqXqXqXqXqXqX",
"qX9XsXA.k..XuXhX:Xa ; [.sX8XqXqXqX9XyXaXXXm.0.T.1XdXgXd.   .iXwX0XqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqX8XiX*Xw 3 XXhXpXOXB.8.U.,XfXqX0XqXqX0X9XfXr.  &.aXfX7X~.t.k.[.yXpX9XqXqXqXqXqX",
"qX0XuX(.e.=.-.E.h = [.sX8XqXqXqX9XpX$X*.*.i.j.w.| v.s.   .iXqX0XqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqX8XpX*X0 7 U.1.@.a.j.q...s.rXeX0XqXqX0X9XfXr.  -.D. .5.f.f.,.O.'.sX9XqXqXqXqX",
"qXqX0XiXcXaX_.k : }.aX8XqXqXqX9XiX}.( Y.pXnXgXiX5Xr.  ..iXqX0XqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqX8XiX-Xw q W.dXgXjXzX9Xw.*.eXwXqXqXqX0X9XfXt.  %.:XkXvXaXuX[._ U.sX9XqXqXqX",
"qX7XzX*.y XXzX%X] 5XwX0XqXqX0XeX5X] }.kX8X:.  ( gXlXw.h.gX8XqXqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqX0XeX .].xXq.> = ;.wXxXy.r.fX9XqXqXqX8XdXJ.*.hXpX[.v O I.bX>X_ #XuX0XqXqX",
"qX0XyXoX6 B pXhXz.p.dX8X0X0X8XdXk.d.jXuXm.p 6 F hX8XqX| -XeX9X0X0X0X0X0X0X0X9XtX}.:XeX9X0X0X0X0X0X0X0X8XsXn.w.BXL.  w.s.  B.fXeX..#XrX9X0X9X0X9X..=XtXeXR l X D.aXfXF.6.fX9XqXqX",
"qXqX8XvX! s fXeX$X*.iXyXyXyXyXsX;.oXyXeX+X5XA S jX6XdXp.D.xXrXyXyXyXyXyXyXyXtXdX.X2XaXyXyXyXyXyXyXyXyXtXgX,.].sX@XS.hX2XX k.gXpXj.M.xXrXyXtXjX`.4.iX0X6X$X_.o G.fX0X2X%.3XeXqXqX",
"qX9XfXN.  C.aX8XeX[ `. X}.}.}.{.' 8XqX0XyXpXv D jX6XgXZ.<.-X].}.}.}.}.}.}.}.}..XR.). X}.}.}.}.}.}.}.}.{.XX` 3XwXrXhXrXM i 5XqXdXK.=.-X].}.[.&Xa.y.jX8X9XjXQ.  G.gX6XpX+.*XyXqXqX",
"qX7XxX#.8.mX9XqX7XX.=X,X>X>X>X>X..3XwXqX0XtXn D jX6XfXM.y.eX-X>X>X>X>X>X>X>X:X2X).+X<X:X>X>X>X>X>X>X>X:X3X..:XwXrX9XG e <XfX7XdXD.9.eX-X>X;X0XB.e.gX8X9XiX!.. G.fX7XtX#.-XtXqXqX",
"qX0XuX(.W.].8XpX).,.aXeXrXrXeXfX3.!.aX8XqXyXc Z kX5XdX5.Q.gXwXrXrXrXrXrXrXrXeXaX X<XiXrXrXrXrXrXrXrXrXeXgX9.I.gX<XV   V.'.#XqXdX9.P.hXwXrXeXsX+X=.rX0X9XaXQ.  F.fXeX#X*.qXwXqXqX",
"qX7XlX>.  p 1XxX8.D.sX9XqXqX9XsXI.-.jX8X9XtXZ L sXuX+X .wXqX0XqXqXqXqXqXqXqX9XyX}.:XrX0XqXqXqXqXqXqXqX9XaX!.+.MXD.  > *   <.zX:XX.5XwX0XqX0X9XaX*.R.hX5XiX^.& H.pXkXt.z.gX9XqXqX",
"qXqXwX4X<XqXzXh.G sX0X0XqXqXqX0XaX,.s.kXpXqX1X6XjX4X' !.aX9XqXqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqX9XsX6.t.fXqX<X,X3XpXrXJ j.zX7XqXqXqX0XuX@X( @XkXtX7X<XrXxXD...wXqXqXqXqX",
"qXqX0XiXtX}.w.N   0.dX9X0XqXqX0XwXwX;.,.].0XaX1XD.' H.sX9XqXqXqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXwXrX1.;.[.wXsX4XJ.+.3 q .XaX8XqXqXqX9XaX_._ b.=XuXyXoXq.o.1XtX0XqXqXqX",
"qX9XdXm.` :.C.iX=.  q.dX9X0XqXqX0XwXpXF.:.O.{ %.8.+XsX9XqXqXqXqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqX0XwXaXK.:.O.{ $.5.+X-X0 3 oXaX8XqXqXqX9XpX,Xy.*.| ..=.b.eXtX0XqXqXqXqX",
"qX0XuX(.$XiXdXqXsX=.  q.dX9X0XqXqX0X0XdXiX=XE.qXsXuX9XqXqXqXqXqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqX0X0XdXiX:XR.0XsXyXiX*Xw 3 oXaX8XqXqXqX9XtXdXrX/.oXuXfXqX0XqXqXqXqXqX",
"qX0XyX}.3XeX9X0X0XsX=.  q.fX9X0XqXqXqX9XeX5X}.uX8X0XqXqXqXqXqXqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqX9XwX7X[.yX8X0X8XiX*Xw 3 oXaX8XqXqXqX0X8XuXoX>XtX8XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqX0X0XsX=.  q.fX9X0XqXqXqXeX3X].tX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X`.tX0XqXqX8XiX*Xw 3 oXaX8XqXqXqX0XyX X-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqX0X0XsX=.  q.dX9X0XqXqXeX3X].yX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqXqXqX8XpX*Xw 3 oXaX8XqXqX0XyX X-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqX0X0XsX=.  q.fX9X0XqXeX3X].yX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqXqXqXqX8XiX*Xw 3 oXaX8XqX0XyX X-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqX0X0XsX=.  q.fX9X0XeX3X].yX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqXqXqXqXqX8XiX*Xw 3 oXaX8X0XyX X-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqX0X0XsX=.  q.dX9XwX3X].yX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqXqXqXqXqXqX8XiX*Xw 3 oXaX7XuX X-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqX0X0XsX=.  q.dXqX2X].yX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqXqXqXqXqXqXqX8XpX*Xw 3 oXpXeX.X-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqX0X0XsX=.  q.hX1X'.yX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqXqXqXqXqXqXqXqX8XiX*Xw 3 XXhX{.-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqXqXqXqXqXqXqX0X0XsX=.  e.rX`.tX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX}.>XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X`.tX0XqXqXqXqXqXqXqXqXqX8XiX*Xw 1 $X$X&XtX0XqXqXqXqXqXqXqX",
"qX0XyX{.1XrXwXwXwXwXwXwXwXwXwXeXqXwXdX;.  5.$XeXqXeXwXwXwXwXwXwXwXwXwXwXwXwXqXiX|.,XyXwXwXwXwXwXwXwXwXwXwXwXwXwXwXrX6X'.yXwXwXwXwXwXwXwXwXwXwXeX9XaX=Xe 7 K.3XeXqXqXqXqXqXqXqXqX",
"qX0XyX].1XwX3X5X5X5X5X5X5X5X5X5X5X4X4XtX$.  u.jX1X5X5X5X5X5X5X5X5X5X5X5X5X5X4X0X].*X8X5X5X5X5X5X5X5X5X5X5X5X5X5X5X6X<X).9X4X5X5X5X5X5X5X5X5X5X5X5X2XeXXXi 7 -XyX0XqXqXqXqXqXqXqX",
"qX7XxX#.< T. X).'.`.'.'.'.'.'.'.`.'.`._.#X:.  @.oX_.`.'.'.'.'.'.'.'.'.'.'.'.`.[.Y.^.].`.'.'.'.'.'.'.'.'.'.'.'.'.`.'.).I.].`.'.'.'.'.'.'.'.'.'.'.'.'./.&X6.  (.sX9XqXqXqXqXqXqXqX",
"qX9XsXD.# g <XsXwXtXtXtXtXtXtXtXtXtXtXrXrXgX/   r.jXeXrXtXtXtXtXtXtXtXtXtXtXrXaX X1XiXrXtXtXtXtXtXtXtXtXtXtXtXtXtXyX9X].pXrXtXtXtXtXtXtXtXtXtXtXtXwXkXS.. J 4XeX0XqXqXqXqXqXqXqX",
"qX0XrX+X~., u ;XuX7X0X0X0X0X0X0X0X0X0X0XwX<XoX,.  0.dX8X9X0X0X0X0X0X0X0X0X0X9XtX}.:XeX0X0X0X0X0X0X0X0X0X0X0X0X0X0XwX4X`.rX9X0X0X0X0X0X0X0X0X0X0X7XdXC.  Q .X:XrXqXqXqXqXqXqXqXqX",
"qX0XyX`.0X.X: p ;XiX8XqXqXqXqXqXqXqXqXqXeX2X'.hX*.  q.fX9X0XqXqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqXqXqXqXqXqXqXqXqX8XfXZ.  U pXoX=XtX0XqXqXqXqXqXqXqX",
"qX0XyX{.:XfX}.> i ;XiX8XqXqXqXqXqXqXqXqXeX3X'.tXaX=.  q.fX9X0XqXqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqXqXqXqXqXqXqXqX8XgXZ.  T wXpX}.-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XqXaX}.: i ;XiX8XqXqXqXqXqXqXqXeX3X].tX9XsX=.  q.dX9X0XqXqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqXqXqXqXqXqXqX8XfXZ.  T eXwXtX.X-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XrX7XsX}.: p ;XiX8XqXqXqXqXqXqXeX3X].yX9X0XsX=.  q.dX9X0XqXqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqXqXqXqXqXqX8XfXZ.  T eXeX8XuX X-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqX8XsX}.> p ;XiX8XqXqXqXqXqXeX3X].yX0X0X0XsX=.  q.fX9X0XqXqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqXqXqXqXqX8XfXZ.  T eXeX9X0XyX X-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqX8XsX}.: i ;XiX8XqXqXqXqXeX3X].yX0XqX0X0XsX=.  q.fX9X0XqXqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqXqXqXqX8XgXZ.  T eXeX9XqX0XyX X-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.<XrXqXqXqX8XsX}.: i ;XiX8XqXqXqXeX3X].yX0XqXqX0X0XsX=.  q.dX9X0XqXqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqXqXqX8XfXZ.  T eXeX9XqXqX0XyX X-XtX0XqXqXqXqXqXqXqX",
"qX0XyX[.>XeXqXqXqXqX8XsX}.> p ;XiX8XqXqXwX1X`.rX9XqXqXqX0X0XsX=.  q.dX9X0XqX0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqXqX8XfXZ.  T eXeX9XqXqXqX9XtX}.*XeX0XqXqXqXqXqXqXqX",
"qX0XtXXXrXpX9X0XqXqXqX8XsX}.> i ;XiX8X9XiXtX#XjXqX9XqXqXqX0X0XsX=.  q.fX9X0X0XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0XqX8XfXZ.  T eXeX9XqXqXqX9X0XjX-X7XdX9X0XqXqXqXqXqXqX",
"qX9XaXF.m.$XsXyX9XqXqXqX8XsX}.: i -XfXaX$XS.u./.6XfX0XqXqXqX0X0XsX=.  q.fX9X9XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX0X8XgXZ.  T eXeX9XqXqXqX9XsXqX].d.m.oXpXuX9XqXqXqXqXqX",
"qX9XiXW.5.&.;.&XpX9XqXqXqX8XsX|.; d (.3.#.q.y.6...n.iXwX0XqXqX0X0XsX=.  q.dX8XyX}.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.tX7XfXZ.  T eXeX9XqXqXqX0XsXK...<.t.e.*.%..XsX9XqXqXqXqX",
"qXqX9XdXbXaXL.) }.iX9XqXqXqX8XaX X1 u Y.eXnXxXsX,X<.5.uX0XqXqXqX0X0XsX=.  q.sXtX{.>XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X`.wXdXZ.  T eXeX9XqXqXqX9XsXf.o.&XlXvXvXfX~.) Q.aX9XqXqXqX",
"qX7XzX-.9 /.nX{.' 4XrX0XqXqX0XqXwX] [.kXqX7.$ [ jXhX3.j.fX9XqXqXqX0X0XsX=.  0.lX{.:XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X).kXC.  T eXeX9XqXqXqX9XsXK.o.sXuX&.: u G.vX*X_ *XyX0XqXqX",
"qX0XtXXXx 2 =XcXg.f.dX8X0X0X8XdXz.p.jXyXM.e 1 D gX9X9X} >XwX9X0X0X0X9X9XaX*.  y.$X;XeX9X0X0X0X0X0X0X0X0X0X0X0X0X0XqX2X+XD.  U wXwX8X0X0X0X9X0XqXX.@XxX5.  f.E +  XcXC.9.fX9XqXqX",
"qXqX8XlX}.  n.xXoX=.iXyXyXyXyXdX:..XyXeX.X,XS S jX6XdXi.G.zXrXyXyXyXuXtXyXjX<.  -.rXuXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXiX0.  ! pXpXtXuXyXyXyXtXjX'.2.uXpXi P VX-Xo q.xX;X%.4XeXqXqX",
"qXqX0XeX#XO u.kX9X{ ].XX X X.X X] 8XqX0XyXaXv D jX6XgXC.2.:X{. X X X X X|. X$X{   Y.@X|. X X X X X X X X X X X X X.X|.X A XXoX}..X X X X X}.-Xd.t.kX6X5 o.fX>X< %.lXtX+.=XtXqXqX",
"qXqX8XgXoX  z.jX5X..%X;X-X-X-X-X .4XwXqX0XtXn D jX6XfXN.e.0X&X-X-X-X-X-X-X=X-X6X].&X;X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X'.3X>X=X-X-X-X-X-X-X*X7Xm.e.gXtXe ! MX>X$ 5.kXwX#.-XtXqXqX",
"qX0XwX2XQ @ +XfX'.,.aXrXtXtXrXfX2./.aX9XqXyXb S kX5XdX7.E.hXeXtXtXtXtXtXtXtXrXeXaXyXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXaXeXrXtXtXtXtXtXtXtXeXdXoX-.wXgX@.  W.=.  ).fX#X*.0XwXqXqX",
"qX7XxX&.o M.yXjX5.Z.dX8X0X0X9XsXJ.,.kX8X9XyXM G dXyX&X| 0XqX0X0X0X0X0X0X0X0X0X0X9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9X0XqX0X0X0X0X0X0X0X0X9XpX%.~.jX:XU o ; u.rXlXp.h.gX9XqXqX",
"qXqXeX<X1XhXlXM.$.yXqXqXqXqXqX0XiX=.z.zXiX0X-X<XhX9X{ Y.sX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuX.X( =XxXiX*X<XgXzXI.} 0XwXqXqXqX",
"qXqX9XpXiXoX0.+.5XrX0XqXqXqXqX0XeX8X$.5.XXeXgX8XI.[ C.dX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XsXQ.) C.,XgXpX#Xp.} ;XyX0XqXqXqX",
"qX9XsXZ.[ -.b.tXrX0XqXqXqXqXqXqX0XeXyXN.*.#.O.$.2.|.dX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XaX*X9.%.O.+.&.h.0XyX9XqXqXqXqX",
"qXqXeX1X=XwXgXqX0XqXqXqXqXqXqXqXqX0X0XfXeX;X*X1XsXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XyXfX4X=X-X0XgXwX0XqXqXqXqXqX",
"qXqXqXrXtXqX9XqXqXqXqXqXqXqXqXqXqXqXqX9XqXtXyXrX9X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9XeXtXtXwX9XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_3[] = {
/* columns rows colors chars-per-pixel */
"64 64 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XqXqXqX9X8X0XqX0X0X0X0X0X0X0X0X0X9X0X0X0X0X0X0X0X0X9XqXqXqX9X9XqXqX0XqXqXqXqX",
"qXqX0XtXuXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXpXwX0X9XaXhXeX0XyXuXrXtXtXtXtXtXtXdXtXtXtXtXtXtXtXtXpXqXqX8XsXsX0X0XyXeXqXqXqX",
"qX0XtX@X|.*X$X$X$X$X$X$X$X$X$X$X$X$X$X$X$X$X$X$X&X).5XrXgXI.n.4XuXoX|.*X$X$X$X$X%X$XF.+X%X$X$X$X$X$X&X).8X0XhXL.P.yXrXXX3XeXqXqX",
"qX0XuX|.OX4X<X1X1X1X1X1X1X1X<X4X|.&X3X<X1X1X1X>XwXi.XX3Xv.;Xy 6.NXp._.7X<X1X1X1X<X3XU.*X2X<X1X1X1X,X9Xe.@XgXm.h XXyXiXp.&XyX0XqX",
"qX0XrX*X4XyXeXrXrXrXrXrXrXrXeXuX%X4XyXeXrXrXrXwXdX~.G.gXS v v [.uXa.wXrXeXrXrXrXeXuX=X7XtXeXrXrXrXwXdXI.E.7X  u H XX3Xy.yX0XqXqX",
"qX0XtX$X<XeX0XqXqXqXqXqXqXqX0XrX#X2XeX0XqXqXqX9XwXiXJ L.hX@XwXsX) h.vX6XqXqXqXqX0XrX+X4XwX0XqXqXqX9XrX0XF I.>X&XXX-X9.|.iX0XqXqX",
"qX0XtX$X1XrXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqXqX0X0XhX` v j.a.H.n.a.5.7 W.hX7XqXqXqX0XrX@X4XeXqXqXqX9XwXsXY C z.l.G.Z.e.[.pX9XqXqXqX",
"qX0XtX$X1XrXqXqXqXqXqXqXqXqX0XtX#X2XeXqXqX0X0XhX` g 8XgX-XS./.8XbXv.> !.hX7XqXqX0XrX@X4XeXqXqX9XwXdXU m tXsX%XZ._.wXiX9XqXqXqXqX",
"qX0XtX$X1XrXqXqXqXqXqXqXqXqX0XtX#X2XeXqX0X0XhX` g 8XyX6XpX:XqXrX6XlXn.> !.hX7XqX0XrX@X4XeXqX9XwXdXI M tXrX7XpX-XeXeX9XqXqXqXqXqX",
"qX0XtX$X1XrXqXqXqXqXqXqXqXqX0XeX+X<XwX9X0XhX` g 8XyX8X0XwXoX<XqXqX7XlXn.> !.hX7X0XrX@X4XeX9XwXdXI M tXrX9X0XqX.X2XqXqXqXqXqXqXqX",
"qX0XtX$X1XrXqXqXqXqXqXqXqX0XwXhX,XeXaXqXhX` g 8XyX9X0XwXhX>XrXsX9XqX7XlXn.> !.hX6XtX+X5XqXwXdXI M tXrX9X0XeXhX;XuXpX9XqXqXqXqXqX",
"qX0XtX$X1XrXqXqXqXqXqXqX9XuX6XJ.u.j._.sX_ g 8XyX9X0XiX4XH.y.k.].pXqXqX7XlXn.> !.gX0X@X3XeXsXI M tXrX9X9XpX1XD.t.x.}.aXqXqXqXqXqX",
"qX0XtX$X1XrXqXqXqXqXqX0XuX@X7.~.%X+XZ.H b 7XyX8X0XiXoX8.^.%X#Xj.j.tXqXqX7XlXn.> Q.zX X6XfXI M tXrX9X0XpX[.7.(.*X+Xd.b.uX0XqXqXqX",
"qX0XtX$X<XeX0X0X0X0X0XwX6X6.<Xf.b o.aXB.x.hX6X0XwX3X6.1Xe.z ,.kXV.N.sX8X0X6XkXn.; ^.3XqXU m rXeX8X0XrX:X5.rXm.z $.gXv.G.sX9XqXqX",
"qX0XtX$X4XuXrXtXtXtXeXhXL.L.pX+.`.a B.nXy.<XiXwXjXF.E.6XY oX0 P.lXr.5XuXrXtXqXvXN.4 V.Q B pXiXwXtXeXjXN.].-X1 ).c r.kXw.7XwXqXqX",
"qX0XtX$X|.%X#X#X#X#X+X,Xu.OXiXVXo.s %XjXb.K.:XOX>Xt.$XeXVXG.z ,XhXx.U.;X@X#X#XOX4Xd.  z >X$X+X$X#X@X;X0.qXZ.v HX:./ CXf.#XyX0XqX",
"qX0XtX$X&X4X2X2X2X2X<XeXk.}.6Xs.<Xb >.BXl._.9X,XwXh.}.hXB.a +XpXuXj.[.8X1X2X2X2X:XfXz.3 U.tX>X2X2X<X0Xp.1X`.6 wXL =.MXa.=XtXqXqX",
"qX0XtX$X3XyXeXeXeXeXwXaX_.V.kXU j c [.wXq.wXrXwXsX^.F.uX= w V /.uXa.wXeXeXeXeXeXeXwX2Xj.1 ^.kX9XeXwXfXU.N.mX..t V 4X:Xt.yX0XqXqX",
"qX0XtX$X<XeX0XqXqXqX9XqXsXH B.kX-XtXeX0.).pX9X9XwXiXF B.3X,X#X3X^ s.vX6XqXqXqXqX0XrXXXsXM.: Q.hX7XqXqXrXr.R.jX,XaX3X6.oXuX0XqXqX",
"qX0XtX$X1XrXqXqXqX0X0XhX_ b x.i.Z.v.0.).aX9X0XqXfX! N v.g.C.m.f.w.6 P.jX7XqXqXqX0XrX@X1XxXn.> !.hX7XqXeX9Xj.p.A.x.q..XpX9XqXqXqX",
"qX0XtX$X1XrXqXqX0X0XhX` g 7XhX<XJ.).wXpX9X0XqXgXQ x qXgX>XH._.0XbXC.: U.jX7XqXqX0XrX@X5X9XlXn.> !.hX7X0XwXaX:XF.{.tXiX9XqXqXqXqX",
"qX0XtX$X1XrXqX0X0XhX` g 8XyX6XiX:X9XrX9X0XqXgXQ x wXtX6XiX;X0XeX6XlXA.: U.jX7XqX0XrX@X5XeX7XlXn.> !.hX7XqX8XiX-XwXwX0XqXqXqXqXqX",
"qX0XtX$X>XwX9X0XhX` g 8XyX8X0XwXOX,XwX9XqXgXQ x wXtX9XqXrX+X2XeXqX7XkXA.: U.jX7X0XwXOX2XqXqX7XlXn.> !.hX7X0XqXXX3XqXqXqXqXqXqXqX",
"qX0XrX&XqXdXwXhX` g 8XyX9X0XwXjX>XwXsXwXfXQ x wXtX9XqX0XtX@X3XeXqXqX7XkXA.: U.jXqXhX;XtXsX0XqX7XlXn.> !.hXqXhX-XyXaX9XqXqXqXqXqX",
"qX9XiX).y.~.pX` g 8XyX9X0XiX3XF.r.h./.iXQ x wXtX9XqXqX0XtX@X3XeXqXqXqX7XkXA.; Y.6XC.r.j._.iXqXqX7XlXm.: !.4XV.w.z.'.pXqXqXqXqXqX",
"qXqXwX5X$XH.K v 8XyX8X0XiXoX6.W.<X.XA.F N qXtX9XqXqXqX0XtX@X3XeXqXqXqXqX7XkXA.4 6.].%X%Xk.j.tXqXqX7XkXM.5 7._.<X Xf.z.uX0XqXqXqX",
"qX9XfXA.^ hXJ.j.gX6X0XwX4X6.3X8XG ~ NXZ.c.gX6XqX0X0X0X9XrX@X3XwX0X0X0X0X0XqX4X1.5X,.j :.lXV.V.sX8X0XwX1X<.rX-Xv #.VXv.S.sX9XqXqX",
"qXqXqXuXt v.NXi.-XpXwXhXK.K.cXQ.^ v tXaXi.,XiXrXtXtXtXrXiX%X6XuXrXtXtXtXeXjXB.`.<X^ %X7 Q.hXr.7XuXwXhXM.(.lXP.K J hX9Xy.7XwXqXqX",
"qX9XaXW.l *XjXB.D.:XoX>Xy.+XeXkXT.b yXsXm.J.;X+X@X@X@X@X%X_..X$X@X@X@X@X+X;Xe.=XeXAXc.M 5XfXk.R.=XOX-Xw.>X0XvXx.J fXiXk.#XyX0XqX",
"qX9XdXF.R.dXiXv.(.qX<XrXz.{.yXsXC.z yXsXx.'.0X2X3X3X3X3X6X.X;X5X3X3X3X3X1XeXg.OXgXf.s *XyXuXf..X8X1XwXd.%XwXgXy.F fXyXf.-XtXqXqX",
"qX9XsXU.i U.fXp.8XrXqXsX].g.jXsXD.B hX2Xw.eXeXeXeXeXeXwXuX$X5XtXeXeXeXeXqXdXQ.M.qX* i N [.qXd.tXwXqXfXT.m.kXfXd.U MX@Xu.uX0XqXqX",
"qXqXwX7X-X4X' 3.bX6XqX9XpXj.V.uXiXqX2X0.'.pX9XqXqXqXqX0XrX@X3XeXqXqXqXqXqX0XyXa.S.6X4X-X<XI j.vX6XqX0XtXu.L.pXuXwX-X6.@XuX0XqXqX",
"qX0XyX+Xk.f.h.7 x.lX7XqXqXtXN.y.v.x.e.].pX9XqXqXqXqXqX0XtX@X3XeXqXqXqXqXqX0XwXwXb.s.M.x.h.y.5 P.jX7XqXeXqXx.i.b.l.r.OXiX9XqXqXqX",
"qX0XuX}.).wXvXQ.> v.lX7XqXqXsX5XU.].tXiX9XqXqXqXqXqXqX0XtX@X3XeXqXqXqXqXqXqX0XwXaX1XL.}.eXbXC.: U.jX7X0XwXaX1XL..XuXuX9XqXqXqXqX",
"qX0XrX*X7XrX6XhX^.> v.lX7XqX8XuX;X8XeX9XqXqXqXqXqXqXqX0XtX@X3XeXqXqXqXqXqXqXqXqX8XuX-X0XwX6XlXA.: U.jX7XqX8XuX=XqXwX0XqXqXqXqXqX",
"qX0XtX$X<XeXqX7XhX^.> v.lX7X0XrX@X1XeXqXqXqXqXqXqXqXqX0XtX@X3XeXqXqXqXqXqXqXqXqX0XrX+X4XwXqX7XkXA.: U.jX7XqXeXOX5XwXqXqXqXqXqXqX",
"qX0XtX$X1XrXqXqX7XhX^.> v.lX6XtX#X2XeXqXqXqXqXqXqXqXqX0XtX@X3XeXqXqXqXqXqXqXqXqX0XrX@X4XeXqXqX7XkXA.: U.jX6XrX+X6XwXqXqXqXqXqXqX",
"qX0XtX$X1XrXqXqXqX7XhX^.> v.kXqX#X2XeXqXqXqXqXqXqXqXqX0XtX@X3XeXqXqXqXqXqXqXqXqX0XrX@X4XeXqXqXqX7XkXA.: U.hX0X+X6XwXqXqXqXqXqXqX",
"qX0XtX$X1XrXqXqXqXqX7XhX^.> v.vXOX2XeXqXqXqXqXqXqXqXqX0XtX@X3XeXqXqXqXqXqXqXqXqX0XrX@X4XeXqXqXqXqX7XkXA.: U.zX.X6XwXqXqXqXqXqXqX",
"qX0XtX#X>XeX0X0X0X0XqX6XgX~.; M.5X:XeX0X0X0X0X0X0X0X0X9XrX@X3XwX0X0X0X0X0X0X0X0X9XrX+X4XwX0X0X0X0X0X6XkXZ.- T.<X3XwXqXqXqXqXqXqX",
"qX0XrX;XrXtXrXrXrXrXrXtX0XlX(.3 t.gXwXrXrXrXrXrXrXrXrXrXiX%X6XyXrXrXrXrXrXrXrXrXrXiX$X8XyXrXrXrXrXrXrX0XcXS.4 m.rX0XqXqXqXqXqXqX",
"qX9XdXJ.u.3XoX+X@X@X@X@X@XXX<XZ.1 j.2XXX@X@X@X@X@X@X@X+X$X)..X#X+X@X@X@X@X@X@X@X+X$X)..X#X+X@X@X@X@X@X@X X8Xr.3 eXqXqXqXqXqXqXqX",
"qX9XsXT.@ a.dX1X4X5X4X4X4X5X<XdXU.+ k.sX1X4X5X4X4X4X4X4X8XXX>X6X4X4X4X4X4X4X4X4X4X8X.X,X6X4X4X4X4X4X5X2X8X4XB Q uX0XqXqXqXqXqXqX",
"qX0XwX2XW.4 j.xX9XeXeXeXeXeXwXwX1XU.2 n.xX9XeXeXeXeXeXwXyX#X5XtXwXeXeXeXeXeXeXeXwXyX#X6XrXwXeXeXeXeX0XyXtXn W -X6XwXqXqXqXqXqXqX",
"qX0XtX+XeX[., h.lX7XqXqXqXqX0XtXOXyX(.: v.lX7XqXqXqXqX0XrX@X3XeXqXqXqXqXqXqXqXqX0XrX+X4XwXqXqXqXqX9XrXrXb T hX@X5XwXqXqXqXqXqXqX",
"qX0XtX%X:XjX`., h.lX7XqXqXqX0XtX$X>XkX~.> v.lX7XqXqXqX0XtX@X3XeXqXqXqXqXqXqXqXqX0XrX@X4XeXqXqXqX9XtXrXb T dXtXoX6XwXqXqXqXqXqXqX",
"qX0XtX$X1X0XfX'.< h.lX7XqXqX0XtX#X2X9XhX^.> v.lX7XqXqX0XtX@X3XeXqXqXqXqXqXqXqXqX0XrX@X4XeXqXqX9XtXrXb R dXqXeX+X6XwXqXqXqXqXqXqX",
"qX0XtX$X1XrX7XgX'., h.lX7XqX0XtX#X2XrX7XhX^.> v.lX7XqX0XtX@X3XeXqXqXqXqXqXqXqXqX0XrX@X4XeXqX9XtXrXb T dXwX9XrX+X6XwXqXqXqXqXqXqX",
"qX0XtX$X,XwXqX7XgX'.< h.lX7X9XeX@X<XqXqX7XhX^.> v.lX7X0XtX@X3XeXqXqXqXqXqXqXqXqX0XrX@X4XeX9XtXrXb R dXqX0X0XwXoX4XqXqXqXqXqXqXqX",
"qX0XtX%X9XgXqX0X7XgX'., h.lXwXjX-X0XfXqXqX7XhX^.> v.lX6XtX@X3XeXqXqXqXqXqXqXqXqX0XrX+X5X0XrXrXb T dXwX0X9XtXhX&XtXsX0XqXqXqXqXqX",
"qX9XiX_.r.P.eXrX0X7XgX'., j.4XM.w.d.U.rXeX0X7XhX^.> v.kXqX@X3XeXqXqXqXqXqXqXqXqX0XrX@X2XyXrXb R dXqX0X9XaX$Xb.r.f.W.uXwX0XqXqXqX",
"qXqXwX3X-XG.w.7XeXqX7XgX'.7 -.{.2X+XV.t.9XwXqX7XhX^.> v.vXoX3XeXqXqXqXqXqXqXqXqX0XtXoX8XtXb T dXwX0X9XaXW.0.XX%X-Xn.d.rXqXqXqXqX",
"qX9XfXF.v 8X).d.pX9X0XqX7X7.0X1Xv P MXY.j.aX8XqX6XgX~.; m.5X<XwX0X0X0X0X0X0X0X0X0XwX&X7Xn T dXqX9X0XtX$X8.dXe.a ^ sXD.N.dX9XqXqX",
"qXqX9XfX( ~ VXp.*XpXqXhXF.U.zX(...n rXpXa.:XiXeXrX0XlX(.4 0.yXrXrXrXrXrXrXrXrXrXwXpX.Xk ! hXtXeXrXwXhXb..X+X2 ;XZ 3.cXe.5XeXqXqX",
"qXqX8XlXn.M NXN.S.-X.X;Xt.+XeXlXY.b yXsXm.G.-XOX+X+X.X<XM.l |.#XOX+X+X+X+X+X+X+XXX1X,.G :X@XoX+X+XoX*X9.wXZ.b HX>./ CXf.#XuX0XqX",
"qXqX9XdX/ ^ VXz.`.eX2XuXv.].yXsXC.l tXsXz.{.wX5X6X6X6X3XrXeX5X6X6X6X6X6X6X6X6X6X6X6X7XiX6X5X6X6X6X4XtXd.:X X4 *XZ 2.mXp.;XtXqXqX",
"qX9XfXS.j ,X7X0.eXeXqXpX|.a.gXdXH.H kX:Xe.tXwXwXwXwXwXeX0XqXeXwXwXwXwXwXwXwXwXwXwXwXwX0XwXwXwXwXwXqXsXQ.z.MX4.t T yX+Xi.iX0XqXqX",
"qXqXqX0XfX1X7.[.iX9XqX9XaXb.c.qXpXyX=X7. XiX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuXd.Z.fXrXgX#X4.*XyX0XqXqX",
"qX0XyXoXi.t. XpX9XqXqXqXqXuXG.t.k.g.y.oXpX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXrXB.t.z.s.i.=XuX9XqXqXqX",
"qXqXeX3X%XyXiX9XqXqXqXqXqX0XsX8X#X*XyXuX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXdX5X@X;XiXyX9XqXqXqXqX",
"qXqXqXeXyX0X0XqXqXqXqXqXqXqX9XwXyXtX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XeXuXtX0X0XqXqXqXqXqX",
"qXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_4[] = {
/* columns rows colors chars-per-pixel */
"48 48 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX0XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXqX9X0XqX0XqX0X0X0X0X0X0X0X0X0XqX0XqXqX0X0XqX0XqXqXqXqX",
"qXqXqXeXtXeXrXrXrXrXrXrXrXrXrXrXrXrXrXtXqX9XpXtXqXyXrXrXrXrXrXuXrXrXrXrXeXtXeX9XtXuX9XrXwXqXqXqX",
"qXqXeX<X&X,X:X:X:X:X:X>X:X:X>X:X:X>X:X$XwXsX'.&XwX@X>X:X:X:X:X{.:X>X:X:X,X%X,XfX,XOXiX>X5XwXqXqX",
"qX0XtX&X%X5X2X2X2X2X2X4X;X#X5X2X2X2X4XM.8Xz./ %.rXN.0X1X2X2X4X`.,X3X2X1X9X~.).,Xa s.kXS.=XtXqXqX",
"qXqXeX,X5XtXeXeXeXeXeXtX6X>XyXwXeX0XhX(.m.L.q.=Xs.*XpXqXeXeXtX>X0XrXeXeXqXaXh.V.>.x._.S.yX0XqXqX",
"qX0XrX:X2XeX0XqXqXqX0XwX3X;XeX0X9XrXrX~ ,.~.^.U.| $.hX0X0X0XeX*X8XqXqX6XlXS.H G.).R.S.7XeXqXqXqX",
"qX0XrX:X2XeXqXqXqXqXqXeX3X;XrX8XrXyXY u.vX:X~.6XlX@.O.jX0X0XeX*X8XwX7XlXA.Z 2XaX`.XXpXeX0XqXqXqX",
"qX0XrX:X2XeXqXqXqXqXqXqX6X<XwXeXyXY i.xX6XeX2XeX8XkX#.O.jX0XwX*X8X7XlXA.Z 3XuX0X5X8XqX0XqXqXqXqX",
"qX0XrX:X2XeXqXqXqXqXqXaX&X{.pXtXY i.xX6XpX3X`.0XyX8XkX#.O.jXwX*X4XzXA.Z 3XuXqXiX X#XaXwX0XqXqXqX",
"qX0XrX:X2XeX0XqXqXqXqXP.K.P.K.E p.zX6XyX/.J.Y.A.|.yX8XkX#.o.lX$XhXS.Z 2XuXqX1XS.Y.P.G.9XwXqXqXqX",
"qX0XrX>X4XtXwXeXqXpXL.Y.6.2.[.0.uXqXpX'.H.t.X.{.Z.&XuXqXzX$.+.iXN.D 4XaX0XeXB.].@.w.[.S.uX0XqXqX",
"qX0XrX:X;X6X3X3X6X=XA.eXJ.C +XXXQ.0X6XV.,X].%.p.9XZ.wX1X2XiX#.C ! &X0X<X0X(.}.y.p.+.R.E.=XtXqXqX",
"qX0XrX:X#X>X;X;X<X|.G.9XP.G L.@XK.4X-Xb.sXK.! 6X1XM.3X-X-X=XyX' / yX*X=X4XU.@X0.d.+.U./.#XyX0XqX",
"qX0XrX>X5XyXeXrXwXpXK.}.1.*.$XZ.,XtXaX/._. .! _.E.oXpXeXrXeXrX0X#.@.zXwXtX4XB.:XX.9.1XM.wXqXqXqX",
"qX0XrX:X2XwX0X8XeXtXE 0..X XZ.`.uX7XlX,.' {.{.T.U l.zX6XqX0XwX&XfX#.o.jX8XyX}.B.XX}.n.*XyX0XqXqX",
"qX0XrX:X2XeX9XrXyXY p.iXQ.K.,XuX7XzX>.[ yX}.F.&X7XK l.zX7XqXeX&X7XlX#.O.jX9XyX2XI.E.8XyX0XqXqXqX",
"qX0XrX>X2X0XrXyXY i.zXqX0X4XtX7XzX>.] fXeXtX1XyXyXqXJ z.xX6XqX*X6X9XkX#.O.jX8XrX5X9XeX0XqXqXqXqX",
"qX0XrX;X2XhXrXY i.xX6XaX5X&XsXlX>.] fXqX0X9X%XwX8XyXqXJ l.cXuX$XwXuX8XkX#.O.hXdX-X1XsX0XqXqXqXqX",
"qX0XrX-XF.(.~ u.xX6XuX'.G.J./.,.[ fXqX0XwX9X*XwXqX9XyX0XI a.).L.G.OXyX8XkX#.#./.H.H.(.yX0XqXqXqX",
"qX0XuX XL.M.,.cX6XpX~.L.W.p.).] yXeX0XwXwX0X*XeXqXwX9XaX=X| G.<.).C.XXyX0XfX| !.V.m.R.T.iX9XqXqX",
"qX0XyX%XN yXP.>XeX5XM.gXv.x gX(.|.tX9X9X0X8X&XqX9X9X8XyXQ.&XJ.<.] uXM.eXeXOXQ.zX+./ mXM.2XeXqXqX",
"qX0XaX/.a.kXD.(.2X'.G.aX7X~ 0X}.G.1X%X*X*X&X}.=X*X*X%X<XN.*XsX1.K.fXb.+X1XP.^.lX_.{ cXP.$XyX0XqX",
"qX9XpX).4.eXL.7XeXqXM.eX8X] pXG.#XuXqXwXeXqX=XeXwXwXqXaX/.}.w.q G.4XP.tXtX-XV.vX/.=.kXn.6XwXqXqX",
"qXqXeX2X^.y. .kX8XyX Xm.&X$XK.L.uX0XqXqXqX9X*XwX0XqXqXqXwXC.H.[.(.+.u.vX5XiXT.G.-X+XB.'.iX0XqXqX",
"qX0XtX%XG.-X%.+.kX8XyX*XF.C.[.tXqXqXqXqXqX9X*XwXqXqXqXqXeXwX_.Z.W.#XU l.xX6XyX|.C.F.#XuX0XqXqXqX",
"qX0XrX>X9XpXhX+.@.kX8XuXwX2XsX0XqXqXqXqXqX9X*XwXqXqXqXqXqXqXsX,XrXdX0XJ z.xX5XsX4X9XpX9XqXqXqXqX",
"qX0XrX:X1XqX0XjX+.@.jXqX1X-XwX0XqXqXqXqXqX9X*XwXqXqXqXqXqX0XqX&X7X9XyXqXJ z.zX9X;X<XwXqXqXqXqXqX",
"qX0XrX:X2XeX0X0XjX+.+.lX3X-XrX0XqXqXqXqXqX9X*XwXqXqXqXqXqXqXeX*X8XwX9XyXqXJ l.vX=X2XeXqXqXqXqXqX",
"qX0XrX:X<XeX0X0X9XjXO.#.pX=XwX0XqXqXqXqXqX9X&XwX0XqXqXqXqX0XwX*X8XqXqX8XtX0XH c.tX,XeXqXqXqXqXqX",
"qXqXeX2XrXeXeXeXwXwXlX%.| rXrXwXeXeXeXeXeXqX=XrXeXeXeXeXeXwXtX-X0XrXeXeX0XiXwXI f.wXqXqXqXqXqXqX",
"qX9XpX/.h.9X$X*X*X&X&X7X{ X.8X&X&X*X*X*X=X&X}.=X*X*X*X*X*X*X-X|.%X=X*X*X=X$X>X;Xa #XuX0XqXqXqXqX",
"qX0XyX%XB z.jX4X8X8X7X8XsX_ +.fX7X7X8X8X8X6X%X9X8X8X8X8X8X8X0X%X5X9X8X8X7X6XgX1.[ 8XwX0XqXqXqXqX",
"qXqXwX3X1XL c.xX8XwXwXeX2XqX%.@.kXqXqXwXwX0X*XeXqXwXwXwXwXqXrX=X9XeXwXqX9XzX1.[ 7X3XwXqXqXqXqXqX",
"qX0XrX-X7XwXG c.zX7XqXeX3X-XzXO.@.kX0X0XqX9X*XwXqXqXqXqXqX0XeX*X8XwX0X8XzX<._ hX>X1XeXqXqXqXqXqX",
"qX0XrX:X,XiX0XH c.zX7XeX3X=XwXjX+.@.kX0XqX0X*XwXqXqXqXqXqXqXeX*X8XqX8XzX<.` dXrX-X1XeXqXqXqXqXqX",
"qXqXeX,X9XqXtX0XH c.zXqX0X3XtX8XkX+.@.kX0X9X*XwXqXqXqXqXqXqXeX*X8X9XzX<.` fX0XrX5X8XrX0XqXqXqXqX",
"qX0XtX&X).wXwXtX0XG c.fX_.W.5XyX9XkX+.@.kX8X&XwXqXqXqXqXqXqXeX*X5XxX<.` fX0XtX6X~._.qXtX0XqXqXqX",
"qX0XrX-XU.B.>XwXtX9XU 0./.Y.m.|.yX8XjXO.@.hX&XqX0X0X0X0X0X0XwX&XhX1._ dX0XtXoXC.I.Y.B.:XtX0XqXqX",
"qX0XiX}...@XV.yX0XdXZ.;Xv.K tXm.3XyXwXzX$.@.3XrXeXrXrXrXrXeXrX5X,.{ hXrXrX6XZ.'.^ #.+XB.eXqXqXqX",
"qXqX7XvX..L.~.[.5X.XS.fX}.Y rX[.U.6X-X:XeX` 6.0X-X>X>X>X>X-X0Xt.E 9X>X-X5XT.@X1.H.2.F.(.$XyX0XqX",
"qX0XwX7XJ |.Y.oX8X%XZ.fX2XG tX_.Q.9X1X1X2X9X%X3X2X2X2X2X2X2X3X%X8X3X1X1X8X^.}.c.@.( ].R.*XtXqXqX",
"qX0XuX|.S.@XC.pXqXaXF.[.uXS.,Xn.8XtXeXeXeXwXyXeXeXeXeXeXeXeXeXyXwXwXeXeXrXwXm.-Xp.A.#XV.yX0XqXqX",
"qXqXeX4XT.Z.7XeX0XwX9XS.F.~.M.*XyX0XqXqXqX0X0XqXqXqXqXqXqXqXqX0X0XqXqXqX0XtX:XN.).W.C.7XeXqXqXqX",
"qXqXwX4X&XuXeX0XqX0XwXuX;X#XwXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXeX#X&XyXeX0XqXqXqX",
"qXqXqXwXtX0XqXqXqXqXqX0XtXyXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXyXtX0XqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_5[] = {
/* columns rows colors chars-per-pixel */
"44 44 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqX0XqXqX0X0XqX0XqXqXqXqX0X0XqXqXqXqX0XqXqX0XqXqXqXqXqXqX",
"qXqXqXwXrXeXeXeXeXeXeXeXeXeXeXeXeXrXeX9XrXtX0XrXeXeXeXeXtXrXeXeXeXeXrXqX9XyXqXwXwXqXqXqX",
"qXqXwX7X>X2X2X2X2X2X2X2X2X2X2X2X2X,X4XaX1X:XwX>X2X2X1X3X=X;X3X1X2X2X,X9XsX;XwX7X7XwXqXqX",
"qX0XrX>X+X3X<X<X<X<X1X>X+X1X<X,X5X`.[.oX=.{ qXG.6X,X,X4X.XoX3X,X,X6XH.9Xa.U yX(..XuX0XqX",
"qXqXeX2X3XyXeXeXeXeXrXqX:XtXeXeXqXdXd.I.y.`.g.+XdXqXeXtX6X3XtXeXqXaX=Xa.a.0.Q.G.8XwXqXqX",
"qXqXeX2X<XeX0XqXqXqXqX8X-XwX0X7XlXm.( E.Q.!.0.| uXeX9XeX1X<XeX9XqXsX@.3.'.Q.L.,XtX0XqXqX",
"qXqXeX2X<XeXqXqXqXqXqX7X*XwX7XzXm.! eXpX X2XvXf.' pXeXwX2X<XwXqXfX..w.cX4X XtXrX0XqXqXqX",
"qXqXeX2X<XeXqXqXqX0XeXtX,XuXzXm.! rXrXyX2XrX9XlXg.` pXtX<X,XrXdX..e.lX9XtX1XiX9XqXqXqXqX",
"qXqXeX2X<XeXqXqX0XeX8X].T.$Xn.~ eXtXeX+XW./.7X0XlXg.` sX3X,XhX..e.lXqX8X_.E. XtXqXqXqXqX",
"qXqXeX2X1XrXqXqXwX9XF.v.l.P._ 9XtXtX~.m.y.E.S.6X9XzXg.[ 0XwXo.e.zX9X8XG.G.a.U.T.yX0XqXqX",
"qXqXeX2X>XqX8X7XyX!.|.U.K (.|.#XtX&XT.`.*.5.;XY.yX4XkXd.` %.e.hX5XyX~.[.2.@.v./.#XtX0XqX",
"qXqXeX2X+X:X-X*X2XA.:X-X( B.@XU.8XQ./.pX-.U.wXV.<X=X%XeXt.f 0X;X%X1XG.@X .G.;.&X'.iX0XqX",
"qXqXeX2X1XtXwXwXuX@X!.x.X.(.P.<XtX8XW.n.N E.)..XiXqXwXqXsXs.[ aXeXuX{.(.C._  XK.,XrXqXqX",
"qXqXeX2X<XeX0X7XzXn._ _.`.L.[.eXpX&XY A.].W.{ f.zX7X0XeX>XuXj.` pXwXeXU.E.}.J.[.uX0XqXqX",
"qXqXeX2X,XeX7XzXm.~ 0X$XT.<XeXaX*XJ _.qXR.#XrX] f.zX7XeX1X;XvXf.` pXeXtX@XR.<XuX0XqXqXqX",
"qXqXeX2X5XqXlXm.! eXtXtX7XtXpX*XJ _.gXeX2XqXtXpX' g.zXqX6X5X0XlXg.` pXeXyX7XuX8XqXqXqXqX",
"qXqXrX<X[.fXm.! rXrXtX=X^.0X&XJ _.hX7XeX=X7X0XeXpX` g.gX X}.rX0XlXg.' yX=X/.5XyX0XqXqXqX",
"qXqXrX,X!.a.( eXeXeXW.J.P.F.T ).gX7X0XwX-X7XwX9XwXiX{ r.J.L.G.2X9XlXg.] !.L.C.XXuX0XqXqX",
"qX0XrX>X&.|.Y.aXyX XQ..XF :XF.0XeXeXwXtX:X9XrXeX0XaX~.U.:.&.oXK.uXyX$XE..XF ,XF.2XeXqXqX",
"qX0XtX;X*.uXD.oX4XD.:XtX' *X.XR.2X=X-X:XXX*X;X-X:X$XS.pX`.&.uXD.OX4XA.,XrX' *X X{.iX0XqX",
"qX0XuXoXy.rXY.4XtXQ.{.nXo.>XQ.OXwX7X7X9X*X4X8X7X7X9XG.;Xx 7.uXY.4XtXW.{.bXo.,XW.$XyX0XqX",
"qXqXwX5XY.t.q.vX9X8XV.XXoXR.I.yXqXwXqXrX;X8XeXwXqXuX@Xv.H.T.r.q.vX9X7XV.XXXXR.I.yX0XqXqX",
"qXqXrX<XE.#X| s.lX0X7XE.G.].eXqXqXqX0XeX-X7XwXqXqX0XyX-XR.R.$X| d.lX0X6XE.G.].eXqXqXqXqX",
"qXqXeX1X4XdXiX' f.lX0XuX<XpX0XqXqXqXqXeX-X8XwXqXqXqX9XuX7X4XdXuX' g.lX0XyX<XpX0XqXqXqXqX",
"qXqXeX2X,XqXwXpX' f.zX4X*XqXqXqXqXqXqXeX-X8XwXqXqXqXqXwX<X,XqXeXpX' f.zX4X=XqXqXqXqXqXqX",
"qXqXeX1X,XeX9XwXpX' f.kX%XwX0XqXqXqX0XeX-X7XqXqXqXqX0XeX1X<XeX9XeXpX` g.kX&XeXqXqXqXqXqX",
"qXqXeX3X4XrXeXwXtXdX{ d.wXqXeXeXeXeXeXtX:X0XrXeXeXeXeXtX3X2XyXeXwXtXsX[ d.0X0XqXqXqXqXqX",
"qX0XtX-XXXqX>X2X<X3X0X! u.sX>X1X2X1X1X3X@X,X2X1X2X2X1X3X*X&X3X1X2X<X2XqX_ k.gX8XqXqXqXqX",
"qX0XtX=XF /.eX;X1X,X1XwX! w.iX:X<X<X<X2X+X:X1X<X<X<X<X2X&X%X2X<X<X,X,XrX] s.hX8XqXqXqXqX",
"qXqXqX9X).U oXjX0XrXwXwX2X} g.vX0XeXeXtX:X0XrXeXeXeXeXyX3X2XyXeXwXrXhXo.e.9X0XqXqXqXqXqX",
"qXqXeX,XwX+XH  XfX7XwX7X:XaX` d.zX7X0XeX-X7XqX0XqXqX0XeX1X<XeX9XqXdX .r.jX*XeXqXqXqXqXqX",
"qXqXeX2X-XgXoXJ .XfX8X8X&XeXpX' d.zX7XeX-X8XwXqXqXqXqXeX1X<XwXqXfX .r.zX4X*XqXqXqXqXqXqX",
"qXqXeX1X6XrXsXOXJ  XjXeX,XuXwXpX' d.zX9X-X8XwXqXqXqXqXeX2X,XrXdX .r.lX0XtX,XpX0XqXqXqXqX",
"qXqXeX2XW.=XeXdXoXP (.[.Y.|.eXeXpX' d.cX&X8XwXqXqXqXqXeX<X1XhX .r.lXqX8X_.E.|.rXqXqXqXqX",
"qX0XtX;XV.H.XXeXhX`.{ `.k.P.T.tXwXaX' h.rX6XwXwXwXwXqXeX4XwXo.r.zX9X8XG.G.a.Y.T.yX0XqXqX",
"qXqXqXqX .(.P.7XrX_.*X#Xn 4XR.#XwX0XuX' 7.tX8X9X9X9X8XqX2X/ r.jX4XyX~.[.1.@.c.(.#XtX0XqX",
"qXqX7XzX*.J.!..X4XA.=XhX| $X.XR.2X&X>X=Xg.-X-X-X-X-X=X,X[.l.7X*X*X1XG.@X .F.;.&X'.iX0XqX",
"qX0XtX=X6.#XJ.yXiX|.Y.zXy.;XH.1XtXwXqXeXgXwXeXeXeXeXeXwXtXfXqXwXqXpX{./.Z._  XK.,XrXqXqX",
"qXqXwX6X'.S.<XrX9XtXY.L.}.S.{.uX9XqXqX0X9XqXqXqXqXqXqXqX0X9XqXqXqX0XtXY.E.}.J.[.uX0XqXqX",
"qXqXwX7X.X9XtX0XqXqXuX&X{.<XuX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXyX%X{.,XuX0XqXqXqX",
"qXqXqXwXuXwX0XqXqXqX0XtXiXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXiXrX0XqXqXqXqX",
"qXqXqXqX0XqXqXqXqXqXqX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_6[] = {
/* columns rows colors chars-per-pixel */
"32 32 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXwXrXeXeXeXeXeXeXeXeXeXrXqXwXyXwXrXeXeXrXtXeXeXeXrX0XrXeXwXwXqX",
"wX6X,X3X2X2X2X2X1X2X2X3X;XqX3X@X7X:X3X2X<X=X3X2X3X-XrX<X4X8X6XwX",
"rX,X:X6X4X4X5X3X*X5X4X7XoX|.q.9..X&X6X5X1X+X6X3X8X.X.X/ U.`.:XrX",
"eX3X6XrXwXeXeXqX,XrX8XlXE.2.T.!.=.=XdX0XwX1XrX8XzXU.<.P.S.].tX0X",
"eX2X4XwX0XqXqX9X,X8XhX!.%.3X@X;XoXo.;XaX6X,X8XhXT.*.6X#X=XtXqXqX",
"eX2X4XeXqX0XeXrX*XhX!.#.0XgX3X6XlX+XX.:XiX&XkXT.%.eXdX1X9XtX0XqX",
"eX2X5XeXqXeX4XU.Y.A.=.0XpX}.D.T.@XjX@Xo.*XsXY.&.eXuX|.K.Y.=XtX0X",
"eX2X3XqX9XqX^.P.-.L.].dX@XT.t.g./.:XdXOX#.c.<.8XiX#XL.3.d.Q.4XeX",
"eX1X*X,X<X$X/.2X^ ).~.3XW.6Xr.Z.;X`.<X7X Xn ].0X>X`.P.1.w.'.&XyX",
"eX2X5XrX8XhXG.b.g.U.|.fX;Xh.<.F.v.eXeX9XaX.X+.-XgX&XW.h.S.!.8XwX",
"eX2X4X8XhX!.=.]./.|.rXuX<.M..X[.t.f.jX9X5X6X@X..-XfX+X(.).1XrX0X",
"eX2X6XlX!.#.0XdX,XfXuX:.F.cX,X7XlXy.s.kXeX,XjX+XX.=XzX3XeXtX0XqX",
"rX,X$XR.%.0XpX#X[.-X<.F.jX8X<X4X0XhXu.d.1X].;XhXOX+..X.X.XqXwXqX",
"yX%XB.6.2XgX|.L.k.z.M.cX8XtX3X7XrXqXhXi.i.s.P.&XxX'.y.B.M._.eX0X",
"pX{.s.;X|.1XR.0X| ,X_.2X<X3X*X;X2X2X=X].%X*.OXE.8X(.wXg.b.#X&XtX",
"iX{.K. X*X6XI.pXz.XX!.9X4X7X;X<X6X5X1X(.B.O.;X'.rX(.*X}.F.}.:XrX",
"eX<XL.*.oXlX+XT.).I.1XtXqXrX2X6XrXqXyX$XV.Y.p.a.zX8XE.(.T.[.tX0X",
"rX>X>X-Xo.OXjX,X{.9XrX0X0XeX1X5XwXqX0XtX4X.XyXi.a.kXwX.X=XyXqXqX",
"eX3X5XdX;XX.+XdX,XeX0XqXqXeX1X5XwXqXqXqXqX2XeXhXu.a.lX3X8XqXqXqX",
"eX2X3X0XaX:XX.+X8X9XwXqXqXeX2X5XeXqXqXwX9X>XwX9XjXu.f.eX5XwXqXqX",
"eX2XwX0X6XiX;X| ].sX6X0X9XqX<X4XqX9X9X0X8X:XqX9X8XdXi.s.eXqXqXqX",
"iX|.7.0X>X-X6XOX] `.9X-X>X<X%X=X<X>X>X,X:X@X,X>X;X2X2XE 2XeXqXqX",
"wX9Xp.t.hXqX9XpXXXO.#XfX8XrX2X6XrXwXwXeX0X,XrX0XyXeX=.I.eXqXqXqX",
"eX1XpXl.q.gX0X5X5X:X..@XsX9X1X5XwXqXqXqX9X>X0XrXeX$.W.eX2XwXqXqX",
"eX2X6XzXj.w.gXtX>XgX;XX.+XgX>X6XwXqXqXqX0X:XyXeX%.R.xX1X0XeX0XqX",
"rX<X X7XlXh.r.=X'.&XfX;X..#XwX2XwX0XqXqX8X3XrX$.R.lX<X}.oXqXwXqX",
"yX#Xx.^.9XlXs.Z.d.U.@XjX>X@.`.eXeXeXeXeXeX-X-.W.xX<XT.y.b.).wXqX",
"wX7X<.(..X,X).eX .;XW.<X9XXX#.-X2X<X1X,X8X3.D.tX2X[.H.3.e.(.=XtX",
"tX&Xp.(.$X9XL.uXc..X^.0X3XqX8X3X6X6X6X6X5X4XeX2XqX XW.3.k./.,XrX",
"rX<XR.[.tXyX&XY.).U.2XtXwXqXwXeXwXwXwXwXwXeXqXwXeXqX^.R.R.|.tX0X",
"wX8X,XtXqX0XtX5X*X0XrX0XqXqXqXqXqXqXqXqXqXqXqXqX0XwXwX:X1XtXqXqX",
"qXwXrX0XqXqX0XeXyXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXrX0XqXqX"
};

/* XPM */
static const char *const xpm_icon_7[] = {
/* columns rows colors chars-per-pixel */
"24 24 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"9X4X5X5X5X5X5X5X4X4XeX>X8X2X6X3X1X5X5X2XrX4XwX8X",
"4X>X6X4X5X1X>X4X8XOXG.r._.5X5X1X&X5X7X-XZ.w. X:X",
"5X6XrXwXeX7X3XyXyXi.D.`.t.}.gX7X3XwXfXx.b.R. XrX",
"5X4XwX0XeX0X7XyXe._.dX5XrX0.}.pX>XfXj.G.hX3XpX0X",
"5X5XeXeX8X!._.i._.fX!.R.2XtXw.[.sXk.G.jX|.T.*XeX",
"5X1X7X8X XZ.9.^.wX|.F.u.`.3X7Xe.3.L.aX$Xn.2.I.3X",
"5X>X3X5XXXG.0.{.,X$XN.p.OX:X7X5X%.[.yX XV.5.T.,X",
"5X4XyXyXp.D.E.%XfXg.n.W.r.#XsX0X4Xw. XsX/.U..XwX",
"4X9XyXe._.uX3XgXp.R.yX2X4X8.#XpX6XtX9.|.gX,XiXqX",
"3X Xu._.dX`.}.d.R.jX7X2XpX5Xr.W./.2XrXt.T.'.*XrX",
"-Xl.E.yX.XK.p.!.qX7X5X<X6XtX.XM.u.^.8X@XP.y.`.<X",
"#XH.@X<X_.*XB.oX*X5X,X-X3X4X.XP.w.+X:X].2XN.{.*X",
"5XI.r.rX1X~.Q.+XyXeX9X4XrXrX6XG.K.u.#XuX^.~.[.wX",
"3X8X|.0.tX6X-XyX0XqX7X2XwX0XeX0X<X2X8.%XiX$XrXqX",
"6X3XgX|.w.8X8XqXqXwX8X3XeXqXwX9X2XpX5X8.+XwX0XqX",
"3X9X8XpX{.5.4XeX7X0X6X<X0X8X9X7X<X8XrX1Xe.}.yX0X",
";Xi.5X1XwX).2.2X5X1X,X=X3X1X2X<X=X3X;XdXb.l.fX8X",
"wXT.i.uXwXyX(.e.wXuX7X4XrXwXeX0X3X0XjXT.d.6XqXqX",
"3XaX(.r.uX9XuX.X9.wXqX<XwX0XqX9X>XhXY.a.aX5XeX0X",
"4XXXfX(.p.#X}.pX Xq.8X7XwXqXwX7XyXR.a.dX:X{.4XeX",
"1Xx.'.gXT.M.k.}.aX}.7.-XeX9X9XrXA.a.aX,XV.q.E.7X",
"4Xs.^.:X|.&Xv.XX#XeX(.'.6X<X1X1XJ.2X5X XM.>.U.:X",
"2XI.+XiX$X_.E..XtX9XuXyXqXwXwXwXaXqXeX4XE.A.'.9X",
"0X2XeX0XrX,X:XwXqXqX0X0XqXqXqXqX9XqX0XeX6X,X9XeX"
};

/* XPM */
static const char *const xpm_icon_8[] = {
/* columns rows colors chars-per-pixel */
"16 16 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"9X6X7X6X6X5X0X>X8X6X4X5X9X4X3X0X",
"6X4X9X5X3XeXU.n.E.yX:XeX$Xk.~.0X",
"7X9XrX9X9XE.).8X!.^.sX*XG.,X7XeX",
"6X5XqX[.x.`.,XV.4X~.Z.K.3XH.Y.wX",
"6X3X9XS.k.7XE.p._.gXB.I.wXg.V.9X",
"7XrXE./.5X[.P.=XJ.OXhX~./.3X1XeX",
"4XJ.`.>XL.Y.yX9XeXK.K.<X^.J.XXwX",
"&XG.6X_.C.:X6X;X8X&Xi.!.6XI.L.7X",
"7XT.!.3X X8XqX4XwX4X].G.oX-X@XwX",
"7XaX^.Q.uXwXqX5XwXeXqXeXH.+XtXqX",
":X}.dXI.L.rX3X,X5X5X>X4XeXj.<XrX",
"8XF.|.kXR.W.iX2X0X0X2XuX`.T.wXqX",
"6X7XI.`.8X/.Q.qXwXwXwX`.R.6X2XeX",
"-XF.3XK.V.6XE.~.qX0X[.K.0Xh.C.qX",
",XR.4X@XW.>XeX3X8X8X3XwX4XS.Y.0X",
"0X0XrX7X5XeXqXeXwXwXeXqXeX9X0XwX"
};

const char *const *const xpm_icons[] = {
    xpm_icon_0,
    xpm_icon_1,
    xpm_icon_2,
    xpm_icon_3,
    xpm_icon_4,
    xpm_icon_5,
    xpm_icon_6,
    xpm_icon_7,
    xpm_icon_8,
};
const int n_xpm_icons = 9;
