#include "gtk.h"
/* XPM */
static const char *const xpm_icon_0[] = {
/* columns rows colors chars-per-pixel */
"128 128 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX9X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X2X3X9XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXwX2X/./././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././././.(.XX6XwXqXqXqXqXqXqX",
"qXqXqXqXqXqXwX3X].].'._._._._._._._._._._._._._._._._._._._._.`.].'.`._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._.`.`.'.`.`.`._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._.}.gXyXqXqXqXqXqXqXqX",
"qXqXqXqXqXqX0XpXFXFXMXwX7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8XsXVXvXtX8X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X0XpXhXjXjXfXrX8X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X9XkXJXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqX0XsXUXJX8X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXZXPXxXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXrXuXgXxXvXzXjXpXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXHXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXrXaX8X}.5XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXhXHXZXsXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXrXuXfXxXvXlXgXuXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX5X{.(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXZXLXlXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXiXgXxXvXlXgXuXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X(.(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXhXJXAXaXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXaXgXxXvXkXdXuXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXyXZXKXkXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXaXhXxXcXkXsXuXwX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXhXLXDXiXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXaXjXkXdXwX6XdXyXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXuXAXKXlXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXrX1X.X<XHXcXpXrXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXeXjXLXFXyXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X=X.XvXPXAXnXjXpXrXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqX0XpXAXKXzXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX3X@X:XLXIXGXnXcXcXkXaXeXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXeXlXPXFXyXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X8X7X7X7X8X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX=X#XzXPXHXsXaXkXcXvXcXkXsXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqX0XsXAXKXcXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XW.@.B l D 1.'.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX3XOX,XKXLXvXrXwXrXpXjXvXvXxXkXsXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXeXxXPXDXuXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XG.N ;     . < U ~.0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX&X$XjXUXGXiXwXqXqXeXtXpXhXcXvXxXjXsXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqX0XdXSXKXvXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX-XK o 0 $.a.^ = *  .7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX2X.X3XJXLXnXrXqXqXqXqXqXwXtXpXgXxXvXxXjXsXtXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXwXvXIXSXiXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXXXi < %.+XiXT.D   g 4XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX$X#XgXUXFXpXwXqXqXqXqXqXqXqXqXrXpXgXzXvXcXjXaXyXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqX0XdXSXKXvXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X*Xq.5._.7XyX}._   g 3XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX3X|.6XGXLXmXeXqXqXqXqXqXqXqXqXqXqXqXrXpXgXzXvXcXjXpXyXwXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXeXmXIXZXpXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXyXyXeXwXqXN.h % ) 7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X#X+XhXUXDXsXwXqXqXqXqXqXqXqXqXqXqXqXqXqXwXpXhXzXvXcXjXpXtXwXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXfXDXLXvXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX9X/.R   k F.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X3X{.8XFXPXmXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXiXhXzXcXcXkXpXrXwXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXwXBXIXCXsXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X_.R + e 9.0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X+X XkXUXSXdXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXyXgXzXcXcXkXpXrXwXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXfXFXPXcXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X(.J + 0 5.5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X4X[.0XDXIXnXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXtXfXxXcXcXkXaXeXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXwXCXIXCXsXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X!.C + 0 0.1XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X+X}.xXUXAXfXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXtXdXzXvXxXkXsXuXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXwXgXGXIXzXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XT.N @ 7 0.-X0X8X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X3X[.9XSXUXvXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXyXsXzXvXzXcXPXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXeXZXPXZXsXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX&XV +   & h D A C ~ {.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X+X{.bXIXSXfXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXuXaXkXnXPXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXwXgXJXIXlXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX.X+               : R.7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X1X[.5XFXUXcXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXuXfXPXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXtXCXPXAXaXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX:X3.-.;.;.;.;.;.=.9.#X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6X+X}.nXPXDXdX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXwXgXLXIXlXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXiXpXpXpXpXpXpXpXiXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X>X[.1XHXUXxXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXuXCXLXDXiXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX5XOX XbXPXGXsXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXwXhXIXPXzXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X;X].>XKXIXcXyXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqX0XiXVXLXFXyXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX3XoXoXcXPXHXpXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqX0XdXFXmXyX<X1X7X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX-X'.>XLXPXcXyXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXwXhXBX8X=X%X%X*X<X7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX3X.XOXzXIXHXiXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqX0XpXVXJXcXuX<X%X#X%X=X1X8X0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX&X`.<XKXLXbXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXlXLXIXFXcXfXyX<X$X#X%X-X1X8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX3X X+XjXUXGXiXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwX0XaXVXUXGXlXkXvXcXdXeX1X$X@X%X;X2X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX%X_.4XJXLXnXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXwXxXLXKXbXeXyXsXjXcXcXdX0X<X$X@X&X>X2X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX3X}.oXhXUXFXaXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXyXiXiXuXtXwXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XqXaXCXUXAXiXwXqXqXyXsXjXcXcXsX8X<X$X@X%X,X4X8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X#X_.7XGXLXmXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX0X*XL.v.k.m.Q.,XrXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXbXLXKXvXqXqXqXqXqXqXrXsXkXxXcXdX6X>X%X@X%X,X5X8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX4X{. XhXUXDXsXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X X3.j @   - A k.;XwXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XsXAXUXCXiXqXqXqXqXqXqXqXwXrXsXlXxXcXsX4X-X%X#X&X,X6X8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X@X_.9XDXPXmXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX#X*.o % e a 4 . y f.eXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.5XMXPXLXcXwXqXqXqXqXqXqXqXqXqXwXeXaXlXxXxXaX3X*X%X#X&X,X7X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X4X].}.lXUXSXdXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXyXR.s 1 P `.<Xp.u   ~ 8XqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.eXDXUXBXiXqXqXqXqXqXqXqXqXqXqXqXqXeXrXpXkXxXzXaX4X&X%X$X*X1X9X0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X+X_.0XSXIXnXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXOXe.u./.5XqX'.D   T <XqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).).xXIXKXzXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXrXpXkXxXlXaX5X$X$X%X*X<X7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X3X'.[.cXUXAXfXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X<X<X7X0X:X7.q % &.eXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).).NXUXMXuXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXtXpXjXxXjXiX4X%X#X%X=X<X7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6X+X_.8XDXUXvXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX0X XG g 3 p <.XXwXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).).VXLXjXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXtXaXhXxXhXuX4X&X#X%X-X1X8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X1X`.[.bXIXSXfXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X].s @   * ! W.tXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).).CXMXuXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXyXaXjXcXgXrX2X&X@X%X:X2X8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6X+X_.4XFXUXcXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXtXeX<XV.g.@.q & [ XXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).).mXhXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXsXjXzXfXwX1X%X@X%X>X3X9XqXqXqXqXqXqXqXqXqXqXqXqXqX0X,X`.}.nXPXDXdXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X=X/.`.4XuXpX>XQ   a n.wXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.sXiXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXsXkXzXfX9X<X%X@X&X,X4X8XqXqXqXqXqXqXqXqXqXqXqX5X+X_.<XHXUXxXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX8XL.U ( m.,X0X*XQ   u h.wXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.7XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXdXkXzXfX7X>X%X#X%X,X5X8XqXqXqXqXqXqXqXqX0X;X_. XbXLXGXaXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXS.1 . l y.M.%.0   Z Q.wXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXrXaXlXxXfX6X:X&X#X%X,X7X9XqXqXqXqXqXqX4XOX_.>XKXIXcXyXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX@X;.4 $     X & p w.0XqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXrXaXlXxXfX6X-X%X$X&X,X7X9XqXqXqXqX=X).oXcXPXHXpXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX;Xj.K t 7 d _ G.5XqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXrXpXkXzXdX6X*X%X$X&X1X8X0XqX3XXX).>XLXPXvXyXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X5X,X-X-X;X<X6XqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXrXpXjXlXsX6X%X$X%X*X<X8X=X_.+XlXIXHXiXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXwXwXwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXwX4X{./.(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXtXpXhXjXpX7X&X$X%X=XoX`.1XKXLXbXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXwX1XQ.!.^.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXyXaXgXhXiX5X&X#XoX=XjXUXJXaXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwXuXyXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXyXsXhXdX6X*XOX9XSXSXlX6X7X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqX0XdXUXUXlXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXpX<XXX.XyXCXyX<X*X;X1X6X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqX0XuXBXzX$X6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X4X{.~.%XzXBXqX1X=X#X$X;X2X7X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX7X@X).4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXwXeXrXrXeXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X&X`._.eXKXLXVXlXqX<X=X#X$X:X4X7X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXwX9X>XOX|.|.OX,X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X{.(.&XcXPXGXZXvXvXzXwX,X*X$X$X:X5X8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqX0X X8.x 1 , x q.OX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X*X_.).yXGXbXiXkXzXcXvXxXlXeX>X&X%X%X:X6X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXeX[.] $ @ $ $ # & ..|.rXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6X[.).=XmXZXsXwXeXrXaXzXcXvXzXkXyX,X%X%X%X:X7X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXrXq.; % I R.T.Q & = s.0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X&X).).aXHXcXwXqXqXqXeXtXpXkXcXvXxXhXuX<X$X%X&X;X6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXwX;X[   h T.0X0X'.s   ' .XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX4X[._.=XBXZXdXqXqXqXqXqXqXwXtXpXjXcXbXxXgXuX1X$X$X*X>X6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXwX@X~   M %XwXwX8Xl   V P.tXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X*X(.`.dXJXxXrXqXqXqXqXqXqXqXqXwXtXpXhXxXbXcXdXyX2X%X$X=X,X5X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X9XKXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXwX2X .  e d.5X3Xv.7   h B.yXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX2X{._.*XCXAXsXqXqXqXqXqXqXqXqXqXqXqXqXtXaXgXxXbXvXdXeX3X&X#X-X<X6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6X;XJXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXrXh.0   f -.*.b 7 o d M.yXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X*X/.[.fXJXzXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXsXhXzXvXbXfXqX2X*X$X;X2X6X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X1X.XHXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXeX#X7.s > & 7 T B * z S.yXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX1X{._.*XSXDXaXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXaXhXzXvXvXgX9X1X*X$X;X2X7X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6X@X}.HXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqX0X2X'.h.+.+.p.OXJ X J W.rXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X&X^.|.hXJXzXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXpXhXzXcXvXjX9X,X*X%X;X3X8X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX<X`..XHXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqX0XS.O.9.).6X3X^.u   #.&XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX>X{.).*XFXFXiXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXeXuXhXxXcXcXlX0X>X*X&X;X5X8X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX5X#X`.;XJXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXuXr., > W U.D.D % 3 v.tXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X%X/.oXhXJXxXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXeXuXhXxXcXxXkXwX:X&X&X;X5X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX:X`.{.0XLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXrX'./ . $ , > # @ ;.#XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X,X}.).>XHXGXuXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXeXtXgXxXvXxXhXrX>X%X&X:X7X9XqXqXqXqXqXqXqXqXqXqXqXqX6X@X`.%XrXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqX8X_.*.t + $ j q.OX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X#X/.OXhXJXcXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXyXdXzXvXzXgXtX>X%X*X:X6X0XqXqXqXqXqXqXqXqXqXwX;X`.}.5XtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXwX8X$XW.K.L.^.;X0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X>X{.(.2XHXFXuXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXyXsXzXbXxXdXtX,X%X*X>X5XqXqXqXqXqXqXqXqX5X+X`.%X0XtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXeXyXyXyXtXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XOX/.@XjXKXbXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXiXdXlXbXxXsXwX<X&X*X,X5XqXqXqXqXqXqX-X'. X4XqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X<X[.(.7XHXSXpXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXiXdXkXvXxXsXqX<X&X=X<X6XqXqXqX6XoX_.%XqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XOX(.@XlXPXnXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXuXdXkXcXxXsX9X<X*X;X2X6X0X:X[. X2XqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X1X[.(.qXHXAXaXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXuXfXkXcXcXdX8X,X*X;X2X#X].&XqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XoX).@XxXPXbXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXyXfXlXcXcXfX7X>X&XOX@X1XwXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X2X`.(.rXGXZXsXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXwXrXdXlXzXlXiX,X%X<X0XqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XoX_.#XbXIXvXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXrXuX0X$X}.<XqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X2X`.).iXHXZXdXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6X[..X4XwXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6XOX_.%XNXIXcXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X;X'.;X9XqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X<X_.`.aXJXAXdXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXtXuXuXyXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X X.X4XqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX5X+X_.%XVXPXzXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX'.e. .[ >.S.6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X:X'.;X9XqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X>X_.].sXJXSXsXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXH.~ 5     % m y.,XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX5X.X|.6XqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.<X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX4X+X).&XAXPXzXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX XT . 1 D P a   7 G.6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X;X`.-X9XqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.%X2X6X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X;X_.|.sXKXFXpXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX8.6 - o.XX-Xb.m j f.:XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX3X|.}.7XqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X)./.+X$X=X2X7X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX4X+X(.*XDXPXzXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XL   G J.OX_.~.M.L.|.5XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X-X'.*X0XqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.-X-X%X$X-X4X9X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX=X_..XaXLXGXiXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X%Xz   / y.S 3 1 H H.2XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX1X|.}.7XqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4X9X3X-X%X%X=X3X9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX5XoX/.-XGXLXxXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X).f   9 d t f * @ t W.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X&X'.&X0XqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwX0X8X3X-X$X&X-X3X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX*X_.OXfXPXGXuXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X!.d   : +..X|.w.& + -.%XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX,X{.|.6XqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqX0X8X3X;X$X&X;X4X9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX5XXX/.<XGXLXvXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X'.g   N H.rXwX|.b @ P {.rXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X$X'.&XqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqX0X7X3X;X%X*X:X3X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X&X`.OXfXIXFXiXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX,Xb   F R.tXrX*XZ # J ].rXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX>X{. X5XqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqX0X7X2X;X&X*X>X4X0XqXqXqXqXqXqXqXqXqXqXqXqX0X5X X(.6XGXKXbXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX' $ 7 0.9X9XB.8 o O.@XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X+X'.&XqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqX7X1X;X&X*X<X4X0XqXqXqXqXqXqXqXqXqXqX0X&X_.OXjXUXDXpXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXF.z   f %.*.m   7 Z.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX>X[..X4XqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqX7X1X-X&X*X1X5X0XqXqXqXqXqXqXqX0X6X}.(.9XFXLXnXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6Xu.v *     # g 3.;XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XoX`.*XqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqX8X1X-X&X=X2X6X9XqXqXqXqXqX9X&X_.OXzXUXAXaXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX6XK.+.L J } b.5XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX>X[..X3XwXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X1X=X&X-X2X7X0XqXqX0X6X{._.eXFXLXbXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXeXeXeXeXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X.X_.-XqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X2X=X&X;X3X9X9X9X-X`.OXbXUXZXsXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X:X]..X3XwXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9X3X-X%X-X3X6XoX`.yXFXPXvXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X|._.;X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X8X3X-X&X%XOX@XMXUXZXdXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X;X`..X4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X7X3X=X#XpXHXUXvXrXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6X|._.:X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X6X6XVXSXxX9X2X8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X;X`. X5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX5X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXfXyX2X*X&X;X1X7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX5X|._.;X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXwX6X).(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X4X<X;X%X@X%X:X2X7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X;X).|.6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqX8X*X(.(.4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X<X-X&X#X$X:X3X7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX3X|.`.-X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXwX1X~.^./.3XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX0X7X<X=X%X@X#X-X2X7X9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X8X*X/.{.6X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XrXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXwX5XoXoXoXwXpXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXtX8X1X1X<X<X6XrXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiX0X=X@X1XuXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXsXLXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqX0XaXPXLXVXKXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXPXPXPXPXIXIXIXIXIXIXIXIXIXIXDXFXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXPXPXPXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXLXZXIXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqX0XaXPXDX3XSXUXIXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXiXaXUXUXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXVX3XKXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXeXtX2X{.VXUXxXyXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuX9XXX8XUXDXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXuXrX>X XHXUXsX0XqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX0X7X2XuXsXrX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX8X3X0XdXiX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX0X6X3XpXsXeXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXwX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX0X0XqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_1[] = {
/* columns rows colors chars-per-pixel */
"96 96 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXqXqXqXqX",
"qXqXqXeXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXyXrXqXqXqXqX",
"qXqXeX2X%X*X&X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X&X&X&X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X&X&X&X&X&X&X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X%X&X#X:XeX0XqXqX",
"qX0XyXoXH.U.T.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.U.U.T.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.T.Y.Y.Y.Y.T.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.R.U._.eX0XqXqXqX",
"qXwX8XgXCXnX6X3X4X3X3X3X3X3X3X3X3X3X3X4X3X4XhXcX9X3X4X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X4X5XtXsXaXiX7X2X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X3X2XcXAX6XwXqXqX",
"qXwX6XbXUXwX9XrXeXeXeXeXeXeXeXeXeXeXeXeXwXaXUXBXwXrXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXrXyXsXcXmXbXkXtXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXwXuXUXBX7XwXqXqX",
"qXqXqXwX1X/.6XwXqXqXqXqXqXqXqXqXqXqXqX0X0XVXHXuX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0X0XuXlXcXcXsXwX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XrXUXBX7XwXqXqX",
"qXqX0XtXXXW.8XwXqXqXqXqXqXqXqXqXqXqXqX0XpXUXmX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXaXlXvXvXpXeX9X9XwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqX0XAXJXyXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXaXkXnXxXdXpX9X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqX0XdXIXMX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XwXaXhXiX-XaXfX9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXwX9XSXPXrXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XrX$X.XHXAXhXuX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXwX9XjXIXnXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXrXrXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX0X}.tXUXJXmXcXjXuX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqXAXUXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXqX5X5X7XrXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrX;X@XFXUXdXpXxXbXvXkXyXqX0XqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXwX8XkXUXvXwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX+X( @   i h.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX0X|.wXUXnXwX0XwXsXxXvXbXjXtXwX0XqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqX0XrXAXIXuX0XqXqXqXqXqXqXqXqXqXqXqXqX0XtX>Xm   f Z +   h.pX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX;XoXKXPXsX9XqXqXqXwXaXlXvXnXhXyXwX0XqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXwX9XjXUXcXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XsXE.  b *XmX2.  v lX7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX0X{.eXUXmXwXqXqXqXqXqX9XwXaXjXbXnXgXuXqX0XqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqX0XtXFXKXpX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXrX&Xf./.0XwXC.  n kX7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX-X XPXKXfX9XqXqXqXqXqXqXwX0XwXpXhXnXbXhXiX0X0XqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqX0XkXUXvX0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXlXiXqX9Xb   w.sX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X].yXUXNX0XwXqXqXqXqXqXqXqXqXqX0XwXiXjXnXvXhXiX0X0XqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqX0XrXLXHXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X5X0X8X!   J 0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX&X{.PXLXfX0XqXqXqXqXqXqXqXqXqXqXqXqX0XwXuXjXnXvXkXuX0XqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqX0XzXUXnX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX3XT   N >XeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X_.aXUXVX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXtXkXvXvXkXyX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXeXIXHXiXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX,XL   R eXcXsXwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX$X|.KXIXsXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXyXkXcXbXhXpXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqX9XnXUXmX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX,XP   7 h.K.N.P.7XqXqXqXqXqXqXqXqXqXqXqXqXqXwX8X(.sXUXCX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuXjXzXMXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXtXPXLXyXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XfXJ.              -XwX0XqXqXqXqXqXqXqXqXqXqX0XrX#X.XFXUXiXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyXhXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXwX8XNXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX%X7.e.w.y.i.r.x.6XqXqXqXqXqXqXqXqXqXqXqXqXeX5X^.aXUXBXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XeXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqX0XiXKXIXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXvXxXcXcXxXcXlXwXqXqXqXqXqXqXqXqXqXqXqXqXwX+XOXDXUXiXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXwX8XNXUXbXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X6X7X6X6X6X6X7XqXqXqXqXqXqXqXqXqXqXqXqXeX5X~.yXUXmXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqX9XsXLXUXyXqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXXX+XGXUXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXwX7XxXZX0X;X1X8XqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X^.rXUXnXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXVXkX>X$XoX&X2X9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX|.oXKXPXsX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XeX8XgXUXUXNXuX3X@XOX=X2X9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX3X(.rXUXmXwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XqXeXVXUXmXgXmXvXuX2XOX@X-X2XqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX}.|.IXKXfX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X8X8X9X0XqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XqXfXUXFXtXqXyXhXmXxXpX>XOX#X=X3XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X(.uXUXNX0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXeXaXhXhXsXrXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.7XrXAXUXzX9XwX0XwXtXjXnXxXpX-X+X#X*X5XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX{.[.IXLXdX0XqXqXqXqXqXqXqXqXqXqXqXqX9XyX%X+.i t / {.iX9XqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.7XjXUXSXtXqXqXqXqXqXwXtXkXvXvXrX*X@X@X-X7X0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX:X~.sXUXVX9XwXqXqXqXqXqXqXqXqXqXqXqX9XtX.Xi   % &   t ).pX9XqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.9XJXUXzX9XwXqXqXqXqXqXqXqXyXkXvXcX0X=X@X+X>X7XqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X}.{.KXIXaXqXqXqXqXqXqXqXqXqXqXqXqXqX7XlXq.O J ,X3XW   7.gX8XqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXE.hXUXDXwXwXqXqXqXqXqXqXqXqXqX0XuXhXbXxX9X-XoX#X1X7XqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX-XE.fXUXCX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XyXXXJ.:XnXaX@.  7.gX8XqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXOXR.DXUXjX0XqXqXqXqXqXqXqXqXqXqXqXqX0XuXhXmXkX8X*XXX%X<X7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X{.XXFXUXiXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXpX1X9.] 4 R XXuX9XqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXOXR.GXGXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX0XqXyXhXnXgX0X%XOX&X,X9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXeX;XT.sXUXBXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X9X+X      | 2XtX9XqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXOXE.GXgX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXyXkXnXgX7X#X+X&X<X0XqXqXqXqXqXqXqXqXqXqXqXwX7X`.@XDXUXiXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXiX0X{.W.L   +.9XwX0XqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXE.nXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXtXkXvXhX3X@X@X%X3X0XqXqXqXqXqXqXqXqXqXwX;XR.uXUXmXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXH.0.oXhXZXF.  l @XuX0XqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.wXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXyXlXcXfX<X$X+X%X5X0XwXqXqXqXqXqXwX7X^.#XGXUXaX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XhX..  v Q.).L   U 1XrX0XqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXyXkXvXaX<X$XOX=X5X9XwXqXqXqXqX%X~.eXUXnXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XyX[.b         e I.sX8XqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XrXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuXkXvXiX1X+X+X;X5XqXwXqX7X~.oXKXPXdX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XiX=Xq.) ( 1.OXpX9XqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX9XiX+XW.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XuXkXxXyX<XXX#X>X4XeX@X/.rXUXmXwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XwXhXcXvXkXrX9XqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qX0XeX<X`.^.7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX0XqXuXkXlXyX:XoX$X:X].|.PXKXfX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X7X5X5X7X9XqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qX0XyXOXU.Q.7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXuXjXlXtX=X@XoXiXUXCXqXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXwX8XzXLXBXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXuXkXtX$X+XFXmX8X,X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXwX7XcXUX4X6XwXqXqXqXqXqXqXqXqXqXqX9X7X7X8X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XwX+X!.>XAX3X&X@X#X<X8XqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqXqXwX;X~.6XwXqXqXqXqXqXqXqXqX0X0XpXkXlXdXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX2X'./.mXUXZXgX4X=XOX$X2X8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XyXXXW.8XwXqXqXqXqXqXqXqX0XeXrX_.M.b.R.0XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwX#XR.3XLXZXVXbXMXdX5X*XXX*X2X9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXOXW.8XwXqXqXqXqXqXqX0X0X0Xt.3     @ $.2XwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX2X`._.mXSX9XiXdXzXmXmXaX8X%XoX-X2XqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqX8XfXF.  X 9.c.q   7.gX7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X#XT.7XUXsXqX0XqXyXsXzXmXbXsX9X@X+X-X2XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqX0XiX*.  3.fXcXU.  m :XeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X^.}.MXAXeXqXqXqX0X0XrXaXxXmXvXgX6X+X#X=X5XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqX0XqXrX@.  u.sXgX'.  w Q.aX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X@XR.7XUXsX9XqXqXqXqXqX0XqXeXaXxXnXcXjX3X#X#X*X6XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X6XUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqX8XhXg.  1 C.E.x   < P.dX8XqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X!. XVXCXeXqXqXqXqXqXqXqXqXqXqXwXaXxXbXbXjX1X%X@X;X8X0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XOXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqX0XeX3XX.O   O L r < E.sX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXXX!.7XUXdX9XwXqXqXqXqXqXqXqXqXqXqXqXqXsXlXvXmXdX2X&XOX>X7XqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX=X[.UXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqX2X[.n.q.Y.=X  j oXyX9XqXqXqXqXqXqXqXqXqXqXqXqXqX3X!.|.GXBXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXaXjXbXMXaX5X%X+X,X7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqX'. XUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqX8XgXb.7 &.rXpX@.  _ aX8XqXqXqXqXqXqXqXqXqXqXqXqXqXwX|.^.0XUXgX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXpXjXmXmXpX6X#X$X<X8XwXqXqXqXqXqXqXqXqXqXqXqXwX*X^.3XUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqX9XpX'.<   k v   < ~.pX8XqXqXqXqXqXqXqXqXqXqXqXqXqX1X/.|.LXVXqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXiXjXMXbXsX5X@X%X<X0XwXqXqXqXqXqXqXqXqXqX9X{.[.rXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqX9XyX(.S     m T.yX9XqXqXqXqXqXqXqXqXqXqXqXqX0XwX X!.uXIXgX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXyXjXmXcXfX3X@X&X<XqXqXqXqXqXqXqX0XrX*X^.,XtXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXtX6X:X:X5XyXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX:X(. XLXAX0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXtXlXnXcXgX1X$X%X2XqXqXqXqXqXqX7X[.}.9XtXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqX0XrXyXyXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXXXW.gXIXfXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXuXlXvXbXdX<X%X%X4X0XwX0XeX*X!.<XwXrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX:X^.@XJXDX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyXjXvXnXaX1X%X=X5XqX8X[. X9X0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XOXW.jXUXaXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XuXgXvXmXpX2X@X-X;X_.:XrX9XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX,XW.=XHXSXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXuXfXnXbXsX7X@X$X8XwX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX8X.X~.hXUXpXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXtXdXpX;X#X0XqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX1XR.:XPXVXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0X0X0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqX3X^.-XrX0XqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX9X{.`.jXUXsX9XqXqXqXqXqXqXqXqXqXqXqXqXqX9XtXaXaXuX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX%X[.qX0XqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX<XW.;XUXNXeXqXqXqXqXqXqXqXqXqXqXqXqXqX0XiX>X.X X*XtXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X`.-XeXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X'.'.cXUXfX9XwXqXqXqXqXqXqXqXqXqXqXqX0XiXT.M     r c.rXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX#X[.qXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.4XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX-X^.;XUXNXwXwXqXqXqXqXqXqXqXqXqXqXqX0XeXOX2   F U     B.uX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6X_.;XwXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXQ.$X<X7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0X'._.NXIXhX9XwXqXqXqXqXqXqXqXqXqXqXqX7XhX .  ^ fXBXA.I z.qXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXOX].0XqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXQ.&X@X%X,X0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX*X/.<XUXCXqXwXqXqXqXqXqXqXqXqXqXqXqXqX8XiXf   R.8.K [ _.wX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX3X_.:XwXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.7X7X=X#X&X<X0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X[./.ZXIXgX0XqXqXqXqXqXqXqXqXqXqXqXqX0XwX;Xe   M - $   5 ].iX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXOX].9XwXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XeXqX5X*X$X%X2X0XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX*X!.9XUXAX0XwXqXqXqXqXqXqXqXqXqXqXqXqX0XeX#Xe   M >X<XR   o.0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX<X/.:XeXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqX5X=X$X&X4X9XwXqXqXqXqXqXqXqXqXqXqXqXwX6X{.).ZXUXdXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0X5Xt   M.rXjXM.. N 4XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXoX[.9XwXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXwXwX4X-X#X=X4XqXwXqXqXqXqXqXqXqXqX0XeX=XE.tXUXAX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XjXZ   =.kXvX6.  W 7XeXqXqXqXqXqXqXqXqXqXqXqXqXqXeX<X^.:XrX0XqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXwXqX4X-X#X;X4XwXwXqXqXqXqXqXqXwX6X'.|.CXUXaXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XuXD.  O 5.2.,   V.uXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X.X}.9XwXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXwXqX4X-X#X;X4XwXqXqXqXqXqXwX=XT.tXUXCXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyXp.9     1 w.9XeX0XqXqXqXqXqXqXqXqXqXqXqXqXwX1X^.;XrX0XqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXwX0X5X=X#X;X6XwXqXqXwX6X).OXZXUXaXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XiX XY.Y.{.tXeX0XqXqXqXqXqXqXqXqXqXqXqXqXwX0X{.}.0XqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqXwXqX5X=X$X>X8XqXqX=XQ.tXUXNXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XpXgXgXpX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX>X/.-XeX0XqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX5X=X%X,X8X].+XFXUXdX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X8X8X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX'.[.qXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX4X=X#X.XrXUXBXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX=X(.;XeXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XtXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXwX<X3XLXKXpX6XeXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X].].qXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX0XrXoXW.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXuXlX6X$X*X1X9XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX&X/.:XwXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qXqX9XuX+XW.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X5X2X:X@X+X*X<X9XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X[.].0XqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXBX7XwXqXqX",
"qX0XrX;X/.~.7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXeX0X6X:X+X@X&X<X0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX%X!.>XwXqXqXqXqXqXqXqXqXqXqXqXqX9XrXUXBX7XwXqXqX",
"qX0XtX#XW.(.8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX0X7X;X@X#X&X<X0XqXqXqXqXqXqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwX4X].{.9XwXqXqXqXqXqXqXqXqXqXqXqXqX0XyXUXBX7XwXqXqX",
"qXwX7XvXUXZXKXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXSXLXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXAXUXBX7XwXqXqX",
"qXwX7XlXKX&XHXUXDXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXDXDXFXFXFXFXFXFXFXFXFXDXFXZX%XUXUXSXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXDXFXMX&XUXCX6XwXqXqX",
"qXqXqX0X2X.XnXMX5X8X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X,XoXVXvX5X8X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X;X@XSXlX8XwXqXqX",
"qXqXqXqXeXyX7X7XwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXeXtX7X8XwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXrXtX6X8XqXqXqXqX",
"qXqXqXqXqX0XwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXwXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_2[] = {
/* columns rows colors chars-per-pixel */
"88 88 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXqXqXqX",
"qXqXqXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X8XqXqXqXqX",
"qXqXwX8XR.T.W.!.!.!.!.!.!.!.!.!.!.!.!.!.!.E.E.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.Q.W.E.W.Q.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.W.XXqXqXqXqX",
"qXqXqXqX6X9X-X+X#X#X#X#X#X#X#X#X#X#X#X@X$X3X<X#X@X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X$X:X,X>X=X+X@X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X#X$XmXtX0XqXqX",
"qXqX0XeXUXSXwXyXtXtXtXtXtXtXtXtXtXtXyXeXkXUXxXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXyXdXzXBXmXlXiXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXyXqXmXUXeX0XqXqX",
"qXqXqXqXtX}.=XrX9X0X0X0X0X0X0X0X0X0X0XqXAXZXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9X9XeXgXcXxXdXwX9X9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX6XnXUXwX0XqXqX",
"qXqXqXeX1XT.,XrX0XqXqXqXqXqXqXqXqXqX0XfXUXgX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XtXjXcXvXaXeX9X0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XnXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXqX0XGXZXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyXhXbXcXfXyX8X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqX9XkXIXhX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXyXdX8X*XbXuX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXwXFXDX0XwXqXqXqXqXqXqXqXqXqXqXqXqX9X7X6X7X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XeX}.7XUXBXjXyXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXwX8XzXUXfXqXqXqXqXqXqXqXqXqXqXqXqXqX9XtXlXcXkXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X.XCXUXnXcXnXhXyXqX0XqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqX0XrXGXDXwXqXqXqXqXqXqXqXqXqXqXqXqX0XiX@Xw.*.u.*XuX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXoX8XUXmX9XiXhXnXbXhXuX0X0XqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqX9XlXUXdX0XqXqXqXqXqXqXqXqXqXqXqX0XyX_.i       k .XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X.XCXUXyXqX0XwXyXhXbXvXjXyX0X0XqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqX0XrXPXZXeXqXqXqXqXqXqXqXqXqXqXqXqX7XjXL   8.0X..  ..zX7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXoX4XUXvXqXqXqXqX0XqXtXhXvXbXkXyXqX0XqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqX9XcXUXhX9XwXqXqXqXqXqXqXqXqXqXqXqX0XtXu.>.3XgXW.  U cX6XqXqXqXqXqXqXqXqXqXqXqXqXqXwX3X{.DXPXiX0XqXqXqXqXqXqX0XtXgXcXnXhXyXqX0XqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXrXPXAXqXwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XgXhXqXaX)   s.fX8XqXqXqXqXqXqXqXqXqXqXqXqX0XeX.X6XUXnX0XqXqXqXqXqXqXqXqXqX0XyXfXvXnXhXuX0X0XqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXwX8XmXUXhX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6X7XuX1.  U wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX<X'.DXPXpX0XqXqXqXqXqXqXqXqXqXqX0X0XtXdXvXvXkXyX0XqX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXwX9XiXKXDX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXtX=.  V >XqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXeX}.9XUXMX9XwXqXqXqXqXqXqXqXqXqXqXqXqX0XqXeXfXcXbXkXuX7XnXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXwX9XnXUXdXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX#.  / sXmXfXqXqXqXqXqXqXqXqXqXqXqXqXqXeX,X[.ZXUXuXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXfXxXmXdXBXUXeX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXwX9XpXIXAXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX8XfX$.  . F ] L 3.0XqX0XqXqXqXqXqXqXqXqXqXqXwX[.8XUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XrXsXlXGXUXeX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXmXUXfX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7XjXN   &       z 9XqX0XqXqXqXqXqXqXqXqXqXeX>X}.ZXUXyXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqX0XNXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqX0XiXUXCXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X6X6X9X0X0X0XqXqXqXqXqXqXqXqXqXqXqXqXqX].5XUXvXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX6XnXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqX9XVXUXhX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXeXeXeXeXeXwXqXqXqXqXqXqXqXqXqXqX0XeX-X].FXPXpX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrXqX0XsXUXFXeXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X].6XUXnX0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrXqX7XMXVX0X,X9XwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX-X_.FXPXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XwXZXtX*X+X#X1X8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX9X).qXUXMX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XtX7XkXUXKXlX0X*XoX&X1X9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX=X].ZXUXyXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XeXeXDXUXjXkXMXjXqX$XOX&X2XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX0X/.8XUXnX0XqXqXqXqXqXqXqXqXqXqXqXqX0XeXyXyXeX0XqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XeXjXUXmX0XwXyXzXbXlX7X$X+X&X4X0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX#X{.AXUXuX0XqXqXqXqXqXqXqXqXqXqXqXqXyX6X,X>X8XyXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.>XyXGXPXpX0XqXqXqXuXlXnXkX4X%XoX=X5X0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X).5XUXvXqXqXqXqXqXqXqXqXqXqXqXqX9XuXP.l     H |.rX0XqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.:XbXUXmX0XwXqXqXqXqXqXiXjXmXfX5X@X+X-X4XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX+X`.GXPXpX9XqXqXqXqXqXqXqXqXqXqX8XpX^.  . &.Q   l wXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.1XPXKXiX0XqXqXqXqXqXqX0XwXyXlXbXdX4X+X@X-X7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.8XUXmX0XwXqXqXqXqXqXqXqXqXqXqX9XsXR.E B.NXuXf   >XeXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX4XU.sXUXnX9XwXqXqXqXqXqXqXqXqX0XqXuXzXcXfX,X@X@X;X7XqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX#X).FXPXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX9X@Xh.2 F qXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX4XI.kXUXrX0XqXqXqXqXqXqXqXqXqXqXqXqX0XiXlXvXpX,X@X+X,X7XwXqXqXqXqXqXqXqXqXqXqXqXqXwX5XW.wXUXMX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX`.    7 ^.aX0XqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX4XI.zXMX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXiXlXbXuX,XOX$X,X8XwXqXqXqXqXqXqXqXqXqXqXwXoX[.ZXUXyXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtXdX2Xz.<.& e +XrX0XqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX4XU.sXsX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXuXzXxXuX-XOX$X<X0XqXwXqXqXqXqXqXqXwX5XQ.8XUXnXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX_.f.*XbXFXQ   i.sX9XqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XT.3XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXiXzXxXrX=X+X#X2X9XwXqXqXqXqXqXwX|.{.AXUXuX0XqXqXqXqXqXqXqXqXqXqXqXqXqX8XgXa.  N W.A.5   H.iX9XqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXiXzXxXwX-XoX%X3X9XwXqX0XeX<X^.5XUXbXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX:X~       $ $.rXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXiXzXlXwX&XOX*X3XqXqXqX|.).GXLXpX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX8XG.i.s.T.uXwXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX4XW.,XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXiXlXjX9X#X@X*X6X<X!.0XUXmX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXfXzXlXsX9XqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXwX9XQ.P.;XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXiXlXjX7X%X#XOX{.SXUXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X8X7X7X8XqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqX0X2X2X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXpXzXdX<XXXyXUXbX5XwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqX0XeXUXGXqXqXqXqXqXqXqXqXqXqXqX0X9X9X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXuX@X`.lXfX%X$X:X7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXqXuX X=XtX0XqXqXqXqXqXqXqX9XtXsXdXyX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX>XE.%XPXMXuX:X@X#X>X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX1XT.,XrX0XqXqXqXqXqXqXqXuX:X`.).*XuXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwX}.(.kXUXDXmXbXuX>X@X@X1X8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqX0XtXP.d     q V.rX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX:X~.*XIXaXtXgXxXmXvXrX<XOX$X1X9XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqX0XyX=X2   6.s.* o {.uX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X|.~.mXCX0XqX0XtXsXxXMXxXuX>XOX&X<XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XnXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqX0XiXF.  @.fXbXf.  5.yXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX:XW.,XLXdX0XwXqX0X0XeXaXxXmXxXiX-X#X%X3XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX6XnXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqX0XiXG.  ' iXvX2.  Q wXwXqXqXqXqXqXqXqXqXqXqXqXqXwX8X}./.mXSX9XqXqXqXqXqXqXqXwXsXzXnXbXtX:X@X%X4X0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX<XkXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqX0XtX-X8   _ ' 4 o F wXwXqXqXqXqXqXqXqXqXqXqXqXqXwX,XR.3XIXaXqXqXqXqXqXqXqXqXqX0XwXsXlXmXvXrX>XOX-X4XqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrX#XtXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqX0XtX[.P * J w.X ^ rXqXqXqXqXqXqXqXqXqXqXqXqXqXqX'.'.MXAXqXqXqXqXqXqXqXqXqXqXqXqX0XeXpXkXMXzXuX-X+X-X5XwXwXqXqXqXqXqXqXqXqXqXqXqXqXwX0X/.yXUXqX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqX0XtX*X<.B.2XlXy.  f.iX0XqXqXqXqXqXqXqXqXqXqXqXwX=XQ.2XUXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXuXzXnXcXuX=X$X=X7XqXwXqXqXqXqXqXqXqXqXqXqXwX;X~.jXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqX0XyX-X-   q.9.  1 %XuX0XqXqXqXqXqXqXqXqXqXqXqX0X].'.AXVXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXpXlXnXbXeX:X@X;X7XwXqXqXqXqXqXqXqXqXqX9X'. XNXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqX0XtXJ.w     e K.yX9XqXqXqXqXqXqXqXqXqXqX0XrX=XQ.7XUXdX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXpXjXmXcXrX;X@X>X7XeXqXqXqXqXqX0XeX&X_.<XMXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXuX%XI.I.%XuX0XqXqXqXqXqXqXqXqXqXqXqXwX7X[.`.GXCX0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XwXyXkXMXzXyX=X$X:X9XwXqXqXqXqX8X].OX9XnXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqX9XyXgXgXyX9XqXqXqXqXqXqXqXqXqXqXqXqXwX;XR.rXIXdX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXyXlXbXxXeX=X$X,X0XwXqXeX&X/.6X8XnXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqX9X8X8X9XqXqXqXqXqXqXqXqXqXqXqXqXwX8X_.|.DXAX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyXkXbXcXqX;X$X1X0X8X'.+XeX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX-XW.tXUXpXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XuXhXmXxXeX=X*X=X'.3XrX6XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X(..XGXCXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXtXhXnXlXeX#X-XqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX$X~.tXUXaX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XtXyX@X@XrXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X`..XIXBXwXqXqXqXqXqXqXqXqXqXqXqXqX0X8X8X8X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X6X_.,XeXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX%XW.fXUXdX9XwXqXqXqXqXqXqXqXqXqXqX0XrXfXlXjXiXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX#X XwXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX6X_.@XLXCX0XwXqXqXqXqXqXqXqXqXqXqX0XyXOX` a A S.pXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX6X'.<XwXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.:XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX&XW.jXUXsX0XqXqXqXqXqXqXqXqXqXqXqXrX@X8   1 o   w.aX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX#X XqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX2XW.+X3X8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X~.*XPXZX0XqXqXqXqXqXqXqXqXqXqXqX9XdX{   3.MXXXn ..5XeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X).<XeXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX2XW.+X#X%X1XqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX@X/.hXUXpXqXqXqXqXqXqXqXqXqXqXqXqX0XtX2 t ^.%.] n.7X9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX+X XwXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.>X9X=X$X*X2XqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX6X/.&XUXNXwXqXqXqXqXqXqXqXqXqXqXqX0XrX%X, - l 5     s.dX8XqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X).<XeX0XqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XtXqX6X-X$X&X4XqXwXqXqXqXqXqXqXqXqXqXqXqX0XeXOX(.cXUXsX9XqXqXqXqXqXqXqXqXqXqXqX0XtXoX3   9.jXU.  g -XtX0XqXqXqXqXqXqXqXqXqXqXqXqXwX.X|.eXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XwXwX5X;X#X-X4XqXwXqXqXqXqXqXqXqXqXqXwX3X(.*XUXBXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqX6X, 2 OXaX0Xf 3  XiX9XqXqXqXqXqXqXqXqXqXqXqXwX1X_.<XeXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXwXqX6X;X#X-X6XwXqXqXqXqXqXqXqXqXwX@X~.NXUXdX9XqXqXqXqXqXqXqXqXqXqXqXqXqX9XdXK   w.mXL.  v ,XrX0XqXqXqXqXqXqXqXqXqXqXqXwX.X|.wXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXwXqX6X-X%X-X7XqXwXqXqXqXqXwX4X!.2XUXCX0XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX`.>   6 O X x.sX8XqXqXqXqXqXqXqXqXqXqX0XeX<X(.2XwXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXwXwX5X-X$X;X7XwXqXqXqXqXOX).MXUXaXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX|.' c T P.pX0XqXqXqXqXqXqXqXqXqXqXqXqX0X X XqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXwXwX5X-X$X>X8XwXwX5X!.5XUXVXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXjXxXkXuX0XqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.1XeX0XqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXqXqXwXqX5X-X&X>XqX+X].NXUXpX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X8X7X7X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX]..XqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXqXqXqXqXwXqX5X-X&X X4XUXBXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX;X(.<XeX0XqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXwX3X:XDXPXuX8XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X]. XeXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX3XR.,XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXjX5X#X=X3XqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX*X/.1XwXqXqXqXqXqXqXqXqXqXwX7XmXUXwX0XqXqX",
"qXqXqXeX5XE.,XtXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqX7X5X,X$X#X=X3XwXeXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXwX9X{. XeXwXwXwXwXwXwXwXwXwXwXeX8XmXUXwX0XqXqX",
"qXqXwX9X~.I.=XqX7X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X9X0X8X5X-X+XoX@X,X6X7X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X0X@XR.,X9X8X8X8X8X8X8X8X8X8X8X9X3XbXUXwX0XqXqX",
"qXqXqX0X:X;XuXjXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXhXsXqX9X8XeXfXgXgXgXgXgXhXjXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXdX2X3XfXgXgXgXgXgXgXgXgXgXgXgXgXdXAXUXeX0XqXqX",
"qXqX0XeXUXSXSXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXZXCXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXDXMXUXeX0XqXqX",
"qXqXqXqXsX|.fXUXiXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXpXpXpXpXpXaXaXaXpXpXdXXXiXUXpXpXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXpXpXpXpXaXaXaXaXaXaXaXaXaXpXpXfX+XeXUXqX0XqXqX",
"qXqXqXqX8X7XwXeX9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X8X7XwXeX9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X8X7XqXrXqXqXqXqX",
"qXqXqXqXqXwXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqX0XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_3[] = {
/* columns rows colors chars-per-pixel */
"64 64 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXeXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXeXqXqXqXqX",
"qXqXqX0X5X5X6X6X6X6X6X6X6X6X6X6X5X5X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X5X5X5X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X4X0XqXqXqX",
"qXqXwX6X`.`.(.).).).).).).).).(._.).(.).).).).).).).).).).).).).).).(._.`.)./.(.).).).).).).).).).).).).).).).).).).^.#XrX0XqXqX",
"qXqX0XyXKXyX7X9X9X9X9X9X9X9X6XsXCX0X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9XeXfXlXaX0X7X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9XFXiX9XqXqX",
"qXqX0XtX2X@XtXqXwXwXwXwXwXwX0XDXzX9XwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXqXuXlXcXdXwX0XqXwXwXwXwXwXwXwXwXwXwXwXwXwXqXrXUXyX0XqXqX",
"qXqX0XtX.XoXtX0XqXqXqXqXqX8XgXFXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXiXlXzXsXrX8X0XqXqXqXqXqXqXqXqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqXqXqXqX0XFXlX8XwXqXqXqXqXqXqXqX9X7X8XqXqXqXqXqXqXqXqXqXqXqX0X0XpXyX1XzXeX0X0XqXqXqXqXqXqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqXqXwX8XjXGXqXqXqXqXqXqXqXqXqX0XiXcXhXqXqXqXqXqXqXqXqXqXqXqXqXqXeX%XwXUXcXgXrX0X0XqXqXqXqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqXqXqXqXGXkX9XqXqXqXqXqXqXqX0XpX_.1.v.0XeX0XqXqXqXqXqXqXqXqXqXwX9X$XJXcXyXzXvXgXrX0X0XqXqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqXwX8XzXGXqXqXqXqXqXqXqXqX9XsXA.o k % ' uX0XqXqXqXqXqXqXqXqX0XrX&XyXPXqX9XqXiXlXcXgXrX0X0XqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqXqXwXJXkX9XqXqXqXqXqXqXqX9XgXr.' ZXp.X 9XeXqXqXqXqXqXqXqXqXwX9X@XJXzX8XwXqX0X0XuXkXcXgXrX0X0XqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXwX8XxXHXqXqXqXqXqXqXqXqXqXqX0XtXfXiXv G pX0XqXqXqXqXqXqXqX0XrX$XtXIXwXqXqXqXqXqX0X0XuXkXcXgXrX0X0XrXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqX0XeXKXkX9XqXqXqXqXqXqXqXqXqX0XeXeXJ 7 ;XtX0XqXqXqXqXqXqXqXwX7XOXJXzX8XqXqXqXqXqXqXqX0X0XyXjXcXgXeXwXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XwX8XvXHXqXqXqXqXqXqXqXqXqXqX0XwX8XP q ,XfX9XqXqXqXqXqXqXqX0XrX#XtXIXwXqXqXqXqXqXqXqXqXqXqX0X0XyXjXxXjXPXuX0XqXqX",
"qXqX0XtXOXoXtX0X0XrXPXkX9XwXqXqXqXqXqXqXqXqX7XlX=.  - 3 8 #XtX0XqXqXqXqXqXwX6XoXKXzX8XwXqXqXqXqXqXqXqXqXqXqXqX0X0XtXkXUXyX0XqXqX",
"qXqX0XtXOXoXtXqX8XnXGXqXqXqXqXqXqXqXqXqXqXqX0XtX|.H.H.C.G.4XwXqXqXqXqXqXqXrXOXtXIXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XwXPXuX0XqXqX",
"qXqX0XtXOXoXtX9XyXUXlX9XqXqXqXqXqXqXqXqXqXqXqX0XiXgXfXhXgXeXqXqXqXqXqXqXwX5X.XKXzX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXPXuX0XqXqX",
"qXqX0XtXOXoXyX7XcXMX5XqXwXqXqXqXqXqXqXqXqXqXqXqX0X8X8X8X8XqXqXqXqXqXqX0XrXXXtXPXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XVXrX#X$X3XwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X XKXzX8XwXqXqXqXqXqXqXqX9X9X0XqXqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXOXeXgXUXmXjX7X#X%X4XwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX.XtXIXwXqXqXqXqXqXqXqXqX0XpXsXwX0XqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtXGXNX9XaXxXjX5X#X&X5XwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX2X|.LXzX8XqXqXqXqXqXqX0XeXpX|.~.7XyX0XqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOX XcXIXuX0X0XqXaXxXhX3X@X=X6XwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXrX|.tXIXwXqXqXqXqXqXqXqXqXwX{ @ $ m =XtX0XqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXOXUXmX9XwXqXqX0XqXaXzXfX1X@X-X8XwXwXqXqXqXqXqXqXqXqXqXqXeX<X}.LXzX8XqXqXqXqXqXqX0XyX-XC W.0X  y.kX8XqXqX0XeXPXuX0XqXqX",
"qXqX0XyX.X1XUXrX0XqXqXqXqXqX0XqXaXzXsX,X@X;X8XeXqXqXqXqXqXqXqXqXqXeX{.tXIXwXqXqXqXqXqXqXqXqXqX9XpX^.! j XXuX9XqXqX0XeXPXuX0XqXqX",
"qXqX0XyX X6XVX7XwXqXqXqXqXqXqXqX9XqXaXlXaX>X@X>X9XwXqXqXqXqXqXqXeX,X}.LXzX8XwXqXqXqXqXqXqXqX0XwXbXl.r u /.uX0XqXqX0XeXPXuX0XqXqX",
"qXqX0XyXXX:XdX9XqXqXqXqXqXqXqXqXqXqX9XqXaXlXiX;X@X,X0XeXqXqXqXqXeX].tXIXwXqXqXqXqXqXqXqXqX0XyX@X3.0XGX< P fX9XqXqX0XeXPXuX0XqXqX",
"qXqX0XtXoXOXtX0XqXqXqXqXqXqXqXqXqXqXqXqX9XqXsXkXyX-X#X<XqXwXqXeX:X{.LXzX8XwXqXqXqXqXqXqXqX0XtX*X3 7 v o v.aX9XqXqX0XeXPXuX0XqXqX",
"qXqX0XyX@XoXtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XwXsXjXrX=X$X1XqXeX`.tXPXwXqXqXqXqXqXqXqXqXqXqX0XuX=X3.+.U.iX0XqXqXqX0XeXPXuX0XqXqX",
"qXqXqXqX)..XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XwXsXhXqX&X*X*X|.PXxX8XqXqXqXqXqXqXqXqXqXqXqX0XtXzXcXaX0XqXqXqXqX0XeXPXuX0XqXqX",
"qXqXwX9X&X2XwXqXqXqXqXqXqX0X0X0XqXqXqXqXqXqXqXqXqXqX9XwXsXgX8X.XpXGX0XwXwXqXqXqXqXqXqXqXqXqXqXqX0X7X7X9XqXqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XuXKXwXqXqXqXqXqXqX9XwXyXwX9XqXqXqXqXqXqXqXqXqXqXqX9XtX2X[.nX8X@X;X9XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XrX+XXXtX0XqXqXqX0XiX9X&X9XiX0XqXqXqXqXqXqXqXqXqXqXqXqX).<XUXmXiX:X#X>X9XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXoXoXtX0XqXqX9XiX~.u @ p (.iX9XqXqXqXqXqXqXqXqX0XeX;X_.BXfXpXcXbXiX:X#X,X0XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqX0XiXd V 3XS d uX0XqXqXqXqXqXqXqXqXqXqX).7XZX8X0XqXiXxXbXiX:X$X<XqXwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqX0XqXeX> +.PX*.X #XuX9XqXqXqXqXqXqX0XrX=X`.ZXaX9XwXqX0XqXiXzXbXiX:X%X1XqXeXqXqXqXqXqXqXqXqXqXqXqXqX7XLXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqX9XsXp.& M c 1 |.iX9XqXqXqXqXqXqXqXqX).0XZX8XqXqXqXqXqX0X0XuXzXbXuX:X&X2XqXwXqXqXqXqXqXqXqXqXqXqX@XHXiX9XqXqX",
"qXqX0XtXOXoXtX0XqXqX0XrXG.u.|.2.4 6XeX0XqXqXqXqXqX0XrX*X'.DXaX9XqXqXqXqXqXqXqX0X0XuXlXvXuX:X&X3XwXwXqXqXqXqXqXqXeX,X|.PXuX9XqXqX",
"qXqX0XtXOXoXtX0XqXqX8XfX} ; ..o 2.gX8XqXqXqXqXqXqXqX0X(.tXZX8XqXqXqXqXqXqXqXqXqXqX0X0XyXlXvXuX:X*X4XwXwXqXqXqXqXwX}.:XUXyX0XqXqX",
"qXqX0XtXOXoXtX0XqXqX0XwX0X7.Z a.rXqXqXqXqXqXqXqX0XrX&X{.GXaX9XqXqXqXqXqXqXqXqXqXqXqXqX0X0XyXlXvXyX:X=X5XwXwX0XeX,X[.eXIXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqXqXqXwXhXvXfXqXqXqXqXqXqXqXqXqX0X/.iXZX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0XyXlXcXyX:X-X6XwXwX}.-XyXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqXqXqXqX8X7X8XqXqXqXqXqXqXqXqXrX$X XJXpX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9XtXkXxXyX:X,X>X}.0XrXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX9X/.sXZX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9XtXkXzXwX#X<XwXeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX#XXXLXpX9XqXqXqXqXqXqXqX9X9X9X9XqXqXqXqXqXqXqXqXqXqXqX0X9XuX:X$XtX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX8X(.hXZX8XqXqXqXqXqXqXqX0XiXfXfXsXqXqXqXqXqXqXqXqXqXqXqXqXqX0X X3XeX0XeXPXuX0XqXqX",
"qXqX0XtXOXOXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX@X+XPXiX0XqXqXqXqXqXqX9XpXI.j e s.tX0XqXqXqXqXqXqXqXqXqX0XrX;X@XrXqX0XeXPXuX0XqXqX",
"qXqX0XtX+X}.0XwXwXqXqXqXqXqXqXqXqXqXqXqXqXwX7X(.zXZX8XqXqXqXqXqXqX0XuXOX&  .t.4 ~.iX9XqXqXqXqXqXqXqXqXqX0X|.4XwXqX0XeXPXuX0XqXqX",
"qXqX0XtX+X[.=X;X8XwXwXqXqXqXqXqXqXqXqXqX0XrXOX$XUXiX0XqXqXqXqXqXqX8XkX8.4 b.} f.6XwXqXqXqXqXqXqXqXqX0XrX=X+XrX0XqX0XeXPXuX0XqXqX",
"qXqX0XtXOXOXwX:X&X>X8XeXqXqXqXqXqXqXqXqXwX6X(.vXCX8XqXqXqXqXqXqXqX9XfX*.  { P + ^.iX9XqXqXqXqXqXqXqXwX9X{.5XwXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXyXwX8X>X&X,X9XeXqXqXqXqXqX0XrXoX*XUXiX0XqXqXqXqXqXqXqX8XjX,.+ aXMX1 <.jX8XqXqXqXqXqXqX0XrX&X+XrX0XqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XwXeX8X>X*X<X0XeXqXqXqXwX6X).MXCX8XqXqXqXqXqXqXqXqX9XaX!.  2.e.  K.aX9XqXqXqXqXqXqXwX8X[.5XwXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqXwXeX8X>X=X1X0XwXqXrX.X:XUXiX0XqXqXqXqXqXqXqXqXqX9XiXz.e 0 i.uX0XqXqXqXqXqXqX0XrX$X+XrX0XqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqXqXqXwXeX8X>X-X2XwX5X_.VXVX9XqXqXqXqXqXqXqXqXqXqXqX0XaXpXiXaXqXqXqXqXqXqXqXqXwX7X].5XwXqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXOXoXtX0XqXqXqXqXqXqXwXwX8X>X>XOX2XUXiX9XqXqXqXqXqXqXqXqXqXqXqXqX9X0X0X9XqXqXqXqXqXqXqXqXrX@XOXrX0XqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XtXoXoXtX0XqXqXqXqXqXqXqXqXwXeX7X:XCXvX6XeXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX5X'.5XwXqXqXqXqXqX0XeXPXuX0XqXqX",
"qXqX0XyX#XOXyXqXwXwXwXwXwXwXwXwXwXwXwXtXyX;X$X,X0XrXeXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXtX+X+XtXqXwXwXwXwXwXqXrXPXuX0XqXqX",
"qXqXwX9XE. XqX6X7X7X7X7X7X7X7X7X7X7X7X6X5X6X,X@XoX*X3X6X7X7X7X7X8X7X7X7X7X7X7X7X7X7X7X7X7X6X8X:X~.2X8X6X7X7X7X7X7X6X0XPXuX0XqXqX",
"qXqXqXwXiXdXBXBXBXBXBXBXBXBXBXBXBXBXBXBXVXVXCXBXcXkXxXBXBXBXVXMXvXNXBXBXBXBXBXBXBXBXBXBXBXBXVXzXhXVXBXBXBXBXBXBXBXVXmXPXuX0XqXqX",
"qXqX0XuXzXdXUXmXBXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXVXCXVXBXNXNXZXtXNXLXmXBXNXNXNXNXNXNXNXNXNXBXNXVXCXNXNXNXNXNXNXNXNXVX9XHXiX9XqXqX",
"qXqXqXqX3X8XrX6X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X6X6X6X7X7X7X7X3XwXqX6X7X7X7X7X7X7X7X7X7X7X7X7X6X6X7X7X7X7X7X7X7X7X6X4XyXwXqXqXqX",
"qXqXqXqXeXqX0XwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXeX0XqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXeX0XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_4[] = {
/* columns rows colors chars-per-pixel */
"48 48 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXeXwXwXwXwXwXwXwXeXwXwXwXwXwXwXwXwXwXwXwXwXwXwXeXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXeXqXqXqXqX",
"qXqXqX0X3X5X6X6X6X6X6X6X5X5X6X6X6X6X6X6X6X6X6X6X6X6X5X4X5X6X6X6X6X6X6X6X6X6X6X6X6X6X6X5X0XqXqXqX",
"qXqXqX0X:XXX X X X X.X|.+X+X X X X X X X X X X X X XoX#XXX}. X.X X X X X X X X X X.X}.>XrX0XqXqX",
"qXqX0XyXdX8XrXwXwXwXwXwXZXpXqXeXwXwXwXwXwXwXwXwXwXwXrXdXzXfXeXqXeXeXeXwXwXwXwXwXwXeXqXAXuX9XqXqX",
"qXqX0XrX|.<XeX0XqXqX8XjXbX7XqXqXqXqXqX9X8X0XqXqXqXqX0X9XwXdXjXaX0X9X0XqXqXqXqXqXqXqX9XZXyX0XqXqX",
"qXqX0XrXoX1XeXqXqXqXqXCXyX0XqXqXqXqX0XaXjXwXqXqXqXqXqXqXqX9XyX4XsXxXwX9X0XqXqXqXqXqX9XZXuX0XqXqX",
"qXqX0XrXoX1XeXqXwX8XlXnX8XwXqXqXqX0XiX).m.8XeX0XqXqXqXqXqXqX0X>XSXjXlXfXeX9X0XqXqXqX9XZXuX0XqXqX",
"qXqX0XrXoX1XeXqXqXwXAXyX0XqXqXqX9XdXJ.s H / aX9XqXqXqXqXqXrX>XfXnX5XrXdXlXfXeX9X0XwX0XZXuX0XqXqX",
"qXqX0XrXoX1XeXqX8XxXnX8XwXqXqXqX0XuX].1XU.f aX0XqXqXqXqXqX0X:XSXrXqXqX9XwXsXlXfXeX0X9XZXuX0XqXqX",
"qXqX0XrXoX1XeX0XeXDXyX0XqXqXqXqXqX7XzX#Xn .XaX9XqXqXqXqXrX;XdXnX7XwXqXqXqX9XwXsXkXfXwXZXuX0XqXqX",
"qXqX0XrXoX1XrX7XvXmX8XwXqXqXqXqX9XiX`.  U G.7XwXqXqXqXqX9X-XDXrX0XqXqXqXqXqXqX9XwXsXjXFXyX0XqXqX",
"qXqX0XrXoX2XwXrXFXtX0XqXqXqXqXqX9XpX!.#.&.<.9XwXqXqX0XrX=XdXnX7XwXqXqXqXqXqXqXqXqX0XqXAXyX0XqXqX",
"qXqX0XrXoX2X0XnXmX0XwXqXqXqXqXqXqX0XiXxXzXlXqXqXqXqXwX9X*XDXrX0XqXqXqXqXqXqXqXqXqXwX9XZXuX0XqXqX",
"qXqX0XrXoX1XtXnX,X>XqXeXqXqXqXqXqXqX0X7X7X8XqXqXqX0XtX&XsXmX7XwXqXqXqX0X8X8X0XqXqXqX0XZXuX0XqXqX",
"qXqX0XrXoX>XzXDXaX5X*X<XqXeXqXqXqXqXqXqXqXqXqXqXqXwX8X&XDXrX0XqXqXqX0XyXzXkXeXqXqXqX9XZXuX0XqXqX",
"qXqX0XrXXX1XHXsXeXhXdX3X*X2XwXeXqXqXqXqXqXqXqXqX0XtX$XsXmX7XwXqXqX0XtX$XO.>.3XeX0XqX9XZXuX0XqXqX",
"qXqX0XtX XaXSX8XqX9XtXgXaX2X=X3XwXwXqXqXqXqXqXqXwX7X$XDXrX0XqXqXqX9XfX>.3.I #.lX8XqX9XZXuX0XqXqX",
"qXqX0XrXXXnXdX9XwXqX0X9XtXgXiX<X=X4XeXwXqXqXqX0XtX@XaXmX7XwXqXqXqXqX9XhXY.p S.dX9XqX9XZXuX0XqXqX",
"qXqX0XrXOXpXwXqXqXqXqXqX0X9XtXfXyX,X-X5XeXwX0XwX6X#XFXeX0XqXqXqXqXqXwX`.OX{ / fX9XqX9XZXuX0XqXqX",
"qXqX0XrXOX2XeXqXqXqXqXqXqXqX0X9XyXdXrX>X;X6XwXtXOXaXmX7XwXqXqXqXqX9XdX%.C M ;.dX9XqX9XZXuX0XqXqX",
"qXqXqXeX{.1XeXqXqXqXqXqXqXqXqXqX0X0XyXdXwX:X>X2X#XGXeX0XqXqXqXqXqXqXwXrXI.R.uXqXqXqX9XZXuX0XqXqX",
"qXqXqX0X&X7XwXqXqXqX9X7X9XqXqXqXqXqX0X0XyXaXqXOXaXzX8XeXqXqXqXqXqXqXqXqXdXsX0XqXqXqX9XZXuX0XqXqX",
"qXqX0XyXgX7XwXqXqX0XsXxXdX0XqXqXqXqXqXqX0XwX9XOXMX3X*X5XwXwXqXqXqXqXqXqX9X9XqXqXqXqX9XZXuX0XqXqX",
"qXqX0XrX|.<XeXqX0XiXV.J n.iX0XqXqXqXqXqXqXtXXX7XVXlXpX,X-X6XeXwXqXqXqXqXqXqXqXqXqXqX9XZXuX0XqXqX",
"qXqX0XrXoX1XeX9XpX`.e L.p W.aX9XqXqXqXqXwX4X.XnX9XqXdXzXpX<X;X6XeXwXqXqXqXqXqXqXqXqX0XZXuX0XqXqX",
"qXqX0XrXoX1XeX9XiX}.f F.p y.jX8XqXqXqX0XtX.XeXkX8XqX9XqXsXzXiX<X:X7XeXwXqXqXqXqXqXeX3XnXiX9XqXqX",
"qXqX0XrXoX1XeX0XwX7X=.w.h M.gX9XqXqXqXeX3XoXMXqXqXqXqXqX9XqXsXlXiX1X>X8XeXwXqXqXqXrX+XcXpX9XqXqX",
"qXqX0XrXoX1XeX0XeX4XT Z U 7XwX0XqXqX0XrX.XuXkX8XqXqXqXqXqXqX9XqXaXkXuX1X,X8XeXqXeX4XXXAXyX0XqXqX",
"qXqX0XrXoX1XeXqX0XtX4XY.7XtX0XqXqXqXeX2X+XBX0XqXqXqXqXqXqXqXqXqX9XqXaXkXuX2X<X8XrX+X>XDXyX0XqXqX",
"qXqX0XrXoX1XeXqXqX0XeXdXwX0XqXqXqX0XrX XsXkX8XqXqXqXqXqXqXqXqXqXqXqX9XqXpXjXyX4X<X+XqXZXuX0XqXqX",
"qXqX0XrXoX1XeXqXqXqX0X9X0XqXqXqXqXeX1X#XCX0XqXqXqXqX0X0XwX9XqXqXqXqXqXqX9X0XaXdX*X6XqXZXuX0XqXqX",
"qXqX0XrXoX1XeXqXqXqXqXqXqXqXqXqXqXrX XgXkX8XqXqXqX0XrXiX7XsXqXqXqXqXqXqXqXqXqX1X&XrX9XZXuX0XqXqX",
"qXqX0XrXXX1XtXqXqXqXqXqXqXqXqXqXeX<X*XAX0XqXqXqX0XwX6X] m 0.yX0XqXqXqXqXqXqXeX$X6XeX9XZXuX0XqXqX",
"qXqX0XrXoX@X4XqXwXqXqXqXqXqXqXqXrX XlXjX8XqXqXqX8XjXy.b 6.<.rXqXqXqXqXqXqXeX1X%XrXqX0XZXuX0XqXqX",
"qXqX0XrXoX,X2X;X3XqXwXqXqXqXqXrX>X-XSX9XqXqXqXqX8XkX..M :.K 8XwX0XqXqXqXqXwX+X6XwXqX9XZXuX0XqXqX",
"qXqX0XrXoX2XtX8X<X:X4XwXwXqXqXeX XcXjX8XqXqXqXqX8XlX2.H ,X6 *XyX0XqXqXqXeX,X$XrX0XqX9XZXuX0XqXqX",
"qXqX0XrXoX1XeXqXeX8X<X>X5XqXtX:X>XDX9XqXqXqXqXqX0XeX1X^ g a.pX9XqXqXqXqXwXOX6XwXqXqX9XZXuX0XqXqX",
"qXqX0XrXoX1XeXqXqXwXeX8X<X,X7X.XmXhX8XqXqXqXqXqXqX0XrXiX0XaXqXqXqXqX0XrX:X$XrX0XqXqX9XZXuX0XqXqX",
"qXqX0XrXoX2XrXqXqXqXqXwXeX9X:XwXCX0XeXqXqXqXqXqXqXwX0X0XwX9XqXqXqXqXwXqXXX6XeXqXqXwX0XZXuX0XqXqX",
"qXqXqXeX}.,XqX8X9X9X9X9X8X0X9XtX,X%X1X8X9X9X9X9X9X9X9X8X8X9X9X9X9X8XqX&X+XwX8X9X9X9X7XCXuX0XqXqX",
"qXqXqX0X$XeXdXsXsXsXsXsXsXsXsXaXsXyX5X8XiXdXsXsXpXsXsXsXsXsXsXsXsXsXsX>XyXdXsXsXsXsXaXDXyX0XqXqX",
"qXqX0XtXjXZXZXBXVXVXVXVXVXVXVXVXVXCXCXBXBXBXZXkXVXAXBXVXVXVXVXVXVXVXBXVXVXVXVXVXBXZXlXMXuX0XqXqX",
"qXqXqXqX6XqX9X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X6XqX9X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X6XqXwXqXqXqX",
"qXqXqXqXwXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_5[] = {
/* columns rows colors chars-per-pixel */
"44 44 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXqXqXqXqX",
"qXqXqX0X1X4X4X4X4X4X4X3X3X4X4X4X4X4X4X4X4X4X4X4X3X3X4X4X4X4X4X4X4X4X4X4X4X4X4X3X0XqXqXqX",
"qXqXqXqX1XOXOXOXOXOXoX&X&XoXOXOXOXOXOXOXOXOXOXOX$X=X+XXXOXOXOXOXOXOXOXOXOXOXXX<XtX0XqXqX",
"qXqX0XyXqX9XrXeXeXrXwXVXsXwXrXeXeXeXeXeXeXeXeXeXyXgXlXdXrXrXeXrXeXeXeXeXeXrXqXCXpX9XqXqX",
"qXqXqXwX|.5XwX0XqX8XgXvX7XqX0X0XqX9X9X8X0XqX0XqX0X9XwXsXdX0XeX9X0XqXqX0X0XqX8XNXiX9XqXqX",
"qXqXqXeXOX6XwXqXqX0XBXyX0XqXqXqXqXdXiXsXwXqXqXqXqXqX0XwX<XdXBXpXwX9XqXqXqXwX8XNXiX9XqXqX",
"qXqXqXeXoX6XwXqX8XjXbX8XwXqXqX9XiXd.N 0.rX0XqXqXqXqXqX0X1XCXtXdXkXsXwX9XqXwX9XNXiX9XqXqX",
"qXqXqXeXoX6XwXqXqXCXyX0XqXqXqXqXqX3.(.i @XuX0XqXqXqXrX,XgXxX7X0XeXsXkXsXwX0X8XBXiX9XqXqX",
"qXqXqXeXoX6XeX8XlXnX8XwXqXqXqXqX9XSXp.-.aX9XqXqXqXqX0X,XCXqXqXqXqX9XwXsXjXdX0XNXiX9XqXqX",
"qXqXqXeXoX6XwXwXAXyX0XqXqXqXqX9XsX#., F.4XwXqXqXqXrX:XfXxX7XwXqXqXqXqX9XwXsXgXAXuX9XqXqX",
"qXqXqXeXoX7X9XxXnX7XwXqXqXqXqXqXeX4.#.$.*XtX0XqXqX9X>XCXqXqXqXqXqXqXqXqXqX0X0XVXiX9XqXqX",
"qXqXqXeXoX5XrXCXyXwXqXqXqXqXqXqXqXjXlXlXrX0XqX0XrX-XdXxX7XwXqXqXqXqXqXqXqXwX8XNXiX9XqXqX",
"qXqXqXeXOX3XjXhX*X2XwXeXqXqXqXqXqX8X7X7X0XqXqXqX9X;XZXqXqXqXqXqX9XeX0X0XqXwX9XNXiX9XqXqX",
"qXqXqXeXoX5XAXxXaX5X-X3XwXwXqXqXqXqXqXqXqXqX0XrX*XdXxX7XwXqXqXqXsX4XyXrX0XwX8XNXiX9XqXqX",
"qXqXqXeX XsXZX7XrXfXaX3X-X4XwXwXqXqXqXqXqXqXwX8X=XZXqXqXqXqX9XaXt.S { 3XeXqX8XNXiX9XqXqX",
"qXqXqXeXoXBXsX9XqX9XrXdXiX2X;X5XeXwXqXqXqX0XtX%XsXxX7XwXqXqX0XeXJ.V.c $XyXqX8XNXiX9XqXqX",
"qXqXqXqX$XlX0XqXqXqX0X9XrXdXyX1X:X6XeXwXqXwX7X&XZXqXqXqXqXqXqX8XgXc.M +XyX0X8XNXiX9XqXqX",
"qXqXqXwX+X9XwXqXqXqXqXqX0X9XtXsXrX,X>X7XwXtX#XsXxX7XwXqXqXqX0XyX[ 4.f P.sX0X8XNXiX9XqXqX",
"qXqXqXwX|.5XwXqXqXqXqXqXqXqX0X0XtXaXwX>X<X3X&XSXqXqXqXqXqXqX0XtXoX3.I.uX0XwX8XNXiX9XqXqX",
"qXqXqX9X*X9XwXqXqX9X9X9XqXqXqXqX0X0XtXpXwX+XaXhX9XeXqXqXqXqXqX0XuXxXsX0XqXwX8XNXiX9XqXqX",
"qXqX0XyXuX7XwXqX0XpXsXdXwXqXqXqXqXqX0XqX0X#XmX2X;X6XeXwXqXqXqXqX0X8X9XqXqXwX8XNXiX9XqXqX",
"qXqXqXwX|.6XwX9XiXR.F -.0XwXqXqXqXqX0XtX+X8XnXjXpX1X:X7XeXwXqXqXqXqXqXqXqXwX8XBXiX9XqXqX",
"qXqXqXeXOX6XwX9XaXj F.J h.hX8XqXqXqXwX6XOXvX8XqXaXkXiX1X>X7XeXwXqXqXqXqXqXwX8XMXiX9XqXqX",
"qXqXqXeXoX6XwX9XaX3.&.Z 7.kX8XqXqX0XtX+XrXgX9XqX9X0XpXkXiX2X,X8XeXqXqXqXqXeX<XhXsX9XqXqX",
"qXqXqXeXoX6XwX9XsXw. .m Y.sX9XqXqXwX5X@XnX0XqXqXqXqX9X0XpXjXuX2X<X9XeXqX0XrXOXzXaX9XqXqX",
"qXqXqXeXoX6XwX0XyX[.o.b.iX0XqXqX0XtXOXiXgX8XqXqXqXqXqXqX9X0XiXhXuX3X1X9XrX3XOXCXiX9XqXqX",
"qXqXqXeXoX6XwXqX0XuXzXfX0XqXqXqXeX4X%XMX9XqXqXqXqXqXqXqXqXqX9X0XiXgXyX3X7X#X<XVXiX9XqXqX",
"qXqXqXeXoX6XwXqXqX0X7X9XqXqXqX0XrXOXsXgX8XqXqXqX9XqXqX0XqXqXqXqX0X0XuXgX4X-X0XNXiX9XqXqX",
"qXqXqXeXoX6XwXqXqXqXqXqXqXqXqXeX3X*XNX9XqXqXqX0XiX0XwXuX0XqXqXqXqXqX9XqX&X8X9XNXiX9XqXqX",
"qXqXqXeXoX5XrXqXqXqXqXqXqXqX0XrXOXhXfX8XqXqX9XpX/.P / OXyX0XqXqXqXqXeX1X-XrX8XNXiX9XqXqX",
"qXqXqXeXXX&X3XqXwXqXqXqXqXqXeX1X;XVX8XqXqXqX9XfXR ! ' }.uX0XqXqXqXqXwX$X7XwX8XNXiX9XqXqX",
"qXqXqXeXoX4X2X:X4XqXwXqXqXqXrXOXlXfX9XqXqXqX0XiXl p.` s.hX8XqXqXqXeX,X=XrXqX8XNXiX9XqXqX",
"qXqXqXeXoX6XrX9X1X>X5XqXwXeX<X,XCX8XqXqXqXqX8XhXX.O.Y k.gX8XqXqXqXqX@X8XwXqX8XNXiX9XqXqX",
"qXqXqXeXoX6XwXqXeX9X2X,X5XeXoXvXdX9XqXqXqXqX0XeX8Xe.b.uX0XqXqX0XrX>X*XrX0XwX8XNXiX9XqXqX",
"qXqXqXeXoX6XwXqXqXwXeX9X4X=X9XAX0XwXqXqXqXqXqXqXwXlXgX0XqXqXqXqXqXOX8XwXqXwX9XNXiX9XqXqX",
"qXqXqXeXXX5XqX0X0X0X9X0XqX6XaX3X<X0XqX0X0X0X0X0X9X7X7X0X0X0X9XeX=X%XeX9X0X0X7XNXiX9XqXqX",
"qXqXwX9XoXwXuXyXyXyXyXyXyXuXrXqX3X1XqXuXyXuXtXyXyXyXyXyXyXyXuXtX&XeXuXyXyXuXtXCXiX9XqXqX",
"qXqX0XtXgXZXVXBXBXBXBXBXBXBXVXCXCXMXMXBXCXzXBXZXBXBXBXBXBXBXVXNXNXVXBXBXBXCXvXMXiX9XqXqX",
"qXqXqXqX8XeX9X9X9X9X9X9X9X9X9X9X9X0X0X9X0X7XwX0X9X9X9X9X9X9X9X9X0X9X9X9X9X0X7XqXwXqXqXqX",
"qXqXqXqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_6[] = {
/* columns rows colors chars-per-pixel */
"32 32 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXqXqX",
"qX0X3X4X4X4X4X3X3X4X4X4X4X4X4X4X4X4X3X3X4X4X4X4X4X4X4X4X4X4X0XqX",
"qXwX3X%X&X&X%X;X1X%X&X&X&X&X&X&X&X*X1X:X%X&X*X&X&X&X&X*X$X5XtX0X",
"qXwX,XwXrXrXqXbXfXwXrXrXwXwXwXrXrXrXuXgXhXyXqXeXrXrXrXrXqXnXpX9X",
"qX0X+X0XqX9XuXcX8XqXqX0XaXpXuX9XqX0X0X9XeX0XzXtX0X9XqXqX8XxXiX9X",
"qXqX$X0XwX9XvXiX9XqX9XiXh.' _.iX9XqXqXeX2XfXhXaXgXyX0X0X8XcXiX9X",
"qXqX#XqX9XpXbX9XwXqXqX9X X4.9.lX8XqXqX0X6XbX8X0XrXdXfXuX7XxXiX9X",
"qXqX#XqX9XnXiX9XqXqX9XpXI.I ;XrX0XqXeX1XfXsX9XwX0X0XrXdXaXnXiX9X",
"qXqX$X8XdXmX8XqXqXqX0XiX,.( P.iX9XqX0X4XbX9XqXqXqXqX0X0XqXmXiX9X",
"qXqX$X9XxXeXqXwXqXqXqX0XfXfXrXqXqXeX<XdXsX9XqXqX9X8XqXwX8XxXiX9X",
"qXqX@XuXnX4X<X7XeXwXqXqX9X8X0XqXqX9X2XbX9XwXqXwXfXfXeXqX8XcXiX9X",
"qXqXOXbXhXuXuX4X<X7XeXwXqXqXqXqXrX>XdXdX9XqXwX9Xy.9.1XrX8XcXiX9X",
"qX9X:XMX9X0XeXpXtX3X1X8XeXqXqXqX9X<XbX9XqXqXwX5XZ.F `.aX7XcXiX9X",
"qX9X,XiX9XwXqX9XeXiXrX2X1X9XwXrX;XsXdX9XqXqXwX4X^.P H.fX6XcXiX9X",
"qX0X+X0XqXqXqX0XqX0XeXiXwX2X3X7X,XnX8XqXqX0XeX2X-.O.XXuX7XcXiX9X",
"qX9X;XqXqX0XwXyX0XqX0X0XrXyXwX&XpXiX0XwXqXqX0XrXiXuXuXqX8XcXiX9X",
"0XrX9X0XqXtX6X+XiX0XqXqX0X0XwX-XcX3X1X8XeXqXqX0X0X9X0XwX8XcXiX9X",
"qX0XOX0XwX8X+./ H.sX9XqX0XrX=X0XgXdXuX4X2X9XeXqXqXqXqXwX8XvXiX9X",
"qXqX$X0XeX1Xo.*...kX8XqXwX8X=XjX8X0XyXdXuX5X3X9XwXqXqXwX7XdXpX9X",
"qXqX#X0X0XtX<.m v.fX8X0XrX=XtXiX9XqX0X0XtXsXyX5X4X0XwXrX;XuXsX9X",
"qXqX#X0X0XtX`.D.wXqXqXwX7X;XlX9XqXqXqXqX0X0XtXsXyX6X5X0X@XxXpX9X",
"qXqX#X0XqX0XpXgXqXqX0XrX*XiXiX9XqX0XqXwX0XqX0X0XtXaXuX,X:XbXiX9X",
"qXqX$XqXqXqX0X8XqXqXwX7X>XzX8XwX0XtX0X5XrX0XqXqX0X0XeX:X9XcXiX9X",
"qXqX@X8XeXqXqXqXqX0XrX*XdXiX9XqXeX2X} X.2XeX0XqXqXrX>X2X0XxXiX9X",
"qXqX+X,X2X9XwXqXqXwX5X1XcX8XwX9XgXB.G @.|.uX0XqXqX9X*XwX8XcXiX9X",
"qX0X$XqX6X<X3X0XwXrX*XhXuX0XqX9XdXP.) _ L.dX9X0XrX:X1XrX8XcXiX9X",
"qXqX#X0XwXwX6X1X6X3X5XvX7XwXqXqX0XyXN.b.rXqXqXwX8X&XwXqX8XcXiX9X",
"qXqX$X0XqXqXwXqX7X>XjXrXqXqXqXqXqX0XfXfXqXqX0XrX-X1XeXqX8XcXiX9X",
"wX9X+XqXwXwXqXwXwXrX0X:X3X0XwXwXqXwX9X9XwXqXeX7X%XeXqXwX9XvXiX9X",
"qXwXaXNXbXnXnXnXnXbXnXbXlXcXmXzXnXmXnXnXnXnXnXcXxXmXnXnXcXnXuX9X",
"qXqXwXeXqXwXqXqXqXqXqXwXeXwXwX0XeXwXqXqXqXqXqXwXwXqXqXwXqXqXwXqX",
"qXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_7[] = {
/* columns rows colors chars-per-pixel */
"24 24 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"0X3X2X2X2X1X2X2X2X2X2X2X2X2X1X1X2X2X2X2X2X2X3XqX",
"wX2X>X,X>X0X1X>X,X;X:X,X>X3X7X1X>X>X,X,X,X:XeXrX",
"8X:XtXwXpXxXwXeXtXjXsXeXrXeXuXaXuXiXwXeXrXwXxXtX",
"7X-XeX8XlXtX9XwX6Xa.I.yX9X0XqX4XjXdXpXeX0X8XkXtX",
"7X;XqXuXlX9XqXwX8Xn.4.aX9XwX8XrXfX7XtXaXaXwXkXtX",
"7X;XqXcXyX0XqXwX9XW c.pX9XwX4XhXeXqX0X0XrXpXvXrX",
"7X=XaXjX9XwX0XwX5X/.XXwXwX7XwXgX8XqX0X0X0X8XlXtX",
"8X*XcXeX2X8XwXqXwXaXuX0XeX2XgXeX0XwXyXiXwX9XkXtX",
"5X2XbXrXyX6X3X9XwX0X0XwX7X0XgX8XqXrXf.t.rX9XkXtX",
"5XqXpX8XwXyXrX6X4X9XwXeX<XgXeX0XqXeXW.~ 5XqXkXtX",
"6X-XeXqX9X8XwXtXwX5X6X5X9XhX8XqXqXqXy.@.5XqXkXtX",
"8X1XwX0XpXaX0X0XwXtXwX:XaX0XwXqX0XeX3X8XrX8XkXtX",
"qX4XqXtX!.Q.yX0XqXqX8X3XjX5X4X9XwXqXeXwXqX9XlXtX",
"6X*XtX2XO./ @XuX0XrX:XtXtXiXyX8X5X0XwXqXqX9XhXtX",
"7X-XwXeX,.Y :XtXqX6X3XsX8X0XeXiXyX8X6X0XrX3XyXuX",
"7X-XwXrX'.[.uX9XeX>XpXwXqX0X8X9XeXuXyX8X9X*XgXyX",
"7X;XeX0XiXiX0XeX5X6XsX9X0XuXdXwX0X0XwXiX8X>XzXtX",
"7X=XeXwX9X9XqXeX,XdXwX9XiX^.i.8XwXqX0XqX,X8XlXtX",
"8X%X3X7XwXqXwX4X9XdX8X0XyX[ !  XiX9XeX4X1XqXkXtX",
"7X;XqX3X3X8XeX,XhXqXqX0XuXq.' [.pX9XwX;X0X0XlXtX",
"7X-XeXwX9X5X,XtXdX9XqX0XeX5X@XtX0XeX2X,XeX8XkXtX",
"6X=XrXqXwXeX0XrX1X8XwXwXqXeXuXqXwXqX=XqXwX0XzXtX",
"0XsXcXlXzXlXzXlXhXgXzXhXzXlXlXzXzXkXhXzXzXlXlXrX",
"qXwXqXqXqXqXqXqXwXwXqX0XwXqXqXqXqXqXwXqXqXqX0XqX"
};

/* XPM */
static const char *const xpm_icon_8[] = {
/* columns rows colors chars-per-pixel */
"16 16 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qX7X5X5X5X6X4X5X6X6X5X6X6X5X6XqX",
"qX2X2X8X7X2X0X6X2X9X6X2X2X1X0XtX",
"8X2XrXhXeXrX).>XyXqXuXfXuXqXsXuX",
"8X,XsXiX8XyX3.(.aX7XiXwXeXyXfXyX",
"8X1XgX0XwXqXY.,XeXqXiX0XiXqXsXuX",
"6X0XdX6X8XqXsXrX7XuXeX9X`.4XaXyX",
"7XqXrXqXqX8X6X9X0XiX0XqX&.'.kXtX",
"9X3XeXaXtXwXqX4XrXwXwX8XR.,XdXyX",
"9X6X9XV.*XtXqX6XiX8X8XqXaXqXaXuX",
"7X2X8X( R.dX3XqXeXwXeX0X7X9XwXyX",
"8X2XeX$X8XqX6XyXqXaXeXrXrX4XqXiX",
"8X<XwXyXeX5XrXrX3XG.6XqXwX2XpXuX",
"8X,X7X8X9X0XyXuX XT %XiX6X3XsXyX",
"7X<XqX6X3XrX0X0X8X X6X0X,X7XpXuX",
"9XeXfXsXdXiXuXsXsXkXfXiXyXsXgXyX",
"qXtXrXrXrXtXtXrXrXeXrXtXtXrXeXqX"
};

const char *const *const xpm_icons[] = {
    xpm_icon_0,
    xpm_icon_1,
    xpm_icon_2,
    xpm_icon_3,
    xpm_icon_4,
    xpm_icon_5,
    xpm_icon_6,
    xpm_icon_7,
    xpm_icon_8,
};
const int n_xpm_icons = 9;
