#include "gtk.h"
/* XPM */
static const char *const xpm_icon_0[] = {
/* columns rows colors chars-per-pixel */
"128 128 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX8X7X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX8X[._.<XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXVXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXJXNXrXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X_.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXAX4X0XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXMXfXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXaX3X}.8XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXlX8X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX<X'._.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXyXpXpXiXwXqXqXqXqXqXqXyXiXpXuXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX<XP.e.w.g.;XeXqXqXqXwX7XL.u.9.k.[.eXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqX0X@Xf.M     8 '.tXqXqXqX0XXXA X   e g.yXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqX1X2.7 @ & o e ].tXqXqXeX>Xy.*     i j.tXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXwX].I 7 G C O e ].tXqXqXqX~.a @ #   i j.tXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXwX'.4.h.Q.Q + e ].tXqXrXOX$.& i I 4 y j.tXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqX3X*X3X}.~ + e ].tXwX9XN.3 = e.@.5 y j.tXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXyX{.~ + e ].tXyX(.R # Z |.,.3 y j.tXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX{.~ + e ].yX2X9.= 7 b.2X1.3 y j.tXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX{.~ + e ].iXG.m $ _ 5XeX4.4 u c.aXwXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX{.~ + e ].0X@.o 4 ..B.M.Q : 5 ..Q.6XqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX{.~ + e [.$XI                   M =X0XqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX{.~ + e [.&X` ; 1 1 1 1 % X o # Y ;X0XqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX{.~ + e ].eX>X%X&X&X&X@X..> e e.1X9XqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX[.! o q ].tXwXeXeXeXeX9X&.> r k.uXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX1X/.C.J.1XwXqXqXqXqXqX0X[.D.K.@XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXtXuXuXwXqXqXqXqXqXqXqXrXuXuXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.>XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.4XaXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXGXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.jXUXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXPXZXLXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXUXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXDX3XAXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXBXiXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXyX1X{.BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXnX9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX4X'.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX3X/.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXxXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXrX<X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXhX:X*X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X=X+X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).MXCX4X`.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.~._.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).bXiX&XoXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOX$X9XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).NXUXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXKXMXeXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXVX,X0XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXnXiXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXtX:X{.8XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrXuXiXuXeXqXqXqXeXtXuXuXuXtXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX<X_._.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrX8XI.*. .$.W.tXwX9XoXh.;.+.;.l..XrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX%Xh.D .   # s.pX6X^.} t     X u #._.0XqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXwX4Xv.e $ 1 O ; g.sX.X}   * c G g $ ; X.4XqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXeX$X<.7 T E ; - g.fXT.k 0 W '.0XD.N   a *X0XqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXeX%Xc.w.$X-., = g.aX+Xd.h.'.5XwX.X'   a oX9XqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X3X3X<X:., = g.iX9X3X4X7X6X-Xy.a $ T 3XqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX1X:., = g.iXqXwX0XoXK y 2 w o.'.0XqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX1X:., = g.iXqXqX9X Xb #   + C N.0XqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX1X:., = g.iXwXtXeX1XS.g.=.g + D ~.7XqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX1X:., = h.tX=X).].3XyXaX3X=.  , 0.2XwXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX1X:., = h.rXP.! ] N.:XqX1X>.  , >.1XwXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX1X:., = h.pXV.> o x d.P.w.l   e n.5XwXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX1X:., = g.aXXX+.: # @ @ @ @ 2 ) >X0XqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX1X<.4 < j.iX0X#Xy.C 4 = 5 S u.&X0XqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X%X`._.;XwXwX0X>X.X`.).`.XX>XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXtXtXeXqXqXqXeXrXtXtXtXrXeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXrXuXiXuXeXqXwXyXiXiXiXiXiXiXuXrXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXtX2XS.@...=.}.uXOXs.|  . . . .} 2.T.wXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXXXt.n     : G.sXY.C             5 r.eXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXwX1Xe.6 $ < o 3 J.dXH.v + i D H G D [ F.wXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXeX.X| q W I $ 3 J.dXV.i 3 t.6XyXtXtXeXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXeXoXd.g.XX .& 3 J.fXv.6 5 g.#X*X*X,X7XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X2X6X%X..& 3 J.fXg.- 1 O.A w u o.^.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXrX%X..& 3 J.gXy.  X , 5 ; O & T E.wXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX%X..& 3 J.sXE. .^ ;.Z.F.~ - * O.0XqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX%X..& 3 J.iX4XOX].&XrXuXY.N   C *X0XqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX%X..& 3 J.wX@X^.}.0XqXwX:X~   n .X0XqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX%X..& 3 K.8Xl.K } !.9XeXE.C   S ;X0XqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX%X..& 3 J.tXq.@ * / P.E.( - - #.wXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX%X..% 2 J.aX~.Q & # % % # # Y /.wXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX%XO.: 8 L.uX9X[.>.h > > g >.`.9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX4X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX7X#X).`.1XwXwX8X=X}._._.}.=X7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXmX0XwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXeX5X(.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXtXtXwXqXqXqXeXtXtXtXtXeXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXUXcX;X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X#XoX/.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.lXCX5X'.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.!.~.^.).BXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.uXtX#X X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.X.XOXCXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXzX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX,X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXxX0XwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXrX<X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.,XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXFXUXgX=X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X@X#X.X^.`.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.>XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXSXbX>X).Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.Q.!._.7XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X`.^.3XdXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXDXsX<X-X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X;X<X0XqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX7X'. XsXJXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXLXtXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqX8XoXpXDXLXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXKXLXGXrXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqX8XwXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_1[] = {
/* columns rows colors chars-per-pixel */
"96 96 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXqXqXqX",
"qXqXqXwXtXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX8X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X7X0XqXqXqX",
"qXqXwX6X=X5XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXhXjXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXgXlXuX0XqXqX",
"qX0XtX;XL.=XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXSXeXqXqXqX",
"qX0XrX>XR.;XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXFXhXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXlXzX#X=XtX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXZX3X8X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X6X0X=XT.:XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX8XeXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXyX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqX0XqXtXyXtXqXqXqXqXqXwXtXyXrXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqX0XtXrX<X;X<XqXqXqXqXwX8X>X;X2XqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqX0XqX9Xj.%   i uX0XqX6XgX8.    L eXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqX0X9XrXO.  *   f uX0X8XiX Xt     T eXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqX9XrX=Xn U *.X a uX0X8XuX .  0 X I eXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqX0XeX<XG.rXf.  a iX6XfXG.  M ,.$ L eXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXwXiXpXw.  a yXrX,XH   K.s.  P eXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqX6XiXt.  a yXiXf.  E rXw.  L 9XqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqX8XpXt.  p jX.X3 4 '.JXi.. E kXwXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqX9XpXt.  a fXp.  = 8 w > . O j 3XqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqX9XpXt.. s sXy.  #       X   - 1XqX0XqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqX8XaX0.  q iX0XqXqX8XgX<.  F rXwXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqX9XiXc.a I yX0XwXwX9XsXf.p O.eXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqX0XiXlXhX0XqXqXqXqX0XpXlXfXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XY.;XyXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXtXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqX9X6X7XqXqXqXqXqXqX9X6X8XqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.*X0X6X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X5X9XUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XI.9XAXnXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXNXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.sXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXyXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXgXuXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXiXaX2X].UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXuX6X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X8X0X/.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXpX9XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXZX6XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XrX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXZX9XyXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXpX:XY.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX+X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X.X{.W.:XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.[.SX+XW.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.E.Q.W.:XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.].UXFXHXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXJXGXaX9XqXqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqX9X9X9X0XqXqX0X9X9X9X9XqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXUXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXPXUXuX,XrXqXqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqX0X9XpXjXhXtX0X0XqXaXgXgXpXqX0XqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXZX5X0X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9XqX2X/.-XtX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqX0XeXsXoXH.J.;XrXeXwX[.R.R.|.wXtX0XqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XY.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXwX9XC.r     N.kX4X1.,     4 i.7XwX0XqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXyXG.  * q   S.bX=.  : *.' #   C.gX8XqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXwXwXw.A C.z   S.mX{ S z.kXuXo.  2.dX9XqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqX+X1X-Xh   G.jX-X-XyXxX%XZ   d.hX8XqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXrXfX{.k   G.hXwXpX).s 0 : 9.5XeX0XqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqX8XuX.Xk   G.jX8XtX~.* X   Z .XiX9XqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqX9XiX.Xk   G.hX9XwXwX6X>X`   ~ 1XrX0XqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqX9XiX.Xk   G.pX<.*..XsXMXk.  l XXiX9XqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqX9XiX.Xx   F.cXp   n L.m.j   ^ 3XeX0XqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqX9XpX}.w   C.cXI.g         z ~.aX9XqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqX0XtX=X3.^ ].pXiXoXq.O.+.t.=XpX8XqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqX9X5X5X7XwX0X6X5X5X5X5X5X8XqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqX0XrXzXmXuX0X9XrXkXvXvXjXwX9XwXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqX7XrXvXnXlX9XrXlXnXbXbXnXnXfX0XqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqX0X6X5X9XqXqX9X6X5X5X7X9XqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqX9XgXXX..W 5.fX*X<.Q ] ' ` ` M.tX0XqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqX9XtXm.i     f bX(.5           ( pX9XqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqX0XwX:Xi o A . V nX!.< j I. X'.'.*XwXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqX9XtXoX{ `.6.  m MXU.  I 0XjXsXhXiXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqX9X9XsX1.  m VXC.  V K . < w.qXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXuX<.  m VXN.  = s b +   6.iX0XqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqX0XuX<.  N jX2XU.H.8XnX0.  2 9X0XqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqX0XuX<.  B fX;X#X9X0X1X3X$ * OXwX0XqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqX0XuX<.  N xXh.4 >.rXNXm.X , 8X0XqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqX0XuX3.  C nXS.    / -.4   %.pX9XqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqX0XiX%.  f hX8Xm.u     = &.2XqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXiX8XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XeX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXwX+XP._.eXeXrX+X!.!.[.0XrXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXUXpX0XeXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXrX(.'.UXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXuXhXsX0X0X0XiXfXgXsXwX0XqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.pXUX3X(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.).!.'.UXZX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9X9XqXqXqX0X9X9X9X0XqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX,XL.iXiX)._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._._.). XUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.=XrXwXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXwXyXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9XeXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXAX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX-XT.>XrX0XqX",
"qX0XrX>XT.-XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXUXZX8XtXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXwXuX:XY.>XrX0XqX",
"qX0XrX>XT.-XyXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXrXUXGX3X5X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X4X3X7X&XE.>XrX0XqX",
"qX0XrX>XT.-XqX6X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X6XqXPXwXW.P.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.P.Y.I.-XtX0XqX",
"qX0XrX,XI.-XzXkXkXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXlXkXcXSX6X4X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X5X6X0XqXqXqX",
"qX0XtX=X}.MXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXcX6XwXqX",
"qXqXeX5X0XzXjXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXhXjXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXgXhXyX0XqXqX",
"qXqXqXwXqX7X7X8X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X7X0XqXqXqX",
"qXqXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_2[] = {
/* columns rows colors chars-per-pixel */
"88 88 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqX0X0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXqXqXqX",
"qXqXqXrXtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8XqXqXqXqX",
"qX0XrX#X XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX8XkXSXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXDXmX0XqXqXqX",
"qX0XtX{.~.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XNXUXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXLX3X0XqXqXqX",
"qX0XtX|.(.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXwX9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XwX$X).wXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtX].).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqX0X0X0XqXqXqXqXqXqX0X0XqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXwXyXiXuXtXqXqXqX0XeXuXuXtXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqX0XyXP.i $ @.kX7X8XpX^.u - ..yX0XqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqX9XiXa.  #   ( zX6XwXqXW     Q uX0XqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqX0XwX9X^ R >.  ` lX6XiXC.  q # Q uX0XqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXwX8X].yXs.  ` lX9X5X6 y t.= E iX0XqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXyXsXq.  ` jXiX:.  i.K.  W tX0XqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqX7XpXe.  ) mX/.# M BXK.O _ xX0XqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqX9XaXe.  ( BXY   @.E.=.O n !.wX0XqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqX9XaXe.  ' sX8         o   - yX9X0XqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqX9XaXw.  ) lXL.M.C.U.&.  b ^.wXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqX9XaXi..  .lXiXkXdXBXG.* | xX0XqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X8X8X8X9XqX9X8XqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.eXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXeX8XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXwXwXqXqXqXqXqXqXwXwXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.9X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X3XmXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).SXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXDXFXZXHXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}._.UXUXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXLXUXsXhXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXmX5X0X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X0X7XQ.iXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXrX4XP.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXrXeXrXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXuX}.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XP.sXUX@X~./././././././././././././././././././././././././././././././././.(.W./.eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XP.aXaXXXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOXOX+X;XwXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.aXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXvX0XqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqXqX0XqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXjXkXuX{.wXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXyXyXeXeXqXqXrXyXyXyXtXtXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUX9X6X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7XqX[.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqX0XrX@XJ o s 3XiX$X[ ; % G (.tX0XqXqXqXqXqXqXqXqXeX4XI.sXUXeXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXtX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqX0XtX{.,   X < 9X=Xi   x v   6 .XiX9XqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqX0XuXk.x a.7 > uX~.p X.0XfX[   N.sX9XqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXwX@X1XXX2 2 7X9X%X8XgX;XD   T.aX9XqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXeXfX~.4 2 6XeXrX<Xz # & s.rXqX0XqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqX8XuX(.4 2 6XtXtX3X! s + g }.tX0XqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqX9XpX(.4 2 6X4X+X8XdXzX7.  [ qXwXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqX9XpX(.4 < tXC.2 %.4XfX:.  ~ 8XwXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqX8XpX/.< - rX].;   d g X   S.iX0XqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqX9XiX`.p u 5XpX[.Q , - C H.tXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqX0X7X6X9XwX9X6X7X7X7X7X9XqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXwXtXtXqXqXtXuXuXuXyXyXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqX8XrXzXcXpX9XaXxXxXxXxXxXpX0XqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XqXqXqXqX0XqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqX9XfX;Xp.8.).sX!.w.y.u.y.u.`.tX0XqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqX9XyXY.l     2.BX|           :.fX8XqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqX0XqX7Xn * A   r.NX^ * h.@X{.`.=XeXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqX0XrX:X4.XX&.  w.VXA 1 {.5X>XaXpX0XqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXuXpXX.  w.CX0   x     | 3XwXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqX0XrXX.  w.MX_ a / s.k   &.gX8XqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXtXX.  r.zX8X,XyXmX'.  a wX0X0XqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXtXX.  t.iX9.j.wXpX=X+ p 0X0X0XqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXtX+.  y.lX8   7.+XT   [ hX7XqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXtX(   4.mXA.r       F %XeX0XqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXMX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXeXI.4.).iXiX.Xj.u.N.3XyX0XqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.).UXBX8XrXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXtX6XI.sXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXdXcXpX0X9XyXkXxXhXwX9XqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}._.UXdX).(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.(.).I.pXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqX8X6X9XqXqX9X7X7X7X0XqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX}.`.xXoX{.|.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}.}. X`.hXUXwX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.0XrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXyXqXBXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.eX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX6XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXwX9X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X9XrX{.).eXqXqXqX",
"qX0XtX|./.wXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XMXUXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXiX}.).eXqXqXqX",
"qX0XtX|./.eXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXeX8XNXUX9X-X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X;X,X].`.eXqXqXqX",
"qX0XtX|./.0X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X3XmXfX~.P.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.I.Y.(.eXqXqXqX",
"qX0XtX}._.bXGXDXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXFXDXPXCXnXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXNXnX0XqXqXqX",
"qX0XrX+XsXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXPXKXIXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXJX9XqXqXqX",
"qXqXqXwXwX9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_3[] = {
/* columns rows colors chars-per-pixel */
"64 64 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXqXqXqX",
"qXqXqXqX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXtXwXqXqXqX",
"qXqXwX9X_.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX8XkXUXPXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXPXUXnX9XqXqXqX",
"qXqXwX9X/.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXHXtXpXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXpXyX X7XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXDX5X0X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9XqX2X^.0XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX5X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqXqXqXqX9X7X9XqXqXqX8X8X0XqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqXqXqXqXdXlXiX0XqXqXgXhXeX0XqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqXqX0XyXh.h k.dX9XqX{ D @XyX0XqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqX9XsXf.<   O.jXaXk.    `.pX9XqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqX0XtXW.].p +.bX*X7 T q ).pX9XqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqXqX9XsX0X7 O.DX~ D /.;  XaX9XqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqXqX0XwX4X6 :.oX  x.b.$ S.uX0XqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqXqX0XrX5X6 <.N.  = O   e 6XwXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqXqX0XrX5X3 O.jX=XwXM.+ ~.pX9XqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.5XeXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXeX8XvXFX7XwXqXqXqXqXqXqX0X.X*XtXeXyX4X|.8XwXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.1X8X6X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X3XxXFX7XwXqXqXqXqXqXqXqXiXtX0XqX0XeXpXwXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X/.gXCXBXBXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXBXVXMXSXSX8XwXqXqXqXqXqXqXqX9X0XqXqXqXqX9XqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X/.DXHXNXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXVXBXZXjXaXGX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.AXfX3X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X6XqX].4XKX6XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXkX9XwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXuX|.6XKX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX|.5XKX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX|.5XJX6XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XwX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX|.5XIXqXyXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXtXuX8X).9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX|.5XbX(.`.`._._._._._._._._._._._._._._._._._._._._.`.(.Q.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXqX0X8X9XqXqX9X8X0XqXqXqXqXqXqX0XyX X2XmXyXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXpXaXiXqXqXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqX0XwXhXaX0XqXdXhXtX9XqXqXqXqXqX0XyX|.4XUXCXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXAXDX0X7XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqX0XiX8XC.(.iXtXR.N.*XiX0XqXqXqXqX0XyX|.5XHX1X7X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X6X8X<X~.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqX0XtX@Xx   K hXY o - 0 _.iX9XqXqXqX0XyX|.5XKX7XwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXeX5X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqX9XaXV.#.b T 0XG /.xX. O.zX7XqXqXqX0XyX|.5XKX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqX9XjXm H hXiX~.E s (.pX9XqXqXqX0XyX|.5XKX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXwXrXv J lXzXb.u 9 G.iX0XqXqXqX0XyX|.5XKX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXqXyXm T 4X4.8XUXa g iXqXqXqXqX0XyX|.5XKX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXqXyXi Z 8X= 0 G   4.aX9XqXqXqX0XyX|.5XKX7XwXqXqXqXqXqXqXqX9X8XqX0X8X8X8X9XqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXqXrX0.u.jXoX#.Q v.rXwXqXqXqXqX0XyX|.5XKX7XwXqXqXqXqXqXqX9XaXhXwXyXhXjXkXaX0XqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXqXqXhXgX9XyXzXvXdXqXqXqXqXqXqX0XyX|.5XKX7XwXqXqXqXqXqXqXdX^.Z.8X@XB.M.c.^.rX0XqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXqXqX8X8XqX0X7X7X9XqXqXqXqXqXqX0XyX|.5XKX7XwXqXqXqXqX0XyX,.    :X,.  7 %  .pX9XqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX|.5XKX7XwXqXqXqX0XeX,XX.$.& 1X] E aXyXdX0XqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX|.5XKX7XwXqXqXqXqXqX6XkXV.  2XH - y 3 P.iX9XqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX|.5XKX7XwXqXqXqXqXqX9XdXz.o &X_.d.XXJ 8 eXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XyX|.5XKX7XwXqXqXqXqXqX8XdXv.# @Xb.!.UXa.+ 5XeX0XqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.SXjX7XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XtX|.5XKX7XwXqXqXqXqXqX8XgXh.  -X{   ! @ [ sX9XqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.AXkXqXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXeXpX X5XKX7XwXqXqXqXqXqX0XiX^. .>XtXu.T y.0XwX0XqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X^.FXdX=X,X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X>X2X_.5XKX6XwXqXqXqXqXqXqX0XuXzXeXqXhXbXhXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X).hX#X(.`._._._._._._._._._._._._._._._._._._._._._.`.~.eXJX7XwXqXqXqXqXqXqXqX0X7XqXqX8X7X8XqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.2XwXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeX8XvXFX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.5XeX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXeX4X(.9XwXqXqX",
"qXqXwX9X(.4XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XcXFX7XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XwX4X(.9XwXqXqX",
"qXqXwX9X(.5XeXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX7XvXFX0XtXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXyX7X).9XwXqXqX",
"qXqXwX9X).4XqX9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X0X5XcXDX=X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X:X;X>X&X~.9XwXqXqX",
"qXqXwX9X~.7XaXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXiXpXtXnXiX}. X|.|.|.|.|.|.|.|.|.|.|.|.|.|.|.|.|.|.|.|.|.|.|.+X0XqXqXqX",
"qXqXwX7X$XAXUXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXIXPXLXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXUXPXeX0XqXqX",
"qXqXqX0XeXtXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXqXqXqXqX",
"qXqXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_4[] = {
/* columns rows colors chars-per-pixel */
"48 48 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXqXqXqXqX",
"qXqXqXwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X7X8XqXqXqXqX",
"qXqXwX6X4XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XsXlXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXlXgXqXqXqXqX",
"qX0XrX-XOXtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XFXVXnXmXmXmXmXmXmXmXmXmXmXmXmXmXmXmXNX0X8XwXqXqX",
"qX0XrX;X@XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XAXwX6X7X7X7X7X7X7X7X7X7X7X7X7X7X7X8X5X].0XqXqXqX",
"qX0XrX;X@XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XAXuXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwX0X|.0XqXqXqX",
"qX0XrX;X@XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XAXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX9X|.0XqXqXqX",
"qX0XrX;X@XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XAXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX9X|.0XqXqXqX",
"qX0XrX;X@XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XAXyX0XqXqXqX0X0XwXqXqXqXwXqXqXqXqXwX9X|.0XqXqXqX",
"qX0XrX;X@XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XAXyX0XqXqX0XuXuX7XqXwXtX8XqXqXqXqXwX9X|.0XqXqXqX",
"qX0XrX;X@XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XAXyX0XqX0XyX@Xz K uXaXG C uX0XqXqXwX9X|.0XqXqXqX",
"qX0XrX;X@XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XAXyX0XqX9XuX/.] c bXx.; C qXqXqXqXwX9X|.0XqXqXqX",
"qX0XrX;X@XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XAXyX0XqXqX8XzXL.M 4XG B.P iX0XqXqXwX9X|.0XqXqXqX",
"qX0XrX;X@XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XAXyX0XqXqX9XfXn.D i.$ H O L.aX9XqXwX9X|.0XqXqXqX",
"qX0XrX;X@XtXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXAXyX0XqXqX9XfXS.S >X~.w.j &XtX0XqXwX9X|.0XqXqXqX",
"qX0XrX;X+XqX7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X6XAXyX0XqXqXqXwX7X1XyXfXuX6XyX0XqXqXwX9X|.0XqXqXqX",
"qX0XtX=X-XbXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXxXlXFXyX0XqXqXqXqXwXrX0X9X0XeX0XqXqXqXwX9X|.0XqXqXqX",
"qX0XtX&XqXJXxXnXbXbXbXbXbXbXbXbXbXbXbXbXbXbXMXyXmXiX9XqXqXqXqXqX0XqXqXqX0XqXqXqXqXwX9X|.0XqXqXqX",
"qX0XtX&XqXzX3X8X7X7X7X7X7X7X7X7X7X7X7X7X7X6XqX{.xXpX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX9X|.0XqXqXqX",
"qX0XtX&XqXvX8XwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXtXoXcXiX9XqX0X0X0X0X0X0X0X0X0X0X0X0X0XqX8X}.0XqXqXqX",
"qX0XtX&XqXvX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XtXXXvXfXrXyXtXtXtXtXtXtXtXtXtXtXtXtXtXyXrX X0XqXqXqX",
"qX0XtX&XqXvX7XwXqXqXqXqX0XqXqX0X0XqXqXqXqX0XrXXXkX>XoX@X+X+X+X+X+X+X+X+X+X+X+X+X+X@XoX_.qXqXqXqX",
"qX0XtX&XqXvX7XwXqXqXqX9XeXeX9XeXrX9XqXqXqX0XrXoXfXhXpXaXaXaXaXaXaXaXaXaXaXaXaXaXaXaXsXyX0XqXqXqX",
"qX0XtX&XqXvX7XwXqXqX0XsX7X4XdX6X<XaX0XqXqX0XtX.XbXbXhXkXkXkXkXkXkXkXkXkXkXkXkXkXjXjXkX:X8XwXqXqX",
"qX0XtX&XqXvX7XwXqX9XpXZ.= t.R.g x x.aX9XqX0XrXXXxXyX6X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X5X].0XqXqXqX",
"qX0XtX&XqXvX7XwXqX0XwXK.Z >.I.U.2.K jX8XqX0XrXXXcXpX0XwXqXqXqXqXqXqXqXqXqXqXqXqXqXwX0X|.0XqXqXqX",
"qX0XtX&XqXvX7XwXqXqX7XNX#. .UXY.2 d.sX9XqX0XrXXXcXpX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX9X|.0XqXqXqX",
"qX0XtX&XqXvX7XwXqXqX8XjX' %.C.U.x.7 eXqX0X0XrXXXcXpX9XqXqXqXqX0X9XqX9X8X9XqXqXqXqXwX9X|.0XqXqXqX",
"qX0XtX&XqXvX7XwXqXqX8XjX%.6.J.a k a.iX9XqX0XrXXXcXpX9XqXqXqX0XeXfXeXpXjXjXrX0XqXqXwX9X|.0XqXqXqX",
"qX0XtX&XqXvX7XwXqXqXqXqX8X7XdX8X3XaXqXqXqX0XrXXXcXpX9XqXqX0XtX4XL.3X].b.m.>XrX0XqXwX9X|.0XqXqXqX",
"qX0XtX&XqXvX7XwXqXqXqXqXwXwX9XwXeX9XqXqXqX0XrXXXcXpX9XqXqXwX8X!   Q.A P +.XXyX0XqXwX9X|.0XqXqXqX",
"qX0XtX&XqXvX7XwXqXqXqXqXqXqXqXqX0XqXqXqXqX0XrXXXcXpX9XqXqXeX2X`.b F.c } l.yX0XqXqXwX9X|.0XqXqXqX",
"qX0XtX&XqXvX7XwXqXqXqXqXqXqXqXqXqXqXqXqXqX0XrXXXcXpX9XqXqX0XeXaXz L.S.P.g d.jX8XqXwX9X|.0XqXqXqX",
"qX0XtX&XqXcX6XqX0X0X0X0X0X0X0X0X0X0X0X0X0X0XrX.XcXpX9XqXqXqXqX0Xi V.@.!.m a.jX8XqXwX9X|.0XqXqXqX",
"qX0XtX&XqXmXqXtXtXtXtXtXtXtXtXtXtXtXtXtXtXrXiXOXcXpX9XqXqXqXqXwXo.}.n.M <.qXqXqXqXwX9X|.0XqXqXqX",
"qX0XtX&XwXiXXX$X#X#X#X#X#X#X#X#X#X#X#X#X#X@X%X_.vXpX9XqXqXqXqXqXdXrXdXjXgXwXqXqXqXwX9X|.0XqXqXqX",
"qX0XrX-X&X4X*X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-X-XCXuX0XqXqXqXqXqX9X0X9X8X8XqXqXqXqXwX9X|.0XqXqXqX",
"qX0XrX;X+XuXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXtXeXSXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX9X|.0XqXqXqX",
"qX0XrX;X@XrX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX9XAXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXwX9X|.0XqXqXqX",
"qX0XrX;X@XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XAXyX9XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X|.0XqXqXqX",
"qX0XrX;X@XrX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX9XSXiXqXeXwXwXwXwXwXwXwXwXwXwXwXwXwXeX0X}.0XqXqXqX",
"qX0XrX-XOXuXwXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXeXMX%X XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX.X}.qXqXqXqX",
"qXqXrX>XrXHXAXSXSXSXSXSXSXSXSXSXSXSXSXSXSXSXSXDXAXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXMXNXmXqXqXqXqX",
"qXqXqXqXrXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXeXqXqXqXqX",
"qXqXqXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_5[] = {
/* columns rows colors chars-per-pixel */
"44 44 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXqXqXqXqXqX",
"qXqXqXwXwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX9X8X8X8X8X8X8X8X8X8X8X8X8X8X8X8X7X8XqXqXqXqX",
"qXqXwX6X5XwXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XsXkXkXkXkXkXkXkXkXkXkXkXkXkXkXkXlXgXqXqXqXqX",
"qX0XrX-X$XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX0XDXnXcXvXvXvXvXvXvXvXvXvXvXvXvXcXnX9X8XwXqXqX",
"qX0XrX;X%XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXVX0X7X7X7X7X7X7X7X7X7X7X7X7X7X7X7X{.8XwXqXqX",
"qX0XrX;X%XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXrXqXwXwXwXwXwXwXwXwXwXwXwXwXwXwXXX8XwXqXqX",
"qX0XrX;X%XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX.X8XwXqXqX",
"qX0XrX;X%XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXrX0XqXqXqX9X8XqXqX8X9XqXqXqXqXqX.X8XwXqXqX",
"qX0XrX;X%XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXrX0XqXqXqXgXgXqX0XjXpX0XqXqXqXqX.X8XwXqXqX",
"qX0XrX;X%XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXrX0XqX0XtXi.[ 9XaXX.g.dX9XqXqXqX.X8XwXqXqX",
"qX0XrX;X%XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXrX0X0XrX,X .w iXS.  +.hX8XqXqXqX.X8XwXqXqX",
"qX0XrX;X%XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXrX0XqX0XyX,XC +X^ 8.:.kX8XqXqXqX.X8XwXqXqX",
"qX0XrX;X%XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXrX0XqX0XuX XM y.w x q 3XeX0XqXqX.X8XwXqXqX",
"qX0XrX;X%XtXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXrX0XqX0XtX*X' -X:Xr.-.iX0XqXqXqX.X8XwXqXqX",
"qX0XrX;X$XeX9X9X9X9X9X9X9X9X9X9X9X9X9X9X9X0XZXrX0XqXqXqXwXuXrXuXpXpXwXqXqXqXqX.X8XwXqXqX",
"qX0XtX*X8XGXBXVXVXVXVXVXVXVXVXVXVXVXVXVXCXmXCXrX0XqXqXqXqX0X0X0X0X0XqXqXqXqXqX.X8XwXqXqX",
"qX0XtX&XuXcXeXuXyXyXyXyXyXyXyXyXyXyXyXyXpX-XzXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX.X8XwXqXqX",
"qX0XtX&XyXhX6X0X0X0X0X0X0X0X0X0X0X0X0X9XeXOXzXyX9XqX0X0X0X0X0X0X0X0X0X0X0X0XqX.X8XwXqXqX",
"qX0XtX&XyXjX8XwXqXqXqXqXqXqXqXqXqXqXqXqXrX@XxXaXrXtXtXtXtXtXtXtXtXtXtXtXtXtXtXoX8XwXqXqX",
"qX0XtX&XyXjX8XqXqXqXqX0X0XqX0X0XqXqXqX0XrX@XjX,X%X*X*X*X*X*X*X*X*X*X*X*X*X*X&X'.9XwXqXqX",
"qX0XtX&XyXjX8XqXqXqX0XtXtX9XiXtX0XqXqX0XrX@XdXaXtXyXtXtXtXtXtXtXtXtXtXtXtXtXyXwX0XqXqXqX",
"qX0XtX&XyXjX8XqXqX0XiX-X;XaX|.-XiX0XqX0XrX+XvXcXhXjXjXjXjXjXjXjXjXjXjXjXjXjXlX<X7XwXqXqX",
"qX0XtX&XyXjX8XqX9XiX^., 3.c.Z A ^.pX9X0XrX@XzXeX7X8X8X8X8X8X8X8X8X8X8X8X8X8X8X}.8XwXqXqX",
"qX0XtX&XyXjX8XqX0XeX:X@.O.-XI.m U.dX9X0XrX@XzXuX0XwXqXqXqXqXqXqXqXqXqXqXqXqXwXXX8XwXqXqX",
"qX0XtX&XyXjX8XqXqX7XbXt.} vXY.n Z.sX9X0XrX@XzXyX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX.X8XwXqXqX",
"qX0XtX&XyXjX8XqXqX8XkX,...7.X.M w.jX8X0XrX@XzXyX0XqXqXqX9XqXqXqXrXeXqXqXqXqXqX.X8XwXqXqX",
"qX0XtX&XyXjX8XqXqX0XuX'.^.5Xr.C.yXqXqX0XrX@XzXyX0XqXqX0XaXwX0X0X>X1XqXqXqXqXqX.X8XwXqXqX",
"qX0XtX&XyXjX8XqXqXqX0XiXpXeXlXfX0XqXqX0XrX@XzXyX0XqX9XpXV.8 U.| D &.8XwXqXqXqX.X8XwXqXqX",
"qX0XtX&XyXjX8XqXqXqXqX0X9XqX8X8XqXqXqX0XrX@XzXyX0XqX0XwXE.n h.N @.!.iX9XqXqXqX.X8XwXqXqX",
"qX0XtX&XyXjX8XqXqXqXqXqXqXqXqXqXqXqXqX0XrX@XzXyX0XqXqX8XnX( k.Z.k.k %XyX0XqXqX.X8XwXqXqX",
"qX0XtX&XyXhX7XqXqXqXqXqXqXqXqXqXqXqXqX0XrX+XzXyX0XqXqX9XfXG u.3.n.k #XyX0XqXqX.X8XwXqXqX",
"qX0XtX&XuXlX9XeXeXeXeXeXeXeXeXeXeXeXeXwXyX@XzXyX0XqXqX0XiXk.].K./ I.iX9XqXqXqX.X8XwXqXqX",
"qX0XtX*XqX2X|.oXXXXXXXXXXXXXXXXXXXXXXXXXoX].vXtX0XqXqXqX0XdXyXsXcXaX0XqXqXqXqX.X8XwXqXqX",
"qX0XrX;X%X9X5X6X6X6X6X6X6X6X6X6X6X6X6X6X6X7XCXrX0XqXqXqXqX9X0X9X7X9XqXqXqXqXqX.X8XwXqXqX",
"qX0XrX;X%XyXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXwXCXrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqX.X8XwXqXqX",
"qX0XrX;X%XrX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXCXeX0X0X0X0X0X0X0X0X0X0X0X0X0X0XqX.X8XwXqXqX",
"qX0XrX;X&XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXZXuXrXrXrXrXrXrXrXrXrXrXrXrXrXrXtXXX8XwXqXqX",
"qX0XrX-X#XtX0XqXqXqXqXqXqXqXqXqXqXqXqXqX0XwXnX%X+X@X@X@X@X@X@X@X@X@X@X@X@X@X+X[.9XqXqXqX",
"qXqXrX>XrXDXVXCXCXCXCXCXCXCXCXCXCXCXCXCXCXZXCXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXzXwXqXqXqX",
"qXqXqXqXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXrXyXyXyXyXyXyXyXyXyXyXyXyXyXyXyXuXtXqXqXqXqX",
"qXqXqXqX0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0X0XqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX",
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX"
};

/* XPM */
static const char *const xpm_icon_6[] = {
/* columns rows colors chars-per-pixel */
"32 32 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qXqXqXqXqXqXqXqXqXqXqXqXqXqXqXqX0X9X0X0X9X9X9X9X9X9X9X0X9X9XqXqX",
"qXqX0XqXqXqXqXqXqXqXqXqXqXqXqXqXeXuXuXuXuXuXuXuXuXuXuXuXuXuXqXqX",
"wX6X&XeXqXqXqXqXqXqXqXqXqXqXqX0XMXmXvXvXvXvXvXvXvXvXvXvXmXpX8XwX",
"wX6X$XeX0XqXqXqXqXqXqXqXqXqXqX0XnXqX8X9X9X9X9X9X9X9X9X8XqX+X5XeX",
"wX6X%XeXqXqXqXqXqXqXqXqXqXqXqX0XnXrXqXwXwXqXqXwXqXqXwXqXrX%X6XwX",
"wX6X%XeX0XqXqXqXqXqXqXqXqXqXqX0XnXeX0XqX0X0XqX0X0XqXqX0XeX%X6XwX",
"wX6X%XeX0XqXqXqXqXqXqXqXqXqXqX0XnXeXqX0XtXtX0XrXyXqXqXqXeX%X6XwX",
"wX6X%XeX0XqXqXqXqXqXqXqXqXqXqX0XnXeX9XtX*XC ~.-Xl $XyX9XeX%X6XwX",
"wX6X%XeX0XqXqXqXqXqXqXqXqXqXqX0XnXeX0XqXqX| V.r.C OXiX9XeX%X6XwX",
"wX6X%XrXqXwXwXwXwXwXwXwXwXwXwXqXnXeXqX7XxX7.@.Y a v.gX8XeX%X6XwX",
"wX6X$X0X8X8X8X8X8X8X8X8X8X8X8X7XbXeX0X9XpXF.Q.4Xu. XuX9XeX%X6XwX",
"eX5X;XlXhXhXhXhXhXhXhXhXhXhXhXhXNXwXqXqX0XiXuXtXgXyX0XqXeX%X6XwX",
"eX2X0XMXfXjXhXhXhXhXhXhXhXhXlXtXjXtX0XqXqX9X0X0X8X0XqX0XeX%X6XwX",
"eX3X9XsX5X8X8X8X8X8X8X8X8X7XqX*XpXuXqXwXwXwXwXwXwXwXwXwXrX%X6XwX",
"eX3X9XgX9XwXwXqX0XqX0XqXwXqXrX:XaXqX5X6X6X6X6X6X6X6X6X6X8XOX5XwX",
"eX3X9XfX8XqX0XrXiXwXsXeX0XqXrX:XtX8X3X4X4X4X4X4X4X4X4X3X4X<X9XwX",
"eX3X9XfX8XqXeX1X X7XQ.3XeX0XrX;XdXcXhXjXjXjXjXjXjXjXjXjXlX0X7XwX",
"eX3X9XfX8X0XaXw.A A.W $.aX9XrX:XaXeX7X8X8X8X8X8X8X8X8X7X0X+X5XwX",
"eX3X9XfX8XqXeX:X/ tXl.[ pX9XrX:XaXyX0XwXwXqXqX0X0XqXwXqXrX%X6XwX",
"eX3X9XfX8XqXrX>X^ x...R qXqXrX:XaXyX0XqX0XyXyXyXfXtX0XqXeX%X6XwX",
"eX3X9XfX8XqXwX7X`.=Xv.@XtX0XrX:XaXyX0X0XtX#X%X$XI.=XrX0XeX%X6XwX",
"eX3X9XfX8XqXqXwXiXtXjXyX0XqXrX:XaXyX0XwXeXR @.^ ( $XyX9XeX%X6XwX",
"eX3X9XfX8XqXqX0X9X0X8X0XqX0XrX;XaXyX0X0XaXK.#.z.*.Z.dX8XeX%X6XwX",
"eX3X9XhX0XrXeXeXeXeXeXeXeXeXyX>XaXyX0X9XgXH.} M.) e.jX7XeX%X6XwX",
"eX2X0XyX>X2X1X1X1X1X1X1X1X1X4X#XaXyX0XqXrX=X.X'.B.wXqX0XeX%X6XwX",
"eX5X-X1X&X=X=X=X=X=X=X=X=X=X=X&XzXrX0XqX0XrXuXiXhXqXqXqXeX%X6XwX",
"wX6X$XtXrXrXrXrXrXrXrXrXrXrXrXeXmXeX0XqXqX0X9X0X8X0XqX0XeX$X6XwX",
"wX6X%XeX0XqXqXqXqXqXqXqXqXqXqX0XnXyXeXrXrXrXrXrXrXrXrXeXyX&X6XwX",
"wX6X@XqX8X9X9X9X9X9X9X9X9X9X9X8XcX2X>X,X,X,X,X,X,X,X,X,X1X.X6XwX",
"eX5X4XMXvXvXvXvXvXvXvXvXvXvXvXbXmXtXtXtXtXtXtXtXtXtXtXtXtXtXqXqX",
"qX0XrXuXuXuXuXuXuXuXuXuXuXuXuXuXyXpXaXaXaXaXaXaXaXaXaXaXaXpXwXqX",
"qXqX0X9X0X9X9X9X9X9X9X9X9X9X9X0X0X9X9X9X9X9X9X9X9X9X9X9X9X9XqXqX"
};

/* XPM */
static const char *const xpm_icon_7[] = {
/* columns rows colors chars-per-pixel */
"24 24 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"0X9XqXqXqXqXqXqXqXqXqXqXyXaXaXaXaXaXaXaXaXaXaXwX",
"3X>XeXqXqXqXqXqXqXqXqXqXbXjXhXhXhXhXhXhXhXkXtX6X",
"3X:XrX0XqXqXqXqXqXqXqXwXzX8X9X8X8X8X8X8X8XqX=X2X",
"3X:XeX0XqXqXqXqXqXqXqXwXxXqXwXwXtXqXrXeXqXrX>X3X",
"3X:XeX0XqXqXqXqXqXqXqXwXxXqXqX0X;XeX3X5XwXeX:X3X",
"3X:XeX0XqXqXqXqXqXqXqXwXxX0XyX#XR ,XT ~.aXwX:X3X",
"3X>XrXqXwXwXwXwXwXwXqXeXxXqX9XyX4.-.z A.dXqX:X3X",
"3X;XqX8X8X8X8X8X8X8X8X0XzXqX0XuXb.W.q.P.aXwX:X3X",
"2X8XxXhXjXjXjXjXjXjXjXkXcX0XqXqXyXaXgXuX0XrX:X3X",
"1XiXfXuXiXiXiXiXiXiXaX0XaXeXqXqX0X0X9X0XqXrX>X3X",
"2XtXeX8X9X7X8X7X8X9XqX,XiXqX9X9X9X9X9X9X8XwX-X2X",
"2XyXyX0XqXsXiXdXuX0XeX2XrX7X5X5X5X5X5X5X5X6X<X7X",
"2XyXtX9XrXE.}.J.oXrXwX<XsXgXsXsXsXsXsXsXsXgXqX5X",
"2XyXtX9XpX&.a.e.>.fX0X1XiXqX8X9X8X8X7X8X8XqX-X2X",
"2XyXtX7XhXF.t.d...pXqX1XpXeXqXqXiXtXsXtX0XrX>X3X",
"2XyXtX9XyX+X(.m._.tXwX1XpXeXqXwXXX=X/.-XeXeX:X3X",
"2XtXtX0X0XtXiXfXiX0XeX1XpXwXqX0X .&.R }.yXwX:X3X",
"2XyXyXqXwXqXqX0XqXwXrX2XpXeX8XhXk.7.3.y.fXqX:X3X",
"2XyX0X5X6X6X6X6X6X5X8X;XiXeX9XiX!.P.u.].yXwX:X3X",
"3X<X>X-X-X-X-X-X-X-X-X-XhXwXqX0XyXpXsXuX0XrX:X3X",
"3X:XyXeXrXrXrXrXrXrXeXtXcXqXwXwXqX0X9XqXwXrX>X3X",
"3X;X0X7X8X8X8X8X8X8X8X9XlX6X7X7X7X7X7X7X6X9X*X2X",
"3X8XlXhXhXhXhXhXhXhXhXkXzX7X8X8X8X8X8X8X8X8X6X9X",
"0XyXaXaXaXaXaXaXaXaXaXaXpXsXsXsXsXsXsXsXsXsXsXeX"
};

/* XPM */
static const char *const xpm_icon_8[] = {
/* columns rows colors chars-per-pixel */
"16 16 256 2 ",
"   c black",
".  c #010101",
"X  c #020202",
"o  c gray1",
"O  c #040404",
"+  c gray2",
"@  c #060606",
"#  c #070707",
"$  c gray3",
"%  c #090909",
"&  c gray4",
"*  c #0B0B0B",
"=  c #0C0C0C",
"-  c gray5",
";  c #0E0E0E",
":  c gray6",
">  c #101010",
",  c #111111",
"<  c gray7",
"1  c #131313",
"2  c gray8",
"3  c #151515",
"4  c #161616",
"5  c gray9",
"6  c #181818",
"7  c #191919",
"8  c gray10",
"9  c #1B1B1B",
"0  c gray11",
"q  c #1D1D1D",
"w  c #1E1E1E",
"e  c gray12",
"r  c #202020",
"t  c gray13",
"y  c #222222",
"u  c #232323",
"i  c gray14",
"p  c #252525",
"a  c gray15",
"s  c #272727",
"d  c #282828",
"f  c gray16",
"g  c #2A2A2A",
"h  c gray17",
"j  c #2C2C2C",
"k  c #2D2D2D",
"l  c gray18",
"z  c #2F2F2F",
"x  c gray19",
"c  c #313131",
"v  c #323232",
"b  c gray20",
"n  c #343434",
"m  c #353535",
"M  c gray21",
"N  c #373737",
"B  c gray22",
"V  c #393939",
"C  c #3A3A3A",
"Z  c gray23",
"A  c #3C3C3C",
"S  c gray24",
"D  c #3E3E3E",
"F  c #3F3F3F",
"G  c gray25",
"H  c #414141",
"J  c gray26",
"K  c #434343",
"L  c #444444",
"P  c gray27",
"I  c #464646",
"U  c gray28",
"Y  c #484848",
"T  c #494949",
"R  c gray29",
"E  c #4B4B4B",
"W  c #4C4C4C",
"Q  c gray30",
"!  c #4E4E4E",
"~  c gray31",
"^  c #505050",
"/  c #515151",
"(  c gray32",
")  c #535353",
"_  c gray33",
"`  c #555555",
"'  c #565656",
"]  c gray34",
"[  c #585858",
"{  c gray35",
"}  c #5A5A5A",
"|  c #5B5B5B",
" . c gray36",
".. c #5D5D5D",
"X. c gray37",
"o. c #5F5F5F",
"O. c #606060",
"+. c gray38",
"@. c #626262",
"#. c gray39",
"$. c #646464",
"%. c #656565",
"&. c gray40",
"*. c #676767",
"=. c #686868",
"-. c DimGray",
";. c #6A6A6A",
":. c gray42",
">. c #6C6C6C",
",. c #6D6D6D",
"<. c gray43",
"1. c #6F6F6F",
"2. c gray44",
"3. c #717171",
"4. c #727272",
"5. c gray45",
"6. c #747474",
"7. c gray46",
"8. c #767676",
"9. c #777777",
"0. c gray47",
"q. c #797979",
"w. c gray48",
"e. c #7B7B7B",
"r. c #7C7C7C",
"t. c gray49",
"y. c #7E7E7E",
"u. c gray50",
"i. c #808080",
"p. c #818181",
"a. c gray51",
"s. c #838383",
"d. c #848484",
"f. c gray52",
"g. c #868686",
"h. c gray53",
"j. c #888888",
"k. c #898989",
"l. c gray54",
"z. c #8B8B8B",
"x. c gray55",
"c. c #8D8D8D",
"v. c #8E8E8E",
"b. c gray56",
"n. c #909090",
"m. c gray57",
"M. c #929292",
"N. c #939393",
"B. c gray58",
"V. c #959595",
"C. c gray59",
"Z. c #979797",
"A. c #989898",
"S. c gray60",
"D. c #9A9A9A",
"F. c #9B9B9B",
"G. c gray61",
"H. c #9D9D9D",
"J. c gray62",
"K. c #9F9F9F",
"L. c #A0A0A0",
"P. c gray63",
"I. c #A2A2A2",
"U. c gray64",
"Y. c #A4A4A4",
"T. c #A5A5A5",
"R. c gray65",
"E. c #A7A7A7",
"W. c gray66",
"Q. c #A9A9A9",
"!. c #AAAAAA",
"~. c gray67",
"^. c #ACACAC",
"/. c gray68",
"(. c #AEAEAE",
"). c #AFAFAF",
"_. c gray69",
"`. c #B1B1B1",
"'. c #B2B2B2",
"]. c gray70",
"[. c #B4B4B4",
"{. c gray71",
"}. c #B6B6B6",
"|. c #B7B7B7",
" X c gray72",
".X c #B9B9B9",
"XX c gray73",
"oX c #BBBBBB",
"OX c #BCBCBC",
"+X c gray74",
"@X c gray",
"#X c gray75",
"$X c #C0C0C0",
"%X c #C1C1C1",
"&X c gray76",
"*X c #C3C3C3",
"=X c gray77",
"-X c #C5C5C5",
";X c #C6C6C6",
":X c gray78",
">X c #C8C8C8",
",X c gray79",
"<X c #CACACA",
"1X c #CBCBCB",
"2X c gray80",
"3X c #CDCDCD",
"4X c #CECECE",
"5X c gray81",
"6X c #D0D0D0",
"7X c gray82",
"8X c #D2D2D2",
"9X c LightGray",
"0X c gray83",
"qX c #D5D5D5",
"wX c gray84",
"eX c #D7D7D7",
"rX c #D8D8D8",
"tX c gray85",
"yX c #DADADA",
"uX c gray86",
"iX c gainsboro",
"pX c #DDDDDD",
"aX c gray87",
"sX c #DFDFDF",
"dX c gray88",
"fX c #E1E1E1",
"gX c #E2E2E2",
"hX c gray89",
"jX c #E4E4E4",
"kX c gray90",
"lX c #E6E6E6",
"zX c #E7E7E7",
"xX c gray91",
"cX c #E9E9E9",
"vX c #EAEAEA",
"bX c gray92",
"nX c #ECECEC",
"mX c gray93",
"MX c #EEEEEE",
"NX c #EFEFEF",
"BX c gray94",
"VX c #F1F1F1",
"CX c gray95",
"ZX c #F3F3F3",
"AX c #F4F4F4",
"SX c gray96",
"DX c #F6F6F6",
"FX c gray97",
"GX c #F8F8F8",
"HX c #F9F9F9",
"JX c gray98",
"KX c #FBFBFB",
"LX c gray99",
"PX c #FDFDFD",
"IX c #FEFEFE",
"UX c white",
/* pixels */
"qX0XqXqXqXqXqXqXtXuXyXyXyXyXyXqX",
"6X3XeXqXqXqXqXeXjXaXpXpXiXdXeX8X",
"6X2XeXqXqXqX0XrXsX7XyXtXtX0X<X6X",
"6X2XeXqXqXqXqXrXsXtX`.{.'.iX2X6X",
"6X1XqX9X9X9X9XwXaXpXL.' :.fX<X6X",
"6X7XaXiXiXiXiXsXdXwX2XoXoXrX2X6X",
"6XuXaXuXuXuXpXtXuX0XeXuXyXwX2X6X",
"7XeX0XeXuXyXqX5XwX8X7X7X6X9X3X8X",
"7XeXiXXXZ.G.0X7XuXyXtXrXrXuX8X7X",
"7XeXiX+Xq.:.2X9XrX9XrXaXrXqX1X6X",
"7XrXwX0X*X#XwX7XrXuX[.d.T.iX1X6X",
"7XtXeXwXyXuXwX7XeXpX].O.r.uX1X6X",
"7X5X1X<X,X,X<X<XuXwX9X=X=XrX2X6X",
"6X<X0X8X8X9X8XwXdX8XqXrXrXqX1X5X",
"6X0XdXaXaXaXaXfXsX8X9X8X8X0X6X9X",
"qXtXuXyXyXyXyXyXtXyXtXtXtXtXyXwX"
};

const char *const *const xpm_icons[] = {
    xpm_icon_0,
    xpm_icon_1,
    xpm_icon_2,
    xpm_icon_3,
    xpm_icon_4,
    xpm_icon_5,
    xpm_icon_6,
    xpm_icon_7,
    xpm_icon_8,
};
const int n_xpm_icons = 9;
